package events

import (
	"time"

	"github.com/coyim/coyim/session/access"
	sdata "github.com/coyim/coyim/session/data"
	"github.com/coyim/coyim/xmpp/data"
	"github.com/coyim/coyim/xmpp/jid"
)

// Event represents a Session event
type Event struct {
	Type    EventType
	Session access.Session
}

// EventType represents the type of Session event
type EventType int

// Session event types
const (
	Disconnected EventType = iota
	Connecting
	Connected
	ConnectionLost

	RosterReceived
	Ping
	PongReceived
)

// Peer represents an event associated to a peer
type Peer struct {
	Session access.Session
	Type    PeerType
	// This can be either with or without Resource depending on the peer type
	From jid.Any
}

// Notification represents a notification event
type Notification struct {
	Session      access.Session
	Peer         jid.Any
	Notification string
}

// DelayedMessageSent represents the event that a delayed message is sent
type DelayedMessageSent struct {
	Session access.Session
	Peer    jid.Any
	Tracer  int
}

// PeerType represents the type of Peer event
type PeerType int

// Peer types
const (
	IQReceived PeerType = iota

	OTREnded
	OTRNewKeys
	OTRRenewedKeys

	SubscriptionRequest
	Subscribed
	Unsubscribe
)

// Presence represents a presence event
type Presence struct {
	Session access.Session
	*data.ClientPresence
	Gone bool
}

// Message represents a message event
type Message struct {
	Session   access.Session
	From      jid.Any
	When      time.Time
	Body      []byte
	Encrypted bool
}

// ChatPresence represents a presence event in a chat room
type ChatPresence struct {
	Session access.Session
	*data.ClientPresence
}

// TODO: This feels like bad naming - it implies it's for any chat, but realistically should only be for group chats.
// The event name should actually take that into account
// ChatMessage represents a message event in a chat room
type ChatMessage struct {
	From jid.WithResource
	When time.Time
	Body string

	*data.ClientMessage // TODO: remove me
}

// LogLevel is the current log level
type LogLevel int

// The different available log levels
const (
	Info LogLevel = iota
	Warn
	Alert
)

// Log contains information one specific log event
type Log struct {
	Level   LogLevel
	Message string
}

// FileTransfer represents an event associated with file transfers
type FileTransfer struct {
	Session access.Session
	Peer    jid.WithResource

	Mime             string
	DateLastModified string
	Name             string
	Size             int64
	Description      string
	IsDirectory      bool

	Answer  chan<- *string // one time use
	Control *sdata.FileTransferControl
}

// SMP is an event related to SMP
type SMP struct {
	Type     SMPType
	Session  access.Session
	From     jid.WithResource
	Resource string
	Body     string
}

// SMPType denotes the type of an SMP event
type SMPType int

// SMP types
const (
	SecretNeeded SMPType = iota
	Success
	Failure
)
