package gui

import (
	"bytes"
	"encoding/base64"
	"io/ioutil"
	"net/http"
	"os"
	"path"
	"sync"
	"time"
)

type _escLocalFS struct{}

var _escLocal _escLocalFS

type _escStaticFS struct{}

var _escStatic _escStaticFS

type _escDirectory struct {
	fs   http.FileSystem
	name string
}

type _escFile struct {
	compressed string
	size       int64
	modtime    int64
	local      string
	isDir      bool

	once sync.Once
	data []byte
	name string
}

func (_escLocalFS) Open(name string) (http.File, error) {
	f, present := _escData[path.Clean(name)]
	if !present {
		return nil, os.ErrNotExist
	}
	return os.Open(f.local)
}

func (_escStaticFS) prepare(name string) (*_escFile, error) {
	f, present := _escData[path.Clean(name)]
	if !present {
		return nil, os.ErrNotExist
	}
	var err error
	f.once.Do(func() {
		f.name = path.Base(name)
		if f.size == 0 {
			return
		}
		b64 := base64.NewDecoder(base64.StdEncoding, bytes.NewBufferString(f.compressed))
		f.data, err = ioutil.ReadAll(b64)
	})
	if err != nil {
		return nil, err
	}
	return f, nil
}

func (fs _escStaticFS) Open(name string) (http.File, error) {
	f, err := fs.prepare(name)
	if err != nil {
		return nil, err
	}
	return f.File()
}

func (dir _escDirectory) Open(name string) (http.File, error) {
	return dir.fs.Open(dir.name + name)
}

func (f *_escFile) File() (http.File, error) {
	type httpFile struct {
		*bytes.Reader
		*_escFile
	}
	return &httpFile{
		Reader:   bytes.NewReader(f.data),
		_escFile: f,
	}, nil
}

func (f *_escFile) Close() error {
	return nil
}

func (f *_escFile) Readdir(count int) ([]os.FileInfo, error) {
	return nil, nil
}

func (f *_escFile) Stat() (os.FileInfo, error) {
	return f, nil
}

func (f *_escFile) Name() string {
	return f.name
}

func (f *_escFile) Size() int64 {
	return f.size
}

func (f *_escFile) Mode() os.FileMode {
	return 0
}

func (f *_escFile) ModTime() time.Time {
	return time.Unix(f.modtime, 0)
}

func (f *_escFile) IsDir() bool {
	return f.isDir
}

func (f *_escFile) Sys() interface{} {
	return f
}

// FS returns a http.Filesystem for the embedded assets. If useLocal is true,
// the filesystem's contents are instead used.
func FS(useLocal bool) http.FileSystem {
	if useLocal {
		return _escLocal
	}
	return _escStatic
}

// Dir returns a http.Filesystem for the embedded assets on a given prefix dir.
// If useLocal is true, the filesystem's contents are instead used.
func Dir(useLocal bool, name string) http.FileSystem {
	if useLocal {
		return _escDirectory{fs: _escLocal, name: name}
	}
	return _escDirectory{fs: _escStatic, name: name}
}

// FSByte returns the named file from the embedded assets. If useLocal is
// true, the filesystem's contents are instead used.
func FSByte(useLocal bool, name string) ([]byte, error) {
	if useLocal {
		f, err := _escLocal.Open(name)
		if err != nil {
			return nil, err
		}
		b, err := ioutil.ReadAll(f)
		f.Close()
		return b, err
	}
	f, err := _escStatic.prepare(name)
	if err != nil {
		return nil, err
	}
	return f.data, nil
}

// FSMustByte is the same as FSByte, but panics if name is not present.
func FSMustByte(useLocal bool, name string) []byte {
	b, err := FSByte(useLocal, name)
	if err != nil {
		panic(err)
	}
	return b
}

// FSString is the string version of FSByte.
func FSString(useLocal bool, name string) (string, error) {
	b, err := FSByte(useLocal, name)
	return string(b), err
}

// FSMustString is the string version of FSMustByte.
func FSMustString(useLocal bool, name string) string {
	return string(FSMustByte(useLocal, name))
}

var _escData = map[string]*_escFile{

	"/definitions/AccountDetails.xml": {
		local:   "definitions/AccountDetails.xml",
		size:    34089,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtMaXN0U3RvcmUiIGlkPSJwcm94aWVzLW1vZGVs
Ij4KICAgIDxjb2x1bW5zPgogICAgICA8IS0tIHByb3h5IC0tPgogICAgICA8Y29sdW1uIHR5cGU9Imdj
aGFyYXJyYXkiLz4KICAgICAgPCEtLSByZWFsIHByb3h5IGRhdGEgLS0+CiAgICAgIDxjb2x1bW4gdHlw
ZT0iZ2NoYXJhcnJheSIvPgogICAgPC9jb2x1bW5zPgogIDwvb2JqZWN0PgogIDxvYmplY3QgY2xhc3M9
Ikd0a0xpc3RTdG9yZSIgaWQ9InBpbnMtbW9kZWwiPgogICAgPGNvbHVtbnM+CiAgICAgIDwhLS0gU3Vi
amVjdCAtLT4KICAgICAgPGNvbHVtbiB0eXBlPSJnY2hhcmFycmF5Ii8+CiAgICAgIDwhLS0gSXNzdWVy
IC0tPgogICAgICA8Y29sdW1uIHR5cGU9ImdjaGFyYXJyYXkiLz4KICAgICAgPCEtLSBGaW5nZXJwcmlu
dCAtLT4KICAgICAgPGNvbHVtbiB0eXBlPSJnY2hhcmFycmF5Ii8+CiAgICA8L2NvbHVtbnM+CiAgPC9v
YmplY3Q+CiAgPG9iamVjdCBjbGFzcz0iR3RrRGlhbG9nIiBpZD0iQWNjb3VudERldGFpbHMiPgogICAg
PHByb3BlcnR5IG5hbWU9InRpdGxlIiB0cmFuc2xhdGFibGU9InllcyI+QWNjb3VudCBEZXRhaWxzPC9w
cm9wZXJ0eT4KICAgIDxzaWduYWwgbmFtZT0iY2xvc2UiIGhhbmRsZXI9Im9uX2NhbmNlbF9zaWduYWwi
IC8+CiAgICA8Y2hpbGQgaW50ZXJuYWwtY2hpbGQ9InZib3giPgogICAgICA8b2JqZWN0IGNsYXNzPSJH
dGtCb3giIGlkPSJWYm94Ij4KICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luIj4xMDwvcHJvcGVy
dHk+CiAgICAgICAgPGNoaWxkPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTm90ZWJvb2siIGlk
PSJub3RlYm9vazEiPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+VHJ1ZTwvcHJv
cGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzaG93LWJvcmRlciI+RmFsc2U8L3Byb3Bl
cnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icGFnZSI+MDwvcHJvcGVydHk+CiAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tYm90dG9tIj4xMDwvcHJvcGVydHk+CiAgICAgICAgICAg
IDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCb3giPgogICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9ImNhbl9mb2N1cyI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9Im9yaWVudGF0aW9uIj52ZXJ0aWNhbDwvcHJvcGVydHk+CiAgICAgICAgICAg
ICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0JveCIgaWQ9Im5v
dGlmaWNhdGlvbi1hcmVhIj4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJs
ZSI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdp
bi10b3AiPjU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJn
aW4tbGVmdCI+NTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1h
cmdpbi1yaWdodCI+NTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
Im9yaWVudGF0aW9uIj5HVEtfT1JJRU5UQVRJT05fVkVSVElDQUw8L3Byb3BlcnR5PgogICAgICAgICAg
ICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAg
ICA8L3BhY2tpbmc+CiAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgICAgPGNoaWxk
PgogICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtHcmlkIiBpZD0iZ3JpZCI+CiAgICAg
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi10b3AiPjE1PC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLWJvdHRvbSI+MTA8L3Byb3BlcnR5
PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tc3RhcnQiPjEwPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLWVuZCI+MTA8L3By
b3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJyb3ctc3BhY2luZyI+MTI8
L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjb2x1bW4tc3BhY2lu
ZyI+NjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAg
ICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJBY2NvdW50TWVzc2FnZUxhYmVsIj4KICAg
ICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9Inll
cyI+WW91ciBhY2NvdW50JiN4QTsoZXhhbXBsZToga2ltNDJAamFiYmVyLm90ci5pbSk8L3Byb3BlcnR5
PgogICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ianVzdGlmeSI+R1RLX0pVU1RJ
RllfUklHSFQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAg
ICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9ImxlZnQtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9InRvcC1hdHRhY2giPjA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgPC9w
YWNraW5nPgogICAgICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgPGNo
aWxkPgogICAgICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrRW50cnkiIGlkPSJhY2Nv
dW50Ij4KICAgICAgICAgICAgICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJhY3RpdmF0ZSIgaGFuZGxl
cj0ib25fc2F2ZV9zaWduYWwiIC8+CiAgICAgICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAg
ICAgICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0ibGVmdC1hdHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0idG9wLWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICA8
L3BhY2tpbmc+CiAgICAgICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgICAgICA8
Y2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9IlBh
c3N3b3JkTGFiZWwiPgogICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwi
IHRyYW5zbGF0YWJsZT0ieWVzIj5QYXNzd29yZDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJoYWxpZ24iPkdUS19BTElHTl9FTkQ8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICAgICAgICA8cGFja2luZz4KICAg
ICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4wPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjE8L3By
b3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICAgICAgICAg
IDwvY2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICAgICAg
PG9iamVjdCBjbGFzcz0iR3RrRW50cnkiIGlkPSJwYXNzd29yZCI+CiAgICAgICAgICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmlsaXR5Ij5mYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAg
ICAgICAgICAgICAgIDxzaWduYWwgbmFtZT0iYWN0aXZhdGUiIGhhbmRsZXI9Im9uX3NhdmVfc2lnbmFs
IiAvPgogICAgICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICAgICAgICA8
cGFja2luZz4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNo
Ij4xPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1h
dHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAg
ICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAg
ICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9ImRpc3BsYXlOYW1lTGFiZWwiPgog
ICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0i
eWVzIj5BY2NvdW50IG5hbWUgKG9wdGlvbmFsKTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJoYWxpZ24iPkdUS19BTElHTl9FTkQ8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICAgICAgICA8cGFja2luZz4KICAg
ICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4wPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjI8L3By
b3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICAgICAgICAg
IDwvY2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICAgICAg
PG9iamVjdCBjbGFzcz0iR3RrRW50cnkiIGlkPSJkaXNwbGF5TmFtZSI+CiAgICAgICAgICAgICAgICAg
ICAgICAgIDxzaWduYWwgbmFtZT0iYWN0aXZhdGUiIGhhbmRsZXI9Im9uX3NhdmVfc2lnbmFsIiAvPgog
ICAgICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICAgICAgICA8cGFja2lu
Zz4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4xPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2gi
PjI8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICAg
ICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAg
ICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJzaG93T3RoZXJTZXR0aW5ncyI+CiAgICAg
ICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMi
PkRpc3BsYXkgYWxsIHNldHRpbmdzPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9Imp1c3RpZnkiPkdUS19KVVNUSUZZX1JJR0hUPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+R1RLX0FMSUdOX0VORDwvcHJvcGVy
dHk+CiAgICAgICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICAgICAgICAgIDxw
YWNraW5nPgogICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2gi
PjA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0
dGFjaCI+MzwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAg
ICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAg
ICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtDaGVja0J1dHRvbiIgaWQ9Im90aGVyU2V0dGluZ3Mi
PgogICAgICAgICAgICAgICAgICAgICAgICA8c2lnbmFsIG5hbWU9InRvZ2dsZWQiIGhhbmRsZXI9Im9u
X3RvZ2dsZV9vdGhlcl9zZXR0aW5ncyIgLz4KICAgICAgICAgICAgICAgICAgICAgIDwvb2JqZWN0Pgog
ICAgICAgICAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4zPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAg
ICAgIDwvcGFja2luZz4KICAgICAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgICAg
ICA8L29iamVjdD4KICAgICAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICA8L3Bh
Y2tpbmc+CiAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAg
ICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQgdHlwZT0idGFiIj4KICAgICAgICAgICAg
ICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9ImxhYmVsLXRhYjEiPgogICAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+QWNjb3VudDwvcHJvcGVydHk+
CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+VHJ1ZTwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9InRhYi1maWxsIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNr
aW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAg
PG9iamVjdCBjbGFzcz0iR3RrR3JpZCIgaWQ9Im90aGVyT3B0aW9uc0dyaWQiPgogICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi10b3AiPjE1PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tYm90dG9tIj4xMDwvcHJvcGVydHk+CiAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLXN0YXJ0Ij4xMDwvcHJvcGVydHk+CiAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLWVuZCI+MTA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9InJvdy1zcGFjaW5nIj4xMjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0iY29sdW1uLXNwYWNpbmciPjY8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAg
PGNoaWxkPgogICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9InNlcnZl
ckxhYmVsIj4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0
YWJsZT0ieWVzIj5TZXJ2ZXIgKGxlYXZlIGVtcHR5IGZvciBkZWZhdWx0KTwvcHJvcGVydHk+CiAgICAg
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imp1c3RpZnkiPkdUS19KVVNUSUZZX1JJR0hUPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICAgIDxwYWNr
aW5nPgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MDwvcHJv
cGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjA8L3By
b3BlcnR5PgogICAgICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgICAgICA8L2NoaWxk
PgogICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJH
dGtFbnRyeSIgaWQ9InNlcnZlciI+CiAgICAgICAgICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJhY3Rp
dmF0ZSIgaGFuZGxlcj0ib25fc2F2ZV9zaWduYWwiIC8+CiAgICAgICAgICAgICAgICAgIDwvb2JqZWN0
PgogICAgICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0ibGVmdC1hdHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJ0b3AtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgPC9wYWNraW5n
PgogICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAg
ICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJwb3J0TGFiZWwiPgogICAgICAgICAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPlBvcnQgKGxl
YXZlIGVtcHR5IGZvciBkZWZhdWx0KTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9ImhhbGlnbiI+R1RLX0FMSUdOX0VORDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAg
IDwvb2JqZWN0PgogICAgICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4xPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAg
PC9wYWNraW5nPgogICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgIDxjaGlsZD4K
ICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrRW50cnkiIGlkPSJwb3J0Ij4KICAgICAg
ICAgICAgICAgICAgICA8c2lnbmFsIG5hbWU9ImFjdGl2YXRlIiBoYW5kbGVyPSJvbl9zYXZlX3NpZ25h
bCIgLz4KICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICAgIDxwYWNraW5n
PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MTwvcHJvcGVy
dHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjE8L3Byb3Bl
cnR5PgogICAgICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgICAgICA8L2NoaWxkPgoK
ICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3Rr
TGFiZWwiIGlkPSJwaW5uaW5nUG9saWN5SW5zdHJ1Y3Rpb25zIj4KICAgICAgICAgICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5UaGUgcGlubmluZyBwb2xpY3kg
Z292ZXJucyB3aGV0aGVyIHdlIHdpbGwgY29uc2lkZXIgc2F2aW5nIGluZm9ybWF0aW9uIGFib3V0IGNl
cnRpZmljYXRlcyB3ZSBoYXZlIHNlZW4gYmVmb3JlLCBhbmQgaG93IHdlIHdpbGwgcmVhY3QgaW4gdGhl
c2UgY2FzZXMuIFRoZSBub25lIHBvbGljeSB0dXJucyBvZmYgdGhpcyBiZWhhdmlvci4gRGVueSB3aWxs
IHRyeSB0byB1c2UgdGhlIHBpbnMgd2UgYWxyZWFkeSBoYXZlLCBidXQgd2lsbCBuZXZlciBjb25uZWN0
IHRvIGEgc2VydmVyIHRoYXQgaGFzIGFub3RoZXIgY2VydGlmaWNhdGUuIFRoZSBhZGQgcG9saWN5IHdp
bGwgYWx3YXlzIGFkZCBuZXcgcGlucyAtIGJ5IGl0c2VsZiwgdGhpcyBpcyBub3Qgc28gdXNlZnVsLCBi
dXQgaWYgeW91IHR1cm4gaXQgb2ZmIG9yIGNoYW5nZSBpdCB0byBkZW55LCB5b3Ugd2lsbCBoYXZlIGEg
bGlzdCBvZiBwaW5zIHRoYXQgeW91IGNhbiBjdXJhdGUgbGF0ZXIuIFRoZSBhZGQgZmlyc3QgYW5kIGFz
ayBhZnRlciBwb2xpY3kgd2lsbCBwaW4gdGhlIGZpcnN0IGNlcnRpZmljYXRlIHdlIGV2ZXIgc2VlLCBh
bmQgdGhlbiBsZXQgdGhlIHVzZXIga25vdyBpZiB3ZSBlbmNvdW50ZXIgb3RoZXIgY2VydGlmYWNhdGVz
LiBUaGUgYWRkIGZpcnN0IGFuZCBkZW55IHBvbGljeSBkb2VzIHRoZSBzYW1lIHRoaW5nIGV4Y2VwdCBp
dCBkb2Vzbid0IGV2ZXIgYXNrIGFmdGVyIHRoZSBmaXJzdCBjZXJ0aWZpY2F0ZS4gRmluYWxseSwgdGhl
IGFzayBwb2xpY3kgd2lsbCBhbHdheXMgYXNrIHdoYXQgdG8gZG8gd2hlbiBzZWVpbmcgY2VydGlmaWNh
dGVzIHdlIGhhdmVuJ3QgYWRkZWQuPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0idmlzaWJsZSI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9IndyYXAiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJtYXgtd2lkdGgtY2hhcnMiPjUwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgPC9v
YmplY3Q+CiAgICAgICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjI8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJ3aWR0aCI+MjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgIDwvcGFja2lu
Zz4KICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAg
ICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0icGlubmluZ1BvbGljeUxhYmVsIj4K
ICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVz
Ij5QaW5uaW5nIHBvbGljeTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9ImhhbGlnbiI+R1RLX0FMSUdOX0VORDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgIDwvb2Jq
ZWN0PgogICAgICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0ibGVmdC1hdHRhY2giPjA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4zPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgPC9wYWNr
aW5nPgogICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAg
ICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQ29tYm9Cb3hUZXh0IiBpZD0icGlubmluZ1BvbGlj
eVZhbHVlIj4KICAgICAgICAgICAgICAgICAgICA8aXRlbXM+CiAgICAgICAgICAgICAgICAgICAgICA8
aXRlbSB0cmFuc2xhdGFibGU9InllcyIgaWQ9Im5vbmUiPk5vbmU8L2l0ZW0+CiAgICAgICAgICAgICAg
ICAgICAgICA8aXRlbSB0cmFuc2xhdGFibGU9InllcyIgaWQ9ImRlbnkiPkRlbnk8L2l0ZW0+CiAgICAg
ICAgICAgICAgICAgICAgICA8aXRlbSB0cmFuc2xhdGFibGU9InllcyIgaWQ9ImFkZCI+QWx3YXlzIGFk
ZDwvaXRlbT4KICAgICAgICAgICAgICAgICAgICAgIDxpdGVtIHRyYW5zbGF0YWJsZT0ieWVzIiBpZD0i
YWRkLWZpcnN0LWFzay1yZXN0Ij5BZGQgdGhlIGZpcnN0LCBhc2sgZm9yIHRoZSByZXN0PC9pdGVtPgog
ICAgICAgICAgICAgICAgICAgICAgPGl0ZW0gdHJhbnNsYXRhYmxlPSJ5ZXMiIGlkPSJhZGQtZmlyc3Qt
ZGVueS1yZXN0Ij5BZGQgdGhlIGZpcnN0LCBkZW55IHRoZSByZXN0PC9pdGVtPgogICAgICAgICAgICAg
ICAgICAgICAgPGl0ZW0gdHJhbnNsYXRhYmxlPSJ5ZXMiIGlkPSJhc2siPkFsd2F5cyBhc2s8L2l0ZW0+
CiAgICAgICAgICAgICAgICAgICAgPC9pdGVtcz4KICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAg
ICAgICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJsZWZ0LWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9InRvcC1hdHRhY2giPjM8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAg
ICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAg
ICAgICA8b2JqZWN0IGNsYXNzPSJHdGtQYW5lZCIgaWQ9ImhwYW5lZFBpbnMiPgogICAgICAgICAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuLWZvY3VzIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjE3NTwvcHJvcGVydHk+CiAgICAgICAg
ICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3Rr
U2Nyb2xsZWRXaW5kb3ciIGlkPSJzY3JvbGxlZHdpbmRvd1BpbnMiPgogICAgICAgICAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0iaHNjcm9sbGJhci1wb2xpY3kiPkdUS19QT0xJQ1lfQVVUT01BVElD
PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZzY3JvbGxi
YXItcG9saWN5Ij5HVEtfUE9MSUNZX0FVVE9NQVRJQzwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ3aWR0aC1yZXF1ZXN0Ij4xNzA8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGVpZ2h0LXJlcXVlc3QiPjE1MDwvcHJvcGVy
dHk+CiAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4iPjU8L3Byb3Bl
cnR5PgogICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+VHJ1ZTwv
cHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoZXhwYW5kIj5U
cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZleHBh
bmQiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
Y2FuLWZvY3VzIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9InNoYWRvdy10eXBlIj5pbjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAgIDxj
aGlsZD4KICAgICAgICAgICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtUcmVlVmlldyIg
aWQ9InBpbnMtdmlldyI+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
bW9kZWwiPnBpbnMtbW9kZWw8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9InZpc2libGUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbi1mb2N1cyI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAg
ICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGVhZGVycy12aXNpYmxlIj5UcnVlPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzaG93LWV4cGFu
ZGVycyI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9InJlb3JkZXJhYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAg
ICAgIDxjaGlsZCBpbnRlcm5hbC1jaGlsZD0ic2VsZWN0aW9uIj4KICAgICAgICAgICAgICAgICAgICAg
ICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrVHJlZVNlbGVjdGlvbiIgaWQ9InBpbnMtc2VsZWN0aW9u
Ij4KICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibW9kZSI+R1RL
X1NFTEVDVElPTl9TSU5HTEU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgICAgICA8
L29iamVjdD4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAg
ICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxvYmplY3Qg
Y2xhc3M9Ikd0a1RyZWVWaWV3Q29sdW1uIiBpZD0icGlucy1zdWJqZWN0LWNvbHVtbiI+CiAgICAgICAg
ICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRpdGxlIj5TdWJqZWN0PC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAg
ICAgICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtDZWxsUmVuZGVyZXJUZXh0IiBpZD0i
cGlucy1zdWJqZWN0LWNvbHVtbi1yZW5kZXJlZCIvPgogICAgICAgICAgICAgICAgICAgICAgICAgICAg
ICAgICAgPGF0dHJpYnV0ZXM+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxhdHRy
aWJ1dGUgbmFtZT0idGV4dCI+MDwvYXR0cmlidXRlPgogICAgICAgICAgICAgICAgICAgICAgICAgICAg
ICAgICAgPC9hdHRyaWJ1dGVzPgogICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgIDwvY2hpbGQ+
CiAgICAgICAgICAgICAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgICAgICAg
ICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAg
ICAgICAgICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrVHJlZVZpZXdDb2x1bW4iIGlk
PSJwaW5zLWlzc3Vlci1jb2x1bW4iPgogICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJ0aXRsZSI+SXNzdWVyPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAg
ICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICA8b2JqZWN0IGNs
YXNzPSJHdGtDZWxsUmVuZGVyZXJUZXh0IiBpZD0icGlucy1pc3N1ZXItY29sdW1uLXJlbmRlcmVkIi8+
CiAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICA8YXR0cmlidXRlcz4KICAgICAgICAgICAg
ICAgICAgICAgICAgICAgICAgICAgICAgPGF0dHJpYnV0ZSBuYW1lPSJ0ZXh0Ij4xPC9hdHRyaWJ1dGU+
CiAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICA8L2F0dHJpYnV0ZXM+CiAgICAgICAgICAg
ICAgICAgICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgICAgICAgICAgICAgICAg
PC9vYmplY3Q+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAg
ICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICAgICAgICAgICAgICA8b2JqZWN0
IGNsYXNzPSJHdGtUcmVlVmlld0NvbHVtbiIgaWQ9InBpbnMtZnByLWNvbHVtbiI+CiAgICAgICAgICAg
ICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRpdGxlIj5GaW5nZXJwcmludDwvcHJv
cGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAg
ICAgICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQ2VsbFJlbmRlcmVyVGV4dCIgaWQ9
InBpbnMtZnByLWNvbHVtbi1yZW5kZXJlZCIvPgogICAgICAgICAgICAgICAgICAgICAgICAgICAgICAg
ICAgPGF0dHJpYnV0ZXM+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxhdHRyaWJ1
dGUgbmFtZT0idGV4dCI+MjwvYXR0cmlidXRlPgogICAgICAgICAgICAgICAgICAgICAgICAgICAgICAg
ICAgPC9hdHRyaWJ1dGVzPgogICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAg
ICAgICAgICAgICAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgICAgICAgICAg
ICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAg
ICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAg
ICAgICAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJyZXNpemUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0ic2hyaW5rIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICA8
L3BhY2tpbmc+CiAgICAgICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgICAgICA8
Y2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCb3giIGlkPSJ2Ym94
LXBpbmJ1dHRvbnMiPgogICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2lu
Ij41PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2li
bGUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
Y2FuLWZvY3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJvcmllbnRhdGlvbiI+dmVydGljYWw8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0ic3BhY2luZyI+NjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAg
ICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtC
dXR0b24iIGlkPSJyZW1vdmVfcGluX2J1dHRvbiI+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5fUmVtb3ZlPC9wcm9wZXJ0eT4K
ICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW4tZm9j
dXMiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9InJlY2VpdmVzLWRlZmF1bHQiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9InVzZV91bmRlcmxpbmUiPlRydWU8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgICAgICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJjbGlja2VkIiBoYW5kbGVyPSJvbl9yZW1v
dmVfcGluX3NpZ25hbCIgLz4KICAgICAgICAgICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAg
ICAgICAgICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJleHBhbmQiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAg
ICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MDwvcHJvcGVydHk+CiAgICAgICAg
ICAgICAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICAgICAgICAgICAgICA8L2NoaWxk
PgogICAgICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICAgICAgICA8cGFj
a2luZz4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InJlc2l6ZSI+RmFsc2U8
L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ic2hyaW5rIj5G
YWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAg
ICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAg
ICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFj
aCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRh
Y2giPjQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ3aWR0aCI+
MjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgICAgIDwv
Y2hpbGQ+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAg
ICAgIDxjaGlsZCB0eXBlPSJ0YWIiPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVs
IiBpZD0ibGFiZWwtdGFiMiI+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRy
YW5zbGF0YWJsZT0ieWVzIj5TZXJ2ZXI8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9InZpc2libGUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAg
ICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InBvc2l0aW9u
Ij4xPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0YWItZmlsbCI+RmFs
c2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgPC9jaGlsZD4K
ICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0JveCIgaWQ9
InZib3gxIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4iPjU8L3Byb3BlcnR5
PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPlRydWU8L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbi1mb2N1cyI+RmFsc2U8L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im9yaWVudGF0aW9uIj52ZXJ0aWNhbDwvcHJvcGVy
dHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ic3BhY2luZyI+NjwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a1Bh
bmVkIiBpZD0iaHBhbmVkMSI+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2li
bGUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW4t
Zm9jdXMiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJw
b3NpdGlvbiI+MTc1PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAg
ICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtTY3JvbGxlZFdpbmRvdyIgaWQ9InNjcm9sbGVk
d2luZG93MSI+CiAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoc2Nyb2xsYmFy
LXBvbGljeSI+R1RLX1BPTElDWV9ORVZFUjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJ2c2Nyb2xsYmFyLXBvbGljeSI+R1RLX1BPTElDWV9BVVRPTUFUSUM8L3By
b3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0id2lkdGgtcmVxdWVz
dCI+MTcwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imhl
aWdodC1yZXF1ZXN0Ij4yMzA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0ibWFyZ2luIj41PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9InZpc2libGUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0iaGV4cGFuZCI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJ2ZXhwYW5kIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAg
ICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbi1mb2N1cyI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAg
ICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzaGFkb3ctdHlwZSI+aW48L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgICAgICAgPG9i
amVjdCBjbGFzcz0iR3RrVHJlZVZpZXciIGlkPSJwcm94aWVzLXZpZXciPgogICAgICAgICAgICAgICAg
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1vZGVsIj5wcm94aWVzLW1vZGVsPC9wcm9wZXJ0eT4K
ICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW4tZm9j
dXMiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9ImhlYWRlcnMtdmlzaWJsZSI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9InNob3ctZXhwYW5kZXJzIj5GYWxzZTwvcHJvcGVydHk+CiAgICAg
ICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icmVvcmRlcmFibGUiPlRydWU8L3By
b3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJyb3ctYWN0aXZh
dGVkIiBoYW5kbGVyPSJvbl9lZGl0X2FjdGl2YXRlX3Byb3h5X3NpZ25hbCIgLz4KICAgICAgICAgICAg
ICAgICAgICAgICAgICAgIDxjaGlsZCBpbnRlcm5hbC1jaGlsZD0ic2VsZWN0aW9uIj4KICAgICAgICAg
ICAgICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrVHJlZVNlbGVjdGlvbiIgaWQ9InNl
bGVjdGlvbiI+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1v
ZGUiPkdUS19TRUxFQ1RJT05fU0lOR0xFPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAg
ICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAg
ICAgICAgICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICAgICAgICAgICAgICA8
b2JqZWN0IGNsYXNzPSJHdGtUcmVlVmlld0NvbHVtbiIgaWQ9InByb3h5LW5hbWUtY29sdW1uIj4KICAg
ICAgICAgICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idGl0bGUiPnByb3h5LW5h
bWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAg
ICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0NlbGxSZW5kZXJlclRl
eHQiIGlkPSJwcm94eS1uYW1lLWNvbHVtbi1yZW5kZXJlZCIvPgogICAgICAgICAgICAgICAgICAgICAg
ICAgICAgICAgICAgPGF0dHJpYnV0ZXM+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICAg
IDxhdHRyaWJ1dGUgbmFtZT0idGV4dCI+MDwvYXR0cmlidXRlPgogICAgICAgICAgICAgICAgICAgICAg
ICAgICAgICAgICAgPC9hdHRyaWJ1dGVzPgogICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgIDwv
Y2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAg
ICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgICAgICAgICAgICA8L29iamVjdD4K
ICAgICAgICAgICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgICAgICAgIDwvb2Jq
ZWN0PgogICAgICAgICAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJyZXNpemUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0ic2hyaW5rIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAg
ICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAg
ICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCb3gi
IGlkPSJ2Ym94MyI+CiAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4i
PjU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJs
ZSI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJj
YW4tZm9jdXMiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9Im9yaWVudGF0aW9uIj52ZXJ0aWNhbDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJzcGFjaW5nIj42PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAg
ICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0J1
dHRvbiIgaWQ9ImFkZF9idXR0b24iPgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+X0FkZC4uLjwvcHJvcGVydHk+CiAgICAgICAg
ICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+VHJ1ZTwvcHJvcGVydHk+
CiAgICAgICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuLWZvY3VzIj5UcnVl
PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJyZWNl
aXZlcy1kZWZhdWx0Ij5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJ1c2VfdW5kZXJsaW5lIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAg
ICAgICAgICAgICAgIDxzaWduYWwgbmFtZT0iY2xpY2tlZCIgaGFuZGxlcj0ib25fYWRkX3Byb3h5X3Np
Z25hbCIgLz4KICAgICAgICAgICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICAg
ICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJleHBhbmQiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJmaWxsIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAg
ICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAg
ICAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgICAgICAgPG9iamVjdCBj
bGFzcz0iR3RrQnV0dG9uIiBpZD0icmVtb3ZlX2J1dHRvbiI+CiAgICAgICAgICAgICAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5fUmVtb3ZlLi4uPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxl
Ij5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJjYW4tZm9jdXMiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9InJlY2VpdmVzLWRlZmF1bHQiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAg
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InVzZV91bmRlcmxpbmUiPlRydWU8L3Byb3BlcnR5
PgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJjbGlja2VkIiBoYW5kbGVy
PSJvbl9yZW1vdmVfcHJveHlfc2lnbmFsIiAvPgogICAgICAgICAgICAgICAgICAgICAgICAgIDwvb2Jq
ZWN0PgogICAgICAgICAgICAgICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgICAg
ICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAg
ICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImZpbGwiPlRydWU8L3Byb3BlcnR5PgogICAg
ICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InBvc2l0aW9uIj4xPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgICAgICAgICAg
ICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICAg
ICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCdXR0b24iIGlkPSJlZGl0X2J1dHRvbiI+CiAgICAg
ICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0i
eWVzIj5fRWRpdC4uLjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0idmlzaWJsZSI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0iY2FuLWZvY3VzIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAg
ICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJyZWNlaXZlcy1kZWZhdWx0Ij5UcnVlPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ1c2UtdW5kZXJsaW5l
Ij5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxzaWduYWwgbmFtZT0i
Y2xpY2tlZCIgaGFuZGxlcj0ib25fZWRpdF9wcm94eV9zaWduYWwiIC8+CiAgICAgICAgICAgICAgICAg
ICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAg
ICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZXhwYW5kIj5GYWxzZTwvcHJvcGVy
dHk+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZmlsbCI+VHJ1ZTwv
cHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icG9zaXRp
b24iPjI8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAg
ICAgICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgICAgICAgIDwvb2JqZWN0Pgog
ICAgICAgICAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJyZXNpemUiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9InNocmluayI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAg
ICAgPC9wYWNraW5nPgogICAgICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgICAg
IDwvb2JqZWN0PgogICAgICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0iZXhwYW5kIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0iZmlsbCI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9InBvc2l0aW9uIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgPC9wYWNr
aW5nPgogICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAg
ICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNoaWxkIHR5cGU9InRhYiI+CiAgICAgICAgICAgICAg
PG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJsYWJlbC10YWIzIj4KICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPlByb3hpZXM8L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPlRydWU8L3Byb3BlcnR5PgogICAg
ICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9InBvc2l0aW9uIj4yPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJ0YWItZmlsbCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvcGFja2lu
Zz4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxv
YmplY3QgY2xhc3M9Ikd0a0JveCIgaWQ9ImVuY3J5cHRpb25PcHRpb25zQm94Ij4KICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJib3JkZXItd2lkdGgiPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJob21vZ2VuZW91cyI+ZmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9Im9yaWVudGF0aW9uIj5HVEtfT1JJRU5UQVRJT05fVkVSVElDQUw8
L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICA8b2JqZWN0
IGNsYXNzPSJHdGtMYWJlbCIgaWQ9ImZpbmdlcnByaW50c01lc3NhZ2UiPgogICAgICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJqdXN0aWZ5Ij5HVEtfSlVTVElGWV9MRUZUPC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05fU1RBUlQ8L3By
b3BlcnR5PgogICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICAgIDwvY2hpbGQ+
CiAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0
a0dyaWQiIGlkPSJlbmNyeXB0aW9uR3JpZCI+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9Im1hcmdpbi1ib3R0b20iPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0ibWFyZ2luLXN0YXJ0Ij4xMDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9Im1hcmdpbi1lbmQiPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0icm93LXNwYWNpbmciPjEyPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0iY29sdW1uLXNwYWNpbmciPjY8L3Byb3BlcnR5PgogICAgICAgICAgICAg
ICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVs
IiBpZD0iZW5jcnlwdGlvbkltcG9ydEluc3RydWN0aW9ucyI+CiAgICAgICAgICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPlRoZSBiZWxvdyBidXR0b25z
IGFsbG93IHlvdSB0byBpbXBvcnQgcHJpdmF0ZSBrZXlzIGFuZCBmaW5nZXJwcmludHMuIEJvdGggb2Yg
dGhlbSBzaG91bGQgYmUgaW4gdGhlIFBpZGdpbi9saWJvdHIgZm9ybWF0LiBJZiB5b3UgaW1wb3J0IHBy
aXZhdGUga2V5cywgeW91ciBleGlzdGluZyBwcml2YXRlIGtleXMgd2lsbCBiZSBkZWxldGVkLCBzaW5j
ZSBjdXJyZW50bHkgdGhlcmUgaXMgbm8gd2F5IHRvIGNob29zZSB3aGljaCBrZXkgdG8gdXNlIGZvciBl
bmNyeXB0ZWQgY2hhdC4KClRoZXJlIGFyZSBzZXZlcmFsIGFwcGxpY2F0aW9ucyB0aGF0IHVzZSB0aGUg
bGlib3RyIGZvcm1hdCwgc3VjaCBhcyBQaWRnaW4sIEFkaXVtIG9yIFRvciBNZXNzZW5nZXIuIERlcGVu
ZGluZyBvbiB5b3VyIHBsYXRmb3JtLCB0aGVzZSBmaWxlcyBjYW4gYmUgZm91bmQgaW4gc2V2ZXJhbCBk
aWZmZXJlbnQgcGxhY2VzLiBSZWZlciB0byB0aGUgZG9jdW1lbnRhdGlvbiBmb3IgdGhlIGFwcGxpY2F0
aW9uIGluIHF1ZXN0aW9uIHRvIGZpbmQgb3V0IHdoZXJlIHRoZSBmaWxlcyBhcmUgbG9jYXRlZCBmb3Ig
eW91ciBwbGF0Zm9ybS4gVGhlIGZpbGVuYW1lcyB0byBsb29rIGZvciBhcmUgIm90ci5maW5nZXJwcmlu
dHMiIGFuZCAib3RyLnByaXZhdGVfa2V5Ii48L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ3cmFwIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAg
ICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1heC13aWR0aC1jaGFycyI+NTA8L3Byb3BlcnR5PgogICAg
ICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICAgICAgICA8cGFja2luZz4K
ICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4wPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjA8
L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0id2lkdGgiPjI8
L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICAgICAg
ICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICAg
ICAgPG9iamVjdCBjbGFzcz0iR3RrQnV0dG9uIiBpZD0iaW1wb3J0X2tleV9idXR0b24iPgogICAgICAg
ICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5J
bXBvcnQgUHJpdmF0ZSBfS2V5cy4uLjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9ImNhbi1mb2N1cyI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJyZWNlaXZlcy1kZWZhdWx0Ij5UcnVlPC9wcm9wZXJ0eT4K
ICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InVzZV91bmRlcmxpbmUiPlRydWU8
L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICA8c2lnbmFsIG5hbWU9ImNsaWNrZWQiIGhh
bmRsZXI9Im9uX2ltcG9ydF9rZXlfc2lnbmFsIiAvPgogICAgICAgICAgICAgICAgICAgICAgPC9vYmpl
Y3Q+CiAgICAgICAgICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAg
ICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAg
ICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQnV0
dG9uIiBpZD0iaW1wb3J0X2Zwcl9idXR0b24iPgogICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5JbXBvcnQgX0ZpbmdlcnByaW50cy4uLjwv
cHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5U
cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbi1m
b2N1cyI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJyZWNlaXZlcy1kZWZhdWx0Ij5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9InVzZV91bmRlcmxpbmUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAg
ICAgICAgICAgICA8c2lnbmFsIG5hbWU9ImNsaWNrZWQiIGhhbmRsZXI9Im9uX2ltcG9ydF9mcHJfc2ln
bmFsIiAvPgogICAgICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICAgICAg
ICA8cGFja2luZz4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0
YWNoIj4xPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRv
cC1hdHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAg
ICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAg
ICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJlbmNyeXB0aW9uRXhwb3J0
SW5zdHJ1Y3Rpb25zIj4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVs
IiB0cmFuc2xhdGFibGU9InllcyI+VGhlIGJlbG93IGJ1dHRvbnMgYWxsb3cgeW91IHRvIGV4cG9ydCBw
cml2YXRlIGtleXMgYW5kIGZpbmdlcnByaW50cy4gQmUgY2FyZWZ1bCB3aXRoIHRoZSBmaWxlcyB0aGF0
IGNvbWUgb3V0IG9mIHRoaXMgcHJvY2VzcyBhcyB0aGV5IGNvbnRhaW4gcG90ZW50aWFsbHkgc2Vuc2l0
aXZlIGRhdGEuIFRoZSBleHBvcnQgd2lsbCBiZSBpbiB0aGUgUGlkZ2luL2xpYm90ciBmb3JtYXQuPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPnRy
dWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0id3JhcCI+
dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXgt
d2lkdGgtY2hhcnMiPjUwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgIDwvb2JqZWN0Pgog
ICAgICAgICAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4yPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9IndpZHRoIj4yPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAg
ICAgIDwvcGFja2luZz4KICAgICAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgICAg
ICAgIDxjaGlsZD4KICAgICAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0J1dHRvbiIg
aWQ9ImV4cG9ydF9rZXlfYnV0dG9uIj4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+RXhwb3J0IFByaXZhdGUgS2V5cy4uLjwvcHJvcGVy
dHk+CiAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbi1mb2N1cyI+
VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJyZWNl
aXZlcy1kZWZhdWx0Ij5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgPHNpZ25h
bCBuYW1lPSJjbGlja2VkIiBoYW5kbGVyPSJvbl9leHBvcnRfa2V5X3NpZ25hbCIgLz4KICAgICAgICAg
ICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAg
ICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MDwvcHJvcGVydHk+
CiAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4zPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgICAgICAgICA8
L2NoaWxkPgogICAgICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICAgICAgICAgIDxv
YmplY3QgY2xhc3M9Ikd0a0J1dHRvbiIgaWQ9ImV4cG9ydF9mcHJfYnV0dG9uIj4KICAgICAgICAgICAg
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+RXhwb3J0
IEZpbmdlcnByaW50cy4uLjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9ImNhbi1mb2N1cyI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJyZWNlaXZlcy1kZWZhdWx0Ij5UcnVlPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJjbGlja2VkIiBoYW5kbGVyPSJvbl9leHBvcnRf
ZnByX3NpZ25hbCIgLz4KICAgICAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAg
ICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJs
ZWZ0LWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJ0b3AtYXR0YWNoIj4zPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgIDwvcGFja2lu
Zz4KICAgICAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgICAgICA8L29iamVjdD4K
ICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAg
IDwvY2hpbGQ+CiAgICAgICAgICAgIDxjaGlsZCB0eXBlPSJ0YWIiPgogICAgICAgICAgICAgIDxvYmpl
Y3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0ibGFiZWwtdGFiNCI+CiAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5FbmNyeXB0aW9uPC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MzwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0idGFiLWZpbGwiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+
CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICA8L2NoaWxkPgog
ICAgICAgIDxjaGlsZCBpbnRlcm5hbC1jaGlsZD0iYWN0aW9uX2FyZWEiPgogICAgICAgICAgPG9iamVj
dCBjbGFzcz0iR3RrQnV0dG9uQm94IiBpZD0iYnV0dG9uX2JveCI+CiAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJvcmllbnRhdGlvbiI+R1RLX09SSUVOVEFUSU9OX0hPUklaT05UQUw8L3Byb3BlcnR5Pgog
ICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQnV0dG9uIiBp
ZD0iY2FuY2VsIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRh
YmxlPSJ5ZXMiPkNhbmNlbDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8c2lnbmFsIG5hbWU9ImNs
aWNrZWQiIGhhbmRsZXI9Im9uX2NhbmNlbF9zaWduYWwiLz4KICAgICAgICAgICAgICA8L29iamVjdD4K
ICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmpl
Y3QgY2xhc3M9Ikd0a0J1dHRvbiIgaWQ9InNhdmUiPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+U2F2ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0iY2FuLWRlZmF1bHQiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAg
ICAgPHNpZ25hbCBuYW1lPSJjbGlja2VkIiBoYW5kbGVyPSJvbl9zYXZlX3NpZ25hbCIvPgogICAgICAg
ICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgPC9vYmplY3Q+CiAg
ICAgICAgPC9jaGlsZD4KICAgICAgPC9vYmplY3Q+CiAgICA8L2NoaWxkPgogIDwvb2JqZWN0Pgo8L2lu
dGVyZmFjZT4K
`,
	},

	"/definitions/AccountRegistration.xml": {
		local:   "definitions/AccountRegistration.xml",
		size:    3248,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtBc3Npc3RhbnQiIGlkPSJhc3Npc3RhbnQiPgog
ICAgPHNpZ25hbCBuYW1lPSJjbG9zZSIgaGFuZGxlcj0ib25fY2FuY2VsX3NpZ25hbCIgLz4KICAgIDxz
aWduYWwgbmFtZT0iY2FuY2VsIiBoYW5kbGVyPSJvbl9jYW5jZWxfc2lnbmFsIiAvPgogICAgPHNpZ25h
bCBuYW1lPSJwcmVwYXJlIiBoYW5kbGVyPSJvbl9wcmVwYXJlIiAvPgogICAgPGNoaWxkPgogICAgICA8
b2JqZWN0IGNsYXNzPSJHdGtCb3giIGlkPSJkaWFsb2ciPgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJt
YXJnaW4iPjEwPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0ic3BhY2luZyI+MTA8L3By
b3BlcnR5PgoJPGNoaWxkPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrR3JpZCI+CiAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tYm90dG9tIj4xMDwvcHJvcGVydHk+CiAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJyb3ctc3BhY2luZyI+MTI8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0iY29sdW1uLXNwYWNpbmciPjY8L3Byb3BlcnR5PgoJICAgIDxjaGlsZD4KICAgICAg
ICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9InNlcnZlci1sYWJlbCI+CiAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5SZWNvbW1lbmRl
ZCBzZXJ2ZXJzOiA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imp1c3Rp
ZnkiPkdUS19KVVNUSUZZX1JJR0hUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L29iamVjdD4KICAg
ICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0
dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFj
aCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxk
PgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQ29tYm9C
b3hUZXh0IiBpZD0ic2VydmVyIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoYXMtZW50
cnkiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgIDxw
YWNraW5nPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4xPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4xPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgIDwvY2hpbGQ+CgkgIDwvb2JqZWN0
PgogICAgICAgIDwvY2hpbGQ+CiAgICAgIDwvb2JqZWN0PgogICAgICA8cGFja2luZz4KICAgICAgICA8
cHJvcGVydHkgbmFtZT0idGl0bGUiIHRyYW5zbGF0YWJsZT0ieWVzIj4xLiBDaG9vc2UgYSBzZXJ2ZXI8
L3Byb3BlcnR5PgogICAgICAgICA8cHJvcGVydHkgbmFtZT0icGFnZS10eXBlIj5HVEtfQVNTSVNUQU5U
X1BBR0VfSU5UUk88L3Byb3BlcnR5PgoJPHByb3BlcnR5IG5hbWU9ImNvbXBsZXRlIj50cnVlPC9wcm9w
ZXJ0eT4KICAgICAgPC9wYWNraW5nPgogICAgPC9jaGlsZD4KICAgIDxjaGlsZD4KICAgICAgPG9iamVj
dCBjbGFzcz0iR3RrQm94IiBpZD0iZm9ybSI+CiAgICAgICAgPGNoaWxkPgogICAgICAgICAgPG9iamVj
dCBjbGFzcz0iR3RrSW1hZ2UiIGlkPSJmb3JtSW1hZ2UiPgogICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0ibWFyZ2luX3JpZ2h0Ij4xNTwvcHJvcGVydHk+CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICA8
L2NoaWxkPgoJPGNoaWxkPgoJICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9ImZvcm1NZXNzYWdl
Ij4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi10b3AiPjIwPC9wcm9wZXJ0eT4KCSAg
PC9vYmplY3Q+CiAgICAgICAgPC9jaGlsZD4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0
IGNsYXNzPSJHdGtTcGlubmVyIiBpZD0ic3Bpbm5lciI+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdp
bi10b3AiPjQwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbl9mb2N1cyI+
RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iYWN0aXZlIj5UcnVlPC9w
cm9wZXJ0eT4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgIDwvY2hpbGQ+CiAgICAgICAgPGNoaWxk
PgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrR3JpZCIgaWQ9ImZvcm1HcmlkIj4KICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi1ib3R0b20iPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9InJvdy1zcGFjaW5nIj4xMjwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJjb2x1bW4tc3BhY2luZyI+NjwvcHJvcGVydHk+CiAgICAgICAgICA8L29iamVjdD4K
ICAgICAgICA8L2NoaWxkPgogICAgICA8L29iamVjdD4KICAgICAgPHBhY2tpbmc+CiAgICAgICAgPHBy
b3BlcnR5IG5hbWU9InRpdGxlIiB0cmFuc2xhdGFibGU9InllcyI+Mi4gQ3JlYXRlIGFjY291bnQ8L3By
b3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwYWdlLXR5cGUiPkdUS19BU1NJU1RBTlRfUEFH
RV9DT05GSVJNPC9wcm9wZXJ0eT4KICAgICAgPC9wYWNraW5nPgogICAgPC9jaGlsZD4KICAgIDxjaGls
ZD4KICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQm94Ij4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8
b2JqZWN0IGNsYXNzPSJHdGtJbWFnZSIgaWQ9ImRvbmVJbWFnZSI+CiAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJtYXJnaW5fcmlnaHQiPjE1PC9wcm9wZXJ0eT4KICAgICAgICAgIDwvb2JqZWN0PgoJPC9j
aGlsZD4KCTxjaGlsZD4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0iZG9uZU1l
c3NhZ2UiIC8+CiAgICAgICAgPC9jaGlsZD4KICAgICAgPC9vYmplY3Q+CiAgICAgIDxwYWNraW5nPgog
ICAgICAgICA8cHJvcGVydHkgbmFtZT0icGFnZS10eXBlIj5HVEtfQVNTSVNUQU5UX1BBR0VfU1VNTUFS
WTwvcHJvcGVydHk+CiAgICAgIDwvcGFja2luZz4KICAgIDwvY2hpbGQ+CiAgPC9vYmplY3Q+CjwvaW50
ZXJmYWNlPgo=
`,
	},

	"/definitions/AddChat.xml": {
		local:   "definitions/AddChat.xml",
		size:    7900,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtMaXN0U3RvcmUiIGlkPSJhY2NvdW50cy1tb2Rl
bCI+CiAgICA8Y29sdW1ucz4KICAgICAgPCEtLSBhY2NvdW50IG5hbWUgLS0+CiAgICAgIDxjb2x1bW4g
dHlwZT0iZ2NoYXJhcnJheSIvPgogICAgICA8IS0tIGFjY291bnQgaWQgLS0+CiAgICAgIDxjb2x1bW4g
dHlwZT0iZ2NoYXJhcnJheSIvPgogICAgPC9jb2x1bW5zPgogIDwvb2JqZWN0PgogIDxvYmplY3QgY2xh
c3M9Ikd0a0RpYWxvZyIgaWQ9ImFkZC1jaGF0LWRpYWxvZyI+CiAgICA8cHJvcGVydHkgbmFtZT0idmlz
aWJsZSI+ZmFsc2U8L3Byb3BlcnR5PgogICAgPHByb3BlcnR5IG5hbWU9IndpbmRvdy1wb3NpdGlvbiI+
R1RLX1dJTl9QT1NfQ0VOVEVSPC9wcm9wZXJ0eT4KICAgIDxwcm9wZXJ0eSBuYW1lPSJib3JkZXJfd2lk
dGgiPjc8L3Byb3BlcnR5PgogICAgPHByb3BlcnR5IG5hbWU9InRpdGxlIiB0cmFuc2xhdGFibGU9Inll
cyI+Sm9pbiBjaGF0IHJvb208L3Byb3BlcnR5PgogICAgPHByb3BlcnR5IG5hbWU9InJlc2l6YWJsZSI+
VHJ1ZTwvcHJvcGVydHk+CiAgICA8cHJvcGVydHkgbmFtZT0iZGVzdHJveS13aXRoLXBhcmVudCI+dHJ1
ZTwvcHJvcGVydHk+CiAgICA8Y2hpbGQgaW50ZXJuYWwtY2hpbGQ9InZib3giPgogICAgICA8b2JqZWN0
IGNsYXNzPSJHdGtCb3giIGlkPSJWYm94Ij4KICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+
dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhvbW9nZW5lb3VzIj5mYWxzZTwv
cHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9Im9yaWVudGF0aW9uIj5HVEtfT1JJRU5UQVRJ
T05fVkVSVElDQUw8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzcGFjaW5nIj42PC9w
cm9wZXJ0eT4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8IS0tVE9ETzogdGhpcyBzZWVtcyB0byBy
ZW5kZXIgd2VpcmRseSAtLT4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0JveCIgaWQ9Im5vdGlm
aWNhdGlvbi1hcmVhIj4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPmZhbHNlPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi10b3AiPjU8L3Byb3BlcnR5
PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLWxlZnQiPjU8L3Byb3BlcnR5PgogICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLXJpZ2h0Ij41PC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9Im9yaWVudGF0aW9uIj5HVEtfT1JJRU5UQVRJT05fVkVSVElDQUw8L3By
b3BlcnR5PgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0iZmlsbCI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJw
b3NpdGlvbiI+MDwvcHJvcGVydHk+CiAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgPC9jaGlsZD4K
ICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtHcmlkIiBpZD0iZm9ybSI+
CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj50cnVlPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi10b3AiPjE1PC9wcm9wZXJ0eT4KICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9Im1hcmdpbi1ib3R0b20iPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9Im1hcmdpbi1zdGFydCI+MTA8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0ibWFyZ2luLWVuZCI+MTA8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0icm93LXNwYWNpbmciPjEyPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNv
bHVtbi1zcGFjaW5nIj42PC9wcm9wZXJ0eT4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAg
IDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0iYWNjb3VudHMtbGFiZWwiID4KCQk8cHJvcGVydHkg
bmFtZT0idmlzaWJsZSI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5BY2NvdW50OjwvcHJvcGVydHk+CiAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0ianVzdGlmeSI+R1RLX0pVU1RJRllfUklHSFQ8L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+R1RLX0FMSUdOX0VORDwvcHJvcGVy
dHk+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjA8L3Byb3BlcnR5PgogICAgICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjA8L3Byb3BlcnR5PgogICAgICAgICAgICAg
IDwvcGFja2luZz4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAg
ICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0NvbWJvQm94IiBpZD0iYWNjb3VudHMiPgoJCTxwcm9wZXJ0
eSBuYW1lPSJ2aXNpYmxlIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJtb2RlbCI+YWNjb3VudHMtbW9kZWw8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9Imhhcy1mb2N1cyI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0iaGV4cGFuZCI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAg
ICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0NlbGxSZW5kZXJlclRleHQiIGlkPSJhY2Nv
dW50LW5hbWUtcmVuZGVyZWQiLz4KICAgICAgICAgICAgICAgICAgPGF0dHJpYnV0ZXM+CiAgICAgICAg
ICAgICAgICAgICAgPGF0dHJpYnV0ZSBuYW1lPSJ0ZXh0Ij4wPC9hdHRyaWJ1dGU+CiAgICAgICAgICAg
ICAgICAgIDwvYXR0cmlidXRlcz4KICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAg
PC9vYmplY3Q+CiAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0ibGVmdC1hdHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9InRvcC1hdHRhY2giPjA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
IndpZHRoIj4zPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgIDwv
Y2hpbGQ+CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtM
YWJlbCIgaWQ9InJvb20tbGFiZWwiID4KCQk8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+dHJ1ZTwvcHJv
cGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0i
eWVzIj5Sb29tOjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ianVzdGlm
eSI+R1RLX0pVU1RJRllfUklHSFQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9ImhhbGlnbiI+R1RLX0FMSUdOX0VORDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9vYmplY3Q+
CiAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVm
dC1hdHRhY2giPjA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1h
dHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgPC9j
aGlsZD4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0Vu
dHJ5IiBpZD0icm9vbSI+CgkJPHByb3BlcnR5IG5hbWU9InZpc2libGUiPnRydWU8L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhleHBhbmQiPlRydWU8L3Byb3BlcnR5PgogICAg
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRleHQiPmNveWltLWRldjwvcHJvcGVydHk+CiAgICAg
ICAgICAgICAgICA8c2lnbmFsIG5hbWU9ImFjdGl2YXRlIiBoYW5kbGVyPSJqb2luX3Jvb21faGFuZGxl
ciIgLz4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAg
ICAgPC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAg
ICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJzZXJ2aWNlLWxhYmVsIiA+CgkJPHBy
b3BlcnR5IG5hbWU9InZpc2libGUiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+QDwvcHJvcGVydHk+CiAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0ianVzdGlmeSI+R1RLX0pVU1RJRllfUklHSFQ8L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+R1RLX0FMSUdOX0VORDwvcHJvcGVy
dHk+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjI8L3Byb3BlcnR5PgogICAgICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAg
IDwvcGFja2luZz4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAg
ICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0VudHJ5IiBpZD0ic2VydmljZSI+CgkJPHByb3BlcnR5IG5h
bWU9InZpc2libGUiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
InBsYWNlaG9sZGVyLXRleHQiPmxlYXZlIGVtcHR5IHRvIHVzZSB5b3VyIHNlcnZlcjwvcHJvcGVydHk+
CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGV4cGFuZCI+VHJ1ZTwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idGV4dCI+Y29uZmVyZW5jZS5yaXNldXAubmV0PC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxzaWduYWwgbmFtZT0iYWN0aXZhdGUiIGhhbmRsZXI9Impv
aW5fcm9vbV9oYW5kbGVyIiAvPgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgIDxw
YWNraW5nPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4zPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4xPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAg
IDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9ImhhbmRsZS1s
YWJlbCIgPgoJCTxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkhhbmRsZTo8L3By
b3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imp1c3RpZnkiPkdUS19KVVNUSUZZ
X1JJR0hUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoYWxpZ24iPkdU
S19BTElHTl9FTkQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAg
IDxwYWNraW5nPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4wPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4yPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAg
ICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtFbnRyeSIgaWQ9ImhhbmRs
ZSI+CgkJPHByb3BlcnR5IG5hbWU9InZpc2libGUiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9ImhleHBhbmQiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAg
PHNpZ25hbCBuYW1lPSJhY3RpdmF0ZSIgaGFuZGxlcj0iam9pbl9yb29tX2hhbmRsZXIiIC8+CiAgICAg
ICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjI8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9IndpZHRoIj4zPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAg
ICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgIDxwYWNraW5nPgogICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0iZXhwYW5kIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9ImZpbGwiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0icG9zaXRpb24iPjI8L3Byb3BlcnR5PgogICAgICAgICAgPC9wYWNraW5nPgogICAgICAgIDwv
Y2hpbGQ+CiAgICAgIDwvb2JqZWN0PgogICAgPC9jaGlsZD4KICAgIDxjaGlsZCBpbnRlcm5hbC1jaGls
ZD0iYWN0aW9uX2FyZWEiPgogICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCdXR0b25Cb3giIGlkPSJidXR0
b24tYm94Ij4KICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+dHJ1ZTwvcHJvcGVydHk+CiAg
ICAgICAgPHByb3BlcnR5IG5hbWU9Im9yaWVudGF0aW9uIj5HVEtfT1JJRU5UQVRJT05fSE9SSVpPTlRB
TDwvcHJvcGVydHk+CiAgICAgICAgPGNoaWxkPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQnV0
dG9uIiBpZD0iYnV0dG9uLWNhbmNlbCI+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxl
Ij50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIj5fQ2FuY2Vs
PC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InVzZS11bmRlcmxpbmUiPlRydWU8
L3Byb3BlcnR5PgogICAgICAgICAgICA8c2lnbmFsIG5hbWU9ImNsaWNrZWQiIGhhbmRsZXI9ImNhbmNl
bF9oYW5kbGVyIiAvPgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgPC9jaGlsZD4KICAgICAgICA8
Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCdXR0b24iIGlkPSJidXR0b24tb2siPgog
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkpvaW4gcm9vbTwvcHJv
cGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ1c2UtdW5kZXJsaW5lIj5UcnVlPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbi1kZWZhdWx0Ij50cnVlPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJjbGlja2VkIiBoYW5kbGVyPSJqb2luX3Jvb21faGFu
ZGxlciIgLz4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgIDwvY2hpbGQ+CiAgICAgIDwvb2JqZWN0
PgogICAgPC9jaGlsZD4KICA8L29iamVjdD4KPC9pbnRlcmZhY2U+Cg==
`,
	},

	"/definitions/AddContact.xml": {
		local:   "definitions/AddContact.xml",
		size:    9364,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtMaXN0U3RvcmUiIGlkPSJhY2NvdW50cy1tb2Rl
bCI+CiAgICA8Y29sdW1ucz4KICAgICAgPCEtLSBhY2NvdW50IG5hbWUgLS0+CiAgICAgIDxjb2x1bW4g
dHlwZT0iZ2NoYXJhcnJheSIvPgogICAgICA8IS0tIGFjY291bnQgaWQgLS0+CiAgICAgIDxjb2x1bW4g
dHlwZT0iZ2NoYXJhcnJheSIvPgogICAgPC9jb2x1bW5zPgogIDwvb2JqZWN0PgogIDxvYmplY3QgY2xh
c3M9Ikd0a1RleHRCdWZmZXIiIGlkPSJzdWJzY3JpcHRpb25Bc2tNZXNzYWdlIj4KICAgIDxwcm9wZXJ0
eSBuYW1lPSJ0ZXh0IiB0cmFuc2xhdGFibGU9InllcyI+SSB3b3VsZCBsaWtlIHRvIGFkZCB5b3UgdG8g
bXkgY29udGFjdCBsaXN0LjwvcHJvcGVydHk+CiAgPC9vYmplY3Q+CiAgPG9iamVjdCBjbGFzcz0iR3Rr
RGlhbG9nIiBpZD0iQWRkQ29udGFjdCI+CiAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+ZmFsc2U8
L3Byb3BlcnR5PgogICAgPHByb3BlcnR5IG5hbWU9IndpbmRvdy1wb3NpdGlvbiI+R1RLX1dJTl9QT1Nf
Q0VOVEVSPC9wcm9wZXJ0eT4KICAgIDxwcm9wZXJ0eSBuYW1lPSJib3JkZXJfd2lkdGgiPjc8L3Byb3Bl
cnR5PgogICAgPHByb3BlcnR5IG5hbWU9InRpdGxlIiB0cmFuc2xhdGFibGU9InllcyI+QWRkIG5ldyBj
b250YWN0PC9wcm9wZXJ0eT4KICAgIDxwcm9wZXJ0eSBuYW1lPSJyZXNpemFibGUiPlRydWU8L3Byb3Bl
cnR5PgogICAgPHByb3BlcnR5IG5hbWU9ImRlZmF1bHQtaGVpZ2h0Ij40MDA8L3Byb3BlcnR5PgogICAg
PHByb3BlcnR5IG5hbWU9ImRlZmF1bHQtd2lkdGgiPjUwMDwvcHJvcGVydHk+CiAgICA8cHJvcGVydHkg
bmFtZT0iZGVzdHJveS13aXRoLXBhcmVudCI+dHJ1ZTwvcHJvcGVydHk+CiAgICA8Y2hpbGQgaW50ZXJu
YWwtY2hpbGQ9InZib3giPgogICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCb3giIGlkPSJWYm94Ij4KICAg
ICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgPHByb3Bl
cnR5IG5hbWU9ImhvbW9nZW5lb3VzIj5mYWxzZTwvcHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5h
bWU9Im9yaWVudGF0aW9uIj5HVEtfT1JJRU5UQVRJT05fVkVSVElDQUw8L3Byb3BlcnR5PgogICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJzcGFjaW5nIj42PC9wcm9wZXJ0eT4KICAgICAgICA8Y2hpbGQ+CiAgICAg
ICAgICA8IS0tVE9ETzogdGhpcyBzZWVtcyB0byByZW5kZXIgd2VpcmRseSAtLT4KICAgICAgICAgIDxv
YmplY3QgY2xhc3M9Ikd0a0JveCIgaWQ9Im5vdGlmaWNhdGlvbi1hcmVhIj4KICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9InZpc2libGUiPmZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9Im1hcmdpbi10b3AiPjU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
bWFyZ2luLWxlZnQiPjU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2lu
LXJpZ2h0Ij41PC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im9yaWVudGF0aW9u
Ij5HVEtfT1JJRU5UQVRJT05fVkVSVElDQUw8L3Byb3BlcnR5PgogICAgICAgICAgPC9vYmplY3Q+CiAg
ICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+RmFsc2U8
L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZmlsbCI+VHJ1ZTwvcHJvcGVydHk+
CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MDwvcHJvcGVydHk+CiAgICAgICAg
ICA8L3BhY2tpbmc+CiAgICAgICAgPC9jaGlsZD4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2Jq
ZWN0IGNsYXNzPSJHdGtHcmlkIiBpZD0iZ3JpZCI+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2
aXNpYmxlIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi10
b3AiPjE1PC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi1ib3R0b20i
PjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi1zdGFydCI+MTA8
L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLWVuZCI+MTA8L3Byb3Bl
cnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icm93LXNwYWNpbmciPjEyPC9wcm9wZXJ0eT4K
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNvbHVtbi1zcGFjaW5nIj42PC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0i
YWNjb3VudHNMYWJlbCIgPgoJCTxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj50cnVlPC9wcm9wZXJ0eT4K
ICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkFj
Y291bnQ6PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJqdXN0aWZ5Ij5H
VEtfSlVTVElGWV9SSUdIVDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
aGFsaWduIj5HVEtfQUxJR05fRU5EPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L29iamVjdD4KICAg
ICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0
dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFj
aCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxk
PgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQ29tYm9C
b3giIGlkPSJhY2NvdW50cyI+CgkJPHByb3BlcnR5IG5hbWU9InZpc2libGUiPnRydWU8L3Byb3BlcnR5
PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1vZGVsIj5hY2NvdW50cy1tb2RlbDwvcHJv
cGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGFzLWZvY3VzIj50cnVlPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoZXhwYW5kIj5UcnVlPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0i
R3RrQ2VsbFJlbmRlcmVyVGV4dCIgaWQ9ImFjY291bnQtbmFtZS1yZW5kZXJlZCIvPgogICAgICAgICAg
ICAgICAgICA8YXR0cmlidXRlcz4KICAgICAgICAgICAgICAgICAgICA8YXR0cmlidXRlIG5hbWU9InRl
eHQiPjA8L2F0dHJpYnV0ZT4KICAgICAgICAgICAgICAgICAgPC9hdHRyaWJ1dGVzPgogICAgICAgICAg
ICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2lu
Zz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MTwvcHJvcGVydHk+
CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+MTwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hp
bGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJhY2NvdW50TGFiZWwi
ID4KCQk8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5Db250YWN0IHRvIGFkZDo8
L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imp1c3RpZnkiPkdUS19KVVNU
SUZZX1JJR0hUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoYWxpZ24i
PkdUS19BTElHTl9FTkQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAg
ICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4w
PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4yPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAg
ICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtFbnRyeSIgaWQ9ImFk
ZHJlc3MiPgoJCTxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwbGFjZWhvbGRlci10ZXh0Ij5zb21lb25lQGphYmJlci5vcmc8
L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhleHBhbmQiPlRydWU8L3By
b3BlcnR5PgogICAgICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJhY3RpdmF0ZSIgaGFuZGxlcj0ib25f
c2F2ZV9zaWduYWwiIC8+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgPHBhY2tp
bmc+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjE8L3Byb3BlcnR5
PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjI8L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNo
aWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0ibmlja25hbWVMYWJl
bCIgPgoJCTxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPk5pY2tuYW1lOjwvcHJv
cGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ianVzdGlmeSI+R1RLX0pVU1RJRllf
UklHSFQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+R1RL
X0FMSUdOX0VORDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAg
PHBhY2tpbmc+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjA8L3By
b3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjM8L3Byb3Bl
cnR5PgogICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAg
ICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0VudHJ5IiBpZD0ibmlja25h
bWUiPgoJCTxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJoZXhwYW5kIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAg
IDxzaWduYWwgbmFtZT0iYWN0aXZhdGUiIGhhbmRsZXI9Im9uX3NhdmVfc2lnbmFsIiAvPgogICAgICAg
ICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4xPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4zPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+
CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgIDxwYWNraW5n
PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZXhwYW5kIj50cnVlPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9ImZpbGwiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0icG9zaXRpb24iPjE8L3Byb3BlcnR5PgogICAgICAgICAgPC9wYWNraW5nPgogICAg
ICAgIDwvY2hpbGQ+CiAgICAgICAgPGNoaWxkPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQ2hl
Y2tCdXR0b24iIGlkPSJhdXRvX2F1dGhvcml6ZV9jaGVja2J1dHRvbiI+CiAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+QV9sbG93IHRoaXMgY29udGFjdCB0byB2aWV3IG15
IHN0YXR1czwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ1c2VfYWN0aW9uX2Fw
cGVhcmFuY2UiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbl9m
b2N1cyI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJyZWNlaXZlc19k
ZWZhdWx0Ij5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJub19zaG93
X2FsbCI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ1c2VfdW5kZXJs
aW5lIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InhhbGlnbiI+MDwv
cHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJhY3RpdmUiPlRydWU8L3Byb3BlcnR5
PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZHJhd19pbmRpY2F0b3IiPlRydWU8L3Byb3BlcnR5
PgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9ImV4cGFuZCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0iZmlsbCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icG9zaXRp
b24iPjI8L3Byb3BlcnR5PgogICAgICAgICAgPC9wYWNraW5nPgogICAgICAgIDwvY2hpbGQ+CiAgICAg
ICAgPGNoaWxkPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrU2Nyb2xsZWRXaW5kb3ciIGlkPSJt
ZXNzYWdlX3Njcm9sbGVkd2luZG93Ij4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUi
PlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuX2ZvY3VzIj5UcnVl
PC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im5vX3Nob3dfYWxsIj5UcnVlPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImJvcmRlcl93aWR0aCI+NjwvcHJvcGVy
dHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzaGFkb3dfdHlwZSI+ZXRjaGVkLWluPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1pbl9jb250ZW50X2hlaWdodCI+NTwvcHJv
cGVydHk+CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtU
ZXh0VmlldyIgaWQ9Im1lc3NhZ2VfdGV4dHZpZXciPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9InZpc2libGUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
ImNhbl9mb2N1cyI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
d3JhcF9tb2RlIj53b3JkPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJi
dWZmZXIiPnN1YnNjcmlwdGlvbkFza01lc3NhZ2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvb2Jq
ZWN0PgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICA8cGFj
a2luZz4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+VHJ1ZTwvcHJvcGVydHk+CiAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9InBvc2l0aW9uIj4zPC9wcm9wZXJ0eT4KICAgICAgICAgIDwvcGFja2luZz4K
ICAgICAgICA8L2NoaWxkPgogICAgICAgIDxjaGlsZD4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0
a0xhYmVsIiBpZD0ic3Vic2NyaXB0aW9uSW5zdHJ1Y3Rpb25zIj4KICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJ2aXNpYmxlIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJl
bCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPldhcm5pbmc6IFRoaXMgbWVzc2FnZSB3aWxsIG5vdCBiZSBlbmQt
dG8tZW5kIGVuY3J5cHRlZC4gRG8gbm90IHJldmVhbCBhbnkgc2Vuc2l0aXZlIGluZm9ybWF0aW9uIGlu
IGl0LjwvcHJvcGVydHk+CiAgICAgICAgICA8cHJvcGVydHkgbmFtZT0id3JhcCI+dHJ1ZTwvcHJvcGVy
dHk+CiAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWF4LXdpZHRoLWNoYXJzIj42NTwvcHJvcGVydHk+
CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICA8L2NoaWxkPgogICAgICA8L29iamVjdD4KICAgIDwv
Y2hpbGQ+CiAgICA8Y2hpbGQgaW50ZXJuYWwtY2hpbGQ9ImFjdGlvbl9hcmVhIj4KICAgICAgPG9iamVj
dCBjbGFzcz0iR3RrQnV0dG9uQm94IiBpZD0iYnV0dG9uX2JveCI+CiAgICAgICAgPHByb3BlcnR5IG5h
bWU9InZpc2libGUiPnRydWU8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJvcmllbnRh
dGlvbiI+R1RLX09SSUVOVEFUSU9OX0hPUklaT05UQUw8L3Byb3BlcnR5PgogICAgICAgIDxjaGlsZD4K
ICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0J1dHRvbiIgaWQ9ImJ1dHRvbl9jYW5jZWwiPgogICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCI+X0NhbmNlbDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJ1c2UtdW5kZXJsaW5lIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHNpZ25h
bCBuYW1lPSJjbGlja2VkIiBoYW5kbGVyPSJvbl9jYW5jZWxfc2lnbmFsIiAvPgogICAgICAgICAgPC9v
YmplY3Q+CiAgICAgICAgPC9jaGlsZD4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNs
YXNzPSJHdGtCdXR0b24iIGlkPSJidXR0b25fb2siPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
dmlzaWJsZSI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIg
dHJhbnNsYXRhYmxlPSJ5ZXMiPkFkZDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJ1c2UtdW5kZXJsaW5lIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
ImNhbi1kZWZhdWx0Ij50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJjbGlj
a2VkIiBoYW5kbGVyPSJvbl9zYXZlX3NpZ25hbCIgLz4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAg
IDwvY2hpbGQ+CiAgICAgIDwvb2JqZWN0PgogICAgPC9jaGlsZD4KICA8L29iamVjdD4KPC9pbnRlcmZh
Y2U+Cg==
`,
	},

	"/definitions/AnswerSMPQuestion.xml": {
		local:   "definitions/AnswerSMPQuestion.xml",
		size:    6378,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtFbnRyeUJ1ZmZlciIgaWQ9InBpbl9lbnRyeV9i
dWZmZXIiPgogICAgPHNpZ25hbCBuYW1lPSJpbnNlcnRlZC10ZXh0IiBoYW5kbGVyPSJ0ZXh0X2NoYW5n
aW5nIi8+CiAgICA8c2lnbmFsIG5hbWU9ImRlbGV0ZWQtdGV4dCIgaGFuZGxlcj0idGV4dF9jaGFuZ2lu
ZyIvPgogIDwvb2JqZWN0PgogIDxvYmplY3QgY2xhc3M9Ikd0a0RpYWxvZyIgaWQ9ImRpYWxvZyI+CiAg
ICA8cHJvcGVydHkgbmFtZT0id2luZG93LXBvc2l0aW9uIj5HVEtfV0lOX1BPU19DRU5URVI8L3Byb3Bl
cnR5PgogICAgPHByb3BlcnR5IG5hbWU9ImJvcmRlci13aWR0aCI+NzwvcHJvcGVydHk+CiAgICA8Y2hp
bGQgaW50ZXJuYWwtY2hpbGQ9InZib3giPgogICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCb3giIGlkPSJi
b3giPgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJib3JkZXItd2lkdGgiPjEwPC9wcm9wZXJ0eT4KICAg
ICAgICA8cHJvcGVydHkgbmFtZT0ib3JpZW50YXRpb24iPkdUS19PUklFTlRBVElPTl9WRVJUSUNBTDwv
cHJvcGVydHk+CiAgICAgICAgPGNoaWxkPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQm94IiBp
ZD0ibm90aWZpY2F0aW9uLWFyZWEiPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ib3JpZW50YXRp
b24iPkdUS19PUklFTlRBVElPTl9WRVJUSUNBTDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9InNwYWNpbmciPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAg
ICAgPHBhY2tpbmc+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJleHBhbmQiPmZhbHNlPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImZpbGwiPnRydWU8L3Byb3BlcnR5PgogICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjA8L3Byb3BlcnR5PgogICAgICAgICAgPC9w
YWNraW5nPgogICAgICAgIDwvY2hpbGQ+CiAgICAgICAgPGNoaWxkPgogICAgICAgICAgPG9iamVjdCAg
Y2xhc3M9Ikd0a0ltYWdlIiBpZD0ic21wX2ltYWdlIj4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
Im1hcmdpbl9ib3R0b20iPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1h
cmdpbl90b3AiPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgIDwvY2hpbGQ+
CiAgICAgICAgPGNoaWxkPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrR3JpZCI+CiAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJoYWxpZ24iPkdUS19BTElHTl9DRU5URVI8L3Byb3BlcnR5PgogICAg
ICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJx
dWVzdGlvbl9mcm9tX3BlZXIiPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0
cmFuc2xhdGFibGU9InllcyIvPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi1i
b3R0b20iPjEyPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgPC9j
aGlsZD4KICAgICAgICAgICA8L29iamVjdD4KICAgICAgICA8L2NoaWxkPgogICAgICAgIDxjaGlsZD4K
ICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0dyaWQiPgogICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0iaGFsaWduIj5HVEtfQUxJR05fQ0VOVEVSPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPGNoaWxkPgog
ICAgICAgICAgICAgIDxvYmplY3QgIGNsYXNzPSJHdGtMYWJlbCI+CiAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5BbnN3ZXI6IDwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ianVzdGlmeSI+R1RLX0pVU1RJRllfUklHSFQ8L3By
b3BlcnR5PgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgIDxwYWNraW5nPgogICAg
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ3aWR0aCI+MjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9w
YWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAg
ICAgPG9iamVjdCAgY2xhc3M9Ikd0a0VudHJ5IiBpZD0iYW5zd2VyIj4KICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJidWZmZXIiPnBpbl9lbnRyeV9idWZmZXI8L3Byb3BlcnR5PgogICAgICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi1ib3R0b20iPjEwPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+
CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICA8L2NoaWxkPgog
ICAgICAgIDxjaGlsZD4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIj4KICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+WW91ciBwZWVyIG11c3Qg
aGF2ZSBwcmV2aW91c2x5IHNoYXJlZCBpdCB3aXRoIHlvdS4gSWYgbm90LCB0cnkgdG86PC9wcm9wZXJ0
eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi10b3AiPjU8L3Byb3BlcnR5PgogICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLWJvdHRvbSI+MTA8L3Byb3BlcnR5PgogICAgICAg
ICAgPC9vYmplY3Q+CiAgICAgICAgPC9jaGlsZD4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgIDxv
YmplY3QgY2xhc3M9Ikd0a0dyaWQiPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY29sdW1uLXNw
YWNpbmciPjY8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icm93LXNwYWNpbmci
PjI8L3Byb3BlcnR5PgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICA8b2JqZWN0IGNs
YXNzPSJHdGtJbWFnZSIgaWQ9InBhZGxvY2tfaW1hZ2UxIi8+CiAgICAgICAgICAgICAgICA8cGFja2lu
Zz4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjA8L3Byb3Bl
cnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4wPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAg
ICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiPgogICAg
ICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5B
c2sgdGhlbSBpbiBwZXJzb248L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJqdXN0aWZ5Ij5HVEtfSlVTVElGWV9MRUZUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05fU1RBUlQ8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAg
ICAgIDwvcGFja2luZz4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNoaWxkPgogICAg
ICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrSW1hZ2UiIGlkPSJwYWRsb2NrX2ltYWdlMiIvPgog
ICAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
ImxlZnQtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0idG9wLWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAg
ICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICAgIDxvYmplY3Qg
Y2xhc3M9Ikd0a0xhYmVsIj4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJs
ZSI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVs
IiB0cmFuc2xhdGFibGU9InllcyI+QXNrIHRoZW0gaW4gYSBkaWZmZXJlbnQgZW5jcnlwdGVkIGNoYW5u
ZWw8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJqdXN0aWZ5Ij5H
VEtfSlVTVElGWV9MRUZUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0iaGFsaWduIj5HVEtfQUxJR05fU1RBUlQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPC9vYmpl
Y3Q+CiAgICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0ibGVmdC1hdHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJ0b3AtYXR0YWNoIj4xPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDwvcGFja2luZz4K
ICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgPG9i
amVjdCBjbGFzcz0iR3RrSW1hZ2UiIGlkPSJhbGVydF9pbWFnZSIvPgogICAgICAgICAgICAgICAgPHBh
Y2tpbmc+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4wPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+Mjwv
cHJvcGVydHk+CiAgICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAg
ICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIj4K
ICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+VHJ1ZTwvcHJvcGVydHk+
CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9Inll
cyI+QXNrIHRoZW0gb3ZlciB0aGUgcGhvbmU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJqdXN0aWZ5Ij5HVEtfSlVTVElGWV9MRUZUPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05fU1RBUlQ8L3Byb3BlcnR5
PgogICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAg
ICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjE8L3Byb3BlcnR5PgogICAg
ICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4yPC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPC9v
YmplY3Q+CiAgICAgICAgPC9jaGlsZD4KICAgICAgICA8Y2hpbGQgaW50ZXJuYWwtY2hpbGQ9ImFjdGlv
bl9hcmVhIj4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0J1dHRvbkJveCIgaWQ9ImJ1dHRvbl9i
b3giPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ib3JpZW50YXRpb24iPkdUS19PUklFTlRBVElP
Tl9IT1JJWk9OVEFMPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxv
YmplY3QgY2xhc3M9Ikd0a0J1dHRvbiIgaWQ9ImJ1dHRvbl9zdWJtaXQiPgogICAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+VmVyaWZ5IEFuc3dlcjwvcHJv
cGVydHk+CiAgICAgICAgICAgICAgICA8c2lnbmFsIG5hbWU9ImNsaWNrZWQiIGhhbmRsZXI9ImNsb3Nl
X3NoYXJlX3BpbiIvPgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICA8L2NoaWxkPgog
ICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgPC9jaGlsZD4KICAgICAgPC9vYmplY3Q+CiAgICA8L2No
aWxkPgogICAgPGFjdGlvbi13aWRnZXRzPgogICAgICA8YWN0aW9uLXdpZGdldCByZXNwb25zZT0ib2si
PmJ1dHRvbl9zdWJtaXQ8L2FjdGlvbi13aWRnZXQ+CiAgICA8L2FjdGlvbi13aWRnZXRzPgogIDwvb2Jq
ZWN0Pgo8L2ludGVyZmFjZT4K
`,
	},

	"/definitions/AskForPassword.xml": {
		local:   "definitions/AskForPassword.xml",
		size:    4786,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtEaWFsb2ciIGlkPSJBc2tGb3JQYXNzd29yZCI+
CiAgICA8cHJvcGVydHkgbmFtZT0id2luZG93LXBvc2l0aW9uIj5HVEtfV0lOX1BPU19DRU5URVI8L3By
b3BlcnR5PgogICAgPHByb3BlcnR5IG5hbWU9InRpdGxlIiB0cmFuc2xhdGFibGU9InllcyI+RW50ZXIg
eW91ciBwYXNzd29yZDwvcHJvcGVydHk+CiAgICA8c2lnbmFsIG5hbWU9ImRlbGV0ZS1ldmVudCIgaGFu
ZGxlcj0ib25fY2FuY2VsX3Bhc3N3b3JkX3NpZ25hbCIgLz4KICAgIDxjaGlsZCBpbnRlcm5hbC1jaGls
ZD0idmJveCI+CiAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0JveCIgaWQ9IlZib3giPgogICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJtYXJnaW4iPjEwPC9wcm9wZXJ0eT4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAg
ICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9Im1lc3NhZ2UiPgogICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj48L3Byb3BlcnR5PgogICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05fU1RBUlQ8L3Byb3BlcnR5PgogICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0ianVzdGlmeSI+R1RLX0pVU1RJRllfTEVGVDwvcHJvcGVydHk+CiAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ3cmFwIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9Im1heC13aWR0aC1jaGFycyI+NTA8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0ic2VsZWN0YWJsZSI+VFJVRTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJtYXJnaW4tYm90dG9tIj4xMDwvcHJvcGVydHk+CiAgICAgICAgICA8L29iamVjdD4KICAg
ICAgICA8L2NoaWxkPgogICAgICAgIDxjaGlsZD4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0dy
aWQiIGlkPSJncmlkIj4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi1ib3R0b20iPjEw
PC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InJvdy1zcGFjaW5nIj4xMjwvcHJv
cGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjb2x1bW4tc3BhY2luZyI+NjwvcHJvcGVy
dHk+CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJl
bCIgaWQ9ImFjY291bnRNZXNzYWdlIiA+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFi
ZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5BY2NvdW50PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJqdXN0aWZ5Ij5HVEtfSlVTVElGWV9SSUdIVDwvcHJvcGVydHk+CiAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05fRU5EPC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgICAgIDxhdHRyaWJ1dGVzPgogICAgICAgICAgICAgICAgICA8YXR0cmlidXRlIG5hbWU9
IndlaWdodCIgdmFsdWU9ImJvbGQiLz4KICAgICAgICAgICAgICAgIDwvYXR0cmlidXRlcz4KICAgICAg
ICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0idG9wLWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNraW5n
PgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9i
amVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJhY2NvdW50TmFtZSIgPgogICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9ImhhbGlnbiI+R1RLX0FMSUdOX1NUQVJUPC9wcm9wZXJ0eT4KICAgICAgICAgICAg
ICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0idG9wLWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAg
ICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBj
bGFzcz0iR3RrTGFiZWwiIGlkPSJwYXNzTWVzc2FnZSIgPgogICAgICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+UGFzc3dvcmQ8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imp1c3RpZnkiPkdUS19KVVNUSUZZX1JJR0hUPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoYWxpZ24iPkdUS19BTElHTl9FTkQ8L3By
b3BlcnR5PgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgIDxwYWNraW5nPgogICAg
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4xPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgIDxjaGlsZD4KICAg
ICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtFbnRyeSIgaWQ9InBhc3N3b3JkIj4KICAgICAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoYXMtZm9jdXMiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2liaWxpdHkiPmZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgICAgIDxzaWduYWwgbmFtZT0iYWN0aXZhdGUiIGhhbmRsZXI9Im9uX2VudGVyZWRfcGFzc3dvcmRf
c2lnbmFsIiAvPgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgIDxwYWNraW5nPgog
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4xPC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4xPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgIDxjaGlsZD4K
ICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9InNhdmVQYXNzd29yZExhYmVs
Ij4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMi
PlNhdmUgcGFzc3dvcmQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imp1
c3RpZnkiPkdUS19KVVNUSUZZX1JJR0hUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJoYWxpZ24iPkdUS19BTElHTl9FTkQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvb2Jq
ZWN0PgogICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
ImxlZnQtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0
b3AtYXR0YWNoIj4yPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAg
IDwvY2hpbGQ+CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJH
dGtDaGVja0J1dHRvbiIgaWQ9InNhdmVQYXNzd29yZCIvPgogICAgICAgICAgICAgIDxwYWNraW5nPgog
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4xPC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4yPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICA8L29iamVjdD4K
ICAgICAgICA8L2NoaWxkPgogICAgICAgIDxjaGlsZCBpbnRlcm5hbC1jaGlsZD0iYWN0aW9uX2FyZWEi
PgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQnV0dG9uQm94IiBpZD0iYnV0dG9uX2JveCI+CiAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJvcmllbnRhdGlvbiI+R1RLX09SSUVOVEFUSU9OX0hPUkla
T05UQUw8L3Byb3BlcnR5PgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBj
bGFzcz0iR3RrQnV0dG9uIiBpZD0iYnV0dG9uX2NhbmNlbCI+CiAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0ibGFiZWwiPkNhbmNlbDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8c2lnbmFsIG5h
bWU9ImNsaWNrZWQiIGhhbmRsZXI9Im9uX2NhbmNlbF9wYXNzd29yZF9zaWduYWwiIC8+CiAgICAgICAg
ICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgIDxjaGlsZD4KICAg
ICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCdXR0b24iIGlkPSJidXR0b25fb2siPgogICAgICAg
ICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+Q29ubmVjdDwv
cHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuLWRlZmF1bHQiPnRydWU8
L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJjbGlja2VkIiBoYW5kbGVyPSJv
bl9lbnRlcmVkX3Bhc3N3b3JkX3NpZ25hbCIgLz4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAg
ICAgICAgPC9jaGlsZD4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgIDwvY2hpbGQ+CiAgICAgIDwv
b2JqZWN0PgogICAgPC9jaGlsZD4KICA8L29iamVjdD4KPC9pbnRlcmZhY2U+Cg==
`,
	},

	"/definitions/AskToEncrypt.xml": {
		local:   "definitions/AskToEncrypt.xml",
		size:    725,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtNZXNzYWdlRGlhbG9nIiBpZD0iQXNrVG9FbmNy
eXB0Ij4KICAgIDxwcm9wZXJ0eSBuYW1lPSJ3aW5kb3ctcG9zaXRpb24iPkdUS19XSU5fUE9TX0NFTlRF
UjwvcHJvcGVydHk+CiAgICA8cHJvcGVydHkgbmFtZT0ibW9kYWwiPnRydWU8L3Byb3BlcnR5PgogICAg
PHByb3BlcnR5IG5hbWU9ImJvcmRlcl93aWR0aCI+NzwvcHJvcGVydHk+CiAgICA8cHJvcGVydHkgbmFt
ZT0idGl0bGUiIHRyYW5zbGF0YWJsZT0ieWVzIj5FbmNyeXB0IHlvdXIgYWNjb3VudCdzIGluZm9ybWF0
aW9uPC9wcm9wZXJ0eT4KICAgIDxwcm9wZXJ0eSBuYW1lPSJ0ZXh0IiB0cmFuc2xhdGFibGU9InllcyI+
CldvdWxkIHlvdSBsaWtlIHRvIGVuY3J5cHQgeW91ciBhY2NvdW50J3MgaW5mb3JtYXRpb24/CiAgICA8
L3Byb3BlcnR5PgogICAgPHByb3BlcnR5IG5hbWU9InNlY29uZGFyeV90ZXh0IiB0cmFuc2xhdGFibGU9
InllcyI+WW91IHdpbGwgbm90IGJlIGFibGUgdG8gYWNjZXNzIHlvdXIgYWNjb3VudCdzIGluZm9ybWF0
aW9uIGZpbGUgaWYgeW91IGxvc2UgdGhlCgpwYXNzd29yZC4gWW91IHdpbGwgb25seSBiZSBhc2tlZCBm
b3IgaXQgd2hlbiBDb3lJTSBzdGFydHMuCiAgICA8L3Byb3BlcnR5PgogICAgPHByb3BlcnR5IG5hbWU9
ImJ1dHRvbnMiPkdUS19CVVRUT05TX1lFU19OTzwvcHJvcGVydHk+CiAgPC9vYmplY3Q+CjwvaW50ZXJm
YWNlPgo=
`,
	},

	"/definitions/AuthorizeSubscription.xml": {
		local:   "definitions/AuthorizeSubscription.xml",
		size:    355,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtNZXNzYWdlRGlhbG9nIiBpZD0iZGlhbG9nIj4K
ICAgIDxwcm9wZXJ0eSBuYW1lPSJ0aXRsZSIgdHJhbnNsYXRhYmxlPSJ5ZXMiPlN1YnNjcmlwdGlvbiBy
ZXF1ZXN0PC9wcm9wZXJ0eT4KICAgIDxwcm9wZXJ0eSBuYW1lPSJ3aW5kb3ctcG9zaXRpb24iPkdUS19X
SU5fUE9TX0NFTlRFUjwvcHJvcGVydHk+CiAgICA8cHJvcGVydHkgbmFtZT0ibWVzc2FnZS10eXBlIj5H
VEtfTUVTU0FHRV9RVUVTVElPTjwvcHJvcGVydHk+CiAgICA8cHJvcGVydHkgbmFtZT0iYnV0dG9ucyI+
R1RLX0JVVFRPTlNfWUVTX05PPC9wcm9wZXJ0eT4KICA8L29iamVjdD4KPC9pbnRlcmZhY2U+Cg==
`,
	},

	"/definitions/CannotVerifyWithSMP.xml": {
		local:   "definitions/CannotVerifyWithSMP.xml",
		size:    2134,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtCb3giIGlkPSJzbXAtZXJyb3ItaW5mb2JhciI+
CiAgICA8cHJvcGVydHkgbmFtZT0iY2FuX2ZvY3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICA8cHJvcGVy
dHkgbmFtZT0ic3BhY2luZyI+Mjc8L3Byb3BlcnR5PgogICAgPGNoaWxkPgogICAgICA8b2JqZWN0IGNs
YXNzPSJHdGtJbWFnZSIgaWQ9InNtcC1lcnJvci1pbWFnZSI+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9
InZpc2libGUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMi
PkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luX2xlZnQiPjEwPC9w
cm9wZXJ0eT4KICAgICAgPC9vYmplY3Q+CiAgICAgIDxwYWNraW5nPgogICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJleHBhbmQiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iZmlsbCI+
VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9InBvc2l0aW9uIj4wPC9wcm9wZXJ0
eT4KICAgICAgPC9wYWNraW5nPgogICAgPC9jaGlsZD4KICAgIDxjaGlsZD4KICAgICAgPG9iamVjdCBj
bGFzcz0iR3RrTGFiZWwiIGlkPSJzbXAtZXJyb3ItbGFiZWwiPgogICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuX2ZvY3Vz
Ij5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+c3RhcnQ8L3By
b3BlcnR5PgogICAgICAgIDxhdHRyaWJ1dGVzPgogICAgICAgICAgPGF0dHJpYnV0ZSBuYW1lPSJ3ZWln
aHQiIHZhbHVlPSJib2xkIi8+CiAgICAgICAgPC9hdHRyaWJ1dGVzPgogICAgICA8L29iamVjdD4KICAg
ICAgPHBhY2tpbmc+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+RmFsc2U8L3Byb3BlcnR5
PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICA8cHJv
cGVydHkgbmFtZT0icG9zaXRpb24iPjE8L3Byb3BlcnR5PgogICAgICA8L3BhY2tpbmc+CiAgICA8L2No
aWxkPgogICAgPGNoaWxkPgogICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCdXR0b24iIGlkPSJzbXAtZXJy
b3ItYnV0dG9uIj4KICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+VHJ1ZTwvcHJvcGVydHk+
CiAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbl9mb2N1cyI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAg
PHByb3BlcnR5IG5hbWU9InJlY2VpdmVzX2RlZmF1bHQiPlRydWU8L3Byb3BlcnR5PgogICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJtYXJnaW5fcmlnaHQiPjEwPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkg
bmFtZT0ibWFyZ2luX3RvcCI+MTQ8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJn
aW5fYm90dG9tIj4xMjwvcHJvcGVydHk+CiAgICAgICAgPGNoaWxkPgogICAgICAgICAgPG9iamVjdCBj
bGFzcz0iR3RrTGFiZWwiPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+VHJ1ZTwv
cHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPkZhbHNlPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbl9sZWZ0Ij4xNTwvcHJvcGVydHk+
CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW5fcmlnaHQiPjE1PC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+VHJ5IGFnYWlu
IGxhdGVyPC9wcm9wZXJ0eT4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgIDwvY2hpbGQ+CiAgICAg
IDwvb2JqZWN0PgogICAgICA8cGFja2luZz4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iZXhwYW5kIj5G
YWxzZTwvcHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9ImZpbGwiPlRydWU8L3Byb3BlcnR5
PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwYWNrX3R5cGUiPmVuZDwvcHJvcGVydHk+CiAgICAgICAg
PHByb3BlcnR5IG5hbWU9InBvc2l0aW9uIj4yPC9wcm9wZXJ0eT4KICAgICAgPC9wYWNraW5nPgogICAg
PC9jaGlsZD4KICA8L29iamVjdD4KPC9pbnRlcmZhY2U+Cg==
`,
	},

	"/definitions/CaptureInitialMasterPassword.xml": {
		local:   "definitions/CaptureInitialMasterPassword.xml",
		size:    2775,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtEaWFsb2ciIGlkPSJDYXB0dXJlSW5pdGlhbE1h
c3RlclBhc3N3b3JkIj4KICAgIDxwcm9wZXJ0eSBuYW1lPSJ3aW5kb3ctcG9zaXRpb24iPkdUS19XSU5f
UE9TX0NFTlRFUjwvcHJvcGVydHk+CiAgICA8cHJvcGVydHkgbmFtZT0iYm9yZGVyX3dpZHRoIj43PC9w
cm9wZXJ0eT4KICAgIDxwcm9wZXJ0eSBuYW1lPSJ0aXRsZSIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkNvbmZp
Z3VyZSB5b3VyIG1hc3RlciBwYXNzd29yZDwvcHJvcGVydHk+CiAgICA8cHJvcGVydHkgbmFtZT0iZGVm
YXVsdC13aWR0aCI+MzAwPC9wcm9wZXJ0eT4KICAgIDxzaWduYWwgbmFtZT0iY2xvc2UiIGhhbmRsZXI9
Im9uX2NhbmNlbF9zaWduYWwiIC8+CiAgICA8Y2hpbGQgaW50ZXJuYWwtY2hpbGQ9InZib3giPgogICAg
ICA8b2JqZWN0IGNsYXNzPSJHdGtCb3giIGlkPSJWYm94Ij4KICAgICAgICA8cHJvcGVydHkgbmFtZT0i
bWFyZ2luIj4xMDwvcHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9InNwYWNpbmciPjEwPC9w
cm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iaG9tb2dlbmVvdXMiPmZhbHNlPC9wcm9wZXJ0
eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0ib3JpZW50YXRpb24iPkdUS19PUklFTlRBVElPTl9WRVJU
SUNBTDwvcHJvcGVydHk+CiAgICAgICAgPGNoaWxkPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3Rr
TGFiZWwiIGlkPSJwYXNzTWVzc2FnZSIgPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwi
IHRyYW5zbGF0YWJsZT0ieWVzIj5QbGVhc2UgZW50ZXIgdGhlIG1hc3RlciBwYXNzd29yZCBmb3IgdGhl
IGNvbmZpZ3VyYXRpb24gZmlsZSB0d2ljZS48L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0id3JhcCI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgIDxw
YWNraW5nPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZXhwYW5kIj5mYWxzZTwvcHJvcGVydHk+
CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9InBvc2l0aW9uIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgIDwvcGFja2lu
Zz4KICAgICAgICA8L2NoaWxkPgogICAgICAgIDxjaGlsZD4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9
Ikd0a0VudHJ5IiBpZD0icGFzc3dvcmQiPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGFzLWZv
Y3VzIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2liaWxpdHki
PmZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJhY3RpdmF0ZSIgaGFuZGxl
cj0ib25fc2F2ZV9zaWduYWwiIC8+CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgIDxwYWNraW5n
PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZXhwYW5kIj5mYWxzZTwvcHJvcGVydHk+CiAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9InBvc2l0aW9uIj4xPC9wcm9wZXJ0eT4KICAgICAgICAgIDwvcGFja2luZz4KICAg
ICAgICA8L2NoaWxkPgogICAgICAgIDxjaGlsZD4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0Vu
dHJ5IiBpZD0icGFzc3dvcmQyIj4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2liaWxpdHki
PmZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJhY3RpdmF0ZSIgaGFuZGxl
cj0ib25fc2F2ZV9zaWduYWwiIC8+CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgIDxwYWNraW5n
PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZXhwYW5kIj5mYWxzZTwvcHJvcGVydHk+CiAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9InBvc2l0aW9uIj4yPC9wcm9wZXJ0eT4KICAgICAgICAgIDwvcGFja2luZz4KICAg
ICAgICA8L2NoaWxkPgogICAgICA8L29iamVjdD4KICAgIDwvY2hpbGQ+CiAgICA8Y2hpbGQgaW50ZXJu
YWwtY2hpbGQ9ImFjdGlvbl9hcmVhIj4KICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQnV0dG9uQm94IiBp
ZD0iYnV0dG9uX2JveCI+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9Im9yaWVudGF0aW9uIj5HVEtfT1JJ
RU5UQVRJT05fSE9SSVpPTlRBTDwvcHJvcGVydHk+CiAgICAgICAgPGNoaWxkPgogICAgICAgICAgPG9i
amVjdCBjbGFzcz0iR3RrQnV0dG9uIiBpZD0iY2FuY2VsIj4KICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+Q2FuY2VsPC9wcm9wZXJ0eT4KICAgICAgICAgICAg
PHNpZ25hbCBuYW1lPSJjbGlja2VkIiBoYW5kbGVyPSJvbl9jYW5jZWxfc2lnbmFsIiAvPgogICAgICAg
ICAgPC9vYmplY3Q+CiAgICAgICAgPC9jaGlsZD4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2Jq
ZWN0IGNsYXNzPSJHdGtCdXR0b24iIGlkPSJzYXZlIj4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+T0s8L3Byb3BlcnR5PgogICAgICAgICAgICA8c2lnbmFs
IG5hbWU9ImNsaWNrZWQiIGhhbmRsZXI9Im9uX3NhdmVfc2lnbmFsIiAvPgogICAgICAgICAgPC9vYmpl
Y3Q+CiAgICAgICAgPC9jaGlsZD4KICAgICAgPC9vYmplY3Q+CiAgICA8L2NoaWxkPgogIDwvb2JqZWN0
Pgo8L2ludGVyZmFjZT4K
`,
	},

	"/definitions/CertificateDialog.xml": {
		local:   "definitions/CertificateDialog.xml",
		size:    21533,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtEaWFsb2ciIGlkPSJkaWFsb2ciPgogICAgPHBy
b3BlcnR5IG5hbWU9InRpdGxlIiB0cmFuc2xhdGFibGU9InllcyI+UHJvYmxlbSB3aXRoIHNlY3VyZSBj
b25uZWN0aW9uIGZvciBBQ0NPVU5UX05BTUU8L3Byb3BlcnR5PgogICAgPHByb3BlcnR5IG5hbWU9Im1v
ZGFsIj50cnVlPC9wcm9wZXJ0eT4KICAgIDxwcm9wZXJ0eSBuYW1lPSJ3aWR0aF9yZXF1ZXN0Ij40ODA8
L3Byb3BlcnR5PgogICAgPHByb3BlcnR5IG5hbWU9ImJvcmRlcl93aWR0aCI+MTA8L3Byb3BlcnR5Pgog
ICAgPGNoaWxkIGludGVybmFsLWNoaWxkPSJ2Ym94Ij4KICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQm94
IiBpZD0iYm94Ij4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iYm9yZGVyLXdpZHRoIj4xMDwvcHJvcGVy
dHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhvbW9nZW5lb3VzIj5mYWxzZTwvcHJvcGVydHk+CiAg
ICAgICAgPHByb3BlcnR5IG5hbWU9Im9yaWVudGF0aW9uIj5HVEtfT1JJRU5UQVRJT05fVkVSVElDQUw8
L3Byb3BlcnR5PgogICAgICAgIDxjaGlsZD4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVs
IiBpZD0ibWVzc2FnZSI+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRh
YmxlPSJ5ZXMiPjwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoYWxpZ24iPkdU
S19BTElHTl9TVEFSVDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJqdXN0aWZ5
Ij5HVEtfSlVTVElGWV9MRUZUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Indy
YXAiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWF4LXdpZHRoLWNo
YXJzIj44MDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzZWxlY3RhYmxlIj5U
UlVFPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi1ib3R0b20iPjEw
PC9wcm9wZXJ0eT4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgIDwvY2hpbGQ+CiAgICAgICAgPGNo
aWxkPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJpc3N1ZWRUb01lc3NhZ2Ui
PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5Jc3N1
ZWQgdG86PC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+R1RLX0FM
SUdOX1NUQVJUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imp1c3RpZnkiPkdU
S19KVVNUSUZZX0xFRlQ8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ic2VsZWN0
YWJsZSI+VFJVRTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxhdHRyaWJ1dGVzPgogICAgICAgICAgICAg
IDxhdHRyaWJ1dGUgbmFtZT0id2VpZ2h0IiB2YWx1ZT0iUEFOR09fV0VJR0hUX0JPTEQiLz4KICAgICAg
ICAgICAgPC9hdHRyaWJ1dGVzPgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgPC9jaGlsZD4KICAg
ICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtHcmlkIiBpZD0iaXNzdWVkVG9H
cmlkIj4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi10b3AiPjI8L3Byb3BlcnR5Pgog
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLWJvdHRvbSI+MTA8L3Byb3BlcnR5PgogICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLXN0YXJ0Ij4xMDwvcHJvcGVydHk+CiAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tZW5kIj4xMDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJyb3ctc3BhY2luZyI+MjwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJjb2x1bW4tc3BhY2luZyI+MTI8L3Byb3BlcnR5PgogICAgICAgICAgICA8Y2hpbGQ+CiAg
ICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJpc3N1ZWRUb0NuTGFiZWwiPgog
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+Q29t
bW9uIE5hbWUgKENOKTo8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imhh
bGlnbiI+R1RLX0FMSUdOX1NUQVJUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJqdXN0aWZ5Ij5HVEtfSlVTVElGWV9MRUZUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJzZWxlY3RhYmxlIj5UUlVFPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L29i
amVjdD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJsZWZ0LWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
dG9wLWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAg
ICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0i
R3RrTGFiZWwiIGlkPSJpc3N1ZWRUb0NuVmFsdWUiPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJoYWxpZ24iPkdUS19BTElHTl9TVEFSVDwvcHJvcGVydHk+CiAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0ianVzdGlmeSI+R1RLX0pVU1RJRllfTEVGVDwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ic2VsZWN0YWJsZSI+VFJVRTwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjA8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvcGFj
a2luZz4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAg
IDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0ic25pc0xhYmVsIj4KICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPk90aGVyIG5hbWVzOjwvcHJvcGVy
dHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05fU1RBUlQ8
L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZhbGlnbiI+R1RLX0FMSUdO
X1NUQVJUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJqdXN0aWZ5Ij5H
VEtfSlVTVElGWV9MRUZUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJz
ZWxlY3RhYmxlIj5UUlVFPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAg
ICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+
MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+MTwv
cHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAg
ICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJz
bmlzVmFsdWUiPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFi
bGU9InllcyI+PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoYWxpZ24i
PkdUS19BTElHTl9TVEFSVDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
anVzdGlmeSI+R1RLX0pVU1RJRllfTEVGVDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0ic2VsZWN0YWJsZSI+VFJVRTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0id3JhcCI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAg
ICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRh
Y2giPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2gi
PjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgPC9jaGlsZD4K
ICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBp
ZD0iaXNzdWVkVG9PTGFiZWwiPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0
cmFuc2xhdGFibGU9InllcyI+T3JnYW5pemF0aW9uIChPKTo8L3Byb3BlcnR5PgogICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+R1RLX0FMSUdOX1NUQVJUPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJqdXN0aWZ5Ij5HVEtfSlVTVElGWV9MRUZUPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzZWxlY3RhYmxlIj5UUlVFPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+MjwvcHJvcGVydHk+CiAgICAgICAgICAgICAg
PC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAg
ICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJpc3N1ZWRUb09WYWx1ZSI+CiAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj48L3Byb3BlcnR5
PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+R1RLX0FMSUdOX1NUQVJUPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJqdXN0aWZ5Ij5HVEtfSlVTVElG
WV9MRUZUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzZWxlY3RhYmxl
Ij5UUlVFPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFj
a2luZz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MTwvcHJvcGVy
dHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+MjwvcHJvcGVydHk+
CiAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8
Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJpc3N1ZWRUb09V
TGFiZWwiPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9
InllcyI+T3JnYW5pemF0aW9uYWwgVW5pdCAoT1UpOjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05fU1RBUlQ8L3Byb3BlcnR5PgogICAgICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9Imp1c3RpZnkiPkdUS19KVVNUSUZZX0xFRlQ8L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InNlbGVjdGFibGUiPlRSVUU8L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4zPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L3Bh
Y2tpbmc+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAg
ICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9Imlzc3VlZFRvT1VWYWx1ZSI+CiAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj48L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+R1RLX0FMSUdOX1NUQVJUPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJqdXN0aWZ5Ij5HVEtfSlVTVElGWV9M
RUZUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzZWxlY3RhYmxlIj5U
UlVFPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2lu
Zz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MTwvcHJvcGVydHk+
CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+MzwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hp
bGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJTTkxhYmVsIj4KICAg
ICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPlNlcmlh
bCBOdW1iZXI6PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoYWxpZ24i
PkdUS19BTElHTl9TVEFSVDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
anVzdGlmeSI+R1RLX0pVU1RJRllfTEVGVDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0ic2VsZWN0YWJsZSI+VFJVRTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9vYmplY3Q+
CiAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVm
dC1hdHRhY2giPjA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1h
dHRhY2giPjQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgPC9j
aGlsZD4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xh
YmVsIiBpZD0iU05WYWx1ZSI+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRy
YW5zbGF0YWJsZT0ieWVzIj48L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
ImhhbGlnbiI+R1RLX0FMSUdOX1NUQVJUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJqdXN0aWZ5Ij5HVEtfSlVTVElGWV9MRUZUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJzZWxlY3RhYmxlIj5UUlVFPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8
L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJsZWZ0LWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0idG9wLWF0dGFjaCI+NDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAg
ICAgICA8L2NoaWxkPgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgPC9jaGlsZD4KICAgICAgICA8
Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9Imlzc3VlZEJ5TWVzc2Fn
ZSI+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPklz
c3VlZCBieTo8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtf
QUxJR05fU1RBUlQ8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ianVzdGlmeSI+
R1RLX0pVU1RJRllfTEVGVDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzZWxl
Y3RhYmxlIj5UUlVFPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPGF0dHJpYnV0ZXM+CiAgICAgICAgICAg
ICAgPGF0dHJpYnV0ZSBuYW1lPSJ3ZWlnaHQiIHZhbHVlPSJQQU5HT19XRUlHSFRfQk9MRCIvPgogICAg
ICAgICAgICA8L2F0dHJpYnV0ZXM+CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICA8L2NoaWxkPgog
ICAgICAgIDxjaGlsZD4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0dyaWQiIGlkPSJpc3N1ZWRC
eUdyaWQiPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLXRvcCI+MjwvcHJvcGVydHk+
CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tYm90dG9tIj4xMDwvcHJvcGVydHk+CiAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tc3RhcnQiPjEwPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi1lbmQiPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9InJvdy1zcGFjaW5nIj4yPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9ImNvbHVtbi1zcGFjaW5nIj4xMjwvcHJvcGVydHk+CiAgICAgICAgICAgIDxjaGlsZD4K
ICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9Imlzc3VlZEJ5Q25MYWJlbCI+
CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5D
b21tb24gTmFtZSAoQ04pOjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
aGFsaWduIj5HVEtfQUxJR05fU1RBUlQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9Imp1c3RpZnkiPkdUS19KVVNUSUZZX0xFRlQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9InNlbGVjdGFibGUiPlRSVUU8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwv
b2JqZWN0PgogICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9ImxlZnQtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJ0b3AtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAg
ICAgIDwvY2hpbGQ+CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNz
PSJHdGtMYWJlbCIgaWQ9Imlzc3VlZEJ5Q25WYWx1ZSI+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj48L3Byb3BlcnR5PgogICAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+R1RLX0FMSUdOX1NUQVJUPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJqdXN0aWZ5Ij5HVEtfSlVTVElGWV9MRUZUPC9wcm9wZXJ0eT4K
ICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzZWxlY3RhYmxlIj5UUlVFPC9wcm9wZXJ0eT4K
ICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9w
YWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAg
ICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJpc3N1ZWRCeU9MYWJlbCI+CiAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5Pcmdhbml6YXRpb24g
KE8pOjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtf
QUxJR05fU1RBUlQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imp1c3Rp
ZnkiPkdUS19KVVNUSUZZX0xFRlQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9InNlbGVjdGFibGUiPlRSVUU8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAg
ICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0
YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNo
Ij4xPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgIDwvY2hpbGQ+
CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIg
aWQ9Imlzc3VlZEJ5T1ZhbHVlIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIg
dHJhbnNsYXRhYmxlPSJ5ZXMiPjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0iaGFsaWduIj5HVEtfQUxJR05fU1RBUlQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9Imp1c3RpZnkiPkdUS19KVVNUSUZZX0xFRlQ8L3Byb3BlcnR5PgogICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9InNlbGVjdGFibGUiPlRSVUU8L3Byb3BlcnR5PgogICAgICAgICAgICAg
IDwvb2JqZWN0PgogICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9ImxlZnQtYXR0YWNoIj4xPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJ0b3AtYXR0YWNoIj4xPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAg
ICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNs
YXNzPSJHdGtMYWJlbCIgaWQ9Imlzc3VlZEJ5T1VMYWJlbCI+CiAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5Pcmdhbml6YXRpb25hbCBVbml0IChPVSk6
PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoYWxpZ24iPkdUS19BTElH
Tl9TVEFSVDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ianVzdGlmeSI+
R1RLX0pVU1RJRllfTEVGVDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
c2VsZWN0YWJsZSI+VFJVRTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAg
ICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2gi
PjA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjI8
L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgPC9jaGlsZD4KICAg
ICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0i
aXNzdWVkQnlPVVZhbHVlIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJh
bnNsYXRhYmxlPSJ5ZXMiPjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
aGFsaWduIj5HVEtfQUxJR05fU1RBUlQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9Imp1c3RpZnkiPkdUS19KVVNUSUZZX0xFRlQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9InNlbGVjdGFibGUiPlRSVUU8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwv
b2JqZWN0PgogICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9ImxlZnQtYXR0YWNoIj4xPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJ0b3AtYXR0YWNoIj4yPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAg
ICAgIDwvY2hpbGQ+CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICA8L2NoaWxkPgogICAgICAgIDxj
aGlsZD4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0idmFsaWRpdHlNZXNzYWdl
Ij4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+VmFs
aWRpdHk6PC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+R1RLX0FM
SUdOX1NUQVJUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJqdXN0aWZ5
Ij5HVEtfSlVTVElGWV9MRUZUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InNl
bGVjdGFibGUiPlRSVUU8L3Byb3BlcnR5PgogICAgICAgICAgICA8YXR0cmlidXRlcz4KICAgICAgICAg
ICAgICA8YXR0cmlidXRlIG5hbWU9IndlaWdodCIgdmFsdWU9IlBBTkdPX1dFSUdIVF9CT0xEIi8+CiAg
ICAgICAgICAgIDwvYXR0cmlidXRlcz4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgIDwvY2hpbGQ+
CiAgICAgICAgPGNoaWxkPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrR3JpZCIgaWQ9InZhbGlk
aXR5R3JpZCI+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tdG9wIj4yPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi1ib3R0b20iPjEwPC9wcm9wZXJ0eT4K
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi1zdGFydCI+MTA8L3Byb3BlcnR5PgogICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLWVuZCI+MTA8L3Byb3BlcnR5PgogICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0icm93LXNwYWNpbmciPjI8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0iY29sdW1uLXNwYWNpbmciPjEyPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPGNoaWxk
PgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0iaXNzdWVkT25MYWJlbCI+
CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5J
c3N1ZWQgb246PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoYWxpZ24i
PkdUS19BTElHTl9TVEFSVDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
anVzdGlmeSI+R1RLX0pVU1RJRllfTEVGVDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0ic2VsZWN0YWJsZSI+VFJVRTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9vYmplY3Q+
CiAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVm
dC1hdHRhY2giPjA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1h
dHRhY2giPjA8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgPC9j
aGlsZD4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xh
YmVsIiBpZD0iaXNzdWVkT25WYWx1ZSI+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFi
ZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj48L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9ImhhbGlnbiI+R1RLX0FMSUdOX1NUQVJUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJqdXN0aWZ5Ij5HVEtfSlVTVElGWV9MRUZUPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzZWxlY3RhYmxlIj5UUlVFPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0idG9wLWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNraW5nPgog
ICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVj
dCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJleHBpcmVzT25MYWJlbCI+CiAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5FeHBpcmVzIG9uOjwvcHJvcGVydHk+
CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05fU1RBUlQ8L3By
b3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imp1c3RpZnkiPkdUS19KVVNUSUZZ
X0xFRlQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InNlbGVjdGFibGUi
PlRSVUU8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgIDxwYWNr
aW5nPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4wPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4xPC9wcm9wZXJ0eT4K
ICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgIDxj
aGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9ImV4cGlyZXNPblZh
bHVlIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5
ZXMiPjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtf
QUxJR05fU1RBUlQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imp1c3Rp
ZnkiPkdUS19KVVNUSUZZX0xFRlQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9InNlbGVjdGFibGUiPlRSVUU8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAg
ICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0
YWNoIj4xPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNo
Ij4xPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgIDwvY2hpbGQ+
CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICA8L2NoaWxkPgogICAgICAgIDxjaGlsZD4KICAgICAg
ICAgIDxvYmplY3QgY2xhc3M9Ikd0a0dyaWQiIGlkPSJmaW5nZXJwcmludHNHcmlkIj4KICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi10b3AiPjI8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0ibWFyZ2luLWJvdHRvbSI+MTA8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0ibWFyZ2luLXN0YXJ0Ij4xMDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJtYXJnaW4tZW5kIj4xMDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJy
b3ctc3BhY2luZyI+MjwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjb2x1bW4t
c3BhY2luZyI+MTI8L3Byb3BlcnR5PgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9i
amVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJzaGExRmluZ2VycHJpbnRMYWJlbCI+CiAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5TSEEtMSBmaW5nZXJw
cmludDo8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+R1RL
X0FMSUdOX1NUQVJUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJqdXN0
aWZ5Ij5HVEtfSlVTVElGWV9MRUZUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJzZWxlY3RhYmxlIj5UUlVFPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L29iamVjdD4KICAg
ICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0
dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFj
aCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxk
PgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwi
IGlkPSJzaGExRmluZ2VycHJpbnRWYWx1ZSI+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
bGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj48L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9ImhhbGlnbiI+R1RLX0FMSUdOX1NUQVJUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJqdXN0aWZ5Ij5HVEtfSlVTVElGWV9MRUZUPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzZWxlY3RhYmxlIj5UUlVFPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgICAgIDxhdHRyaWJ1dGVzPgogICAgICAgICAgICAgICAgICA8YXR0cmlidXRlIG5hbWU9ImZv
bnQtZGVzYyIgdmFsdWU9Ik1vbm9zcGFjZSIvPgogICAgICAgICAgICAgICAgPC9hdHRyaWJ1dGVzPgog
ICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4xPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L3Bh
Y2tpbmc+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAg
ICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9InNoYTI1NkZpbmdlcnByaW50TGFiZWwiPgogICAg
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+U0hBLTI1
NiBmaW5nZXJwcmludDo8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imhh
bGlnbiI+R1RLX0FMSUdOX1NUQVJUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJqdXN0aWZ5Ij5HVEtfSlVTVElGWV9MRUZUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJzZWxlY3RhYmxlIj5UUlVFPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L29i
amVjdD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJsZWZ0LWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
dG9wLWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAg
ICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0i
R3RrTGFiZWwiIGlkPSJzaGEyNTZGaW5nZXJwcmludFZhbHVlIj4KICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPjwvcHJvcGVydHk+CiAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05fU1RBUlQ8L3Byb3BlcnR5PgogICAg
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imp1c3RpZnkiPkdUS19KVVNUSUZZX0xFRlQ8L3Byb3Bl
cnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InNlbGVjdGFibGUiPlRSVUU8L3Byb3Bl
cnR5PgogICAgICAgICAgICAgICAgPGF0dHJpYnV0ZXM+CiAgICAgICAgICAgICAgICAgIDxhdHRyaWJ1
dGUgbmFtZT0iZm9udC1kZXNjIiB2YWx1ZT0iTW9ub3NwYWNlIi8+CiAgICAgICAgICAgICAgICA8L2F0
dHJpYnV0ZXM+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgPHBhY2tpbmc+CiAg
ICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjE8L3Byb3BlcnR5PgogICAg
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNoaWxkPgog
ICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0ic2hhM18yNTZGaW5nZXJwcmlu
dExhYmVsIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxl
PSJ5ZXMiPlNIQTMtMjU2IGZpbmdlcnByaW50OjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05fU1RBUlQ8L3Byb3BlcnR5PgogICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9Imp1c3RpZnkiPkdUS19KVVNUSUZZX0xFRlQ8L3Byb3BlcnR5PgogICAg
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InNlbGVjdGFibGUiPlRSVUU8L3Byb3BlcnR5PgogICAg
ICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4yPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L3BhY2tp
bmc+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8
b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9InNoYTNfMjU2RmluZ2VycHJpbnRWYWx1ZSI+CiAgICAg
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj48L3Byb3Bl
cnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+R1RLX0FMSUdOX1NUQVJU
PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJqdXN0aWZ5Ij5HVEtfSlVT
VElGWV9MRUZUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzZWxlY3Rh
YmxlIj5UUlVFPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxhdHRyaWJ1dGVzPgogICAgICAgICAg
ICAgICAgICA8YXR0cmlidXRlIG5hbWU9ImZvbnQtZGVzYyIgdmFsdWU9Ik1vbm9zcGFjZSIvPgogICAg
ICAgICAgICAgICAgPC9hdHRyaWJ1dGVzPgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAg
ICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4x
PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4yPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAg
ICAgICA8L29iamVjdD4KICAgICAgICA8L2NoaWxkPgogICAgICAgIDxjaGlsZD4KICAgICAgICAgIDxv
YmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0iYnV0dG9uSW5zdHJ1Y3Rpb25zTWVzc2FnZSI+CiAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkRlcGVuZGluZyBv
biB3aGV0aGVyIHlvdSBqdWRnZSB0aGUgYWJvdmUgaW5mb3JtYXRpb24gY29ycmVjdCwgb3IgaG93IG11
Y2ggcmlzayB5b3UgYXJlIHdpbGxpbmcgdG8gdGFrZSwgeW91IGhhdmUgdGhyZWUgY2hvaWNlczogRGVj
aWRlIHRvIG5vdCBjb25uZWN0LCBjb25uZWN0IHVzaW5nIHRoaXMgaW5mb3JtYXRpb24gdGhpcyB0aW1l
IGJ1dCByZXF1ZXN0IHRoYXQgd2UgYXNrIGFnYWluIG5leHQgdGltZSwgb3IgY29ubmVjdCB1c2luZyB0
aGlzIGluZm9ybWF0aW9uIGFuZCByZW1lbWJlciBpdCBmb3IgdGhlIGZ1dHVyZS48L3Byb3BlcnR5Pgog
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05fU1RBUlQ8L3Byb3BlcnR5
PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ianVzdGlmeSI+R1RLX0pVU1RJRllfTEVGVDwvcHJv
cGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzZWxlY3RhYmxlIj5UUlVFPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9IndyYXAiPnRydWU8L3Byb3BlcnR5PgogICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0ibWF4LXdpZHRoLWNoYXJzIj44MDwvcHJvcGVydHk+CiAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tYm90dG9tIj4xMDwvcHJvcGVydHk+CiAgICAgICAgICA8
L29iamVjdD4KICAgICAgICA8L2NoaWxkPgogICAgICAgIDxjaGlsZCBpbnRlcm5hbC1jaGlsZD0iYWN0
aW9uX2FyZWEiPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQnV0dG9uQm94IiBpZD0iYnV0dG9u
X2JveCI+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJvcmllbnRhdGlvbiI+R1RLX09SSUVOVEFU
SU9OX0hPUklaT05UQUw8L3Byb3BlcnR5PgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAg
PG9iamVjdCBjbGFzcz0iR3RrQnV0dG9uIiBpZD0iYnV0dG9uX2RvbnQiPgogICAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9ImNhbi1kZWZhdWx0Ij50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkRvbid0IGNvbm5lY3Q8L3By
b3BlcnR5PgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAg
ICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQnV0dG9uIiBpZD0iYnV0
dG9uX3RlbXBfb2siPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xh
dGFibGU9InllcyI+QWNjZXB0IHRlbXBvcmFyaWx5PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L29i
amVjdD4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAg
IDxvYmplY3QgY2xhc3M9Ikd0a0J1dHRvbiIgaWQ9ImJ1dHRvbl9wZXJtX29rIj4KICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkFjY2VwdCBwZXJtYW5l
bnRseTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgIDwvY2hpbGQ+
CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICA8L2NoaWxkPgogICAgICA8L29iamVjdD4KICAgIDwv
Y2hpbGQ+CiAgICA8YWN0aW9uLXdpZGdldHM+CiAgICAgIDxhY3Rpb24td2lkZ2V0IHJlc3BvbnNlPSJj
YW5jZWwiIGRlZmF1bHQ9InRydWUiPmJ1dHRvbl9kb250PC9hY3Rpb24td2lkZ2V0PgogICAgICA8YWN0
aW9uLXdpZGdldCByZXNwb25zZT0ib2siPmJ1dHRvbl90ZW1wX29rPC9hY3Rpb24td2lkZ2V0PgogICAg
ICA8YWN0aW9uLXdpZGdldCByZXNwb25zZT0iYWNjZXB0Ij5idXR0b25fcGVybV9vazwvYWN0aW9uLXdp
ZGdldD4KICAgIDwvYWN0aW9uLXdpZGdldHM+CiAgPC9vYmplY3Q+CjwvaW50ZXJmYWNlPgo=
`,
	},

	"/definitions/ChatRoom.xml": {
		local:   "definitions/ChatRoom.xml",
		size:    7669,
		modtime: 1489449600,
		compressed: `
PD94bWwgdmVyc2lvbj0iMS4wIiBlbmNvZGluZz0idXRmLTgiPz4KPGludGVyZmFjZT4KICA8b2JqZWN0
IGNsYXNzPSJHdGtUZXh0VGFnVGFibGUiIGlkPSJjaGF0LWJ1ZmZlci10YWdzIj4KICAgIDxjaGlsZCB0
eXBlPSJ0YWciPgogICAgICA8b2JqZWN0IGNsYXNzPSJHdGtUZXh0VGFnIj4KCSAgPHByb3BlcnR5IG5h
bWU9Im5hbWUiPnRpbWVzdGFtcDwvcHJvcGVydHk+CgkgIDxwcm9wZXJ0eSBuYW1lPSJmb3JlZ3JvdW5k
Ij4jYWFhYWFhPC9wcm9wZXJ0eT4KICAgICAgPC9vYmplY3Q+CiAgICA8L2NoaWxkPgogICAgPGNoaWxk
IHR5cGU9InRhZyI+CiAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a1RleHRUYWciPgoJICA8cHJvcGVydHkg
bmFtZT0ibmFtZSI+aW5jb21pbmdVc2VyPC9wcm9wZXJ0eT4KCSAgPHByb3BlcnR5IG5hbWU9ImZvcmVn
cm91bmQiPmJsdWU8L3Byb3BlcnR5PgogICAgICA8L29iamVjdD4KICAgIDwvY2hpbGQ+CiAgICA8Y2hp
bGQgdHlwZT0idGFnIj4KICAgICAgPG9iamVjdCBjbGFzcz0iR3RrVGV4dFRhZyI+CgkgIDxwcm9wZXJ0
eSBuYW1lPSJuYW1lIj5pbmNvbWluZ1RleHQ8L3Byb3BlcnR5PgoJICA8cHJvcGVydHkgbmFtZT0iZm9y
ZWdyb3VuZCI+YmxhY2s8L3Byb3BlcnR5PgogICAgICA8L29iamVjdD4KICAgIDwvY2hpbGQ+CiAgPC9v
YmplY3Q+CiAgPG9iamVjdCBjbGFzcz0iR3RrVGV4dEJ1ZmZlciIgaWQ9ImNoYXQtYnVmZmVyIj4KICAg
IDxwcm9wZXJ0eSBuYW1lPSJ0YWctdGFibGUiPmNoYXQtYnVmZmVyLXRhZ3M8L3Byb3BlcnR5PgogIDwv
b2JqZWN0PgogIDxvYmplY3QgY2xhc3M9Ikd0a0xpc3RTdG9yZSIgaWQ9Im9jY3VwYW50cyI+CiAgICA8
Y29sdW1ucz4KICAgICAgPCEtLSBoYW5kbGUgLS0+CiAgICAgIDxjb2x1bW4gdHlwZT0iZ2NoYXJhcnJh
eSIvPgogICAgICA8IS0tIHJvbGUgLS0+CiAgICAgIDxjb2x1bW4gdHlwZT0iZ2NoYXJhcnJheSIvPgog
ICAgICA8IS0tIGFmZmlsaWF0aW9uIC0tPgogICAgICA8Y29sdW1uIHR5cGU9ImdjaGFyYXJyYXkiLz4K
ICAgIDwvY29sdW1ucz4KICA8L29iamVjdD4KCiAgPG9iamVjdCBjbGFzcz0iR3RrV2luZG93IiBpZD0i
bXVjLXdpbmRvdyI+CiAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+ZmFsc2U8L3Byb3BlcnR5Pgog
ICAgPHByb3BlcnR5IG5hbWU9IndpbmRvdy1wb3NpdGlvbiI+R1RLX1dJTl9QT1NfQ0VOVEVSPC9wcm9w
ZXJ0eT4KICAgIDxwcm9wZXJ0eSBuYW1lPSJib3JkZXJfd2lkdGgiPjY8L3Byb3BlcnR5PgogICAgPHBy
b3BlcnR5IG5hbWU9InJlc2l6YWJsZSI+dHJ1ZTwvcHJvcGVydHk+CiAgICA8cHJvcGVydHkgbmFtZT0i
ZGVmYXVsdC1oZWlnaHQiPjQwMDwvcHJvcGVydHk+CiAgICA8cHJvcGVydHkgbmFtZT0iZGVmYXVsdC13
aWR0aCI+NzAwPC9wcm9wZXJ0eT4KICAgIDxwcm9wZXJ0eSBuYW1lPSJkZXN0cm95LXdpdGgtcGFyZW50
Ij50cnVlPC9wcm9wZXJ0eT4KICAgIDxzaWduYWwgbmFtZT0iZGVzdHJveSIgaGFuZGxlcj0ibGVhdmVf
cm9vbV9oYW5kbGVyIiAvPgogICAgPGNoaWxkPgogICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCb3giPgoJ
PHByb3BlcnR5IG5hbWU9InZpc2libGUiPnRydWU8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJvcmllbnRhdGlvbiI+R1RLX09SSUVOVEFUSU9OX1ZFUlRJQ0FMPC9wcm9wZXJ0eT4KICAgICAg
ICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9InN1YmplY3QiPgog
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+ZmFsc2U8L3Byb3BlcnR5PgogICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0iaGV4cGFuZCI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJtYXJnaW4tYm90dG9tIj42PC9wcm9wZXJ0eT4KICAgICAgICAgIDwvb2JqZWN0
PgogICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgPC9jaGlsZD4K
ICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCb3giIGlkPSJtZW51LWJv
eCI+CgkgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgIDxj
aGlsZD4KICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTWVudUJhciI+CiAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9InZpc2libGUiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDxjaGls
ZD4KICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a01lbnVJdGVtIj4KICAgICAgICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAg
ICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiPgog
ICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPlRydWU8L3Byb3BlcnR5
PgogICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9
InllcyI+U2V0dGluZ3M8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9InVzZS11bmRlcmxpbmUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDwvb2Jq
ZWN0PgogICAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgICAgICA8Y2hpbGQgdHlw
ZT0ic3VibWVudSI+CiAgICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTWVudSI+CiAg
ICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+dHJ1ZTwvcHJvcGVydHk+
CiAgICAgICAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgICAgIDxvYmpl
Y3QgY2xhc3M9Ikd0a01lbnVJdGVtIj4KICAgICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0idmlzaWJsZSI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+Q29uZmlndXJlIFJvb208L3Byb3Bl
cnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgIDxzaWduYWwgbmFtZT0iYWN0aXZhdGUiIGhhbmRs
ZXI9Im9uX2NoYW5nZV9yb29tX2NvbmZpZyIgLz4KICAgICAgICAgICAgICAgICAgICAgICAgPC9vYmpl
Y3Q+CiAgICAgICAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgICAgICAgIDwvb2Jq
ZWN0PgogICAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAg
ICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICA8L2NoaWxk
PgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9ImV4cGFuZCI+ZmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0iZmlsbCI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwYWNrLXR5
cGUiPkdUS19QQUNLX1NUQVJUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InBv
c2l0aW9uIj4xPC9wcm9wZXJ0eT4KICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICA8L2NoaWxkPgoJ
PGNoaWxkPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrR3JpZCI+CiAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJoZXhwYW5kIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9InZleHBhbmQiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlz
aWJsZSI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjb2x1bW4taG9t
b2dlbmVvdXMiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icm93LWhv
bW9nZW5lb3VzIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNvbHVt
bi1zcGFjaW5nIj42PC9wcm9wZXJ0eT4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxv
YmplY3QgY2xhc3M9Ikd0a1Njcm9sbGVkV2luZG93IiBpZD0iY2hhdC1ib3giPgogICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAg
PGNoaWxkPgogICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtUZXh0VmlldyI+CiAgICAg
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPnRydWU8L3Byb3BlcnR5PgogICAg
ICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ3cmFwLW1vZGUiPkdUS19XUkFQX1dPUkRfQ0hB
UjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImVkaXRhYmxlIj5G
YWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImN1cnNvci12
aXNpYmxlIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
ImJ1ZmZlciI+Y2hhdC1idWZmZXI8L3Byb3BlcnR5PgogICAgICAgICAgICAJICAgIDxzaWduYWwgbmFt
ZT0ic2l6ZS1hbGxvY2F0ZSIgaGFuZGxlcj0ic2Nyb2xsX2hpc3RvcnlfdG9fYm90dG9tX2hhbmRsZXIi
IC8+CiAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgICAgPC9jaGlsZD4KICAg
ICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0id2lkdGgiPjM8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvcGFja2luZz4KICAg
ICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3Qg
Y2xhc3M9Ikd0a1Njcm9sbGVkV2luZG93IiBpZD0ib2NjdXBhbnRzLWxpc3QiPgogICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAg
PGNoaWxkPgogICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtUcmVlVmlldyIgaWQ9Im9j
Y3VwYW50cy12aWV3Ij4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+
dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1vZGVsIj5v
Y2N1cGFudHM8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAg
ICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a1RyZWVWaWV3Q29sdW1uIj4KICAgICAgICAgICAgICAg
ICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRpdGxlIj5Vc2VyPC9wcm9wZXJ0eT4KICAgICAgICAgICAg
ICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9
Ikd0a0NlbGxSZW5kZXJlclRleHQiLz4KICAgICAgICAgICAgICAgICAgICAgICAgICA8YXR0cmlidXRl
cz4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxhdHRyaWJ1dGUgbmFtZT0idGV4dCI+MDwvYXR0
cmlidXRlPgogICAgICAgICAgICAgICAgICAgICAgICAgIDwvYXR0cmlidXRlcz4KICAgICAgICAgICAg
ICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAg
ICAgICAgICAgICAgIDwvY2hpbGQ+CgogICAgICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAg
ICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a1RyZWVWaWV3Q29sdW1uIj4KICAgICAgICAgICAg
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRpdGxlIj5Sb2xlPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xh
c3M9Ikd0a0NlbGxSZW5kZXJlclRleHQiLz4KICAgICAgICAgICAgICAgICAgICAgICAgICA8YXR0cmli
dXRlcz4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxhdHRyaWJ1dGUgbmFtZT0idGV4dCI+MTwv
YXR0cmlidXRlPgogICAgICAgICAgICAgICAgICAgICAgICAgIDwvYXR0cmlidXRlcz4KICAgICAgICAg
ICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAg
ICAgICAgICAgICAgICAgIDwvY2hpbGQ+CgoKICAgICAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAg
ICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtUcmVlVmlld0NvbHVtbiI+CiAgICAgICAg
ICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0aXRsZSI+QWZmaWxpYXRpb248L3Byb3BlcnR5
PgogICAgICAgICAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgICAgICAg
PG9iamVjdCBjbGFzcz0iR3RrQ2VsbFJlbmRlcmVyVGV4dCIvPgogICAgICAgICAgICAgICAgICAgICAg
ICAgIDxhdHRyaWJ1dGVzPgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPGF0dHJpYnV0ZSBuYW1l
PSJ0ZXh0Ij4yPC9hdHRyaWJ1dGU+CiAgICAgICAgICAgICAgICAgICAgICAgICAgPC9hdHRyaWJ1dGVz
PgogICAgICAgICAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgICAgICAgICAgPC9v
YmplY3Q+CiAgICAgICAgICAgICAgICAgICAgPC9jaGlsZD4KCiAgICAgICAgICAgICAgICAgIDwvb2Jq
ZWN0PgogICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAg
ICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFj
aCI+MzwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+
MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0id2lkdGgiPjE8L3Byb3Bl
cnR5PgogICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAg
IDwvb2JqZWN0PgoJPC9jaGlsZD4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNz
PSJHdGtFbnRyeSIgaWQ9InRleHQtYm94Ij4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2li
bGUiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGV4cGFuZCI+dHJ1
ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tdG9wIj42PC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imhhcy1mb2N1cyI+dHJ1ZTwvcHJvcGVydHk+
CgkgICAgPHNpZ25hbCBuYW1lPSJhY3RpdmF0ZSIgaGFuZGxlcj0ic2VuZF9tZXNzYWdlX2hhbmRsZXIi
IC8+CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgPC9wYWNr
aW5nPgogICAgICAgIDwvY2hpbGQ+CiAgICAgIDwvb2JqZWN0PgogICAgICA8cGFja2luZz4KICAgICAg
PC9wYWNraW5nPgogICAgPC9jaGlsZD4KICA8L29iamVjdD4KPC9pbnRlcmZhY2U+Cg==
`,
	},

	"/definitions/ChooseKeyToImport.xml": {
		local:   "definitions/ChooseKeyToImport.xml",
		size:    2047,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtEaWFsb2ciIGlkPSJkaWFsb2ciPgogICAgPHBy
b3BlcnR5IG5hbWU9InRpdGxlIiB0cmFuc2xhdGFibGU9InllcyI+Q2hvb3NlIGEga2V5IHRvIGltcG9y
dDwvcHJvcGVydHk+CiAgICA8c2lnbmFsIG5hbWU9ImNsb3NlIiBoYW5kbGVyPSJvbl9jYW5jZWxfc2ln
bmFsIiAvPgogICAgPHNpZ25hbCBuYW1lPSJkZWxldGUtZXZlbnQiIGhhbmRsZXI9Im9uX2NhbmNlbF9z
aWduYWwiIC8+CiAgICA8Y2hpbGQgaW50ZXJuYWwtY2hpbGQ9InZib3giPgogICAgICA8b2JqZWN0IGNs
YXNzPSJHdGtCb3giIGlkPSJWYm94Ij4KICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luIj4xMDwv
cHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9InNwYWNpbmciPjEwPC9wcm9wZXJ0eT4KICAg
ICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9Im1lc3NhZ2Ui
PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5UaGUg
ZmlsZSB5b3UgY2hvb3NlIGNvbnRhaW5zIG1vcmUgdGhhbiBvbmUgcHJpdmF0ZSBrZXkuIENob29zZSBm
cm9tIHRoZSBsaXN0IGJlbG93IHRoZSBrZXkgeW91IHdvdWxkIGxpa2UgdG8gaW1wb3J0LjwvcHJvcGVy
dHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ3cmFwIj50cnVlPC9wcm9wZXJ0eT4KICAgICAg
ICAgIDwvb2JqZWN0PgogICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJleHBhbmQiPmZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImZpbGwi
PnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgPC9wYWNraW5nPgogICAgICAgIDwvY2hpbGQ+CiAgICAg
ICAgPGNoaWxkPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQ29tYm9Cb3hUZXh0IiBpZD0ia2V5
cyI+CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICA8L2NoaWxkPgogICAgICAgIDxjaGlsZCBpbnRl
cm5hbC1jaGlsZD0iYWN0aW9uX2FyZWEiPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQnV0dG9u
Qm94IiBpZD0iYmJveCI+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJvcmllbnRhdGlvbiI+R1RL
X09SSUVOVEFUSU9OX0hPUklaT05UQUw8L3Byb3BlcnR5PgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAg
ICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQnV0dG9uIiBpZD0iYnRuLWNhbmNlbCI+CiAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5DYW5jZWw8L3By
b3BlcnR5PgogICAgICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJjbGlja2VkIiBoYW5kbGVyPSJvbl9j
YW5jZWxfc2lnbmFsIiAvPgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICA8L2NoaWxk
PgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQnV0dG9u
IiBpZD0iYnRuLWltcG9ydCI+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRy
YW5zbGF0YWJsZT0ieWVzIj5JbXBvcnQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9ImNhbi1kZWZhdWx0Ij50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxzaWduYWwg
bmFtZT0iY2xpY2tlZCIgaGFuZGxlcj0ib25faW1wb3J0X3NpZ25hbCIgLz4KICAgICAgICAgICAgICA8
L29iamVjdD4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgIDwv
Y2hpbGQ+CiAgICAgIDwvb2JqZWN0PgogICAgPC9jaGlsZD4KICAgIDxhY3Rpb24td2lkZ2V0cz4KICAg
ICAgPGFjdGlvbi13aWRnZXQgcmVzcG9uc2U9ImNhbmNlbCI+YnRuLWNhbmNlbDwvYWN0aW9uLXdpZGdl
dD4KICAgICAgPGFjdGlvbi13aWRnZXQgcmVzcG9uc2U9Im9rIiBkZWZhdWx0PSJ0cnVlIj5idG4taW1w
b3J0PC9hY3Rpb24td2lkZ2V0PgogICAgPC9hY3Rpb24td2lkZ2V0cz4KICA8L29iamVjdD4KPC9pbnRl
cmZhY2U+Cg==
`,
	},

	"/definitions/ConfigureRoom.xml": {
		local:   "definitions/ConfigureRoom.xml",
		size:    2489,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtEaWFsb2ciIGlkPSJkaWFsb2ciPgogICAgPHBy
b3BlcnR5IG5hbWU9IndpbmRvdy1wb3NpdGlvbiI+R1RLX1dJTl9QT1NfQ0VOVEVSPC9wcm9wZXJ0eT4K
ICAgIDxwcm9wZXJ0eSBuYW1lPSJib3JkZXItd2lkdGgiPjc8L3Byb3BlcnR5PgogICAgPHByb3BlcnR5
IG5hbWU9InRpdGxlIiB0cmFuc2xhdGFibGU9InllcyI+Q29uZmlndXJlIFJvb208L3Byb3BlcnR5Pgog
ICAgPGNoaWxkIGludGVybmFsLWNoaWxkPSJ2Ym94Ij4KICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQm94
IiBpZD0iYm94Ij4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iYm9yZGVyLXdpZHRoIj4xMDwvcHJvcGVy
dHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9Im9yaWVudGF0aW9uIj5HVEtfT1JJRU5UQVRJT05fVkVS
VElDQUw8L3Byb3BlcnR5PgogICAgICAgIDxjaGlsZD4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0
a0xhYmVsIiBpZD0ibGFiZWwiPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5z
bGF0YWJsZT0ieWVzIj5Db25maWd1cmUgc29tZSByb29tIHNwZWNpZmljIHNldHRpbmdzOjwvcHJvcGVy
dHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJqdXN0aWZ5Ij5HVEtfSlVTVElGWV9SSUdIVDwv
cHJvcGVydHk+CiAgICAgICAgICAgIDxhdHRyaWJ1dGVzPgogICAgICAgICAgICAgIDxhdHRyaWJ1dGUg
bmFtZT0id2VpZ2h0IiB2YWx1ZT0ic2VtaWJvbGQiLz4KICAgICAgICAgICAgPC9hdHRyaWJ1dGVzPgog
ICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9ImV4cGFuZCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
ZmlsbCI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+
MDwvcHJvcGVydHk+CiAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgPC9jaGlsZD4KICAgICAgICA8
Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtHcmlkIiBpZD0iZ3JpZCI+CiAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tdG9wIj4yMDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJtYXJnaW4tYm90dG9tIj4yMDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJtYXJnaW4tc3RhcnQiPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9Im1hcmdpbi1lbmQiPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
InJvdy1zcGFjaW5nIj4xMDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjb2x1
bW4tc3BhY2luZyI+MzA8L3Byb3BlcnR5PgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICA8cGFj
a2luZz4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+RmFsc2U8L3Byb3BlcnR5Pgog
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZmlsbCI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MTwvcHJvcGVydHk+CiAgICAgICAgICA8L3BhY2tpbmc+
CiAgICAgICAgPC9jaGlsZD4KICAgICAgICA8Y2hpbGQgaW50ZXJuYWwtY2hpbGQ9ImFjdGlvbl9hcmVh
Ij4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0J1dHRvbkJveCIgaWQ9ImJ1dHRvbl9ib3giPgog
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ib3JpZW50YXRpb24iPkdUS19PUklFTlRBVElPTl9IT1JJ
Wk9OVEFMPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3Qg
Y2xhc3M9Ikd0a0J1dHRvbiIgaWQ9ImNhbmNlbCI+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5DYW5jZWw8L3Byb3BlcnR5PgogICAgICAgICAgICAg
ICAgPHNpZ25hbCBuYW1lPSJjbGlja2VkIiBoYW5kbGVyPSJvbl9jYW5jZWxfc2lnbmFsIiAvPgogICAg
ICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+
CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQnV0dG9uIiBpZD0ic2F2ZSI+CiAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5TYXZlPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW4tZGVmYXVsdCI+dHJ1ZTwvcHJv
cGVydHk+CiAgICAgICAgICAgICAgICA8c2lnbmFsIG5hbWU9ImNsaWNrZWQiIGhhbmRsZXI9Im9uX3Nh
dmVfc2lnbmFsIi8+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAg
ICAgICAgICA8L29iamVjdD4KICAgICAgICA8L2NoaWxkPgogICAgICA8L29iamVjdD4KICAgIDwvY2hp
bGQ+CiAgPC9vYmplY3Q+CjwvaW50ZXJmYWNlPgo=
`,
	},

	"/definitions/ConfirmAccountRemoval.xml": {
		local:   "definitions/ConfirmAccountRemoval.xml",
		size:    557,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtNZXNzYWdlRGlhbG9nIiBpZD0iUmVtb3ZlQWNj
b3VudCI+CiAgICA8cHJvcGVydHkgbmFtZT0idGl0bGUiIHRyYW5zbGF0YWJsZT0ieWVzIj5Db25maXJt
IGFjY291bnQgcmVtb3ZhbDwvcHJvcGVydHk+CiAgICA8cHJvcGVydHkgbmFtZT0iYm9yZGVyX3dpZHRo
Ij43PC9wcm9wZXJ0eT4KICAgIDxwcm9wZXJ0eSBuYW1lPSJ0ZXh0IiB0cmFuc2xhdGFibGU9InllcyI+
QXJlIHlvdSBzdXJlIHlvdSB3YW50IHRvIHJlbW92ZSB0aGlzIGFjY291bnQ/PC9wcm9wZXJ0eT4KICAg
IDxwcm9wZXJ0eSBuYW1lPSJ3aW5kb3ctcG9zaXRpb24iPkdUS19XSU5fUE9TX0NFTlRFUjwvcHJvcGVy
dHk+CiAgICA8cHJvcGVydHkgbmFtZT0ibW9kYWwiPnRydWU8L3Byb3BlcnR5PgogICAgPHByb3BlcnR5
IG5hbWU9Im1lc3NhZ2UtdHlwZSI+R1RLX01FU1NBR0VfUVVFU1RJT048L3Byb3BlcnR5PgogICAgPHBy
b3BlcnR5IG5hbWU9ImJ1dHRvbnMiPkdUS19CVVRUT05TX1lFU19OTzwvcHJvcGVydHk+CiAgPC9vYmpl
Y3Q+CjwvaW50ZXJmYWNlPgo=
`,
	},

	"/definitions/ConnectingAccountInfo.xml": {
		local:   "definitions/ConnectingAccountInfo.xml",
		size:    873,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtJbmZvQmFyIiBpZD0iaW5mb2JhciI+CiAgICA8
cHJvcGVydHkgbmFtZT0ibWVzc2FnZS10eXBlIj5HVEtfTUVTU0FHRV9JTkZPPC9wcm9wZXJ0eT4KICAg
IDxjaGlsZCBpbnRlcm5hbC1jaGlsZD0iY29udGVudF9hcmVhIj4KICAgICAgPG9iamVjdCBjbGFzcz0i
R3RrQm94IiBpZD0iYm94Ij4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iaG9tb2dlbmVvdXMiPmZhbHNl
PC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0ib3JpZW50YXRpb24iPkdUS19PUklFTlRB
VElPTl9IT1JJWk9OVEFMPC9wcm9wZXJ0eT4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0
IGNsYXNzPSJHdGtMYWJlbCIgaWQ9Im1lc3NhZ2UiPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
ZWxsaXBzaXplIj5QQU5HT19FTExJUFNJWkVfRU5EPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9IndyYXAiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAg
PC9jaGlsZD4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtTcGlubmVy
IiBpZD0ic3Bpbm5lciI+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJhY3RpdmUiPnRydWU8L3By
b3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05fRU5EPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhleHBhbmQiPnRydWU8L3Byb3BlcnR5
PgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgPC9jaGlsZD4KICAgICAgPC9vYmplY3Q+CiAgICA8
L2NoaWxkPgogIDwvb2JqZWN0Pgo8L2ludGVyZmFjZT4K
`,
	},

	"/definitions/ConnectionFailureNotification.xml": {
		local:   "definitions/ConnectionFailureNotification.xml",
		size:    3114,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtJbmZvQmFyIiBpZD0iaW5mb2JhciI+CiAgICA8
cHJvcGVydHkgbmFtZT0ibWVzc2FnZS10eXBlIj5HVEtfTUVTU0FHRV9FUlJPUjwvcHJvcGVydHk+CiAg
ICA8cHJvcGVydHkgbmFtZT0ic2hvdy1jbG9zZS1idXR0b24iPmZhbHNlPC9wcm9wZXJ0eT4KICAgIDxj
aGlsZCBpbnRlcm5hbC1jaGlsZD0iY29udGVudF9hcmVhIj4KICAgICAgPG9iamVjdCBjbGFzcz0iR3Rr
Qm94IiBpZD0iYm94Ij4KICAgICAgICA8cHJvcGVydHkgbmFtZT0ic3BhY2luZyI+MDwvcHJvcGVydHk+
CiAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhvbW9nZW5lb3VzIj5mYWxzZTwvcHJvcGVydHk+CiAgICAg
ICAgPHByb3BlcnR5IG5hbWU9Im9yaWVudGF0aW9uIj5HVEtfT1JJRU5UQVRJT05fSE9SSVpPTlRBTDwv
cHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhleHBhbmQiPlRydWU8L3Byb3BlcnR5Pgog
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tc3RhcnQiPjA8L3Byb3BlcnR5PgogICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJtYXJnaW4tZW5kIj4wPC9wcm9wZXJ0eT4KICAgICAgICA8Y2hpbGQ+CiAgICAg
ICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9Im1lc3NhZ2UiPgogICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0ic2VsZWN0YWJsZSI+VFJVRTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJlbGxpcHNpemUiPlBBTkdPX0VMTElQU0laRV9FTkQ8L3Byb3BlcnR5PgogICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0iaGV4cGFuZCI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJ3cmFwIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
Im1hcmdpbi1yaWdodCI+MTA8L3Byb3BlcnR5PgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICA8
cGFja2luZz4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+VHJ1ZTwvcHJvcGVydHk+
CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MDwvcHJvcGVydHk+CiAgICAgICAg
ICA8L3BhY2tpbmc+CiAgICAgICAgPC9jaGlsZD4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2Jq
ZWN0IGNsYXNzPSJHdGtCdXR0b24iIGlkPSJidXR0b25fbW9yZV9pbmZvIj4KICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9ImhleHBhbmQiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9ImhhbGlnbiI+ZW5kPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InJl
bGllZiI+bm9uZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2YWxpZ24iPmNl
bnRlcjwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJib3JkZXJfd2lkdGgiPjA8
L3Byb3BlcnR5PgogICAgICAgICAgICA8c3R5bGU+CiAgICAgICAgICAgICAgPGNsYXNzIG5hbWU9InJh
aXNlZCIvPgogICAgICAgICAgICAgIDxjbGFzcyBuYW1lPSJjbG9zZSIvPgogICAgICAgICAgICA8L3N0
eWxlPgogICAgICAgICAgICA8c2lnbmFsIG5hbWU9ImNsaWNrZWQiIGhhbmRsZXI9Im9uX21vcmVfaW5m
b19zaWduYWwiIC8+CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNz
PSJHdGtJbWFnZSIgaWQ9Im1vcmUtaW5mby1idG4iPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9Imljb24tbmFtZSI+aGVscC1mYXE8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvb2JqZWN0Pgog
ICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICA8cGFja2luZz4K
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjE8L3Byb3BlcnR5PgogICAgICAgICAgPC9wYWNr
aW5nPgogICAgICAgIDwvY2hpbGQ+CiAgICAgICAgPGNoaWxkPgogICAgICAgICAgPG9iamVjdCBjbGFz
cz0iR3RrQnV0dG9uIiBpZD0iYnV0dG9uX2Nsb3NlIj4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
ImNhbl9mb2N1cyI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJib3JkZXJf
d2lkdGgiPjI8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icmVsaWVmIj5ub25l
PC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZhbGlnbiI+Y2VudGVyPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgPHN0eWxlPgogICAgICAgICAgICAgIDxjbGFzcyBuYW1lPSJyYWlzZWQi
Lz4KICAgICAgICAgICAgICA8Y2xhc3MgbmFtZT0iY2xvc2UiLz4KICAgICAgICAgICAgPC9zdHlsZT4K
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhleHBhbmQiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+ZW5kPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9InZhbGlnbiI+Y2VudGVyPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHNpZ25hbCBu
YW1lPSJjbGlja2VkIiBoYW5kbGVyPSJvbl9jbG9zZV9zaWduYWwiIC8+CiAgICAgICAgICAgIDxjaGls
ZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtJbWFnZSIgaWQ9ImNsb3NlX2ltYWdlIj4K
ICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJpY29uX25hbWUiPndpbmRvdy1jbG9zZS1zeW1ib2xp
YzwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAg
ICAgICAgICA8L29iamVjdD4KICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0iZXhwYW5kIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJw
b3NpdGlvbiI+MjwvcHJvcGVydHk+CiAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgPC9jaGlsZD4K
ICAgICAgPC9vYmplY3Q+CiAgICA8L2NoaWxkPgogIDwvb2JqZWN0Pgo8L2ludGVyZmFjZT4K
`,
	},

	"/definitions/ConnectionInformation.xml": {
		local:   "definitions/ConnectionInformation.xml",
		size:    6424,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtEaWFsb2ciIGlkPSJDb25uZWN0aW9uSW5mb3Jt
YXRpb24iPgogICAgPHByb3BlcnR5IG5hbWU9InRpdGxlIiB0cmFuc2xhdGFibGU9InllcyI+Q29ubmVj
dGlvbiBJbmZvcm1hdGlvbjwvcHJvcGVydHk+CiAgICA8c2lnbmFsIG5hbWU9ImNsb3NlIiBoYW5kbGVy
PSJvbl9jbG9zZV9zaWduYWwiIC8+CiAgICA8Y2hpbGQgaW50ZXJuYWwtY2hpbGQ9InZib3giPgogICAg
ICA8b2JqZWN0IGNsYXNzPSJHdGtCb3giIGlkPSJWYm94Ij4KICAgICAgICA8cHJvcGVydHkgbmFtZT0i
bWFyZ2luIj4xMDwvcHJvcGVydHk+CiAgICAgICAgPGNoaWxkPgogICAgICAgICAgPG9iamVjdCBjbGFz
cz0iR3RrR3JpZCIgaWQ9ImNvbm5lY3Rpb25JbmZvR3JpZCI+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJtYXJnaW4tdG9wIj4yPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1h
cmdpbi1ib3R0b20iPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdp
bi1zdGFydCI+MTA8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLWVu
ZCI+MTA8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icm93LXNwYWNpbmciPjI8
L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY29sdW1uLXNwYWNpbmciPjEyPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0
a0xhYmVsIiBpZD0ic2VydmVyTGFiZWwiPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imxh
YmVsIiB0cmFuc2xhdGFibGU9InllcyI+U2VydmVyIGNvbm5lY3RlZCB0bzo8L3Byb3BlcnR5PgogICAg
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+R1RLX0FMSUdOX1NUQVJUPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJqdXN0aWZ5Ij5HVEtfSlVTVElGWV9MRUZU
PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzZWxlY3RhYmxlIj5UUlVF
PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2luZz4K
ICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MDwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAg
ICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+
CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJzZXJ2ZXJWYWx1ZSI+CiAg
ICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj48L3By
b3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+R1RLX0FMSUdOX1NU
QVJUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJqdXN0aWZ5Ij5HVEtf
SlVTVElGWV9MRUZUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzZWxl
Y3RhYmxlIj5UUlVFPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAg
ICA8cGFja2luZz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MTwv
cHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+MDwvcHJv
cGVydHk+CiAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAg
ICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJ0bHNW
ZXJzaW9uTGFiZWwiPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xh
dGFibGU9InllcyI+U2VjdXJlIGNvbm5lY3Rpb24gdmVyc2lvbjo8L3Byb3BlcnR5PgogICAgICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+R1RLX0FMSUdOX1NUQVJUPC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJqdXN0aWZ5Ij5HVEtfSlVTVElGWV9MRUZUPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzZWxlY3RhYmxlIj5UUlVFPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAg
ICAgPC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAg
ICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJ0bHNWZXJzaW9uVmFsdWUiPgogICAg
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+PC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoYWxpZ24iPkdUS19BTElHTl9TVEFS
VDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ianVzdGlmeSI+R1RLX0pV
U1RJRllfTEVGVDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ic2VsZWN0
YWJsZSI+VFJVRTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAg
PHBhY2tpbmc+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjE8L3By
b3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjE8L3Byb3Bl
cnR5PgogICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAg
ICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0idGxzQWxn
b0xhYmVsIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxl
PSJ5ZXMiPlNlY3VyZSBjb25uZWN0aW9uIGFsZ29yaXRobTo8L3Byb3BlcnR5PgogICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+R1RLX0FMSUdOX1NUQVJUPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJqdXN0aWZ5Ij5HVEtfSlVTVElGWV9MRUZUPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzZWxlY3RhYmxlIj5UUlVFPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+MjwvcHJvcGVydHk+CiAgICAgICAgICAgICAg
PC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAg
ICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJ0bHNBbGdvVmFsdWUiPgogICAgICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+PC9wcm9wZXJ0eT4K
ICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoYWxpZ24iPkdUS19BTElHTl9TVEFSVDwvcHJv
cGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ianVzdGlmeSI+R1RLX0pVU1RJRllf
TEVGVDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ic2VsZWN0YWJsZSI+
VFJVRTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgPHBhY2tp
bmc+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjE8L3Byb3BlcnR5
PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjI8L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNo
aWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0idGxzRmluZ2VycHJp
bnRMYWJlbCI+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJs
ZT0ieWVzIj5TZWN1cmUgY29ubmVjdGlvbiBmaW5nZXJwcmludDo8L3Byb3BlcnR5PgogICAgICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9InZhbGlnbiI+R1RLX0FMSUdOX1NUQVJUPC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoYWxpZ24iPkdUS19BTElHTl9TVEFSVDwvcHJvcGVy
dHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ianVzdGlmeSI+R1RLX0pVU1RJRllfTEVG
VDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ic2VsZWN0YWJsZSI+VFJV
RTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgPHBhY2tpbmc+
CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjA8L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjM8L3Byb3BlcnR5PgogICAg
ICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNoaWxk
PgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0idGxzRmluZ2VycHJpbnRW
YWx1ZSI+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0i
eWVzIj48L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+R1RL
X0FMSUdOX1NUQVJUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJqdXN0
aWZ5Ij5HVEtfSlVTVElGWV9MRUZUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJzZWxlY3RhYmxlIj5UUlVFPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxhdHRyaWJ1dGVz
PgogICAgICAgICAgICAgICAgICA8YXR0cmlidXRlIG5hbWU9ImZvbnQtZGVzYyIgdmFsdWU9Ik1vbm9z
cGFjZSIvPgogICAgICAgICAgICAgICAgPC9hdHRyaWJ1dGVzPgogICAgICAgICAgICAgIDwvb2JqZWN0
PgogICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imxl
ZnQtYXR0YWNoIj4xPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3At
YXR0YWNoIj4zPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgIDwv
Y2hpbGQ+CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtC
dXR0b24iIGlkPSJwaW4tY2VydCI+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwi
IHRyYW5zbGF0YWJsZT0ieWVzIj5QaW4gY2VydGlmaWNhdGU8L3Byb3BlcnR5PgogICAgICAgICAgICAg
ICAgPHNpZ25hbCBuYW1lPSJjbGlja2VkIiBoYW5kbGVyPSJvbl9waW5fc2lnbmFsIi8+CiAgICAgICAg
ICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9InRvcC1hdHRhY2giPjQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvcGFja2luZz4K
ICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgIDwvY2hpbGQ+CiAg
ICAgIDwvb2JqZWN0PgogICAgPC9jaGlsZD4KICAgIDxjaGlsZCBpbnRlcm5hbC1jaGlsZD0iYWN0aW9u
X2FyZWEiPgogICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCdXR0b25Cb3giIGlkPSJidXR0b25fYm94Ij4K
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ib3JpZW50YXRpb24iPkdUS19PUklFTlRBVElPTl9IT1JJWk9O
VEFMPC9wcm9wZXJ0eT4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtC
dXR0b24iIGlkPSJjbG9zZSI+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNs
YXRhYmxlPSJ5ZXMiPkNsb3NlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJjbGlj
a2VkIiBoYW5kbGVyPSJvbl9jbG9zZV9zaWduYWwiLz4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAg
IDwvY2hpbGQ+CiAgICAgIDwvb2JqZWN0PgogICAgPC9jaGlsZD4KICA8L29iamVjdD4KPC9pbnRlcmZh
Y2U+Cg==
`,
	},

	"/definitions/ContactPopupMenu.xml": {
		local:   "definitions/ContactPopupMenu.xml",
		size:    2549,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtNZW51IiBpZD0iY29udGFjdE1lbnUiPgogICAg
PGNoaWxkPgogICAgICA8b2JqZWN0IGNsYXNzPSJHdGtNZW51SXRlbSIgaWQ9ImVkaXRDb250YWN0TWVu
dUl0ZW0iPgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkVk
aXQuLi48L3Byb3BlcnR5PgogICAgICAgIDxzaWduYWwgbmFtZT0iYWN0aXZhdGUiIGhhbmRsZXI9Im9u
X2VkaXRfY29udGFjdCIgLz4KICAgICAgPC9vYmplY3Q+CiAgICA8L2NoaWxkPgogICAgPGNoaWxkPgog
ICAgICA8b2JqZWN0IGNsYXNzPSJHdGtNZW51SXRlbSIgaWQ9InJlbW92ZUNvbnRhY3RNZW51SXRlbSI+
CiAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+UmVtb3ZlPC9w
cm9wZXJ0eT4KICAgICAgICA8c2lnbmFsIG5hbWU9ImFjdGl2YXRlIiBoYW5kbGVyPSJvbl9yZW1vdmVf
Y29udGFjdCIgLz4KICAgICAgPC9vYmplY3Q+CiAgICA8L2NoaWxkPgogICAgPGNoaWxkPgogICAgICA8
b2JqZWN0IGNsYXNzPSJHdGtTZXBhcmF0b3JNZW51SXRlbSIgLz4KICAgIDwvY2hpbGQ+CiAgICA8Y2hp
bGQ+CiAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a01lbnVJdGVtIiBpZD0icmVzb3VyY2VzTWVudUl0ZW0i
PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAg
ICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5UYWxrIHRvLi4uPC9wcm9w
ZXJ0eT4KICAgICAgPC9vYmplY3Q+CiAgICA8L2NoaWxkPgogICAgPGNoaWxkPgogICAgICA8b2JqZWN0
IGNsYXNzPSJHdGtNZW51SXRlbSIgaWQ9InNlbmRGaWxlTWVudUl0ZW0iPgogICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPlNlbmQgZmlsZS4uLjwvcHJvcGVydHk+CiAg
ICAgICAgPHNpZ25hbCBuYW1lPSJhY3RpdmF0ZSIgaGFuZGxlcj0ib25fc2VuZF9maWxlX3RvX2NvbnRh
Y3QiIC8+CiAgICAgIDwvb2JqZWN0PgogICAgPC9jaGlsZD4KICAgIDxjaGlsZD4KICAgICAgPG9iamVj
dCBjbGFzcz0iR3RrTWVudUl0ZW0iIGlkPSJzZW5kRGlyZWN0b3J5TWVudUl0ZW0iPgogICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPlNlbmQgZGlyZWN0b3J5Li4uPC9w
cm9wZXJ0eT4KICAgICAgICA8c2lnbmFsIG5hbWU9ImFjdGl2YXRlIiBoYW5kbGVyPSJvbl9zZW5kX2Rp
cmVjdG9yeV90b19jb250YWN0IiAvPgogICAgICA8L29iamVjdD4KICAgIDwvY2hpbGQ+CiAgICA8Y2hp
bGQ+CiAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a1NlcGFyYXRvck1lbnVJdGVtIiAvPgogICAgPC9jaGls
ZD4KICAgIDxjaGlsZD4KICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTWVudUl0ZW0iIGlkPSJhc2tDb250
YWN0VG9TZWVTdGF0dXNNZW51SXRlbSI+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFu
c2xhdGFibGU9InllcyI+QXNrIGNvbnRhY3QgdG8gc2VlIHRoZWlyIHN0YXR1czwvcHJvcGVydHk+CiAg
ICAgICAgPHNpZ25hbCBuYW1lPSJhY3RpdmF0ZSIgaGFuZGxlcj0ib25fYXNrX2NvbnRhY3RfdG9fc2Vl
X3N0YXR1cyIgLz4KICAgICAgPC9vYmplY3Q+CiAgICA8L2NoaWxkPgogICAgPGNoaWxkPgogICAgICA8
b2JqZWN0IGNsYXNzPSJHdGtNZW51SXRlbSIgaWQ9ImFsbG93Q29udGFjdFRvU2VlU3RhdHVzTWVudUl0
ZW0iPgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkFsbG93
IGNvbnRhY3QgdG8gc2VlIG15IHN0YXR1czwvcHJvcGVydHk+CiAgICAgICAgPHNpZ25hbCBuYW1lPSJh
Y3RpdmF0ZSIgaGFuZGxlcj0ib25fYWxsb3dfY29udGFjdF90b19zZWVfc3RhdHVzIiAvPgogICAgICA8
L29iamVjdD4KICAgIDwvY2hpbGQ+CiAgICA8Y2hpbGQ+CiAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a01l
bnVJdGVtIiBpZD0iZm9yYmlkQ29udGFjdFRvU2VlU3RhdHVzTWVudUl0ZW0iPgogICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkZvcmJpZCBjb250YWN0IHRvIHNlZSBt
eSBzdGF0dXM8L3Byb3BlcnR5PgogICAgICAgIDxzaWduYWwgbmFtZT0iYWN0aXZhdGUiIGhhbmRsZXI9
Im9uX2ZvcmJpZF9jb250YWN0X3RvX3NlZV9zdGF0dXMiIC8+CiAgICAgIDwvb2JqZWN0PgogICAgPC9j
aGlsZD4KICAgIDxjaGlsZD4KICAgICAgPG9iamVjdCBjbGFzcz0iR3RrU2VwYXJhdG9yTWVudUl0ZW0i
IC8+CiAgICA8L2NoaWxkPgogICAgPGNoaWxkPgogICAgICA8b2JqZWN0IGNsYXNzPSJHdGtNZW51SXRl
bSIgaWQ9ImR1bXBJbmZvTWVudUl0ZW0iPgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJh
bnNsYXRhYmxlPSJ5ZXMiPkR1bXAgaW5mbzwvcHJvcGVydHk+CiAgICAgICAgPHNpZ25hbCBuYW1lPSJh
Y3RpdmF0ZSIgaGFuZGxlcj0ib25fZHVtcF9pbmZvIiAvPgogICAgICA8L29iamVjdD4KICAgIDwvY2hp
bGQ+CiAgPC9vYmplY3Q+CjwvaW50ZXJmYWNlPgo=
`,
	},

	"/definitions/Conversation.xml": {
		local:   "definitions/Conversation.xml",
		size:    664,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtXaW5kb3ciIGlkPSJjb252ZXJzYXRpb24iPgog
ICAgPHByb3BlcnR5IG5hbWU9IndpbmRvdy1wb3NpdGlvbiI+R1RLX1dJTl9QT1NfTk9ORTwvcHJvcGVy
dHk+CiAgICA8cHJvcGVydHkgbmFtZT0iZGVmYXVsdC1oZWlnaHQiPjUwMDwvcHJvcGVydHk+CiAgICA8
cHJvcGVydHkgbmFtZT0iZGVmYXVsdC13aWR0aCI+NDAwPC9wcm9wZXJ0eT4KICAgIDxwcm9wZXJ0eSBu
YW1lPSJkZXN0cm95LXdpdGgtcGFyZW50Ij50cnVlPC9wcm9wZXJ0eT4KICAgIDxzaWduYWwgbmFtZT0i
ZW5hYmxlIiBoYW5kbGVyPSJvbl9jb25uZWN0IiAvPgogICAgPHNpZ25hbCBuYW1lPSJkaXNhYmxlIiBo
YW5kbGVyPSJvbl9kaXNjb25uZWN0IiAvPgogICAgPGNoaWxkPgogICAgICA8b2JqZWN0IGNsYXNzPSJH
dGtCb3giIGlkPSJib3giPgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj50cnVlPC9wcm9w
ZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iaG9tb2dlbmVvdXMiPmZhbHNlPC9wcm9wZXJ0eT4K
ICAgIAk8cHJvcGVydHkgbmFtZT0ib3JpZW50YXRpb24iPkdUS19PUklFTlRBVElPTl9WRVJUSUNBTDwv
cHJvcGVydHk+CiAgICAgIDwvb2JqZWN0PgogICAgPC9jaGlsZD4KICA8L29iamVjdD4KPC9pbnRlcmZh
Y2U+Cg==
`,
	},

	"/definitions/ConversationPane.xml": {
		local:   "definitions/ConversationPane.xml",
		size:    14106,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtCb3giIGlkPSJib3giPgogICAgPHByb3BlcnR5
IG5hbWU9InZpc2libGUiPnRydWU8L3Byb3BlcnR5PgogICAgPHByb3BlcnR5IG5hbWU9ImhvbW9nZW5l
b3VzIj5mYWxzZTwvcHJvcGVydHk+CiAgICA8cHJvcGVydHkgbmFtZT0ib3JpZW50YXRpb24iPkdUS19P
UklFTlRBVElPTl9WRVJUSUNBTDwvcHJvcGVydHk+CiAgICA8Y2hpbGQ+CiAgICAgIDxvYmplY3QgY2xh
c3M9Ikd0a01lbnVCYXIiIGlkPSJtZW51YmFyIj4KICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJs
ZSI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgPGNoaWxkPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0i
R3RrTWVudUl0ZW0iIGlkPSJjb252ZXJzYXRpb25NZW51Ij4KICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9InZpc2libGUiPnRydWU8L3Byb3BlcnR5PgoJICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2Jq
ZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9Im1lbnVUYWciPgogICAgICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9InZpc2libGUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+U2VjdXJlIENoYXQ8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgICAgPHByb3BlcnR5IG5hbWU9InVzZS11bmRlcmxpbmUiPlRydWU8L3Byb3BlcnR5PgogICAg
ICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQg
dHlwZT0ic3VibWVudSI+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTWVudSIgaWQ9Im1l
bnUiPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPnRydWU8L3Byb3BlcnR5
PgogICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJH
dGtNZW51SXRlbSIgaWQ9InN0YXJ0T1RSTWVudSI+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9InZpc2libGUiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPlN0YXJ0IGVuY3J5cHRlZCBjaGF0PC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8c2lnbmFsIG5hbWU9ImFjdGl2YXRlIiBoYW5kbGVyPSJv
bl9zdGFydF9vdHJfc2lnbmFsIiAvPgogICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAg
ICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgIDxv
YmplY3QgY2xhc3M9Ikd0a01lbnVJdGVtIiBpZD0iZW5kT1RSTWVudSI+CiAgICAgICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkVuZCBlbmNyeXB0ZWQg
Y2hhdDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJhY3RpdmF0ZSIg
aGFuZGxlcj0ib25fZW5kX290cl9zaWduYWwiIC8+CiAgICAgICAgICAgICAgICAgIDwvb2JqZWN0Pgog
ICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAg
ICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTWVudUl0ZW0iIGlkPSJ2ZXJpZnlGaW5nZXJNZW51Ij4KICAg
ICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+dHJ1ZTwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+
VmVyaWZ5IGZpbmdlcnByaW50PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8c2lnbmFsIG5h
bWU9ImFjdGl2YXRlIiBoYW5kbGVyPSJvbl92ZXJpZnlfZnBfc2lnbmFsIiAvPgogICAgICAgICAgICAg
ICAgICA8L29iamVjdD4KICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgPC9vYmpl
Y3Q+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICA8L2NoaWxk
PgogICAgICAgIDxjaGlsZD4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a01lbnVJdGVtIiBpZD0i
c2VuZEZpbGVNZW51SXRlbSI+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj50cnVl
PC9wcm9wZXJ0eT4KCSAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFi
ZWwiIGlkPSJzZW5kRmlsZVRhZyI+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJs
ZSI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRy
YW5zbGF0YWJsZT0ieWVzIj5TZW5kIEZpbGU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9InVzZS11bmRlcmxpbmUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvb2Jq
ZWN0PgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8c2lnbmFsIG5hbWU9InNlbGVjdCIg
aGFuZGxlcj0ib25fc2VuZF9maWxlX3RvX2NvbnRhY3QiIC8+CiAgICAgICAgICA8L29iamVjdD4KICAg
ICAgPC9jaGlsZD4KICAgICAgPC9vYmplY3Q+CiAgICAgIDxwYWNraW5nPgogICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJleHBhbmQiPmZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iZmls
bCI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9InBhY2stdHlwZSI+R1RLX1BB
Q0tfU1RBUlQ8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MDwvcHJv
cGVydHk+CiAgICAgIDwvcGFja2luZz4KICAgIDwvY2hpbGQ+CiAgICA8Y2hpbGQ+CiAgICAgIDxvYmpl
Y3QgY2xhc3M9Ikd0a0JveCIgaWQ9Im5vdGlmaWNhdGlvbi1hcmVhIj4KICAgICAgICA8cHJvcGVydHkg
bmFtZT0iY2FuX2ZvY3VzIj5GYWxzZTwvcHJvcGVydHk+Cgk8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+
dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9Im9yaWVudGF0aW9uIj5HVEtfT1JJ
RU5UQVRJT05fVkVSVElDQUw8L3Byb3BlcnR5PgoJPGNoaWxkPgogICAgICAgICAgPG9iamVjdCBjbGFz
cz0iR3RrQm94IiBpZD0ic2VjdXJpdHktd2FybmluZyI+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJ2aXNpYmxlIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5f
Zm9jdXMiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InNwYWNpbmci
PjM3PC9wcm9wZXJ0eT4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xh
c3M9Ikd0a0ltYWdlIiBpZD0iaW1hZ2Utc2VjdXJpdHktd2FybmluZyI+CiAgICAgICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0iY2FuX2ZvY3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0ibWFyZ2luX2xlZnQiPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L29iamVj
dD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJl
eHBhbmQiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxs
Ij5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+
MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgog
ICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlk
PSJsYWJlbC1zZWN1cml0eS13YXJuaW5nIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2
aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5f
Zm9jdXMiPkZhbHNlPC9wcm9wZXJ0eT4KCQk8YXR0cmlidXRlcz4KICAgICAgICAgICAgICAgICAgPGF0
dHJpYnV0ZSBuYW1lPSJ3ZWlnaHQiIHZhbHVlPSJib2xkIi8+CiAgICAgICAgICAgICAgICA8L2F0dHJp
YnV0ZXM+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAg
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZXhwYW5kIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0iZmlsbCI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvcGFj
a2luZz4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAg
IDxvYmplY3QgY2xhc3M9Ikd0a0J1dHRvbiIgaWQ9ImJ1dHRvbi1zZWN1cml0eS13YXJuaW5nIj4KICAg
ICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW5fcmlnaHQiPjEwPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW5fdG9wIj4xNDwvcHJvcGVydHk+CiAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luX2JvdHRvbSI+MTI8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIg
aWQ9ImJ1dHRvbi1sYWJlbC1zZWN1cml0eS13YXJuaW5nIj4KICAgICAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0idmlzaWJsZSI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9ImNhbl9mb2N1cyI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW5fbGVmdCI+MjM8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW5fcmlnaHQiPjIzPC9wcm9wZXJ0eT4KICAgICAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5TZWN1cmUgQ2hh
dDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgICAgPC9j
aGlsZD4KICAgICAgICAgICAgICAgIDxzaWduYWwgbmFtZT0iY2xpY2tlZCIgaGFuZGxlcj0ib25fc3Rh
cnRfb3RyX3NpZ25hbCIgLz4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFj
a2luZz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJleHBhbmQiPkZhbHNlPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj5UcnVlPC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwYWNrX3R5cGUiPmVuZDwvcHJvcGVydHk+CiAgICAg
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjI8L3Byb3BlcnR5PgogICAgICAgICAg
ICAgIDwvcGFja2luZz4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgIDwvb2JqZWN0PgogICAg
ICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJleHBhbmQiPkZhbHNlPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImZpbGwiPlRydWU8L3Byb3BlcnR5Pgog
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icGFja190eXBlIj5zdGFydDwvcHJvcGVydHk+CiAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MDwvcHJvcGVydHk+CiAgICAgICAgICA8L3Bh
Y2tpbmc+CiAgICAgICAgPC9jaGlsZD4KCTxjaGlsZD4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0
a0JveCIgaWQ9ImZpbGUtdHJhbnNmZXIiPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJs
ZSI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuX2ZvY3VzIj5G
YWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzcGFjaW5nIj4zNzwvcHJv
cGVydHk+CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtJ
bWFnZSIgaWQ9ImltYWdlLWZpbGUtdHJhbnNmZXIiPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9InZpc2libGUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
ImNhbl9mb2N1cyI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
Im1hcmdpbl9sZWZ0Ij4xMDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAg
ICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZXhwYW5kIj5GYWxz
ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZmlsbCI+RmFsc2U8L3By
b3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InBvc2l0aW9uIj4wPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAg
IDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCb3giIGlkPSJpbmZvLWZpbGUt
dHJhbnNmZXIiPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPlRydWU8L3By
b3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbl9mb2N1cyI+RmFsc2U8L3By
b3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbl90b3AiPjE0PC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJvcmllbnRhdGlvbiI+dmVydGljYWw8
L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICA8b2JqZWN0
IGNsYXNzPSJHdGtMYWJlbCIgaWQ9ImxhYmVsLWZpbGUtdHJhbnNmZXIiPgogICAgICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0iY2FuX2ZvY3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAg
ICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+c3RhcnQ8L3Byb3BlcnR5PgogICAgICAgICAg
ICAgICAgICAgIDxhdHRyaWJ1dGVzPgogICAgICAgICAgICAgICAgICAgICAgPGF0dHJpYnV0ZSBuYW1l
PSJ3ZWlnaHQiIHZhbHVlPSJib2xkIi8+CiAgICAgICAgICAgICAgICAgICAgPC9hdHRyaWJ1dGVzPgog
ICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAg
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+RmFsc2U8L3Byb3BlcnR5PgogICAg
ICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjA8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAg
ICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtQcm9ncmVzc0JhciIg
aWQ9ImJhci1maWxlLXRyYW5zZmVyIj4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
dmlzaWJsZSI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
ImNhbl9mb2N1cyI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJtYXJnaW5fdG9wIj42PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0ic2hvd190ZXh0Ij5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+
CiAgICAgICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJleHBhbmQiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0iZmlsbCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJwb3NpdGlvbiI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgIDwvcGFja2luZz4KICAg
ICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAg
PHBhY2tpbmc+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZXhwYW5kIj5GYWxzZTwvcHJv
cGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZmlsbCI+RmFsc2U8L3Byb3BlcnR5
PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InBvc2l0aW9uIj4xPC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgIDxjaGls
ZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCb3giPgogICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9InZpc2libGUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9ImNhbl9mb2N1cyI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9Im1hcmdpbl90b3AiPjE0PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJvcmllbnRhdGlvbiI+dmVydGljYWw8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPGNo
aWxkPgogICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCdXR0b24iIGlkPSJidXR0b24t
ZmlsZS10cmFuc2ZlciI+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUi
PlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9j
dXMiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJn
aW5fcmlnaHQiPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
bWFyZ2luX3RvcCI+NzwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
Im1hcmdpbl9ib3R0b20iPjIwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAg
ICAgICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9ImJ1dHRvbi1sYWJl
bC1maWxlLXRyYW5zZmVyIj4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZp
c2libGUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0iY2FuX2ZvY3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJtYXJnaW5fbGVmdCI+Mzc8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luX3JpZ2h0Ij4zNzwvcHJvcGVydHk+CiAgICAgICAgICAgICAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW5fdG9wIj41PC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbl9ib3R0b20iPjU8L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAg
ICAgICAgICA8c2lnbmFsIG5hbWU9ImNsaWNrZWQiIGhhbmRsZXI9Im9uX2Rlc3Ryb3lfZmlsZV90cmFu
c2ZlciIgLz4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8L2NoaWxkPgogICAg
ICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9ImZpbGwiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJwYWNrX3R5cGUiPmVuZDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0icG9zaXRpb24iPjI8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAg
ICAgICAgPC9jaGlsZD4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgPHBhY2tpbmc+CiAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJleHBhbmQiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9ImZpbGwiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9InBhY2tfdHlwZSI+ZW5kPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
InBvc2l0aW9uIj4xPC9wcm9wZXJ0eT4KICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICA8L2NoaWxk
PgogICAgICA8L29iamVjdD4KICAgICAgPHBhY2tpbmc+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9ImV4
cGFuZCI+ZmFsc2U8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj5mYWxzZTwv
cHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9InBhY2stdHlwZSI+c3RhcnQ8L3Byb3BlcnR5
PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MTwvcHJvcGVydHk+CiAgICAgIDwvcGFj
a2luZz4KICAgIDwvY2hpbGQ+CiAgICA8Y2hpbGQ+CiAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a1Njcm9s
bGVkV2luZG93IiBpZD0iaGlzdG9yeVNjcm9sbCI+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2li
bGUiPnRydWU8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2c2Nyb2xsYmFyLXBvbGlj
eSI+R1RLX1BPTElDWV9BVVRPTUFUSUM8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJo
c2Nyb2xsYmFyLXBvbGljeSI+R1RLX1BPTElDWV9BVVRPTUFUSUM8L3Byb3BlcnR5PgogICAgICAgIDxj
aGlsZD4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a1RleHRWaWV3IiBpZD0iaGlzdG9yeSI+CiAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9IndyYXAtbW9kZSI+MjwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJlZGl0YWJsZSI+ZmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0iY3Vyc29yLXZpc2libGUiPmZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9InBpeGVscy1iZWxvdy1saW5lcyI+NTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJsZWZ0LW1hcmdpbiI+NTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJyaWdodC1tYXJnaW4iPjU8L3Byb3BlcnR5PgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgPC9j
aGlsZD4KICAgICAgPC9vYmplY3Q+CiAgICAgIDxwYWNraW5nPgogICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJleHBhbmQiPnRydWU8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj50cnVl
PC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0icGFjay10eXBlIj5HVEtfUEFDS19FTkQ8
L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MjwvcHJvcGVydHk+CiAg
ICAgIDwvcGFja2luZz4KICAgIDwvY2hpbGQ+CiAgICA8Y2hpbGQ+CiAgICAgIDxvYmplY3QgY2xhc3M9
Ikd0a1Njcm9sbGVkV2luZG93IiBpZD0icGVuZGluZ1Njcm9sbCI+CiAgICAgICAgPHByb3BlcnR5IG5h
bWU9InZzY3JvbGxiYXItcG9saWN5Ij5HVEtfUE9MSUNZX0FVVE9NQVRJQzwvcHJvcGVydHk+CiAgICAg
ICAgPHByb3BlcnR5IG5hbWU9ImhzY3JvbGxiYXItcG9saWN5Ij5HVEtfUE9MSUNZX0FVVE9NQVRJQzwv
cHJvcGVydHk+Cgk8cHJvcGVydHkgbmFtZT0ibWluLWNvbnRlbnQtaGVpZ2h0Ij42MDwvcHJvcGVydHk+
CiAgICAgICAgPHByb3BlcnR5IG5hbWU9InNoYWRvdy10eXBlIj5pbjwvcHJvcGVydHk+CiAgICAgICAg
PGNoaWxkPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrVGV4dFZpZXciIGlkPSJwZW5kaW5nIj4K
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPnRydWU8L3Byb3BlcnR5PgogICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0id3JhcC1tb2RlIj4yPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9ImVkaXRhYmxlIj5mYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJjdXJzb3ItdmlzaWJsZSI+ZmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0icGl4ZWxzLWJlbG93LWxpbmVzIj41PC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9ImxlZnQtbWFyZ2luIj41PC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9InJpZ2h0LW1hcmdpbiI+NTwvcHJvcGVydHk+CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICA8
L2NoaWxkPgogICAgICA8L29iamVjdD4KICAgICAgPHBhY2tpbmc+CiAgICAgICAgPHByb3BlcnR5IG5h
bWU9ImV4cGFuZCI+ZmFsc2U8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj50
cnVlPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0icGFjay10eXBlIj5HVEtfUEFDS19F
TkQ8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MTwvcHJvcGVydHk+
CiAgICAgIDwvcGFja2luZz4KICAgIDwvY2hpbGQ+CiAgICA8Y2hpbGQ+CiAgICAgIDxvYmplY3QgY2xh
c3M9Ikd0a1Njcm9sbGVkV2luZG93IiBpZD0ibWVzc2FnZVNjcm9sbCI+CiAgICAgICAgPHByb3BlcnR5
IG5hbWU9InZpc2libGUiPnRydWU8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2c2Ny
b2xsYmFyLXBvbGljeSI+R1RLX1BPTElDWV9BVVRPTUFUSUM8L3Byb3BlcnR5PgogICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJoc2Nyb2xsYmFyLXBvbGljeSI+R1RLX1BPTElDWV9BVVRPTUFUSUM8L3Byb3BlcnR5
PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzaGFkb3ctdHlwZSI+aW48L3Byb3BlcnR5PgogICAgICAg
IDxjaGlsZD4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a1RleHRWaWV3IiBpZD0ibWVzc2FnZSI+
CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj50cnVlPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9Imhhcy1mb2N1cyI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJ3cmFwLW1vZGUiPkdUS19XUkFQX1dPUkRfQ0hBUjwvcHJvcGVydHk+CiAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJlZGl0YWJsZSI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LW1hcmdpbiI+MzwvcHJvcGVydHk+CiAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJyaWdodC1tYXJnaW4iPjM8L3Byb3BlcnR5PgogICAgICAgICAgPC9vYmplY3Q+
CiAgICAgICAgPC9jaGlsZD4KICAgICAgPC9vYmplY3Q+CiAgICAgIDxwYWNraW5nPgogICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJleHBhbmQiPmZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFt
ZT0iZmlsbCI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9InBhY2stdHlwZSI+
R1RLX1BBQ0tfRU5EPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjA8
L3Byb3BlcnR5PgogICAgICA8L3BhY2tpbmc+CiAgICA8L2NoaWxkPgogIDwvb2JqZWN0Pgo8L2ludGVy
ZmFjZT4K
`,
	},

	"/definitions/EditProxy.xml": {
		local:   "definitions/EditProxy.xml",
		size:    6724,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtEaWFsb2ciIGlkPSJFZGl0UHJveHkiPgogICAg
PHByb3BlcnR5IG5hbWU9InRpdGxlIiB0cmFuc2xhdGFibGU9InllcyI+RWRpdCBQcm94eTwvcHJvcGVy
dHk+CiAgICA8cHJvcGVydHkgbmFtZT0ibW9kYWwiPlRydWU8L3Byb3BlcnR5PgogICAgPHNpZ25hbCBu
YW1lPSJjbG9zZSIgaGFuZGxlcj0ib25fY2FuY2VsX3NpZ25hbCIgLz4KICAgIDxjaGlsZCBpbnRlcm5h
bC1jaGlsZD0idmJveCI+CiAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0JveCIgaWQ9ImJveCI+CiAgICAg
ICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbiI+MTA8L3Byb3BlcnR5PgogICAgICAgIDxjaGlsZD4KICAg
ICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0dyaWQiIGlkPSJncmlkIj4KICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9InZpc2libGUiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0ibWFyZ2luLWJvdHRvbSI+MTA8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
cm93LXNwYWNpbmciPjEyPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNvbHVt
bi1zcGFjaW5nIj42PC9wcm9wZXJ0eT4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxv
YmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0idHlwZUxhYmVsIj4KICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPlByb3h5IHR5cGU6PC9wcm9wZXJ0eT4K
ICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJqdXN0aWZ5Ij5HVEtfSlVTVElGWV9SSUdIVDwv
cHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05f
RU5EPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2lu
Zz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MDwvcHJvcGVydHk+
CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+MDwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hp
bGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQ29tYm9Cb3hUZXh0IiBpZD0icHJvdG9j
b2wtdHlwZSI+CiAgICAgICAgICAgICAgICA8aXRlbXM+CiAgICAgICAgICAgICAgICA8L2l0ZW1zPgog
ICAgICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJjaGFuZ2VkIiBoYW5kbGVyPSJvbl9wcm90b2NvbF90
eXBlX2NoYW5nZWQiIHN3YXBwZWQ9Im5vIi8+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAg
ICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2gi
PjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjA8
L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgPC9jaGlsZD4KICAg
ICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0i
dXNlckxhYmVsIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRh
YmxlPSJ5ZXMiPlVzZXI6PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJq
dXN0aWZ5Ij5HVEtfSlVTVElGWV9SSUdIVDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05fRU5EPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L29i
amVjdD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJsZWZ0LWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
dG9wLWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAg
ICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0i
R3RrRW50cnkiIGlkPSJ1c2VyIj4KICAgICAgICAgICAgICAgIDxzaWduYWwgbmFtZT0iYWN0aXZhdGUi
IGhhbmRsZXI9Im9uX3NhdmVfc2lnbmFsIiAvPgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAg
ICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNo
Ij4xPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4x
PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAg
ICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9
InBhc3NMYWJlbCI+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0
YWJsZT0ieWVzIj5QYXNzd29yZDo8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9Imp1c3RpZnkiPkdUS19KVVNUSUZZX1JJR0hUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJoYWxpZ24iPkdUS19BTElHTl9FTkQ8L3Byb3BlcnR5PgogICAgICAgICAgICAg
IDwvb2JqZWN0PgogICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9ImxlZnQtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJ0b3AtYXR0YWNoIj4yPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAg
ICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNs
YXNzPSJHdGtFbnRyeSIgaWQ9InBhc3N3b3JkIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJ2aXNpYmlsaXR5Ij5mYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8c2lnbmFsIG5hbWU9
ImFjdGl2YXRlIiBoYW5kbGVyPSJvbl9zYXZlX3NpZ25hbCIgLz4KICAgICAgICAgICAgICA8L29iamVj
dD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJs
ZWZ0LWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idG9w
LWF0dGFjaCI+MjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICA8
L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3Rr
TGFiZWwiIGlkPSJzZXJ2ZXJMYWJlbCI+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFi
ZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5TZXJ2ZXI6PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJqdXN0aWZ5Ij5HVEtfSlVTVElGWV9SSUdIVDwvcHJvcGVydHk+CiAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05fRU5EPC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+MzwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNr
aW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAg
PG9iamVjdCBjbGFzcz0iR3RrRW50cnkiIGlkPSJzZXJ2ZXIiPgogICAgICAgICAgICAgICAgPHNpZ25h
bCBuYW1lPSJhY3RpdmF0ZSIgaGFuZGxlcj0ib25fc2F2ZV9zaWduYWwiIC8+CiAgICAgICAgICAgICAg
PC9vYmplY3Q+CiAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0ibGVmdC1hdHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9InRvcC1hdHRhY2giPjM8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAg
ICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xh
c3M9Ikd0a0xhYmVsIiBpZD0icG9ydExhYmVsIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPlBvcnQ6PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJqdXN0aWZ5Ij5HVEtfSlVTVElGWV9SSUdIVDwvcHJvcGVydHk+CiAgICAg
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05fRU5EPC9wcm9wZXJ0eT4K
ICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+NDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9w
YWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAg
ICAgPG9iamVjdCBjbGFzcz0iR3RrRW50cnkiIGlkPSJwb3J0Ij4KICAgICAgICAgICAgICAgIDxzaWdu
YWwgbmFtZT0iYWN0aXZhdGUiIGhhbmRsZXI9Im9uX3NhdmVfc2lnbmFsIiAvPgogICAgICAgICAgICAg
IDwvb2JqZWN0PgogICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9ImxlZnQtYXR0YWNoIj4xPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJ0b3AtYXR0YWNoIj40PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAg
ICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNs
YXNzPSJHdGtMYWJlbCIgaWQ9InBhdGhMYWJlbCI+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5QYXRoOjwvcHJvcGVydHk+CiAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0ianVzdGlmeSI+R1RLX0pVU1RJRllfUklHSFQ8L3Byb3BlcnR5PgogICAg
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+R1RLX0FMSUdOX0VORDwvcHJvcGVydHk+
CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjA8L3Byb3BlcnR5PgogICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjU8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwv
cGFja2luZz4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAg
ICAgIDxvYmplY3QgY2xhc3M9Ikd0a0VudHJ5IiBpZD0icGF0aCI+CiAgICAgICAgICAgICAgICA8c2ln
bmFsIG5hbWU9ImFjdGl2YXRlIiBoYW5kbGVyPSJvbl9zYXZlX3NpZ25hbCIgLz4KICAgICAgICAgICAg
ICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0idG9wLWF0dGFjaCI+NTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAg
ICAgICAgICA8L2NoaWxkPgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgPC9jaGlsZD4KICAgICAg
ICA8Y2hpbGQgaW50ZXJuYWwtY2hpbGQ9ImFjdGlvbl9hcmVhIj4KICAgICAgICAgIDxvYmplY3QgY2xh
c3M9Ikd0a0J1dHRvbkJveCIgaWQ9ImJ1dHRvbl9ib3giPgogICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0ib3JpZW50YXRpb24iPkdUS19PUklFTlRBVElPTl9IT1JJWk9OVEFMPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0J1dHRvbiIgaWQ9ImNh
bmNlbCI+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0i
eWVzIj5DYW5jZWw8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJjbGlja2Vk
IiBoYW5kbGVyPSJvbl9jYW5jZWxfc2lnbmFsIi8+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAg
ICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNs
YXNzPSJHdGtCdXR0b24iIGlkPSJzYXZlIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJs
YWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPlNhdmU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9ImNhbi1kZWZhdWx0Ij50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxz
aWduYWwgbmFtZT0iY2xpY2tlZCIgaGFuZGxlcj0ib25fc2F2ZV9zaWduYWwiLz4KICAgICAgICAgICAg
ICA8L29iamVjdD4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAg
IDwvY2hpbGQ+CiAgICAgIDwvb2JqZWN0PgogICAgPC9jaGlsZD4KICA8L29iamVjdD4KPC9pbnRlcmZh
Y2U+Cg==
`,
	},

	"/definitions/ErrorNotification.xml": {
		local:   "definitions/ErrorNotification.xml",
		size:    382,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtCb3giIGlkPSJpbmZvYmFyIj4KICAgIDxwcm9w
ZXJ0eSBuYW1lPSJvcmllbnRhdGlvbiI+R1RLX09SSUVOVEFUSU9OX0hPUklaT05UQUw8L3Byb3BlcnR5
PgogICAgPGNoaWxkPgogICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9Im1lc3NhZ2UiPgog
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ3cmFwIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVy
dHkgbmFtZT0ibWFyZ2luLWxlZnQiPjEwPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0i
bWFyZ2luLXJpZ2h0Ij4xMDwvcHJvcGVydHk+CiAgICAgIDwvb2JqZWN0PgogICAgPC9jaGlsZD4KICA8
L29iamVjdD4KPC9pbnRlcmZhY2U+Cg==
`,
	},

	"/definitions/Feedback.xml": {
		local:   "definitions/Feedback.xml",
		size:    2777,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtEaWFsb2ciIGlkPSJkaWFsb2ciPgogICAgPHBy
b3BlcnR5IG5hbWU9IndpbmRvdy1wb3NpdGlvbiI+R1RLX1dJTl9QT1NfQ0VOVEVSPC9wcm9wZXJ0eT4K
ICAgIDxwcm9wZXJ0eSBuYW1lPSJ0aXRsZSIgdHJhbnNsYXRhYmxlPSJ5ZXMiPldlIHdvdWxkIGxpa2Ug
dG8gcmVjZWl2ZSB5b3VyIGZlZWRiYWNrPC9wcm9wZXJ0eT4KICAgIDxwcm9wZXJ0eSBuYW1lPSJib3Jk
ZXJfd2lkdGgiPjc8L3Byb3BlcnR5PgogICAgPGNoaWxkIGludGVybmFsLWNoaWxkPSJ2Ym94Ij4KICAg
ICAgPG9iamVjdCBjbGFzcz0iR3RrQm94IiBpZD0iYm94Ij4KICAgICAgICA8cHJvcGVydHkgbmFtZT0i
Ym9yZGVyLXdpZHRoIj4xMDwvcHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9Im9yaWVudGF0
aW9uIj5HVEtfT1JJRU5UQVRJT05fVkVSVElDQUw8L3Byb3BlcnR5PgoJPGNoaWxkPgogICAgICAgICAg
PG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuX2Zv
Y3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tYm90
dG9tIj4zPC9wcm9wZXJ0eT4KCSAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0i
eWVzIj4KVmlzaXQgdGhlIHdlYnNpdGUgdG8gZ2V0IGluIHRvdWNoIHdpdGggdXM6CiAgICAgICAgICAg
IDwvcHJvcGVydHk+CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgIDxwYWNraW5nPgogICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0iZXhwYW5kIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJmaWxsIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9InBvc2l0aW9uIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICA8L2No
aWxkPgoJPGNoaWxkPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJtZXNzYWdl
Ij4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbl9mb2N1cyI+RmFsc2U8L3Byb3BlcnR5Pgog
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLXRvcCI+MTA8L3Byb3BlcnR5PgoJICAgIDxw
cm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPmh0dHBzOi8vY295LmltPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InNlbGVjdGFibGUiPlRydWU8L3Byb3BlcnR5
PgogICAgICAgICAgICA8YXR0cmlidXRlcz4KICAgICAgICAgICAgICA8YXR0cmlidXRlIG5hbWU9ImZv
bnQtZGVzYyIgdmFsdWU9IiZsdDtFbnRlciBWYWx1ZSZndDsgMTQiLz4KCSAgICAgIDxhdHRyaWJ1dGUg
bmFtZT0id2VpZ2h0IiB2YWx1ZT0ic2VtaWJvbGQiLz4KICAgICAgICAgICAgPC9hdHRyaWJ1dGVzPgog
ICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9ImV4cGFuZCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
ZmlsbCI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+
MTwvcHJvcGVydHk+CiAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgPC9jaGlsZD4KICAgICAgICA8
Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCI+CiAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9Im1hcmdpbi10b3AiPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
Im1hcmdpbi1ib3R0b20iPjEwPC9wcm9wZXJ0eT4KCSAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRy
YW5zbGF0YWJsZT0ieWVzIj4KTGV0IHVzIGtub3cgd2hhdCB5b3UgdGhpbmsgb2YgQ295SU0uCgpUaGlz
IGlzIHRoZSBvbmx5IHdheSB3ZSBjYW4gY3JlYXRlIGEgYmV0dGVyIHByaXZhY3kgdG9vbC4KICAgICAg
ICAgICAgPC9wcm9wZXJ0eT4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgPHBhY2tpbmc+CiAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJleHBhbmQiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9ImZpbGwiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0icG9zaXRpb24iPjI8L3Byb3BlcnR5PgogICAgICAgICAgPC9wYWNraW5nPgogICAgICAg
IDwvY2hpbGQ+CiAgICAgIDwvb2JqZWN0PgogICAgPC9jaGlsZD4KICAgIDxjaGlsZCBpbnRlcm5hbC1j
aGlsZD0iYWN0aW9uX2FyZWEiPgogICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCdXR0b25Cb3giIGlkPSJi
dXR0b25fYm94Ij4KICAgICAgICA8cHJvcGVydHkgbmFtZT0ib3JpZW50YXRpb24iPkdUS19PUklFTlRB
VElPTl9IT1JJWk9OVEFMPC9wcm9wZXJ0eT4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0
IGNsYXNzPSJHdGtCdXR0b24iIGlkPSJjbG9zZSI+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJs
YWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkNsb3NlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHNpZ25h
bCBuYW1lPSJjbGlja2VkIiBoYW5kbGVyPSJvbl9jbG9zZV9zaWduYWwiIC8+CiAgICAgICAgICA8L29i
amVjdD4KICAgICAgICA8L2NoaWxkPgogICAgICA8L29iamVjdD4KICAgIDwvY2hpbGQ+CiAgPC9vYmpl
Y3Q+CjwvaW50ZXJmYWNlPgo=
`,
	},

	"/definitions/FeedbackInfo.xml": {
		local:   "definitions/FeedbackInfo.xml",
		size:    1128,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtJbmZvQmFyIiBpZD0iZmVlZGJhY2tJbmZvIj4K
ICAgIDxwcm9wZXJ0eSBuYW1lPSJtZXNzYWdlLXR5cGUiPkdUS19NRVNTQUdFX09USEVSPC9wcm9wZXJ0
eT4KICAgIDxwcm9wZXJ0eSBuYW1lPSJzaG93LWNsb3NlLWJ1dHRvbiI+dHJ1ZTwvcHJvcGVydHk+CiAg
ICA8c2lnbmFsIG5hbWU9InJlc3BvbnNlIiBoYW5kbGVyPSJoYW5kbGVSZXNwb25zZSIgLz4KICAgIDxj
aGlsZCBpbnRlcm5hbC1jaGlsZD0iY29udGVudF9hcmVhIj4KICAgICAgPG9iamVjdCBjbGFzcz0iR3Rr
Qm94IiBpZD0iYm94Ij4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iaG9tb2dlbmVvdXMiPmZhbHNlPC9w
cm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0ib3JpZW50YXRpb24iPkdUS19PUklFTlRBVElP
Tl9IT1JJWk9OVEFMPC9wcm9wZXJ0eT4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNs
YXNzPSJHdGtMYWJlbCIgaWQ9ImZlZWRiYWNrTWVzc2FnZSI+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkFyZSB5b3UgbGlraW5nIGl0PzwvcHJvcGVydHk+
CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICA8L2NoaWxkPgogICAgICA8L29iamVjdD4KICAgIDwv
Y2hpbGQ+CiAgICA8Y2hpbGQgaW50ZXJuYWwtY2hpbGQ9ImFjdGlvbl9hcmVhIj4KICAgICAgPG9iamVj
dCBjbGFzcz0iR3RrQm94IiBpZD0iYnV0dG9uX2JveCI+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9Im9y
aWVudGF0aW9uIj5HVEtfT1JJRU5UQVRJT05fSE9SSVpPTlRBTDwvcHJvcGVydHk+CiAgICAgICAgPGNo
aWxkPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQnV0dG9uIiBpZD0iZmVlZGJhY2tCdXR0b24i
PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5GZWVk
YmFjazwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoYWxpZ24iPkdUS19BTElH
Tl9FTkQ8L3Byb3BlcnR5PgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgPC9jaGlsZD4KICAgICAg
PC9vYmplY3Q+CiAgICA8L2NoaWxkPgogIDwvb2JqZWN0Pgo8L2ludGVyZmFjZT4K
`,
	},

	"/definitions/FileTransferAskToReceive.xml": {
		local:   "definitions/FileTransferAskToReceive.xml",
		size:    465,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtNZXNzYWdlRGlhbG9nIiBpZD0iRmlsZVRyYW5z
ZmVyQXNrVG9SZWNlaXZlIj4KICAgIDxwcm9wZXJ0eSBuYW1lPSJ3aW5kb3ctcG9zaXRpb24iPkdUS19X
SU5fUE9TX0NFTlRFUl9BTFdBWVM8L3Byb3BlcnR5PgogICAgPHByb3BlcnR5IG5hbWU9Im1vZGFsIj5m
YWxzZTwvcHJvcGVydHk+CiAgICA8cHJvcGVydHkgbmFtZT0iYm9yZGVyX3dpZHRoIj43PC9wcm9wZXJ0
eT4KICAgIDxwcm9wZXJ0eSBuYW1lPSJ0aXRsZSIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkZpbGUgdHJhbnNm
ZXI8L3Byb3BlcnR5PgogICAgPHByb3BlcnR5IG5hbWU9InNlY29uZGFyeV90ZXh0IiB0cmFuc2xhdGFi
bGU9InllcyI+PC9wcm9wZXJ0eT4KICAgIDxwcm9wZXJ0eSBuYW1lPSJidXR0b25zIj5HVEtfQlVUVE9O
U19ZRVNfTk88L3Byb3BlcnR5PgogIDwvb2JqZWN0Pgo8L2ludGVyZmFjZT4K
`,
	},

	"/definitions/FileTransferNotification.xml": {
		local:   "definitions/FileTransferNotification.xml",
		size:    1832,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtCb3giIGlkPSJhcmVhLWZpbGUtdHJhbnNmZXIt
aW5mbyI+CiAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj50cnVlPC9wcm9wZXJ0eT4KICAgPHByb3Bl
cnR5IG5hbWU9ImhvbW9nZW5lb3VzIj5mYWxzZTwvcHJvcGVydHk+CiAgIDxwcm9wZXJ0eSBuYW1lPSJv
cmllbnRhdGlvbiI+R1RLX09SSUVOVEFUSU9OX0hPUklaT05UQUw8L3Byb3BlcnR5PgogICA8cHJvcGVy
dHkgbmFtZT0ic3BhY2luZyI+OTwvcHJvcGVydHk+CiAgIDxjaGlsZD4KICAgICA8b2JqZWN0IGNsYXNz
PSJHdGtMYWJlbCIgaWQ9Im5hbWUtZmlsZS10cmFuc2Zlci1pbmZvIj4KICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9j
dXMiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoYWxpZ24iPnN0YXJ0PC9w
cm9wZXJ0eT4KICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCI+c3RhcnQ8L3Byb3BlcnR5PgogICAg
ICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi10b3AiPjQ8L3Byb3BlcnR5PgogICAgICAgPGF0dHJpYnV0
ZXM+CiAgICAgICAgIDxhdHRyaWJ1dGUgbmFtZT0iZm9udC1kZXNjIiB2YWx1ZT0iJmx0O0VudGVyIFZh
bHVlJmd0OyA5Ii8+CiAgICAgICAgIDxhdHRyaWJ1dGUgbmFtZT0id2VpZ2h0IiB2YWx1ZT0idGhpbiIv
PgogICAgICAgPC9hdHRyaWJ1dGVzPgogICAgIDwvb2JqZWN0PgogICAgIDxwYWNraW5nPgogICAgICAg
PHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgPHByb3BlcnR5IG5h
bWU9ImZpbGwiPlRydWU8L3Byb3BlcnR5PgogICAgICAgPHByb3BlcnR5IG5hbWU9InBvc2l0aW9uIj4w
PC9wcm9wZXJ0eT4KICAgICA8L3BhY2tpbmc+CiAgIDwvY2hpbGQ+CiAgIDxjaGlsZD4KICAgICA8b2Jq
ZWN0IGNsYXNzPSJHdGtFdmVudEJveCI+CiAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+VHJ1
ZTwvcHJvcGVydHk+CiAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuX2ZvY3VzIj5GYWxzZTwvcHJvcGVy
dHk+CiAgICAgICA8c2lnbmFsIG5hbWU9ImJ1dHRvbi1wcmVzcy1ldmVudCIgaGFuZGxlcj0ib25fZGVz
dHJveV9zaW5nbGVfZmlsZV90cmFuc2ZlciIgc3dhcHBlZD0ibm8iLz4KICAgICAgPGNoaWxkPgogICAg
ICAgIDxvYmplY3QgY2xhc3M9Ikd0a0ltYWdlIiBpZD0iaW1hZ2UtZmlsZS10cmFuc2Zlci1pbmZvIj4K
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJwaXhlbF9zaXplIj4xMjwvcHJvcGVydHk+CiAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0iaWNvbl9uYW1lIj53aW5kb3ctY2xvc2U8L3Byb3BlcnR5PgogICAgICAgICAgPHByb3BlcnR5IG5h
bWU9Imljb25fc2l6ZSI+MTwvcHJvcGVydHk+CiAgICAgICAgPC9vYmplY3Q+CiAgICAgIDwvY2hpbGQ+
CiAgICAgPC9vYmplY3Q+CiAgICAgPHBhY2tpbmc+CiAgICAgICA8cHJvcGVydHkgbmFtZT0iZXhwYW5k
Ij5GYWxzZTwvcHJvcGVydHk+CiAgICAgICA8cHJvcGVydHkgbmFtZT0iZmlsbCI+VHJ1ZTwvcHJvcGVy
dHk+CiAgICAgICA8cHJvcGVydHkgbmFtZT0icGFjay10eXBlIj5lbmQ8L3Byb3BlcnR5PgogICAgICAg
PHByb3BlcnR5IG5hbWU9InBvc2l0aW9uIj4xPC9wcm9wZXJ0eT4KICAgICA8L3BhY2tpbmc+CiAgIDwv
Y2hpbGQ+CiAgPC9vYmplY3Q+CjwvaW50ZXJmYWNlPgo=
`,
	},

	"/definitions/FirstAccountDialog.xml": {
		local:   "definitions/FirstAccountDialog.xml",
		size:    2919,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtEaWFsb2ciIGlkPSJkaWFsb2ciPgogICAgPHBy
b3BlcnR5IG5hbWU9IndpbmRvdy1wb3NpdGlvbiI+R1RLX1dJTl9QT1NfQ0VOVEVSPC9wcm9wZXJ0eT4K
ICAgIDxwcm9wZXJ0eSBuYW1lPSJ0aXRsZSIgdHJhbnNsYXRhYmxlPSJ5ZXMiPlNldHVwIHlvdXIgZmly
c3QgYWNjb3VudDwvcHJvcGVydHk+CiAgICA8c2lnbmFsIG5hbWU9ImRlbGV0ZS1ldmVudCIgaGFuZGxl
cj0ib25fY2FuY2VsX3NpZ25hbCIgLz4KICAgIDxjaGlsZCBpbnRlcm5hbC1jaGlsZD0idmJveCI+CiAg
ICAgIDxvYmplY3QgY2xhc3M9Ikd0a0JveCIgaWQ9IlZib3giPgogICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJvcmllbnRhdGlvbiI+dmVydGljYWw8L3Byb3BlcnR5PgoJPHByb3BlcnR5IG5hbWU9Im1hcmdpbiI+
MjA8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW5fdG9wIj4zMDwvcHJvcGVy
dHk+Cgk8cHJvcGVydHkgbmFtZT0ic3BhY2luZyI+MjwvcHJvcGVydHk+CiAgICAgICAgPGNoaWxkPgog
ICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJtZXNzYWdlIj4KICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9IndyYXAiPnRydWU8L3Byb3BlcnR5PgoJICAgIDxwcm9wZXJ0eSBuYW1lPSJs
YWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPldlbGNvbWUgdG8gQ295SU0hPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbl9ib3R0b20iPjEwPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgPGF0dHJpYnV0ZXM+CiAgICAgICAgICAgICAgPGF0dHJpYnV0ZSBuYW1lPSJmb250LWRlc2MiIHZh
bHVlPSImbHQ7RW50ZXIgVmFsdWUmZ3Q7IDE0Ii8+CiAgICAgICAgICAgICAgPGF0dHJpYnV0ZSBuYW1l
PSJzdHlsZSIgdmFsdWU9Im5vcm1hbCIvPgogICAgICAgICAgICAgIDxhdHRyaWJ1dGUgbmFtZT0id2Vp
Z2h0IiB2YWx1ZT0ic2VtaWJvbGQiLz4KICAgICAgICAgICAgPC9hdHRyaWJ1dGVzPgogICAgICAgICAg
PC9vYmplY3Q+CiAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImV4
cGFuZCI+ZmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZmlsbCI+dHJ1
ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MDwvcHJvcGVy
dHk+CiAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgPC9jaGlsZD4KICAgICAgICA8Y2hpbGQgaW50
ZXJuYWwtY2hpbGQ9ImFjdGlvbl9hcmVhIj4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0J1dHRv
bkJveCIgaWQ9ImJ1dHRvbl9ib3giPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ib3JpZW50YXRp
b24iPkdUS19PUklFTlRBVElPTl9WRVJUSUNBTDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJob21vZ2VuZW91cyI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJtYXJnaW5fdG9wIj4yMDwvcHJvcGVydHk+CgkgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxv
YmplY3QgY2xhc3M9Ikd0a0J1dHRvbiIgaWQ9ImJ1dHRvbl9yZWdpc3RlciI+CiAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5DcmVhdGUgYSBuZXcgYWNj
b3VudDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luX2JvdHRv
bSI+NTwvcHJvcGVydHk+CgkJPHNpZ25hbCBuYW1lPSJjbGlja2VkIiBoYW5kbGVyPSJvbl9yZWdpc3Rl
cl9zaWduYWwiIC8+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAg
ICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCdXR0b24iIGlk
PSJidXR0b25fZXhpc3RpbmciPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0
cmFuc2xhdGFibGU9InllcyI+QWRkIGFuIGV4aXN0aW5nIGFjY291bnQ8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbl9ib3R0b20iPjU8L3Byb3BlcnR5PgoJCTxzaWdu
YWwgbmFtZT0iY2xpY2tlZCIgaGFuZGxlcj0ib25fZXhpc3Rpbmdfc2lnbmFsIiAvPgogICAgICAgICAg
ICAgIDwvb2JqZWN0PgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAg
ICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQnV0dG9uIiBpZD0iYnV0dG9uX2ltcG9ydCI+CiAgICAg
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5JbXBvcnQg
YWNjb3VudCBmcm9tIHlvdXIgY29tcHV0ZXI8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9Im1hcmdpbl9ib3R0b20iPjU8L3Byb3BlcnR5PgoJCTxzaWduYWwgbmFtZT0iY2xpY2tl
ZCIgaGFuZGxlcj0ib25faW1wb3J0X3NpZ25hbCIgLz4KICAgICAgICAgICAgICA8L29iamVjdD4KICAg
ICAgICAgICAgPC9jaGlsZD4KCSAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0i
R3RrQnV0dG9uIiBpZD0iYnV0dG9uX2NhbmNlbCI+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0ibGFiZWwiPkNhbmNlbDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
bWFyZ2luX2JvdHRvbSI+NTwvcHJvcGVydHk+CgkJPHNpZ25hbCBuYW1lPSJjbGlja2VkIiBoYW5kbGVy
PSJvbl9jYW5jZWxfc2lnbmFsIiAvPgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICA8
L2NoaWxkPgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgPC9jaGlsZD4KICAgICAgPC9vYmplY3Q+
CiAgICA8L2NoaWxkPgogIDwvb2JqZWN0Pgo8L2ludGVyZmFjZT4K
`,
	},

	"/definitions/GeneratePIN.xml": {
		local:   "definitions/GeneratePIN.xml",
		size:    5168,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtEaWFsb2ciIGlkPSJkaWFsb2ciPgogICAgPHBy
b3BlcnR5IG5hbWU9IndpbmRvdy1wb3NpdGlvbiI+R1RLX1dJTl9QT1NfQ0VOVEVSPC9wcm9wZXJ0eT4K
ICAgIDxwcm9wZXJ0eSBuYW1lPSJib3JkZXItd2lkdGgiPjc8L3Byb3BlcnR5PgogICAgPHByb3BlcnR5
IG5hbWU9InRpdGxlIiB0cmFuc2xhdGFibGU9InllcyI+PC9wcm9wZXJ0eT4KICAgIDxjaGlsZCBpbnRl
cm5hbC1jaGlsZD0idmJveCI+CiAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQm94IiBpZD0ibm90aWZp
Y2F0aW9uLWFyZWEiPgogICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImJvcmRlci13aWR0aCI+MTA8L3By
b3BlcnR5PgogICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im9yaWVudGF0aW9uIj5HVEtfT1JJRU5UQVRJ
T05fVkVSVElDQUw8L3Byb3BlcnR5PgogICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICA8b2JqZWN0
IGNsYXNzPSJHdGtCb3giPgogICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJvcmllbnRhdGlvbiI+
R1RLX09SSUVOVEFUSU9OX1ZFUlRJQ0FMPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0ic3BhY2luZyI+MTA8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAg
ICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrSW1hZ2UiIGlkPSJzbXBfaW1hZ2UiPgogICAgICAgICAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW5fdG9wIj4xMDwvcHJvcGVydHk+CgkJICA8L29i
amVjdD4KCSAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAg
ICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9InByb21wdCI+CiAgICAgICAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLXRvcCI+NTwvcHJvcGVydHk+CgkJICAgICAgPHByb3BlcnR5
IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAg
ICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICA8Y2hpbGQ+CiAg
ICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0icGluIj4KICAgICAgICAg
ICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzZWxlY3RhYmxlIj5UcnVlPC9wcm9wZXJ0eT4K
ICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAg
ICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0dyaWQiPgogICAg
ICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY29sdW1uLXNwYWNpbmciPjY8L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icm93LXNwYWNpbmciPjI8L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNz
PSJHdGtJbWFnZSIgaWQ9InBhZGxvY2tfaW1hZ2UxIi8+CiAgICAgICAgICAgICAgICAgICAgICA8cGFj
a2luZz4KICAgICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2gi
PjA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3At
YXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAg
ICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAg
ICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiPgogICAgICAgICAgICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5TaGFyZSBpbiBw
ZXJzb248L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJq
dXN0aWZ5Ij5HVEtfSlVTVElGWV9MRUZUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05fU1RBUlQ8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICAgICAgICA8cGFja2luZz4KICAg
ICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjE8L3Byb3Bl
cnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4w
PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgICAg
ICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICAgICAg
PG9iamVjdCBjbGFzcz0iR3RrSW1hZ2UiIGlkPSJwYWRsb2NrX2ltYWdlMiIvPgogICAgICAgICAgICAg
ICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
ImxlZnQtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0idG9wLWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICA8L3Bh
Y2tpbmc+CiAgICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgICAgIDxjaGlsZD4K
ICAgICAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIj4KICAgICAgICAgICAg
ICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+VHJ1ZTwvcHJvcGVydHk+CiAgICAg
ICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9Inll
cyI+U2hhcmUgdGhyb3VnaCBhIGRpZmZlcmVudCBlbmNyeXB0ZWQgY2hhbm5lbDwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imp1c3RpZnkiPkdUS19KVVNUSUZZ
X0xFRlQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJo
YWxpZ24iPkdUS19BTElHTl9TVEFSVDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICA8L29i
amVjdD4KICAgICAgICAgICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAg
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAg
ICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtJ
bWFnZSIgaWQ9ImFsZXJ0X2ltYWdlIi8+CiAgICAgICAgICAgICAgICAgICAgICA8cGFja2luZz4KICAg
ICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjA8L3Byb3Bl
cnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4y
PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgICAg
ICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICAgICAg
PG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiPgogICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5TaGFyZSBpbiBhIHBob25lIGNh
bGw8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJqdXN0
aWZ5Ij5HVEtfSlVTVElGWV9MRUZUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05fU1RBUlQ8L3Byb3BlcnR5PgogICAgICAgICAg
ICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAg
ICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjE8L3Byb3BlcnR5
PgogICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4yPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgICAgICAg
PC9jaGlsZD4KICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgPC9jaGlsZD4K
ICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgPGNoaWxkIGlu
dGVybmFsLWNoaWxkPSJhY3Rpb25fYXJlYSI+CiAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0J1
dHRvbkJveCIgaWQ9ImJ1dHRvbl9ib3giPgogICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJvcmll
bnRhdGlvbiI+R1RLX09SSUVOVEFUSU9OX0hPUklaT05UQUw8L3Byb3BlcnR5PgogICAgICAgICAgICAg
IDxjaGlsZD4KICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0J1dHRvbiIgaWQ9ImJ1dHRv
bl9maW5pc2hlZCI+CiAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW4tZGVmYXVsdCI+
dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJh
bnNsYXRhYmxlPSJ5ZXMiPkNsb3NlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgPHNpZ25hbCBu
YW1lPSJjbGlja2VkIiBoYW5kbGVyPSJjbG9zZV9zaGFyZV9waW4iLz4KICAgICAgICAgICAgICAgIDwv
b2JqZWN0PgogICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAg
ICAgPC9jaGlsZD4KICAgICAgICA8L29iamVjdD4KICAgIDwvY2hpbGQ+CiAgPC9vYmplY3Q+CjwvaW50
ZXJmYWNlPgo=
`,
	},

	"/definitions/GlobalPreferences.xml": {
		local:   "definitions/GlobalPreferences.xml",
		size:    19890,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtMaXN0U3RvcmUiIGlkPSJub3RpZmljYXRpb24t
dHlwZS1tb2RlbCI+CiAgICA8Y29sdW1ucz4KICAgICAgPCEtLSBub3RpZmljYXRpb24gbmFtZSAtLT4K
ICAgICAgPGNvbHVtbiB0eXBlPSJnY2hhcmFycmF5Ii8+CiAgICAgIDwhLS0gbm90aWZpY2F0aW9uIGlk
IC0tPgogICAgICA8Y29sdW1uIHR5cGU9ImdjaGFyYXJyYXkiLz4KICAgIDwvY29sdW1ucz4KICAgIDxk
YXRhPgogICAgICA8cm93PgogICAgICAgIDxjb2wgaWQ9IjAiIHRyYW5zbGF0YWJsZT0ieWVzIj5ObyBu
b3RpZmljYXRpb25zPC9jb2w+CiAgICAgICAgPGNvbCBpZD0iMSI+b2ZmPC9jb2w+CiAgICAgIDwvcm93
PgogICAgICA8cm93PgogICAgICAgIDxjb2wgaWQ9IjAiIHRyYW5zbGF0YWJsZT0ieWVzIj5Pbmx5IHNo
b3cgdGhhdCBhIG5ldyBtZXNzYWdlIGFycml2ZWQ8L2NvbD4KICAgICAgICA8Y29sIGlkPSIxIj5vbmx5
LXByZXNlbmNlLW9mLW5ldy1pbmZvcm1hdGlvbjwvY29sPgogICAgICA8L3Jvdz4KICAgICAgPHJvdz4K
ICAgICAgICA8Y29sIGlkPSIwIiB0cmFuc2xhdGFibGU9InllcyI+U2hvdyB3aG8gc2VudCB0aGUgbWVz
c2FnZTwvY29sPgogICAgICAgIDxjb2wgaWQ9IjEiPndpdGgtYXV0aG9yLWJ1dC1uby1jb250ZW50PC9j
b2w+CiAgICAgIDwvcm93PgogICAgICA8cm93PgogICAgICAgIDxjb2wgaWQ9IjAiIHRyYW5zbGF0YWJs
ZT0ieWVzIj5TaG93IG1lc3NhZ2U8L2NvbD4KICAgICAgICA8Y29sIGlkPSIxIj53aXRoLWNvbnRlbnQ8
L2NvbD4KICAgICAgPC9yb3c+CiAgICA8L2RhdGE+CiAgPC9vYmplY3Q+CiAgPG9iamVjdCBjbGFzcz0i
R3RrQWRqdXN0bWVudCIgaWQ9ImFkanVzdG1lbnQxIj4KICAgIDxwcm9wZXJ0eSBuYW1lPSJsb3dlciI+
MDwvcHJvcGVydHk+CiAgICA8cHJvcGVydHkgbmFtZT0idXBwZXIiPjM2MDA8L3Byb3BlcnR5PgogICAg
PHByb3BlcnR5IG5hbWU9InN0ZXBfaW5jcmVtZW50Ij42MDwvcHJvcGVydHk+CiAgICA8cHJvcGVydHkg
bmFtZT0icGFnZV9pbmNyZW1lbnQiPjYwPC9wcm9wZXJ0eT4KICA8L29iamVjdD4KICA8b2JqZWN0IGNs
YXNzPSJHdGtEaWFsb2ciIGlkPSJHbG9iYWxQcmVmZXJlbmNlcyI+CiAgICA8cHJvcGVydHkgbmFtZT0i
dGl0bGUiIHRyYW5zbGF0YWJsZT0ieWVzIj5BY2NvdW50IERldGFpbHM8L3Byb3BlcnR5PgogICAgPHNp
Z25hbCBuYW1lPSJjbG9zZSIgaGFuZGxlcj0ib25fY2FuY2VsX3NpZ25hbCIgLz4KICAgIDxjaGlsZCBp
bnRlcm5hbC1jaGlsZD0idmJveCI+CiAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0JveCIgaWQ9IlZib3gi
PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4iPjEwPC9wcm9wZXJ0eT4KICAgICAgICA8Y2hp
bGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCb3giIGlkPSJub3RpZmljYXRpb24tYXJlYSI+
CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj50cnVlPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9Im9yaWVudGF0aW9uIj5HVEtfT1JJRU5UQVRJT05fVkVSVElDQUw8
L3Byb3BlcnR5PgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgPC9jaGlsZD4KICAgICAgICA8Y2hp
bGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtOb3RlYm9vayIgaWQ9Im5vdGVib29rMSI+CiAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9InNob3ctYm9yZGVyIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJwYWdlIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9Im1hcmdpbi1ib3R0b20iPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAg
ICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0dyaWQiIGlkPSJnZW5lcmFsR3JpZCI+CiAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLXRvcCI+MTU8L3Byb3BlcnR5PgogICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi1ib3R0b20iPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tc3RhcnQiPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tZW5kIj4xMDwvcHJvcGVydHk+CiAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0icm93LXNwYWNpbmciPjEyPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJjb2x1bW4tc3BhY2luZyI+NjwvcHJvcGVydHk+CiAgICAgICAgICAgICAg
ICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0ic2lu
Z2xlV2luZG93TGFiZWwiPgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIg
dHJhbnNsYXRhYmxlPSJ5ZXMiPlVuaWZ5IGNvbnZlcnNhdGlvbnMgaW4gb25lIHdpbmRvdzwvcHJvcGVy
dHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imp1c3RpZnkiPkdUS19KVVNUSUZZ
X1JJR0hUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGFsaWdu
Ij5HVEtfQUxJR05fRU5EPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAg
ICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJs
ZWZ0LWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
InRvcC1hdHRhY2giPjA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAg
ICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAg
ICA8b2JqZWN0IGNsYXNzPSJHdGtDaGVja0J1dHRvbiIgaWQ9InNpbmdsZVdpbmRvdyI+CiAgICAgICAg
ICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAgICAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAg
IDxjaGlsZD4KICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJzZW5k
V2l0aFNoaWZ0RW50ZXJMYWJlbCI+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imxh
YmVsIiB0cmFuc2xhdGFibGU9InllcyI+U2VuZCBtZXNzYWdlcyB3aXRoIFNoaWZ0LUVudGVyPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ianVzdGlmeSI+R1RLX0pVU1RJ
RllfUklHSFQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoYWxp
Z24iPkdUS19BTElHTl9FTkQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAg
ICAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
ImxlZnQtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0idG9wLWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgIDwvcGFja2luZz4KICAg
ICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAg
ICAgIDxvYmplY3QgY2xhc3M9Ikd0a0NoZWNrQnV0dG9uIiBpZD0ic2VuZFdpdGhTaGlmdEVudGVyIj4K
ICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICAgIDxwYWNraW5nPgogICAg
ICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MTwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjE8L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAg
ICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIg
aWQ9InNob3dFbXB0eUdyb3Vwc0xhYmVsIj4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5EaXNwbGF5IGVtcHR5IGdyb3VwczwvcHJvcGVydHk+
CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imp1c3RpZnkiPkdUS19KVVNUSUZZX1JJ
R0hUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGFsaWduIj5H
VEtfQUxJR05fRU5EPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAg
ICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0
LWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRv
cC1hdHRhY2giPjI8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAg
ICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICA8
b2JqZWN0IGNsYXNzPSJHdGtDaGVja0J1dHRvbiIgaWQ9InNob3dFbXB0eUdyb3VwcyI+CiAgICAgICAg
ICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAgICAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4yPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAg
IDxjaGlsZD4KICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJlbWFj
c0tleWJvYXJkTGFiZWwiPgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIg
dHJhbnNsYXRhYmxlPSJ5ZXMiPlVzZSBFbWFjcyBrZXkgYmluZGluZ3M8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJqdXN0aWZ5Ij5HVEtfSlVTVElGWV9SSUdIVDwvcHJv
cGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+R1RLX0FMSUdO
X0VORDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgICAg
ICA8cGFja2luZz4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2gi
PjA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNo
Ij4zPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICAgICAg
PC9jaGlsZD4KICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICAgICAgPG9iamVjdCBj
bGFzcz0iR3RrQ2hlY2tCdXR0b24iIGlkPSJlbWFjc0tleWJvYXJkIj4KICAgICAgICAgICAgICAgICAg
PC9vYmplY3Q+CiAgICAgICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjM8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICA8
L3BhY2tpbmc+CiAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgIDwvb2JqZWN0Pgog
ICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQgdHlwZT0idGFiIj4KICAgICAgICAg
ICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9ImxhYmVsLWdlbmVyYWwtdGFiIj4KICAgICAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkdlbmVyYWw8
L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPlRydWU8L3By
b3BlcnR5PgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgIDxwYWNraW5nPgogICAg
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InBvc2l0aW9uIj4wPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0YWItZmlsbCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAg
ICAgIDwvcGFja2luZz4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNoaWxkPgogICAg
ICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0dyaWQiIGlkPSJub3RpZmljYXRpb25zR3JpZCI+CiAg
ICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLXRvcCI+MTU8L3Byb3BlcnR5PgogICAg
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi1ib3R0b20iPjEwPC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tc3RhcnQiPjEwPC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tZW5kIj4xMDwvcHJvcGVydHk+CiAgICAg
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icm93LXNwYWNpbmciPjEyPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjb2x1bW4tc3BhY2luZyI+NjwvcHJvcGVydHk+CiAgICAg
ICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVs
IiBpZD0ibm90aWZpY2F0aW9uVHlwZUluc3RydWN0aW9ucyI+CiAgICAgICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+Q295SU0gc3VwcG9ydHMgbm90aWZ5
aW5nIHlvdSB3aGVuIGEgbmV3IG1lc3NhZ2UgYXJyaXZlcyAtIHRoZXNlIG5vdGlmaWNhdGlvbnMgY2Fu
IGJlIHR1cm5lZCBvZmYsIG9yIHNldCB0byBzaG93IGRpZmZlcmVudCBhbW91bnRzIG9mIGluZm9ybWF0
aW9uLiBJdCdzIGltcG9ydGFudCB0byByZW1lbWJlciB0aGF0IHNob3dpbmcgdGhlIHRleHQgb2YgdGhl
IG1lc3NhZ2UsIG9yIGV2ZW4gdGhlIGF1dGhvciwgY291bGQgcG90ZW50aWFsbHkgYmUgYSBwcml2YWN5
IGxlYWsgaWYgeW91J3JlIGV2ZXIgaW4gc2V0dGluZ3Mgd2hlcmUgc29tZW9uZSBjYW4gc2VlIHlvdXIg
c2NyZWVuLjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2li
bGUiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ3cmFw
Ij50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWF4LXdp
ZHRoLWNoYXJzIj41MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAg
ICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVm
dC1hdHRhY2giPjA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0
b3AtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
d2lkdGgiPjI8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAg
ICAgICA8L2NoaWxkPgogICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICA8b2Jq
ZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9Im5vdGlmaWNhdGlvblR5cGVMYWJlbCI+CiAgICAgICAgICAg
ICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+Tm90aWZpY2F0
aW9uIHR5cGU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJqdXN0
aWZ5Ij5HVEtfSlVTVElGWV9SSUdIVDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9ImhhbGlnbiI+R1RLX0FMSUdOX0VORDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAg
IDwvb2JqZWN0PgogICAgICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4xPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAg
PC9wYWNraW5nPgogICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgIDxjaGlsZD4K
ICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQ29tYm9Cb3giIGlkPSJub3RpZmljYXRp
b25zVHlwZSI+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1vZGVsIj5ub3RpZmlj
YXRpb24tdHlwZS1tb2RlbDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9Imhhcy1mb2N1cyI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9ImhleHBhbmQiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxjaGlsZD4K
ICAgICAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0NlbGxSZW5kZXJlclRleHQiIGlk
PSJub3RpZmljYXRpb24tbmFtZS1yZW5kZXJlZCIvPgogICAgICAgICAgICAgICAgICAgICAgPGF0dHJp
YnV0ZXM+CiAgICAgICAgICAgICAgICAgICAgICAgIDxhdHRyaWJ1dGUgbmFtZT0idGV4dCI+MDwvYXR0
cmlidXRlPgogICAgICAgICAgICAgICAgICAgICAgPC9hdHRyaWJ1dGVzPgogICAgICAgICAgICAgICAg
ICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgICAgICA8
cGFja2luZz4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjE8
L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4x
PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICAgICAgPC9j
aGlsZD4KICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFz
cz0iR3RrTGFiZWwiIGlkPSJub3RpZmljYXRpb25VcmdlbmN5SW5zdHJ1Y3Rpb25zIj4KICAgICAgICAg
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5Ob3RpZmlj
YXRpb25zIGNhbiBiZSBzZXQgdG8gZGlzcGxheSB1cmdlbnRseSAtIHRoaXMgaXMgdXNlZnVsIGlmIHlv
dSB3b3JrIGluIGZ1bGxzY3JlZW4gbW9kZS4gSWYgdGhlIG5vdGlmaWNhdGlvbiBpcyBub3QgdXJnZW50
LCBpdCB3aWxsIG5vdCBkaXNwbGF5IG92ZXIgYSBmdWxsc2NyZWVuIHdpbmRvdy48L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj50cnVlPC9wcm9wZXJ0eT4K
ICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0id3JhcCI+dHJ1ZTwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1heC13aWR0aC1jaGFycyI+NTA8L3Byb3Bl
cnR5PgogICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICAgICAgPHBhY2tpbmc+
CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4wPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+MjwvcHJvcGVy
dHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9IndpZHRoIj4yPC9wcm9wZXJ0eT4K
ICAgICAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAg
ICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwi
IGlkPSJub3RpZmljYXRpb25VcmdlbnRMYWJlbCI+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+U2hvdWxkIG5vdGlmaWNhdGlvbnMgYmUgZGlz
cGxheWVkIHVyZ2VudGx5PzwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9Imp1c3RpZnkiPkdUS19KVVNUSUZZX1JJR0hUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05fRU5EPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAg
ICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjM8L3Byb3BlcnR5PgogICAgICAgICAg
ICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgICAg
PGNoaWxkPgogICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtDaGVja0J1dHRvbiIgaWQ9
Im5vdGlmaWNhdGlvblVyZ2VudCI+CiAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAg
ICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1h
dHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3At
YXR0YWNoIj4zPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAg
ICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICAgICAgPG9i
amVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJub3RpZmljYXRpb25FeHBpcnlJbnN0cnVjdGlvbnMiPgog
ICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMi
Pk5vdGlmaWNhdGlvbnMgY2FuIHN0YXkgb24gdGhlIHNjcmVlbiB1bnRpbCB5b3UndmUgZ29uZSBiYWNr
IHRvIENveUlNLCBvciB0aGV5IGNhbiBleHBpcmUgYWZ0ZXIgYSB3aGlsZS4gVGhlIGJlbG93IHNldHRp
bmcgY2hhbmdlcyB0aGlzIGJlaGF2aW9yLjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9InZpc2libGUiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJ3cmFwIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0ibWF4LXdpZHRoLWNoYXJzIj41MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAg
IDwvb2JqZWN0PgogICAgICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj40PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0id2lkdGgiPjI8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICA8L3Bh
Y2tpbmc+CiAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgICAgPGNoaWxkPgogICAg
ICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9Im5vdGlmaWNhdGlvbkV4cGly
ZXNMYWJlbCI+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xh
dGFibGU9InllcyI+U2hvdWxkIG5vdGlmaWNhdGlvbnMgZXhwaXJlPzwvcHJvcGVydHk+CiAgICAgICAg
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imp1c3RpZnkiPkdUS19KVVNUSUZZX1JJR0hUPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05f
RU5EPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICAg
IDxwYWNraW5nPgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+
MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2gi
PjU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgICAgICA8
L2NoaWxkPgogICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICA8b2JqZWN0IGNs
YXNzPSJHdGtDaGVja0J1dHRvbiIgaWQ9Im5vdGlmaWNhdGlvbkV4cGlyZXMiPgogICAgICAgICAgICAg
ICAgICA8L29iamVjdD4KICAgICAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4xPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+NTwvcHJvcGVydHk+CiAgICAgICAgICAgICAg
ICAgIDwvcGFja2luZz4KICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgICA8Y2hp
bGQ+CiAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0ibm90aWZpY2F0
aW9uQ29tbWFuZEluc3RydWN0aW9ucyI+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+WW91IGNhbiBzcGVjaWZ5IGEgY3VzdG9tIGNvbW1hbmQg
dG8gcnVuIHdoZW5ldmVyIGEgbWVzc2FnZSBpcyByZWNlaXZlZC4gSWYgc3BlY2lmaWNlZCwgdGhpcyBj
b21tYW5kIHdpbGwgcnVuIG9uIGV2ZXJ5IG1lc3NhZ2VzIGV4Y2VwdCBpdCB3aWxsIHdhaXQgZm9yIGEg
dGltZW91dCBwZXJpb2QgYmVmb3JlIHJ1bm5pbmcgdGhlIG5leHQgdGltZS4gVGhlIGNvbW1hbmQgYW5k
IHRpbWVvdXQgaXMgc3BlY2lmaWVkIGJlbG93LiA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0id3JhcCI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9Im1heC13aWR0aC1jaGFycyI+NTA8L3Byb3BlcnR5PgogICAgICAgICAgICAg
ICAgICA8L29iamVjdD4KICAgICAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+NjwvcHJvcGVydHk+CiAgICAgICAgICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9IndpZHRoIj4yPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAg
PC9wYWNraW5nPgogICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgIDxjaGlsZD4K
ICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJub3RpZmljYXRpb25D
b21tYW5kTGFiZWwiPgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJh
bnNsYXRhYmxlPSJ5ZXMiPk5vdGlmaWNhdGlvbiBjb21tYW5kPC9wcm9wZXJ0eT4KICAgICAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ianVzdGlmeSI+R1RLX0pVU1RJRllfUklHSFQ8L3Byb3BlcnR5
PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoYWxpZ24iPkdUS19BTElHTl9FTkQ8
L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICAgICAgPHBh
Y2tpbmc+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4wPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+Nzwv
cHJvcGVydHk+CiAgICAgICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgICAgIDwvY2hp
bGQ+CiAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9
Ikd0a0VudHJ5IiBpZD0ibm90aWZpY2F0aW9uQ29tbWFuZCI+CiAgICAgICAgICAgICAgICAgICAgPHNp
Z25hbCBuYW1lPSJhY3RpdmF0ZSIgaGFuZGxlcj0ib25fc2F2ZV9zaWduYWwiIC8+CiAgICAgICAgICAg
ICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj43PC9wcm9wZXJ0eT4KICAgICAgICAgICAg
ICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgIDxj
aGlsZD4KICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJub3RpZmlj
YXRpb25UaW1lb3V0TGFiZWwiPgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJl
bCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPk1pbmltdW0gdGltZSBiZXR3ZWVuIG5vdGlmaWNhdGlvbnMgaW4g
c2Vjb25kczwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imp1c3Rp
ZnkiPkdUS19KVVNUSUZZX1JJR0hUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05fRU5EPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAg
PC9vYmplY3Q+CiAgICAgICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjg8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICA8
L3BhY2tpbmc+CiAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgICAgPGNoaWxkPgog
ICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtTcGluQnV0dG9uIiBpZD0ibm90aWZpY2F0
aW9uVGltZW91dCI+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNsaW1iLXJhdGUi
PjYwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iYWRqdXN0bWVu
dCI+YWRqdXN0bWVudDE8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxzaWduYWwgbmFtZT0i
YWN0aXZhdGUiIGhhbmRsZXI9Im9uX3NhdmVfc2lnbmFsIiAvPgogICAgICAgICAgICAgICAgICA8L29i
amVjdD4KICAgICAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4xPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0idG9wLWF0dGFjaCI+ODwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgIDwvcGFj
a2luZz4KICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAg
ICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgIDxjaGlsZCB0eXBlPSJ0YWIiPgogICAgICAgICAgICAg
IDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0ibGFiZWwtbm90aWZpY2F0aW9ucy10YWIiPgogICAg
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+Tm90aWZp
Y2F0aW9uczwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+
VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgPHBhY2tp
bmc+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjE8L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRhYi1maWxsIj5GYWxzZTwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hp
bGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrR3JpZCIgaWQ9ImRlYnVnZ2luZ0dyaWQi
PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi10b3AiPjE1PC9wcm9wZXJ0eT4K
ICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tYm90dG9tIj4xMDwvcHJvcGVydHk+
CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLXN0YXJ0Ij4xMDwvcHJvcGVydHk+
CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLWVuZCI+MTA8L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InJvdy1zcGFjaW5nIj4xMjwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY29sdW1uLXNwYWNpbmciPjY8L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtM
YWJlbCIgaWQ9InJhd0xvZ0ZpbGVJbnN0cnVjdGlvbnMiPgogICAgICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPklmIHlvdSBzZXQgdGhpcyBwcm9wZXJ0
eSB0byBhIGZpbGUgbmFtZSwgbG93IGxldmVsIGluZm9ybWF0aW9uIHdpbGwgYmUgbG9nZ2VkIHRoZXJl
LiBCZSB2ZXJ5IGNhcmVmdWwgLSB0aGlzIGluZm9ybWF0aW9uIGlzIHNlbnNpdGl2ZSBhbmQgY291bGQg
cG90ZW50aWFsbHkgY29udGFpbiB2ZXJ5IHByaXZhdGUgaW5mb3JtYXRpb24uIE9ubHkgdHVybiB0aGlz
IHNldHRpbmcgb24gaWYgeW91IGFic29sdXRlbHkgbmVlZCBpdCBmb3IgZGVidWdnaW5nLiBUaGlzIGZp
bGUgd2lsbCBzcGVjaWZpY2FsbHkgbG9nIFhNUFAgdHJhZmZpYyBpbmZvcm1hdGlvbi4gVGhpcyBzZXR0
aW5nIHdpbGwgb25seSB0YWtlIGVmZmVjdCBhZnRlciBhIHJlc3RhcnQgb2YgQ295SU0uPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+dHJ1ZTwvcHJvcGVy
dHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9IndyYXAiPnRydWU8L3Byb3BlcnR5
PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXgtd2lkdGgtY2hhcnMiPjUwPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICAgIDxwYWNr
aW5nPgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MDwvcHJv
cGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjA8L3By
b3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ3aWR0aCI+MjwvcHJvcGVy
dHk+CiAgICAgICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAg
ICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xh
YmVsIiBpZD0icmF3TG9nRmlsZUxhYmVsIj4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5SYXcgbG9nIGZpbGU8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJqdXN0aWZ5Ij5HVEtfSlVTVElGWV9SSUdIVDwvcHJv
cGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+R1RLX0FMSUdO
X0VORDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgICAg
ICA8cGFja2luZz4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2gi
PjA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNo
Ij4xPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICAgICAg
PC9jaGlsZD4KICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICAgICAgPG9iamVjdCBj
bGFzcz0iR3RrRW50cnkiIGlkPSJyYXdMb2dGaWxlIj4KICAgICAgICAgICAgICAgICAgICA8c2lnbmFs
IG5hbWU9ImFjdGl2YXRlIiBoYW5kbGVyPSJvbl9zYXZlX3NpZ25hbCIgLz4KICAgICAgICAgICAgICAg
ICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAg
ICA8L3BhY2tpbmc+CiAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgIDwvb2JqZWN0
PgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQgdHlwZT0idGFiIj4KICAgICAg
ICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9ImxhYmVsLWRlYnVnZ2luZy10YWIiPgog
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+RGVi
dWdnaW5nPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5U
cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2lu
Zz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MjwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idGFiLWZpbGwiPkZhbHNlPC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICA8L29iamVj
dD4KICAgICAgICA8L2NoaWxkPgogICAgICAgIDxjaGlsZCBpbnRlcm5hbC1jaGlsZD0iYWN0aW9uX2Fy
ZWEiPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQnV0dG9uQm94IiBpZD0iYnV0dG9uX2JveCI+
CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJvcmllbnRhdGlvbiI+R1RLX09SSUVOVEFUSU9OX0hP
UklaT05UQUw8L3Byb3BlcnR5PgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVj
dCBjbGFzcz0iR3RrQnV0dG9uIiBpZD0iY2FuY2VsIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkNhbmNlbDwvcHJvcGVydHk+CiAgICAgICAgICAg
ICAgICA8c2lnbmFsIG5hbWU9ImNsaWNrZWQiIGhhbmRsZXI9Im9uX2NhbmNlbF9zaWduYWwiLz4KICAg
ICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNoaWxk
PgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0J1dHRvbiIgaWQ9InNhdmUiPgogICAgICAg
ICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+U2F2ZTwvcHJv
cGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuLWRlZmF1bHQiPnRydWU8L3By
b3BlcnR5PgogICAgICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJjbGlja2VkIiBoYW5kbGVyPSJvbl9z
YXZlX3NpZ25hbCIvPgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICA8L2NoaWxkPgog
ICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgPC9jaGlsZD4KICAgICAgPC9vYmplY3Q+CiAgICA8L2No
aWxkPgogIDwvb2JqZWN0Pgo8L2ludGVyZmFjZT4K
`,
	},

	"/definitions/GroupDetails.xml": {
		local:   "definitions/GroupDetails.xml",
		size:    2084,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtEaWFsb2ciIGlkPSJkaWFsb2ciPgogICAgPHBy
b3BlcnR5IG5hbWU9IndpbmRvdy1wb3NpdGlvbiI+R1RLX1dJTl9QT1NfQ0VOVEVSPC9wcm9wZXJ0eT4K
ICAgIDxwcm9wZXJ0eSBuYW1lPSJ0aXRsZSIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkFkZCBHcm91cDwvcHJv
cGVydHk+CiAgICA8Y2hpbGQgaW50ZXJuYWwtY2hpbGQ9InZib3giPgogICAgICA8b2JqZWN0IGNsYXNz
PSJHdGtCb3giIGlkPSJWYm94Ij4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iaG9tb2dlbmVvdXMiPmZh
bHNlPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0ib3JpZW50YXRpb24iPkdUS19PUklF
TlRBVElPTl9WRVJUSUNBTDwvcHJvcGVydHk+CiAgICAgICAgPGNoaWxkPgogICAgICAgICAgPG9iamVj
dCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJncm91cC1uYW1lLWxhYmVsIiA+CiAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkdyb3VwIG5hbWU8L3Byb3BlcnR5Pgog
ICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9ImV4cGFuZCI+ZmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
ZmlsbCI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+
MTwvcHJvcGVydHk+CiAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgPC9jaGlsZD4KICAgICAgICA8
Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtFbnRyeSIgaWQ9Imdyb3VwLW5hbWUiPgog
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGFzLWZvY3VzIj50cnVlPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9ImFjdGl2YXRlcy1kZWZhdWx0Ij5UcnVlPC9wcm9wZXJ0eT4KICAg
ICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJleHBhbmQiPmZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImZp
bGwiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjI8
L3Byb3BlcnR5PgogICAgICAgICAgPC9wYWNraW5nPgogICAgICAgIDwvY2hpbGQ+CiAgICAgICAgPGNo
aWxkIGludGVybmFsLWNoaWxkPSJhY3Rpb25fYXJlYSI+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJH
dGtCdXR0b25Cb3giIGlkPSJiYm94Ij4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im9yaWVudGF0
aW9uIj5HVEtfT1JJRU5UQVRJT05fSE9SSVpPTlRBTDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxjaGls
ZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCdXR0b24iIGlkPSJidG4tY2FuY2VsIj4K
ICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkNh
bmNlbDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgIDwvY2hpbGQ+
CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCdXR0b24i
IGlkPSJidG4tb2siPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xh
dGFibGU9InllcyI+T0s8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNh
bi1kZWZhdWx0Ij50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAg
ICAgPC9jaGlsZD4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgIDwvY2hpbGQ+CiAgICAgIDwvb2Jq
ZWN0PgogICAgPC9jaGlsZD4KICAgIDxhY3Rpb24td2lkZ2V0cz4KICAgICAgPGFjdGlvbi13aWRnZXQg
cmVzcG9uc2U9ImNhbmNlbCI+YnRuLWNhbmNlbDwvYWN0aW9uLXdpZGdldD4KICAgICAgPGFjdGlvbi13
aWRnZXQgcmVzcG9uc2U9Im9rIiBkZWZhdWx0PSJ0cnVlIj5idG4tb2s8L2FjdGlvbi13aWRnZXQ+CiAg
ICA8L2FjdGlvbi13aWRnZXRzPgogIDwvb2JqZWN0Pgo8L2ludGVyZmFjZT4=
`,
	},

	"/definitions/Importer.xml": {
		local:   "definitions/Importer.xml",
		size:    4858,
		modtime: 1489449600,
		compressed: `
PD94bWwgdmVyc2lvbj0iMS4wIiBlbmNvZGluZz0idXRmLTgiPz4KPGludGVyZmFjZT4KICA8b2JqZWN0
IGNsYXNzPSJHdGtMaXN0U3RvcmUiIGlkPSJpbXBvcnRBY2NvdW50c1N0b3JlIj4KICAgIDxjb2x1bW5z
PgogICAgICA8Y29sdW1uIHR5cGU9ImdjaGFyYXJyYXkiLz4KICAgICAgPGNvbHVtbiB0eXBlPSJnY2hh
cmFycmF5Ii8+CiAgICAgIDxjb2x1bW4gdHlwZT0iZ2Jvb2xlYW4iLz4KICAgIDwvY29sdW1ucz4KICA8
L29iamVjdD4KICA8b2JqZWN0IGNsYXNzPSJHdGtEaWFsb2ciIGlkPSJpbXBvcnRlcldpbmRvdyI+CiAg
ICA8cHJvcGVydHkgbmFtZT0id2luZG93LXBvc2l0aW9uIj5HVEtfV0lOX1BPU19DRU5URVI8L3Byb3Bl
cnR5PgogICAgPHByb3BlcnR5IG5hbWU9InRpdGxlIiB0cmFuc2xhdGFibGU9InllcyI+SW1wb3J0IEFj
Y291bnRzPC9wcm9wZXJ0eT4KICAgIDxwcm9wZXJ0eSBuYW1lPSJ3aWR0aF9yZXF1ZXN0Ij40NTA8L3By
b3BlcnR5PgogICAgPHByb3BlcnR5IG5hbWU9ImhlaWdodF9yZXF1ZXN0Ij42MDA8L3Byb3BlcnR5Pgog
ICAgPHByb3BlcnR5IG5hbWU9ImJvcmRlcl93aWR0aCI+MTA8L3Byb3BlcnR5PgogICAgPGNoaWxkIGlu
dGVybmFsLWNoaWxkPSJ2Ym94Ij4KICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQm94IiBpZD0iYm94Ij4K
ICAgICAgICA8cHJvcGVydHkgbmFtZT0iaG9tb2dlbmVvdXMiPmZhbHNlPC9wcm9wZXJ0eT4KICAgICAg
ICA8cHJvcGVydHkgbmFtZT0ib3JpZW50YXRpb24iPkdUS19PUklFTlRBVElPTl9WRVJUSUNBTDwvcHJv
cGVydHk+CiAgICAgICAgPGNoaWxkPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlk
PSJsYWJlbCI+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5
ZXMiPkNob29zZSB0aGUgYWNjb3VudHMgeW91IHdvdWxkIGxpa2UgdG8gaW1wb3J0IGFuZCBzdGFydCB1
c2luZyBmcm9tIENveUlNLjwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ3cmFw
Ij50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9IndyYXAtbW9kZSI+UEFO
R09fV1JBUF9XT1JEX0NIQVI8L3Byb3BlcnR5PgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICA8
cGFja2luZz4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+ZmFsc2U8L3Byb3BlcnR5
PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZmlsbCI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MDwvcHJvcGVydHk+CiAgICAgICAgICA8L3BhY2tp
bmc+CiAgICAgICAgPC9jaGlsZD4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNz
PSJHdGtTY3JvbGxlZFdpbmRvdyIgaWQ9ImltcG9ydGVyU2Nyb2xsIj4KICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9InZzY3JvbGxiYXItcG9saWN5Ij5HVEtfUE9MSUNZX0FVVE9NQVRJQzwvcHJvcGVydHk+
CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoc2Nyb2xsYmFyLXBvbGljeSI+R1RLX1BPTElDWV9B
VVRPTUFUSUM8L3Byb3BlcnR5PgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVj
dCBjbGFzcz0iR3RrVHJlZVZpZXciIGlkPSJpbXBvcnRlclRyZWVWaWV3Ij4KICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJtb2RlbCI+aW1wb3J0QWNjb3VudHNTdG9yZTwvcHJvcGVydHk+CiAgICAg
ICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a1RyZWVW
aWV3Q29sdW1uIiBpZD0iaW1wb3J0LXRoaXMtYWNjb3VudC1jb2x1bW4iPgogICAgICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJ0aXRsZSIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkltcG9ydCB0aGlzPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ic29ydF9jb2x1bW5faWQi
PjI8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICAgICAg
ICAgIDxvYmplY3QgY2xhc3M9Ikd0a0NlbGxSZW5kZXJlclRvZ2dsZSIgaWQ9ImltcG9ydC10aGlzLWFj
Y291bnQtcmVuZGVyZXIiPgogICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iYWN0
aXZhdGFibGUiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAg
ICAgICAgICAgICAgICAgICAgICA8YXR0cmlidXRlcz4KICAgICAgICAgICAgICAgICAgICAgICAgPGF0
dHJpYnV0ZSBuYW1lPSJhY3RpdmUiPjI8L2F0dHJpYnV0ZT4KICAgICAgICAgICAgICAgICAgICAgIDwv
YXR0cmlidXRlcz4KICAgICAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgICAgICA8
L29iamVjdD4KICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAg
ICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a1RyZWVWaWV3Q29sdW1uIiBpZD0iZnJvbS1h
cHBsaWNhdGlvbi1jb2x1bW4iPgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0aXRs
ZSIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkZyb20gYXBwbGljYXRpb248L3Byb3BlcnR5PgogICAgICAgICAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzb3J0X2NvbHVtbl9pZCI+MDwvcHJvcGVydHk+CiAgICAg
ICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0i
R3RrQ2VsbFJlbmRlcmVyVGV4dCIgaWQ9ImZyb20tYXBwbGljYXRpb24tcmVuZGVyZXIiLz4KICAgICAg
ICAgICAgICAgICAgICAgIDxhdHRyaWJ1dGVzPgogICAgICAgICAgICAgICAgICAgICAgICA8YXR0cmli
dXRlIG5hbWU9InRleHQiPjA8L2F0dHJpYnV0ZT4KICAgICAgICAgICAgICAgICAgICAgIDwvYXR0cmli
dXRlcz4KICAgICAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgICAgICA8L29iamVj
dD4KICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAg
ICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a1RyZWVWaWV3Q29sdW1uIiBpZD0iYWNjb3VudC1uYW1l
LWNvbHVtbiI+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRpdGxlIiB0cmFuc2xh
dGFibGU9InllcyI+QWNjb3VudCBuYW1lPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0ic29ydF9jb2x1bW5faWQiPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAg
IDxjaGlsZD4KICAgICAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0NlbGxSZW5kZXJl
clRleHQiIGlkPSJhY2NvdW50LW5hbWUtcmVuZGVyZXIiLz4KICAgICAgICAgICAgICAgICAgICAgIDxh
dHRyaWJ1dGVzPgogICAgICAgICAgICAgICAgICAgICAgICA8YXR0cmlidXRlIG5hbWU9InRleHQiPjE8
L2F0dHJpYnV0ZT4KICAgICAgICAgICAgICAgICAgICAgIDwvYXR0cmlidXRlcz4KICAgICAgICAgICAg
ICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICAg
IDwvY2hpbGQ+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAg
ICAgICA8L29iamVjdD4KICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0iZXhwYW5kIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImZpbGwi
PnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjE8L3By
b3BlcnR5PgogICAgICAgICAgPC9wYWNraW5nPgogICAgICAgIDwvY2hpbGQ+CiAgICAgICAgPGNoaWxk
IGludGVybmFsLWNoaWxkPSJhY3Rpb25fYXJlYSI+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtC
dXR0b25Cb3giIGlkPSJidXR0b25fYm94Ij4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im9yaWVu
dGF0aW9uIj5HVEtfT1JJRU5UQVRJT05fSE9SSVpPTlRBTDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJtYXJnaW4tdG9wIj4xMDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxjaGlsZD4K
ICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCdXR0b24iIGlkPSJidXR0b25fY2FuY2VsIj4K
ICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkNh
bmNlbDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgIDwvY2hpbGQ+
CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCdXR0b24i
IGlkPSJidXR0b25fb2siPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFu
c2xhdGFibGU9InllcyI+SW1wb3J0PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJjYW4tZGVmYXVsdCI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAg
ICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICA8L2NoaWxkPgogICAg
ICA8L29iamVjdD4KICAgIDwvY2hpbGQ+CiAgICA8YWN0aW9uLXdpZGdldHM+CiAgICAgIDxhY3Rpb24t
d2lkZ2V0IHJlc3BvbnNlPSJjYW5jZWwiPmJ1dHRvbl9jYW5jZWw8L2FjdGlvbi13aWRnZXQ+CiAgICAg
IDxhY3Rpb24td2lkZ2V0IHJlc3BvbnNlPSJvayIgZGVmYXVsdD0idHJ1ZSI+YnV0dG9uX29rPC9hY3Rp
b24td2lkZ2V0PgogICAgPC9hY3Rpb24td2lkZ2V0cz4KICA8L29iamVjdD4KPC9pbnRlcmZhY2U+Cg==
`,
	},

	"/definitions/ListChatRooms.xml": {
		local:   "definitions/ListChatRooms.xml",
		size:    10099,
		modtime: 1489449600,
		compressed: `
PD94bWwgdmVyc2lvbj0iMS4wIiBlbmNvZGluZz0iVVRGLTgiPz4KPGludGVyZmFjZT4KICA8cmVxdWly
ZXMgbGliPSJndGsrIiB2ZXJzaW9uPSIzLjAiLz4KICA8b2JqZWN0IGNsYXNzPSJHdGtMaXN0U3RvcmUi
IGlkPSJyb29tcyI+CiAgICA8Y29sdW1ucz4KICAgICAgPGNvbHVtbiB0eXBlPSJnY2hhcmFycmF5Ii8+
CiAgICAgIDxjb2x1bW4gdHlwZT0iZ2NoYXJhcnJheSIvPgogICAgPC9jb2x1bW5zPgogIDwvb2JqZWN0
PgogIDxvYmplY3QgY2xhc3M9Ikd0a0xpc3RTdG9yZSIgaWQ9ImFjY291bnRzLW1vZGVsIj4KICAgIDxj
b2x1bW5zPgogICAgICA8IS0tIGFjY291bnQgbmFtZSAtLT4KICAgICAgPGNvbHVtbiB0eXBlPSJnY2hh
cmFycmF5Ii8+CiAgICAgIDwhLS0gYWNjb3VudCBpZCAtLT4KICAgICAgPGNvbHVtbiB0eXBlPSJnY2hh
cmFycmF5Ii8+CiAgICA8L2NvbHVtbnM+CiAgPC9vYmplY3Q+CiAgPG9iamVjdCBjbGFzcz0iR3RrRGlh
bG9nIiBpZD0ibGlzdC1jaGF0LXJvb21zIj4KICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPkZh
bHNlPC9wcm9wZXJ0eT4KICAgIDxwcm9wZXJ0eSBuYW1lPSJ0eXBlX2hpbnQiPmRpYWxvZzwvcHJvcGVy
dHk+CiAgICA8Y2hpbGQgaW50ZXJuYWwtY2hpbGQ9InZib3giPgogICAgICA8b2JqZWN0IGNsYXNzPSJH
dGtCb3giPgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPkZhbHNlPC9wcm9wZXJ0eT4K
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ib3JpZW50YXRpb24iPnZlcnRpY2FsPC9wcm9wZXJ0eT4KICAg
ICAgICA8cHJvcGVydHkgbmFtZT0ic3BhY2luZyI+MzwvcHJvcGVydHk+CiAgICAgICAgPGNoaWxkPgog
ICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQm94Ij4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
InZpc2libGUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuX2Zv
Y3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJvcmllbnRhdGlv
biI+dmVydGljYWw8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGVpZ2h0LXJl
cXVlc3QiPjQwMDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2Jq
ZWN0IGNsYXNzPSJHdGtHcmlkIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxl
Ij5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMi
PkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tdG9w
Ij4xNTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLXN0YXJ0
Ij4xMDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLWVuZCI+
MTA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi1ib3R0b20i
PjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjb2x1bW4tc3BhY2lu
ZyI+MTA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InJvdy1zcGFjaW5n
Ij4xMjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgIDxv
YmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0iYWNjb3VudHMtbGFiZWwiID4KICAgIAkJICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkFjY291bnQ6PC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ianVzdGlmeSI+R1RLX0pVU1RJ
RllfUklHSFQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoYWxp
Z24iPkdUS19BTElHTl9TVEFSVDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgIDwvb2JqZWN0Pgog
ICAgICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0idG9wLWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9ImxlZnQtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0id2lkdGgiPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAg
ICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAg
ICA8b2JqZWN0IGNsYXNzPSJHdGtDb21ib0JveCIgaWQ9ImFjY291bnRzIj4KICAgICAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9Im1vZGVsIj5hY2NvdW50cy1tb2RlbDwvcHJvcGVydHk+CiAgICAg
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imhhcy1mb2N1cyI+dHJ1ZTwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhleHBhbmQiPlRydWU8L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xh
c3M9Ikd0a0NlbGxSZW5kZXJlclRleHQiIGlkPSJhY2NvdW50LW5hbWUtcmVuZGVyZWQiLz4KICAgICAg
ICAgICAgICAgICAgICAgIDxhdHRyaWJ1dGVzPgogICAgICAgICAgICAgICAgICAgICAgICA8YXR0cmli
dXRlIG5hbWU9InRleHQiPjA8L2F0dHJpYnV0ZT4KICAgICAgICAgICAgICAgICAgICAgIDwvYXR0cmli
dXRlcz4KICAgICAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgICAgICA8L29iamVj
dD4KICAgICAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9InRvcC1hdHRhY2giPjA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9IndpZHRoIj4zPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgPC9wYWNraW5nPgog
ICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAg
ICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJzZXJ2aWNlLWxhYmVsIiA+CiAgICAgICAg
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPnRydWU8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPlNlcnZp
Y2U6PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ianVzdGlmeSI+
R1RLX0pVU1RJRllfUklHSFQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJoYWxpZ24iPkdUS19BTElHTl9TVEFSVDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgIDwv
b2JqZWN0PgogICAgICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0idG9wLWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0id2lkdGgiPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICA8L3BhY2tp
bmc+CiAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAg
ICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtFbnRyeSIgaWQ9InNlcnZpY2UiPgogICAgICAgICAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuX2ZvY3VzIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0id2lkdGhfY2hhcnMiPjQwPC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idGV4dCI+Y29uZmVyZW5jZS5yaXNldXAubmV0
PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICAgIDxw
YWNraW5nPgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3BfYXR0YWNoIj4xPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdF9hdHRhY2giPjE8
L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ3aWR0aCI+MzwvcHJv
cGVydHk+CiAgICAgICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgICAgIDwvY2hpbGQ+
CiAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0
a0J1dHRvbiIgaWQ9ImZldGNoLXJvb21zIj4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5GZXRjaCByb29tczwvcHJvcGVydHk+CiAgICAgICAg
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPlRydWU8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPlRydWU8L3Byb3BlcnR5PgogICAg
ICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJyZWNlaXZlc19kZWZhdWx0Ij5UcnVlPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8c2lnbmFsIG5hbWU9ImNsaWNrZWQiIGhhbmRsZXI9ImZl
dGNoX3Jvb21zX2hhbmRsZXIiIHN3YXBwZWQ9Im5vIi8+CiAgICAgICAgICAgICAgICAgIDwvb2JqZWN0
PgogICAgICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0idG9wX2F0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9ImxlZnRfYXR0YWNoIj40PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgPC9wYWNraW5n
PgogICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAg
ICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJleHBhbmQiPkZhbHNl
PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj5UcnVlPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MDwvcHJvcGVydHk+
CiAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8
Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJlbXB0eS1saXN0
LWxhYmVsIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCI+PC9wcm9wZXJ0eT4K
ICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5GYWxzZTwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLXRvcCI+NDA8L3Byb3BlcnR5PgogICAg
ICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+
CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrU2Nyb2xsZWRXaW5kb3ciIGlkPSJyb29tLWxp
c3Qtc2Nyb2xsIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPlRydWU8L3By
b3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InNoYWRvd190eXBlIj5pbjwvcHJv
cGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ib3ZlcmxheS1zY3JvbGxpbmciPkZh
bHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtaW4tY29udGVudC1o
ZWlnaHQiPjQwMDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAg
ICAgIDxvYmplY3QgY2xhc3M9Ikd0a1RyZWVWaWV3IiBpZD0icm9vbXMtbGlzdC12aWV3Ij4KICAgICAg
ICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+VHJ1ZTwvcHJvcGVydHk+CiAgICAg
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbl9mb2N1cyI+VHJ1ZTwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1vZGVsIj5yb29tczwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImVuYWJsZV9ncmlkX2xpbmVzIj5ib3RoPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgICA8
b2JqZWN0IGNsYXNzPSJHdGtUcmVlVmlld0NvbHVtbiI+CiAgICAgICAgICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJ0aXRsZSI+Um9vbSBOYW1lPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAg
ICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0Nl
bGxSZW5kZXJlclRleHQiLz4KICAgICAgICAgICAgICAgICAgICAgICAgICA8YXR0cmlidXRlcz4KICAg
ICAgICAgICAgICAgICAgICAgICAgICAgIDxhdHRyaWJ1dGUgbmFtZT0idGV4dCI+MDwvYXR0cmlidXRl
PgogICAgICAgICAgICAgICAgICAgICAgICAgIDwvYXR0cmlidXRlcz4KICAgICAgICAgICAgICAgICAg
ICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAg
ICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAg
ICAgICAgPG9iamVjdCBjbGFzcz0iR3RrVHJlZVZpZXdDb2x1bW4iPgogICAgICAgICAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0idGl0bGUiPkRlc2NyaXB0aW9uPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xh
c3M9Ikd0a0NlbGxSZW5kZXJlclRleHQiLz4KICAgICAgICAgICAgICAgICAgICAgICAgICA8YXR0cmli
dXRlcz4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxhdHRyaWJ1dGUgbmFtZT0idGV4dCI+MTwv
YXR0cmlidXRlPgogICAgICAgICAgICAgICAgICAgICAgICAgIDwvYXR0cmlidXRlcz4KICAgICAgICAg
ICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAg
ICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgPGNoaWxkIGludGVybmFs
LWNoaWxkPSJzZWxlY3Rpb24iPgogICAgICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3Rr
VHJlZVNlbGVjdGlvbiIgaWQ9InJvb20tc2VsZWN0aW9uIi8+CiAgICAgICAgICAgICAgICAgICAgPC9j
aGlsZD4KICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICA8L2NoaWxkPgog
ICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9ImZpbGwiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9InBhZGRpbmciPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJwb3NpdGlvbiI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAg
ICAgICA8L2NoaWxkPgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICA8cGFja2luZz4KICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0iZmlsbCI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJwb3NpdGlvbiI+MDwvcHJvcGVydHk+CiAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgPC9j
aGlsZD4KICAgICAgICA8Y2hpbGQgaW50ZXJuYWwtY2hpbGQ9ImFjdGlvbl9hcmVhIj4KICAgICAgICAg
IDxvYmplY3QgY2xhc3M9Ikd0a0J1dHRvbkJveCIgaWQ9ImRpYWxvZy1hY3Rpb25fYXJlYTEiPgogICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuX2ZvY3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJsYXlvdXRfc3R5bGUiPmVuZDwvcHJvcGVydHk+CiAgICAgICAgICAg
IDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCdXR0b24iIGlkPSJjYW5jZWwt
YnV0dG9uIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxl
PSJ5ZXMiPkNhbmNlbDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlz
aWJsZSI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuX2Zv
Y3VzIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJyZWNlaXZl
c19kZWZhdWx0Ij5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxzaWduYWwgbmFtZT0iY2xp
Y2tlZCIgaGFuZGxlcj0iY2FuY2VsX2hhbmRsZXIiIHN3YXBwZWQ9Im5vIi8+CiAgICAgICAgICAgICAg
PC9vYmplY3Q+CiAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0iZXhwYW5kIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJmaWxsIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3Np
dGlvbiI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICA8L2No
aWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQnV0
dG9uIiBpZD0iam9pbi1yb29tIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIg
dHJhbnNsYXRhYmxlPSJ5ZXMiPkpvaW4gcm9vbTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0idmlzaWJsZSI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0iY2FuX2ZvY3VzIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJyZWNlaXZlc19kZWZhdWx0Ij5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxz
aWduYWwgbmFtZT0iY2xpY2tlZCIgaGFuZGxlcj0iam9pbl9zZWxlY3RlZF9yb29tX2hhbmRsZXIiIHN3
YXBwZWQ9Im5vIi8+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgPHBhY2tpbmc+
CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZXhwYW5kIj5UcnVlPC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAg
PC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAg
ICA8cGFja2luZz4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+RmFsc2U8L3Byb3Bl
cnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZmlsbCI+RmFsc2U8L3Byb3BlcnR5PgogICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjE8L3Byb3BlcnR5PgogICAgICAgICAgPC9w
YWNraW5nPgogICAgICAgIDwvY2hpbGQ+CiAgICAgIDwvb2JqZWN0PgogICAgPC9jaGlsZD4KICA8L29i
amVjdD4KPC9pbnRlcmZhY2U+Cg==
`,
	},

	"/definitions/Main.xml": {
		local:   "definitions/Main.xml",
		size:    11894,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtBcHBsaWNhdGlvbldpbmRvdyIgaWQ9Im1haW5X
aW5kb3ciPgogICAgPHByb3BlcnR5IG5hbWU9ImNhbl9mb2N1cyI+RmFsc2U8L3Byb3BlcnR5PgogICAg
PHByb3BlcnR5IG5hbWU9InRpdGxlIj5Db3lJTTwvcHJvcGVydHk+CiAgICA8cHJvcGVydHkgbmFtZT0i
ZGVmYXVsdF93aWR0aCI+MjAwPC9wcm9wZXJ0eT4KICAgIDxwcm9wZXJ0eSBuYW1lPSJkZWZhdWx0X2hl
aWdodCI+NjAwPC9wcm9wZXJ0eT4KICAgIDxzaWduYWwgbmFtZT0iZGVzdHJveSIgaGFuZGxlcj0ib25f
Y2xvc2Vfd2luZG93X3NpZ25hbCIgc3dhcHBlZD0ibm8iLz4KICAgIDxjaGlsZD4KICAgICAgPG9iamVj
dCBjbGFzcz0iR3RrQm94IiBpZD0iSGJveCI+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUi
PlRydWU8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPkZhbHNlPC9w
cm9wZXJ0eT4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCb3giIGlk
PSJWYm94Ij4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbl9mb2N1cyI+RmFsc2U8L3Byb3Bl
cnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ib3JpZW50YXRpb24iPnZlcnRpY2FsPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a01l
bnVCYXIiIGlkPSJtZW51YmFyIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9j
dXMiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICAg
ICAgPG9iamVjdCBjbGFzcz0iR3RrTWVudUl0ZW0iIGlkPSJDb250YWN0c01lbnUiPgogICAgICAgICAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5fQ29u
dGFjdHM8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ1c2UtdW5k
ZXJsaW5lIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8Y2hpbGQgdHlwZT0ic3Vi
bWVudSI+CiAgICAgICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtNZW51IiBpZD0ibWVu
dSI+CiAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPkZhbHNl
PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAg
ICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a01lbnVJdGVtIiBpZD0iYWRkTWVudSI+CiAgICAgICAg
ICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuX2ZvY3VzIj5GYWxzZTwvcHJvcGVy
dHk+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5z
bGF0YWJsZT0ieWVzIj5BZGQuLi48L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgICAg
PHNpZ25hbCBuYW1lPSJhY3RpdmF0ZSIgaGFuZGxlcj0ib25fYWRkX2NvbnRhY3Rfd2luZG93X3NpZ25h
bCIgc3dhcHBlZD0ibm8iLz4KICAgICAgICAgICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAg
ICAgICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgICAgICAgICAgPGNoaWxkPgog
ICAgICAgICAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a01lbnVJdGVtIiBpZD0ibmV3
Q29udk1lbnUiPgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbl9m
b2N1cyI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+TmV3IENvbnZlcnNhdGlvbi4uLjwvcHJvcGVy
dHk+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICA8c2lnbmFsIG5hbWU9ImFjdGl2YXRlIiBoYW5k
bGVyPSJvbl9uZXdfY29udmVyc2F0aW9uX3NpZ25hbCIgc3dhcHBlZD0ibm8iLz4KICAgICAgICAgICAg
ICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICAgICAgICAgICAgPC9jaGlsZD4KICAg
ICAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAg
ICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAg
ICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTWVudUl0ZW0i
IGlkPSJBY2NvdW50c01lbnUiPgogICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5f
Zm9jdXMiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
bGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5fQWNjb3VudHM8L3Byb3BlcnR5PgogICAgICAgICAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ1c2UtdW5kZXJsaW5lIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAg
ICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtNZW51SXRlbSIgaWQ9
IlZpZXdNZW51Ij4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuX2ZvY3VzIj5G
YWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0
cmFuc2xhdGFibGU9InllcyI+X1ZpZXc8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJ1c2UtdW5kZXJsaW5lIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAg
ICA8Y2hpbGQgdHlwZT0ic3VibWVudSI+CiAgICAgICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNz
PSJHdGtNZW51IiBpZD0ibWVudTIiPgogICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0iY2FuX2ZvY3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAgIDxjaGls
ZD4KICAgICAgICAgICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtDaGVja01lbnVJdGVt
IiBpZD0iQ2hlY2tJdGVtTWVyZ2UiPgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9ImNhbl9mb2N1cyI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+TWVyZ2UgQWNjb3VudHM8
L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJ0b2dnbGVk
IiBoYW5kbGVyPSJvbl90b2dnbGVkX2NoZWNrX0l0ZW1fTWVyZ2Vfc2lnbmFsIiBzd2FwcGVkPSJubyIv
PgogICAgICAgICAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgICAgICAgICAg
ICA8L2NoaWxkPgogICAgICAgICAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAg
ICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQ2hlY2tNZW51SXRlbSIgaWQ9IkNoZWNrSXRlbVNob3dP
ZmZsaW5lIj4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9j
dXMiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPlNob3cgT2ZmbGluZSBDb250YWN0czwvcHJvcGVy
dHk+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICA8c2lnbmFsIG5hbWU9InRvZ2dsZWQiIGhhbmRs
ZXI9Im9uX3RvZ2dsZWRfY2hlY2tfSXRlbV9TaG93X09mZmxpbmVfc2lnbmFsIiBzd2FwcGVkPSJubyIv
PgogICAgICAgICAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgICAgICAgICAg
ICA8L2NoaWxkPgogICAgICAgICAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAg
ICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQ2hlY2tNZW51SXRlbSIgaWQ9IkNoZWNrSXRlbVNob3dX
YWl0aW5nIj4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9j
dXMiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPlNob3cgV2FpdGluZyBDb250YWN0czwvcHJvcGVy
dHk+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICA8c2lnbmFsIG5hbWU9InRvZ2dsZWQiIGhhbmRs
ZXI9Im9uX3RvZ2dsZWRfY2hlY2tfSXRlbV9TaG93X1dhaXRpbmdfc2lnbmFsIiBzd2FwcGVkPSJubyIv
PgogICAgICAgICAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgICAgICAgICAg
ICA8L2NoaWxkPgogICAgICAgICAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAg
ICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQ2hlY2tNZW51SXRlbSIgaWQ9IkNoZWNrSXRlbVNvcnRT
dGF0dXMiPgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbl9mb2N1
cyI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+U29ydCBCeSBTdGF0dXM8L3Byb3BlcnR5PgogICAg
ICAgICAgICAgICAgICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJ0b2dnbGVkIiBoYW5kbGVyPSJvbl90
b2dnbGVkX2NoZWNrX0l0ZW1fU29ydF9CeV9TdGF0dXNfc2lnbmFsIiBzd2FwcGVkPSJubyIvPgogICAg
ICAgICAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgICAgICAgICAgICA8L2No
aWxkPgogICAgICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICAgICAgPC9j
aGlsZD4KICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICA8L2NoaWxkPgog
ICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtN
ZW51SXRlbSIgaWQ9Ik9wdGlvbnNNZW51Ij4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0iY2FuLWZvY3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+X09wdGlvbnM8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ1c2UtdW5kZXJsaW5lIj5UcnVlPC9wcm9wZXJ0eT4K
ICAgICAgICAgICAgICAgICAgICA8Y2hpbGQgdHlwZT0ic3VibWVudSI+CiAgICAgICAgICAgICAgICAg
ICAgICA8b2JqZWN0IGNsYXNzPSJHdGtNZW51IiBpZD0ib3B0aW9uc19zdWJtZW51Ij4KICAgICAgICAg
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbl9mb2N1cyI+RmFsc2U8L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgICAgICAgPG9i
amVjdCBjbGFzcz0iR3RrQ2hlY2tNZW51SXRlbSIgaWQ9IkVuY3J5cHRDb25maWd1cmF0aW9uRmlsZUNo
ZWNrTWVudUl0ZW0iPgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNh
bl9mb2N1cyI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+RW5jcnlwdCBjb25maWd1cmF0aW9uIGZp
bGU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJ0b2dn
bGVkIiBoYW5kbGVyPSJvbl90b2dnbGVkX2VuY3J5cHRfY29uZmlndXJhdGlvbl9maWxlX3NpZ25hbCIg
c3dhcHBlZD0ibm8iLz4KICAgICAgICAgICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAg
ICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgICAgICAgICAgPGNoaWxkPgogICAg
ICAgICAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a01lbnVJdGVtIiBpZD0icHJlZmVy
ZW5jZXNNZW51SXRlbSI+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
Y2FuX2ZvY3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5QcmVmZXJlbmNlcy4uLjwvcHJvcGVy
dHk+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idXNlLXVuZGVybGlu
ZSI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICA8c2lnbmFsIG5hbWU9
ImFjdGl2YXRlIiBoYW5kbGVyPSJvbl9wcmVmZXJlbmNlc19zaWduYWwiIHN3YXBwZWQ9Im5vIi8+CiAg
ICAgICAgICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICAgICAgICAgIDwv
Y2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICAgICAgICA8
L2NoaWxkPgogICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICAgIDwvY2hpbGQ+
CgogICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJH
dGtNZW51SXRlbSIgaWQ9Im11Yy1tb2NrdXAtbWVudSI+CiAgICAgICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9InZpc2libGUiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5DaGF0PC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgICAgICAgICA8Y2hpbGQgdHlwZT0ic3VibWVudSI+CiAgICAgICAgICAgICAgICAgICAgICA8
b2JqZWN0IGNsYXNzPSJHdGtNZW51IiBpZD0ibXVjLW1vY2t1cC1tZW51LW9wdGlvbnMiPgogICAgICAg
ICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuX2ZvY3VzIj5GYWxzZTwvcHJvcGVydHk+
CiAgICAgICAgICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICAgICAgICAgICAgICA8
b2JqZWN0IGNsYXNzPSJHdGtNZW51SXRlbSIgaWQ9Im11Yy1hZGQtY2hhdC1yb29tLW1lbnUiPgogICAg
ICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9
InllcyI+Sm9pbiBDaGF0IFJvb20uLi48L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9InVzZS11bmRlcmxpbmUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAg
ICAgICAgICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJhY3RpdmF0ZSIgaGFuZGxlcj0ib25fam9pbl9j
aGF0X3Jvb20iIHN3YXBwZWQ9Im5vIi8+CiAgICAgICAgICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+
CiAgICAgICAgICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgICAgIDxj
aGlsZD4KICAgICAgICAgICAgICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtNZW51SXRlbSIg
aWQ9Im11Yy1saXN0LWNoYXQtcm9vbXMtbWVudSI+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5MaXN0IENoYXQgUm9vbXMuLi48
L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InVzZS11
bmRlcmxpbmUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHNpZ25h
bCBuYW1lPSJhY3RpdmF0ZSIgaGFuZGxlcj0ib25fbGlzdF9jaGF0X3Jvb21zIiBzd2FwcGVkPSJubyIv
PgogICAgICAgICAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgICAgICAgICAg
ICA8L2NoaWxkPgogICAgICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICAg
ICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICA8L2No
aWxkPgoKICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFz
cz0iR3RrTWVudUl0ZW0iIGlkPSJIZWxwTWVudSI+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9ImNhbl9mb2N1cyI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPl9IZWxwPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idXNlLXVuZGVybGluZSI+VHJ1ZTwvcHJvcGVydHk+
CiAgICAgICAgICAgICAgICAgICAgPGNoaWxkIHR5cGU9InN1Ym1lbnUiPgogICAgICAgICAgICAgICAg
ICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTWVudSIgaWQ9Im1lbnUzIj4KICAgICAgICAgICAgICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbl9mb2N1cyI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAg
ICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFz
cz0iR3RrTWVudUl0ZW0iIGlkPSJmZWVkYmFja01lbnUiPgogICAgICAgICAgICAgICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9ImNhbl9mb2N1cyI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAg
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+RmVl
ZGJhY2s8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJh
Y3RpdmF0ZSIgaGFuZGxlcj0ib25fZmVlZGJhY2tfZGlhbG9nX3NpZ25hbCIgc3dhcHBlZD0ibm8iLz4K
ICAgICAgICAgICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICAgICAgICAgICAg
PC9jaGlsZD4KICAgICAgICAgICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICAg
ICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a01lbnVJdGVtIiBpZD0iYWJvdXRNZW51Ij4KICAgICAgICAg
ICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPkZhbHNlPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNs
YXRhYmxlPSJ5ZXMiPkFib3V0PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxz
aWduYWwgbmFtZT0iYWN0aXZhdGUiIGhhbmRsZXI9Im9uX2Fib3V0X2RpYWxvZ19zaWduYWwiIHN3YXBw
ZWQ9Im5vIi8+CiAgICAgICAgICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAg
ICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAg
ICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAg
ICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgPHBhY2tpbmc+
CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZXhwYW5kIj5GYWxzZTwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZmlsbCI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjA8L3Byb3BlcnR5PgogICAgICAgICAgICAg
IDwvcGFja2luZz4KICAgICAgICAgICAgPC9jaGlsZD4KCSAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAg
PG9iamVjdCBjbGFzcz0iR3RrQm94IiBpZD0ibm90aWZpY2F0aW9uLWFyZWEiPgogICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9ImNhbl9mb2N1cyI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9Im9yaWVudGF0aW9uIj52ZXJ0aWNhbDwvcHJvcGVydHk+CiAgICAgICAg
ICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0iZXhwYW5kIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0iZmlsbCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9InBhY2tfdHlwZSI+ZW5kPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJwb3NpdGlvbiI+MjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAg
ICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0i
R3RrQm94IiBpZD0ic2VhcmNoLWJveCI+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2Fu
X2ZvY3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ib3Jp
ZW50YXRpb24iPnZlcnRpY2FsPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAg
ICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrU2VhcmNoQmFyIiBpZD0ic2VhcmNoLWFyZWEiPgog
ICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPkZhbHNlPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ic2hvdy1jbG9zZS1idXR0b24iPkZh
bHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ic2VhcmNoLW1v
ZGUtZW5hYmxlZCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxjaGlsZD4KICAg
ICAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0VudHJ5IiBpZD0ic2VhcmNoLWVudHJ5
Ij4KICAgICAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPlRydWU8L3By
b3BlcnR5PgogICAgICAgICAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ic2hhZG93X3R5cGUi
PmV0Y2hlZC1pbjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJzZWNvbmRhcnlfaWNvbl9uYW1lIj5lZGl0LWZpbmQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0icGxhY2Vob2xkZXJfdGV4dCIgdHJhbnNsYXRhYmxlPSJ5ZXMi
PlNlYXJjaC4uLjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAg
ICAgICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAg
ICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgPHBhY2tp
bmc+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icGFja190eXBlIj5lbmQ8L3Byb3BlcnR5
PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InBvc2l0aW9uIj4xPC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgICA8L3BhY2tpbmc+CgkgICAgPC9jaGlsZD4KCSAgPC9vYmplY3Q+CiAgICAgICAgICA8
cGFja2luZz4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+VHJ1ZTwvcHJvcGVydHk+
CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9InBvc2l0aW9uIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgIDwvcGFja2lu
Zz4KICAgICAgICA8L2NoaWxkPgogICAgICA8L29iamVjdD4KICAgIDwvY2hpbGQ+CiAgPC9vYmplY3Q+
CjwvaW50ZXJmYWNlPgo=
`,
	},

	"/definitions/MasterPassword.xml": {
		local:   "definitions/MasterPassword.xml",
		size:    2357,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtEaWFsb2ciIGlkPSJNYXN0ZXJQYXNzd29yZCI+
CiAgICA8cHJvcGVydHkgbmFtZT0id2luZG93LXBvc2l0aW9uIj5HVEtfV0lOX1BPU19DRU5URVI8L3By
b3BlcnR5PgogICAgPHByb3BlcnR5IG5hbWU9InRpdGxlIiB0cmFuc2xhdGFibGU9InllcyI+RW50ZXIg
bWFzdGVyIHBhc3N3b3JkPC9wcm9wZXJ0eT4KICAgIDxwcm9wZXJ0eSBuYW1lPSJkZWZhdWx0LXdpZHRo
Ij4zMDA8L3Byb3BlcnR5PgogICAgPHNpZ25hbCBuYW1lPSJkZWxldGUtZXZlbnQiIGhhbmRsZXI9Im9u
X2NhbmNlbF9zaWduYWwiIC8+CiAgICA8Y2hpbGQgaW50ZXJuYWwtY2hpbGQ9InZib3giPgogICAgICA8
b2JqZWN0IGNsYXNzPSJHdGtCb3giIGlkPSJWYm94Ij4KICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFy
Z2luIj4xMDwvcHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9InNwYWNpbmciPjEwPC9wcm9w
ZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iaG9tb2dlbmVvdXMiPmZhbHNlPC9wcm9wZXJ0eT4K
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ib3JpZW50YXRpb24iPkdUS19PUklFTlRBVElPTl9WRVJUSUNB
TDwvcHJvcGVydHk+CiAgICAgICAgPGNoaWxkPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFi
ZWwiIGlkPSJwYXNzTWVzc2FnZSIgPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRy
YW5zbGF0YWJsZT0ieWVzIj5QbGVhc2UgZW50ZXIgdGhlIG1hc3RlciBwYXNzd29yZCBmb3IgdGhlIGNv
bmZpZ3VyYXRpb24gZmlsZS4gWW91IHdpbGwgbm90IGJlIGFza2VkIGZvciB0aGlzIHBhc3N3b3JkIGFn
YWluIHVudGlsIHlvdSByZXN0YXJ0IENveUlNLjwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJ3cmFwIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAg
PHBhY2tpbmc+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJleHBhbmQiPmZhbHNlPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImZpbGwiPnRydWU8L3Byb3BlcnR5PgogICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjA8L3Byb3BlcnR5PgogICAgICAgICAgPC9wYWNr
aW5nPgogICAgICAgIDwvY2hpbGQ+CiAgICAgICAgPGNoaWxkPgogICAgICAgICAgPG9iamVjdCBjbGFz
cz0iR3RrRW50cnkiIGlkPSJwYXNzd29yZCI+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoYXMt
Zm9jdXMiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJpbGl0
eSI+ZmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICA8c2lnbmFsIG5hbWU9ImFjdGl2YXRlIiBoYW5k
bGVyPSJvbl9zYXZlX3NpZ25hbCIgLz4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgPHBhY2tp
bmc+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJleHBhbmQiPmZhbHNlPC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImZpbGwiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjE8L3Byb3BlcnR5PgogICAgICAgICAgPC9wYWNraW5nPgog
ICAgICAgIDwvY2hpbGQ+CiAgICAgIDwvb2JqZWN0PgogICAgPC9jaGlsZD4KICAgIDxjaGlsZCBpbnRl
cm5hbC1jaGlsZD0iYWN0aW9uX2FyZWEiPgogICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCdXR0b25Cb3gi
IGlkPSJidXR0b25fYm94Ij4KICAgICAgICA8cHJvcGVydHkgbmFtZT0ib3JpZW50YXRpb24iPkdUS19P
UklFTlRBVElPTl9IT1JJWk9OVEFMPC9wcm9wZXJ0eT4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8
b2JqZWN0IGNsYXNzPSJHdGtCdXR0b24iIGlkPSJjYW5jZWwiPgogICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5DYW5jZWw8L3Byb3BlcnR5PgogICAgICAgICAg
ICA8c2lnbmFsIG5hbWU9ImNsaWNrZWQiIGhhbmRsZXI9Im9uX2NhbmNlbF9zaWduYWwiIC8+CiAgICAg
ICAgICA8L29iamVjdD4KICAgICAgICA8L2NoaWxkPgogICAgICAgIDxjaGlsZD4KICAgICAgICAgIDxv
YmplY3QgY2xhc3M9Ikd0a0J1dHRvbiIgaWQ9InNhdmUiPgogICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5PSzwvcHJvcGVydHk+CiAgICAgICAgICAgIDxzaWdu
YWwgbmFtZT0iY2xpY2tlZCIgaGFuZGxlcj0ib25fc2F2ZV9zaWduYWwiIC8+CiAgICAgICAgICA8L29i
amVjdD4KICAgICAgICA8L2NoaWxkPgogICAgICA8L29iamVjdD4KICAgIDwvY2hpbGQ+CiAgPC9vYmpl
Y3Q+CjwvaW50ZXJmYWNlPgo=
`,
	},

	"/definitions/NewCustomConversation.xml": {
		local:   "definitions/NewCustomConversation.xml",
		size:    4728,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtMaXN0U3RvcmUiIGlkPSJhY2NvdW50cy1tb2Rl
bCI+CiAgICA8Y29sdW1ucz4KICAgICAgPCEtLSBhY2NvdW50IG5hbWUgLS0+CiAgICAgIDxjb2x1bW4g
dHlwZT0iZ2NoYXJhcnJheSIvPgogICAgICA8IS0tIGFjY291bnQgaWQgLS0+CiAgICAgIDxjb2x1bW4g
dHlwZT0iZ2NoYXJhcnJheSIvPgogICAgPC9jb2x1bW5zPgogIDwvb2JqZWN0PgogIDxvYmplY3QgY2xh
c3M9Ikd0a0RpYWxvZyIgaWQ9Ik5ld0N1c3RvbUNvbnZlcnNhdGlvbiI+CiAgICA8cHJvcGVydHkgbmFt
ZT0id2luZG93LXBvc2l0aW9uIj5HVEtfV0lOX1BPU19DRU5URVI8L3Byb3BlcnR5PgogICAgPHByb3Bl
cnR5IG5hbWU9ImJvcmRlcl93aWR0aCI+NjwvcHJvcGVydHk+CiAgICA8cHJvcGVydHkgbmFtZT0idGl0
bGUiIHRyYW5zbGF0YWJsZT0ieWVzIj5TdGFydCBuZXcgY29udmVyc2F0aW9uPC9wcm9wZXJ0eT4KICAg
IDxwcm9wZXJ0eSBuYW1lPSJyZXNpemFibGUiPlRydWU8L3Byb3BlcnR5PgogICAgPHByb3BlcnR5IG5h
bWU9ImRlZmF1bHQtaGVpZ2h0Ij4yMDA8L3Byb3BlcnR5PgogICAgPHByb3BlcnR5IG5hbWU9ImRlZmF1
bHQtd2lkdGgiPjUwMDwvcHJvcGVydHk+CiAgICA8cHJvcGVydHkgbmFtZT0iZGVzdHJveS13aXRoLXBh
cmVudCI+dHJ1ZTwvcHJvcGVydHk+CiAgICA8Y2hpbGQgaW50ZXJuYWwtY2hpbGQ9InZib3giPgogICAg
ICA8b2JqZWN0IGNsYXNzPSJHdGtCb3giIGlkPSJWYm94Ij4KICAgICAgICA8cHJvcGVydHkgbmFtZT0i
aG9tb2dlbmVvdXMiPmZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0ib3JpZW50
YXRpb24iPkdUS19PUklFTlRBVElPTl9WRVJUSUNBTDwvcHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5
IG5hbWU9InNwYWNpbmciPjY8L3Byb3BlcnR5PgogICAgICAgIDxjaGlsZD4KICAgICAgICAgIDxvYmpl
Y3QgY2xhc3M9Ikd0a0dyaWQiIGlkPSJncmlkIj4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1h
cmdpbi10b3AiPjE1PC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi1i
b3R0b20iPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi1zdGFy
dCI+MTA8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLWVuZCI+MTA8
L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icm93LXNwYWNpbmciPjEyPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNvbHVtbi1zcGFjaW5nIj42PC9wcm9wZXJ0
eT4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVs
IiBpZD0iYWNjb3VudHNMYWJlbCIgPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVs
IiB0cmFuc2xhdGFibGU9InllcyI+RnJvbTo8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9Imp1c3RpZnkiPkdUS19KVVNUSUZZX1JJR0hUPC9wcm9wZXJ0eT4KICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJoYWxpZ24iPkdUS19BTElHTl9FTkQ8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+
CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2Jq
ZWN0IGNsYXNzPSJHdGtDb21ib0JveCIgaWQ9ImFjY291bnRzIj4KICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJtb2RlbCI+YWNjb3VudHMtbW9kZWw8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9Imhhcy1mb2N1cyI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0iaGV4cGFuZCI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8Y2hp
bGQ+CiAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0NlbGxSZW5kZXJlclRleHQiIGlk
PSJhY2NvdW50LW5hbWUtcmVuZGVyZWQiLz4KICAgICAgICAgICAgICAgICAgPGF0dHJpYnV0ZXM+CiAg
ICAgICAgICAgICAgICAgICAgPGF0dHJpYnV0ZSBuYW1lPSJ0ZXh0Ij4wPC9hdHRyaWJ1dGU+CiAgICAg
ICAgICAgICAgICAgIDwvYXR0cmlidXRlcz4KICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAg
ICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9InRvcC1hdHRhY2giPjA8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvcGFja2luZz4K
ICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmpl
Y3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0icGVlckxhYmVsIiA+CiAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5XaXRoOjwvcHJvcGVydHk+CiAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ianVzdGlmeSI+R1RLX0pVU1RJRllfUklHSFQ8L3Byb3BlcnR5
PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+R1RLX0FMSUdOX0VORDwvcHJv
cGVydHk+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAg
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRhY2giPjA8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAgICAg
ICAgIDwvcGFja2luZz4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNoaWxkPgogICAg
ICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0VudHJ5IiBpZD0iYWRkcmVzcyI+CiAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0icGxhY2Vob2xkZXItdGV4dCI+c29tZW9uZUBqYWJiZXIub3JnPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoZXhwYW5kIj5UcnVlPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICAgIDxzaWduYWwgbmFtZT0iYWN0aXZhdGUiIGhhbmRsZXI9Im9uX3N0
YXJ0X3NpZ25hbCIgLz4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2lu
Zz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MTwvcHJvcGVydHk+
CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+MTwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgPC9vYmpl
Y3Q+CiAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+
dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj50cnVlPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InBvc2l0aW9uIj4xPC9wcm9wZXJ0eT4KICAg
ICAgICAgIDwvcGFja2luZz4KICAgICAgICA8L2NoaWxkPgogICAgICAgIDxjaGlsZCBpbnRlcm5hbC1j
aGlsZD0iYWN0aW9uX2FyZWEiPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQnV0dG9uQm94IiBp
ZD0iYnV0dG9uX2JveCI+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJvcmllbnRhdGlvbiI+R1RL
X09SSUVOVEFUSU9OX0hPUklaT05UQUw8L3Byb3BlcnR5PgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAg
ICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQnV0dG9uIiBpZD0iYnV0dG9uX2NhbmNlbCI+CiAgICAg
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiPl9DYW5jZWw8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgICAgPHByb3BlcnR5IG5hbWU9InVzZS11bmRlcmxpbmUiPlRydWU8L3Byb3BlcnR5PgogICAg
ICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJjbGlja2VkIiBoYW5kbGVyPSJvbl9jbG9zZV9zaWduYWwi
IC8+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAg
IDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCdXR0b24iIGlkPSJidXR0b25f
b2siPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9Inll
cyI+U3RhcnQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InVzZS11bmRl
cmxpbmUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbi1k
ZWZhdWx0Ij50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxzaWduYWwgbmFtZT0iY2xpY2tl
ZCIgaGFuZGxlcj0ib25fc3RhcnRfc2lnbmFsIiAvPgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAg
ICAgICAgICA8L2NoaWxkPgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgPC9jaGlsZD4KICAgICAg
PC9vYmplY3Q+CiAgICA8L2NoaWxkPgogIDwvb2JqZWN0Pgo8L2ludGVyZmFjZT4K
`,
	},

	"/definitions/PeerDetails.xml": {
		local:   "definitions/PeerDetails.xml",
		size:    9959,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtMaXN0U3RvcmUiIGlkPSJjdXJyZW50LWdyb3Vw
cyI+CiAgICA8Y29sdW1ucz4KICAgICAgPGNvbHVtbiB0eXBlPSJnY2hhcmFycmF5Ii8+CiAgICA8L2Nv
bHVtbnM+CiAgPC9vYmplY3Q+CiAgPG9iamVjdCBjbGFzcz0iR3RrTWVudUl0ZW0iIGlkPSJhZGRHcm91
cCI+CiAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+VHJ1ZTwvcHJvcGVydHk+CiAgICA8cHJvcGVy
dHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5OZXcgR3JvdXAuLi48L3Byb3BlcnR5Pgog
ICAgPHNpZ25hbCBuYW1lPSJhY3RpdmF0ZSIgaGFuZGxlcj0ib24tYWRkLW5ldy1ncm91cCIgLz4KICA8
L29iamVjdD4KICA8b2JqZWN0IGNsYXNzPSJHdGtNZW51IiBpZD0iZ3JvdXBzLW1lbnUiPgogICAgPHBy
b3BlcnR5IG5hbWU9InZpc2libGUiPlRydWU8L3Byb3BlcnR5PgogIDwvb2JqZWN0PgogIDxvYmplY3Qg
Y2xhc3M9Ikd0a0RpYWxvZyIgaWQ9ImRpYWxvZyI+CiAgICA8cHJvcGVydHkgbmFtZT0idGl0bGUiIHRy
YW5zbGF0YWJsZT0ieWVzIj5FZGl0IGNvbnRhY3Q8L3Byb3BlcnR5PgogICAgPHNpZ25hbCBuYW1lPSJj
bG9zZSIgaGFuZGxlcj0ib24tY2FuY2VsIiAvPgogICAgPGNoaWxkIGludGVybmFsLWNoaWxkPSJ2Ym94
Ij4KICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQm94IiBpZD0iVmJveCI+CiAgICAgICAgPHByb3BlcnR5
IG5hbWU9Im1hcmdpbiI+MTA8L3Byb3BlcnR5PgogICAgICAgIDxjaGlsZD4KICAgICAgICAgIDxvYmpl
Y3QgY2xhc3M9Ikd0a0dyaWQiIGlkPSJncmlkIj4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1h
cmdpbi10b3AiPjE1PC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi1i
b3R0b20iPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi1zdGFy
dCI+MTA8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLWVuZCI+MTA8
L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icm93LXNwYWNpbmciPjEyPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNvbHVtbi1zcGFjaW5nIj42PC9wcm9wZXJ0
eT4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVs
IiBpZD0iYWNjb3VudC1sYWJlbCI+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwi
IHRyYW5zbGF0YWJsZT0ieWVzIj5BY2NvdW50OjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05fRU5EPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJzZWxlY3RhYmxlIj5UUlVFPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8
L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJsZWZ0LWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0idG9wLWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAg
ICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFz
cz0iR3RrTGFiZWwiIGlkPSJhY2NvdW50LW5hbWUiPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9ImhhbGlnbiI+R1RLX0FMSUdOX1NUQVJUPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJzZWxlY3RhYmxlIj5UUlVFPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L29iamVj
dD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJs
ZWZ0LWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idG9w
LWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICA8
L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3Rr
TGFiZWwiIGlkPSJqaWQtbGFiZWwiPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVs
IiB0cmFuc2xhdGFibGU9InllcyI+Q29udGFjdDo8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9ImhhbGlnbiI+R1RLX0FMSUdOX0VORDwvcHJvcGVydHk+CiAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0ic2VsZWN0YWJsZSI+VFJVRTwvcHJvcGVydHk+CiAgICAgICAgICAgICAg
PC9vYmplY3Q+CiAgICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0ibGVmdC1hdHRhY2giPjA8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9InRvcC1hdHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAg
ICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xh
c3M9Ikd0a0xhYmVsIiBpZD0iamlkIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoYWxp
Z24iPkdUS19BTElHTl9TVEFSVDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0ic2VsZWN0YWJsZSI+VFJVRTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAg
ICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1hdHRh
Y2giPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRhY2gi
PjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgPC9jaGlsZD4K
ICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBp
ZD0ibmlja25hbWUtbGFiZWwiPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0
cmFuc2xhdGFibGU9InllcyI+Tmlja25hbWU6PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJoYWxpZ24iPkdUS19BTElHTl9FTkQ8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwv
b2JqZWN0PgogICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9ImxlZnQtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJ0b3AtYXR0YWNoIj4yPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAg
ICAgIDwvY2hpbGQ+CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNz
PSJHdGtFbnRyeSIgaWQ9Im5pY2tuYW1lIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJh
Y3RpdmF0ZXMtZGVmYXVsdCI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9vYmplY3Q+CiAg
ICAgICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGVmdC1h
dHRhY2giPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRvcC1hdHRh
Y2giPjI8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICAgICAgPC9jaGls
ZD4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVs
IiBpZD0icmVxdWlyZS1lbmNyeXB0aW9uLWxhYmVsIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPlJlcXVpcmUgZW5jcnlwdGlvbiB3aXRoIHRoaXMg
cGVlcjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtf
QUxJR05fRU5EPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzZWxlY3Rh
YmxlIj5UUlVFPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8
cGFja2luZz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MDwvcHJv
cGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+MzwvcHJvcGVy
dHk+CiAgICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAg
ICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQ2hlY2tCdXR0b24iIGlkPSJy
ZXF1aXJlLWVuY3J5cHRpb24iLz4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+MzwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNr
aW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAg
PG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJncm91cHMiPgogICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+R3JvdXBzOjwvcHJvcGVydHk+CiAgICAg
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05fU1RBUlQ8L3Byb3BlcnR5
PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InNlbGVjdGFibGUiPlRSVUU8L3Byb3BlcnR5
PgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj40PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJ3aWR0aCI+MjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNraW5n
PgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8IS0tIFRPRE86IFRoaXMgc2hvdWxkIGJl
IGluIGEgc2Nyb2xsZWQgd2luZG93IHdpdGggbWluL21heCBzaXplICAtLT4KICAgICAgICAgICAgPGNo
aWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a1RyZWVWaWV3IiBpZD0iZ3JvdXBzLXZp
ZXciPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1vZGVsIj5jdXJyZW50LWdyb3Vwczwv
cHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+VHJ1ZTwvcHJv
cGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuLWZvY3VzIj5UcnVlPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoZWFkZXJzLXZpc2libGUiPkZhbHNl
PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzaG93LWV4cGFuZGVycyI+
RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InJlb3JkZXJhYmxl
Ij5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxjaGlsZCBpbnRlcm5hbC1jaGlsZD0ic2Vs
ZWN0aW9uIj4KICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrVHJlZVNlbGVjdGlvbiIg
aWQ9InNlbGVjdGlvbiI+CiAgICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1vZGUiPkdU
S19TRUxFQ1RJT05fU0lOR0xFPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICA8c2lnbmFsIG5h
bWU9ImNoYW5nZWQiIGhhbmRsZXI9Im9uLWdyb3VwLXNlbGVjdGlvbi1jaGFuZ2VkIiBzd2FwcGVkPSJu
byIvPgogICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAg
ICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a1Ry
ZWVWaWV3Q29sdW1uIiBpZD0iZ3JvdXAtbmFtZS1jb2x1bW4iPgogICAgICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJ0aXRsZSI+Z3JvdXAtbmFtZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAg
ICAgPGNoaWxkPgogICAgICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQ2VsbFJlbmRl
cmVyVGV4dCIgaWQ9Imdyb3VwLW5hbWUtY29sdW1uLXJlbmRlcmVkIi8+CiAgICAgICAgICAgICAgICAg
ICAgICA8YXR0cmlidXRlcz4KICAgICAgICAgICAgICAgICAgICAgICAgPGF0dHJpYnV0ZSBuYW1lPSJ0
ZXh0Ij4wPC9hdHRyaWJ1dGU+CiAgICAgICAgICAgICAgICAgICAgICA8L2F0dHJpYnV0ZXM+CiAgICAg
ICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAg
ICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgIDxwYWNr
aW5nPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4wPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj41PC9wcm9wZXJ0eT4K
ICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ3aWR0aCI+MjwvcHJvcGVydHk+CiAgICAgICAg
ICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAg
ICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQnV0dG9uQm94IiBpZD0iZ3JvdXBzLWJ1dHRvbnMi
PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im9yaWVudGF0aW9uIj5HVEtfT1JJRU5UQVRJ
T05fSE9SSVpPTlRBTDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAg
ICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a01lbnVCdXR0b24iIGlkPSJhZGQtYnRuIj4KICAgICAgICAg
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAg
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InBvcHVwIj5ncm91cHMtbWVudTwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImRpcmVjdGlvbiI+R1RLX0FSUk9XX0RPV048
L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICAgICAgICAg
IDxvYmplY3QgY2xhc3M9Ikd0a0ltYWdlIiBpZD0iYWRkLWJ0bi1pbWciPgogICAgICAgICAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0iaWNvbi1uYW1lIj5saXN0LWFkZDwvcHJvcGVydHk+CiAgICAg
ICAgICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICAgICAgICA8L2NoaWxkPgogICAg
ICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAg
ICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0J1dHRvbiIgaWQ9
InJlbW92ZS1idG4iPgogICAgICAgICAgICAgICAgICAgIDxzaWduYWwgbmFtZT0iY2xpY2tlZCIgaGFu
ZGxlcj0ib24tcmVtb3ZlLWdyb3VwIi8+CiAgICAgICAgICAgICAgICAgICAgPGNoaWxkPgogICAgICAg
ICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrSW1hZ2UiIGlkPSJyZW1vdmUtYnRuLWltZyI+
CiAgICAgICAgICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJpY29uLW5hbWUiPmxpc3QtcmVt
b3ZlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAg
ICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgICAg
PC9jaGlsZD4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAg
ICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAg
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+NjwvcHJvcGVydHk+CiAgICAgICAg
ICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgPC9vYmplY3Q+CiAg
ICAgICAgPC9jaGlsZD4KICAgICAgICA8Y2hpbGQgaW50ZXJuYWwtY2hpbGQ9ImFjdGlvbl9hcmVhIj4K
ICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0J1dHRvbkJveCIgaWQ9ImJ1dHRvbl9ib3giPgogICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ib3JpZW50YXRpb24iPkdUS19PUklFTlRBVElPTl9IT1JJWk9O
VEFMPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xh
c3M9Ikd0a0J1dHRvbiIgaWQ9ImJ0bi1jYW5jZWwiPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+Q2FuY2VsPC9wcm9wZXJ0eT4KICAgICAgICAgICAg
ICAgIDxzaWduYWwgbmFtZT0iY2xpY2tlZCIgaGFuZGxlcj0ib24tY2FuY2VsIi8+CiAgICAgICAgICAg
ICAgPC9vYmplY3Q+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAg
ICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCdXR0b24iIGlkPSJidG4tc2F2ZSI+CiAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5TYXZlPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW4tZGVmYXVsdCI+dHJ1ZTwvcHJvcGVy
dHk+CiAgICAgICAgICAgICAgICA8c2lnbmFsIG5hbWU9ImNsaWNrZWQiIGhhbmRsZXI9Im9uLXNhdmUi
Lz4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgIDwv
b2JqZWN0PgogICAgICAgIDwvY2hpbGQ+CiAgICAgICA8IS0tIFRPRE86IFRoaXMgc2hvdWxkIGJlIGlu
IGEgc2Nyb2xsZWQgd2luZG93IHdpdGggbWluL21heCBzaXplICAtLT4KICAgICAgIDxjaGlsZD4KICAg
ICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCb3giIGlkPSJib3giPgogICAgICAgICAgPHByb3BlcnR5IG5h
bWU9ImJvcmRlci13aWR0aCI+MTA8L3Byb3BlcnR5PgogICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imhv
bW9nZW5lb3VzIj5mYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ib3JpZW50
YXRpb24iPkdUS19PUklFTlRBVElPTl9WRVJUSUNBTDwvcHJvcGVydHk+CiAgICAgICAgICA8Y2hpbGQ+
CiAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0iZmluZ2VycHJpbnRzSW5mb3Jt
YXRpb24iPgogICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzZWxlY3RhYmxlIj50cnVlPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgPGNo
aWxkPgogICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtHcmlkIiBpZD0iZmluZ2VycHJpbnRzR3Jp
ZCI+CiAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi10b3AiPjE1PC9wcm9wZXJ0eT4K
ICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLWJvdHRvbSI+MTA8L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tc3RhcnQiPjEwPC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLWVuZCI+MTA8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJyb3ctc3BhY2luZyI+MTI8L3Byb3BlcnR5PgogICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJjb2x1bW4tc3BhY2luZyI+NjwvcHJvcGVydHk+CiAgICAgICAgICAg
IDwvb2JqZWN0PgogICAgICAgICAgPC9jaGlsZD4KICAgICAgICA8L29iamVjdD4KICAgICAgICA8L2No
aWxkPgogICAgICA8L29iamVjdD4KICAgIDwvY2hpbGQ+CiAgPC9vYmplY3Q+CjwvaW50ZXJmYWNlPgo=
`,
	},

	"/definitions/PeerRequestsSMP.xml": {
		local:   "definitions/PeerRequestsSMP.xml",
		size:    4259,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtCb3giIGlkPSJzbXAtcmVxdWVzdGVkLWluZm9i
YXIiPgogICAgPHByb3BlcnR5IG5hbWU9ImNhbl9mb2N1cyI+RmFsc2U8L3Byb3BlcnR5PgogICAgPHBy
b3BlcnR5IG5hbWU9Im9yaWVudGF0aW9uIj52ZXJ0aWNhbDwvcHJvcGVydHk+CiAgICA8Y2hpbGQ+CiAg
ICAgIDxvYmplY3QgY2xhc3M9Ikd0a0JveCIgaWQ9InNtcC1yZXF1ZXN0ZWQtY2xvc2UtaW5mb2JhciI+
CiAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPkZhbHNlPC9wcm9wZXJ0eT4KCTxjaGlsZD4KICAgICAgICAg
IDxvYmplY3QgY2xhc3M9Ikd0a0V2ZW50Qm94Ij4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZp
c2libGUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuX2ZvY3Vz
Ij5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxzaWduYWwgbmFtZT0iYnV0dG9uLXByZXNzLWV2
ZW50IiBoYW5kbGVyPSJvbl9wcmVzc19jbG9zZV9pbWFnZSIgc3dhcHBlZD0ibm8iLz4KCSAgICA8Y2hp
bGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrSW1hZ2UiPgogICAgICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9InZpc2libGUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9ImNhbl9mb2N1cyI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9InBpeGVsX3NpemUiPjEyPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJpY29uX25hbWUiPndpbmRvdy1jbG9zZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0iaWNvbl9zaXplIj4zPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJtYXJnaW5fcmlnaHQiPjM8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9Im1hcmdpbl90b3AiPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9Im1hcmdpbl9ib3R0b20iPjE8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvb2JqZWN0
PgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICA8cGFja2lu
Zz4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+RmFsc2U8L3Byb3BlcnR5PgogICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0iZmlsbCI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJwYWNrX3R5cGUiPmVuZDwvcHJvcGVydHk+CgkgICAgPHByb3BlcnR5IG5hbWU9
InBvc2l0aW9uIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICA8L2NoaWxk
PgogICAgICA8L29iamVjdD4KICAgICAgPHBhY2tpbmc+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9ImV4
cGFuZCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj5UcnVlPC9w
cm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjA8L3Byb3BlcnR5PgogICAg
ICA8L3BhY2tpbmc+CiAgICA8L2NoaWxkPgogICAgPGNoaWxkPgogICAgICA8b2JqZWN0IGNsYXNzPSJH
dGtCb3giIGlkPSJzbXAtcmVxdWVzdGVkLW5vdGlmaWNhdGlvbiI+CiAgICAgICAgPHByb3BlcnR5IG5h
bWU9InZpc2libGUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9j
dXMiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0ic3BhY2luZyI+MzM8L3By
b3BlcnR5PgogICAgICAgIDxjaGlsZD4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0ltYWdlIiBp
ZD0ic21wLXJlcXVlc3RlZC1pbWFnZSI+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxl
Ij5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbl9mb2N1cyI+RmFs
c2U8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luX2xlZnQiPjEwPC9w
cm9wZXJ0eT4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJleHBhbmQiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9ImZpbGwiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
cG9zaXRpb24iPjA8L3Byb3BlcnR5PgogICAgICAgICAgPC9wYWNraW5nPgogICAgICAgIDwvY2hpbGQ+
CiAgICAgICAgPGNoaWxkPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrTGFiZWwiIGlkPSJzbXAt
cmVxdWVzdGVkLW1lc3NhZ2UiPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+VHJ1
ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPkZhbHNlPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+c3RhcnQ8L3Byb3BlcnR5
PgogICAgICAgICAgICA8YXR0cmlidXRlcz4KICAgICAgICAgICAgICA8YXR0cmlidXRlIG5hbWU9Indl
aWdodCIgdmFsdWU9ImJvbGQiLz4KICAgICAgICAgICAgPC9hdHRyaWJ1dGVzPgoJICA8L29iamVjdD4K
ICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZXhwYW5kIj5GYWxz
ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj5UcnVlPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InBvc2l0aW9uIj4xPC9wcm9wZXJ0eT4KICAgICAg
ICAgIDwvcGFja2luZz4KICAgICAgICA8L2NoaWxkPgogICAgICAgIDxjaGlsZD4KICAgICAgICAgIDxv
YmplY3QgY2xhc3M9Ikd0a0J1dHRvbiIgaWQ9InNtcC1yZXF1ZXN0ZWQtYnV0dG9uIj4KICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0iY2FuX2ZvY3VzIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9InJlY2VpdmVzX2RlZmF1bHQiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0ibWFyZ2luX3JpZ2h0Ij4xMDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJtYXJnaW5fdG9wIj4xNDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJt
YXJnaW5fYm90dG9tIj4xMjwvcHJvcGVydHk+CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAg
ICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCI+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
dmlzaWJsZSI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2Fu
X2ZvY3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFy
Z2luX2xlZnQiPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJn
aW5fcmlnaHQiPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJl
bCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPkZpbmlzaCBWZXJpZmljYXRpb248L3Byb3BlcnR5PgogICAgICAg
ICAgICAgIDwvb2JqZWN0PgoJICAgIDwvY2hpbGQ+CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICAg
IDxwYWNraW5nPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZXhwYW5kIj5GYWxzZTwvcHJvcGVy
dHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9InBhY2tfdHlwZSI+ZW5kPC9wcm9wZXJ0eT4KCSAgICA8cHJvcGVy
dHkgbmFtZT0icG9zaXRpb24iPjI8L3Byb3BlcnR5PgogICAgICAgICAgPC9wYWNraW5nPgogICAgICAg
IDwvY2hpbGQ+CiAgICAgIDwvb2JqZWN0PgogICAgICA8cGFja2luZz4KICAgICAgICA8cHJvcGVydHkg
bmFtZT0iZXhwYW5kIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9ImZpbGwi
PlRydWU8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MTwvcHJvcGVy
dHk+CiAgICAgIDwvcGFja2luZz4KICAgIDwvY2hpbGQ+CiAgPC9vYmplY3Q+CjwvaW50ZXJmYWNlPgo=
`,
	},

	"/definitions/RegistrationForm.xml": {
		local:   "definitions/RegistrationForm.xml",
		size:    2185,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtEaWFsb2ciIGlkPSJkaWFsb2ciPgogICAgPHNp
Z25hbCBuYW1lPSJyZXNwb25zZSIgaGFuZGxlcj0icmVzcG9uc2UtaGFuZGxlciIgLz4KICAgIDxzaWdu
YWwgbmFtZT0iY2xvc2UiIGhhbmRsZXI9ImNsb3NlLWhhbmRsZXIiIC8+CiAgICA8Y2hpbGQgaW50ZXJu
YWwtY2hpbGQ9InZib3giPgogICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCb3giIGlkPSJWYm94Ij4KICAg
ICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgPHByb3Bl
cnR5IG5hbWU9Im1hcmdpbi1ib3R0b20iPjEwPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFt
ZT0ibWFyZ2luIj4xMDwvcHJvcGVydHk+CiAgICAgICAgPGNoaWxkPgogICAgICAgICAgPG9iamVjdCBj
bGFzcz0iR3RrR3JpZCIgaWQ9ImdyaWQiPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJs
ZSI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tYm90dG9t
Ij4xMDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJyb3ctc3BhY2luZyI+MTI8
L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY29sdW1uLXNwYWNpbmciPjY8L3By
b3BlcnR5PgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3Rr
TGFiZWwiIGlkPSJpbnN0cnVjdGlvbnMiPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZp
c2libGUiPnRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAg
IDxwYWNraW5nPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxlZnQtYXR0YWNoIj4wPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b3AtYXR0YWNoIj4wPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ3aWR0aCI+MjwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgPC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgPC9vYmpl
Y3Q+CiAgICAgICAgPC9jaGlsZD4KICAgICAgICA8Y2hpbGQgaW50ZXJuYWwtY2hpbGQ9ImFjdGlvbl9h
cmVhIj4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0J1dHRvbkJveCIgaWQ9ImJib3giPgogICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJvcmllbnRhdGlvbiI+R1RLX09SSUVOVEFUSU9OX0hPUklaT05UQUw8L3By
b3BlcnR5PgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3Rr
QnV0dG9uIiBpZD0iYnRuLWNhbmNlbCI+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlz
aWJsZSI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwi
IHRyYW5zbGF0YWJsZT0ieWVzIj5DYW5jZWw8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvb2JqZWN0
PgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9i
amVjdCBjbGFzcz0iR3RrQnV0dG9uIiBpZD0iYnRuLXJlZ2lzdGVyIj4KICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPlJlZ2lzdGVyPC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW4tZGVmYXVsdCI+dHJ1ZTwvcHJvcGVydHk+CiAg
ICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICA8L29iamVj
dD4KICAgICAgICA8L2NoaWxkPgogICAgICA8L29iamVjdD4KICAgIDwvY2hpbGQ+CiAgICA8YWN0aW9u
LXdpZGdldHM+CiAgICAgIDxhY3Rpb24td2lkZ2V0IHJlc3BvbnNlPSJjYW5jZWwiPmJ0bi1jYW5jZWw8
L2FjdGlvbi13aWRnZXQ+CiAgICAgIDxhY3Rpb24td2lkZ2V0IHJlc3BvbnNlPSJhcHBseSIgZGVmYXVs
dD0idHJ1ZSI+YnRuLXJlZ2lzdGVyPC9hY3Rpb24td2lkZ2V0PgogICAgPC9hY3Rpb24td2lkZ2V0cz4K
ICA8L29iamVjdD4KPC9pbnRlcmZhY2U+Cg==
`,
	},

	"/definitions/Roster.xml": {
		local:   "definitions/Roster.xml",
		size:    2681,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtUcmVlU3RvcmUiIGlkPSJyb3N0ZXItbW9kZWwi
PgogICAgPGNvbHVtbnM+CiAgICAgIDwhLS0gamlkIC0tPgogICAgICA8Y29sdW1uIHR5cGU9ImdjaGFy
YXJyYXkiLz4KICAgICAgPCEtLSBkaXNwbGF5IG5hbWUgLS0+CiAgICAgIDxjb2x1bW4gdHlwZT0iZ2No
YXJhcnJheSIvPgogICAgICA8IS0tIGFjY291bnQgaWQgLS0+CiAgICAgIDxjb2x1bW4gdHlwZT0iZ2No
YXJhcnJheSIvPgogICAgICA8IS0tIHN0YXR1cyBjb2xvciAtLT4KICAgICAgPGNvbHVtbiB0eXBlPSJn
Y2hhcmFycmF5Ii8+CiAgICAgIDwhLS0gYmFja2dyb3VuZCBjb2xvciAtLT4KICAgICAgPGNvbHVtbiB0
eXBlPSJnY2hhcmFycmF5Ii8+CiAgICAgIDwhLS0gd2VpZ2h0IG9mIGZvbnQgLS0+CiAgICAgIDxjb2x1
bW4gdHlwZT0iZ2ludCIvPgogICAgICA8IS0tIHRvb2x0aXAgLS0+CiAgICAgIDxjb2x1bW4gdHlwZT0i
Z2NoYXJhcnJheSIvPgogICAgICA8IS0tIGljb24gLS0+CiAgICAgIDxjb2x1bW4gdHlwZT0iR2RrUGl4
YnVmIi8+CiAgICAgIDwhLS0gcm93IHR5cGUgLS0+CiAgICAgIDxjb2x1bW4gdHlwZT0iZ2NoYXJhcnJh
eSIvPgogICAgPC9jb2x1bW5zPgogIDwvb2JqZWN0PgogIDxvYmplY3QgY2xhc3M9Ikd0a1Njcm9sbGVk
V2luZG93IiBpZD0icm9zdGVyIj4KICAgIDxwcm9wZXJ0eSBuYW1lPSJoc2Nyb2xsYmFyLXBvbGljeSI+
R1RLX1BPTElDWV9ORVZFUjwvcHJvcGVydHk+CiAgICA8cHJvcGVydHkgbmFtZT0idnNjcm9sbGJhci1w
b2xpY3kiPkdUS19QT0xJQ1lfQVVUT01BVElDPC9wcm9wZXJ0eT4KICAgIDxjaGlsZD4KICAgICAgPG9i
amVjdCBjbGFzcz0iR3RrVHJlZVZpZXciIGlkPSJyb3N0ZXItdHJlZSI+CiAgICAgICAgPHByb3BlcnR5
IG5hbWU9Im1vZGVsIj5yb3N0ZXItbW9kZWw8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJoZWFkZXJzLXZpc2libGUiPmZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0i
c2hvdy1leHBhbmRlcnMiPmZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGV2
ZWwtaW5kZW50YXRpb24iPjM8L3Byb3BlcnR5PgogICAgICAgIDwhLS0gVE9ETyByZW1vdmUgbWFnaWMg
bnVtYmVyIC0tPgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ0b29sdGlwLWNvbHVtbiI+NjwvcHJvcGVy
dHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9InNlYXJjaC1jb2x1bW4iPjY8L3Byb3BlcnR5PgogICAg
ICAgIDxzaWduYWwgbmFtZT0icm93LWFjdGl2YXRlZCIgaGFuZGxlcj0ib25fYWN0aXZhdGVfYnVkZHki
IC8+CiAgICAgICAgPHNpZ25hbCBuYW1lPSJidXR0b24tcHJlc3MtZXZlbnQiIGhhbmRsZXI9Im9uX2J1
dHRvbl9wcmVzcyIgLz4KICAgICAgICA8Y2hpbGQgaW50ZXJuYWwtY2hpbGQ9InNlbGVjdGlvbiI+CiAg
ICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtUcmVlU2VsZWN0aW9uIiBpZD0ic2VsZWN0aW9uIj4KICAg
ICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1vZGUiPkdUS19TRUxFQ1RJT05fU0lOR0xFPC9wcm9wZXJ0
eT4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgIDwvY2hpbGQ+CiAgICAgICAgPGNoaWxkPgogICAg
ICAgICAgPG9iamVjdCBjbGFzcz0iR3RrVHJlZVZpZXdDb2x1bW4iIGlkPSJpY29uLWNvbHVtbiI+CiAg
ICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtDZWxsUmVuZGVy
ZXJQaXhidWYiIGlkPSJpY29uLWNvbHVtbi1yZW5kZXJlZCIvPgogICAgICAgICAgICAgIDxhdHRyaWJ1
dGVzPgogICAgICAgICAgICAgICAgPGF0dHJpYnV0ZSBuYW1lPSJwaXhidWYiPjc8L2F0dHJpYnV0ZT4K
ICAgICAgICAgICAgICAgIDxhdHRyaWJ1dGUgbmFtZT0iY2VsbC1iYWNrZ3JvdW5kIj40PC9hdHRyaWJ1
dGU+CiAgICAgICAgICAgICAgPC9hdHRyaWJ1dGVzPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAg
ICAgPC9vYmplY3Q+CiAgICAgICAgPC9jaGlsZD4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2Jq
ZWN0IGNsYXNzPSJHdGtUcmVlVmlld0NvbHVtbiIgaWQ9Im5hbWUtY29sdW1uIj4KICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9InRpdGxlIj5uYW1lPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPGNoaWxkPgog
ICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0NlbGxSZW5kZXJlclRleHQiIGlkPSJuYW1lLWNv
bHVtbi1yZW5kZXJlZCIvPgogICAgICAgICAgICAgIDxhdHRyaWJ1dGVzPgogICAgICAgICAgICAgICAg
PCEtLSBUT0RPIHJlbW92ZSBtYWdpYyBudW1iZXJzIC0tPgogICAgICAgICAgICAgICAgPGF0dHJpYnV0
ZSBuYW1lPSJ0ZXh0Ij4xPC9hdHRyaWJ1dGU+CiAgICAgICAgICAgICAgICA8YXR0cmlidXRlIG5hbWU9
ImZvcmVncm91bmQiPjM8L2F0dHJpYnV0ZT4KICAgICAgICAgICAgICAgIDxhdHRyaWJ1dGUgbmFtZT0i
YmFja2dyb3VuZCI+NDwvYXR0cmlidXRlPgogICAgICAgICAgICAgICAgPGF0dHJpYnV0ZSBuYW1lPSJ3
ZWlnaHQiPjU8L2F0dHJpYnV0ZT4KICAgICAgICAgICAgICA8L2F0dHJpYnV0ZXM+CiAgICAgICAgICAg
IDwvY2hpbGQ+CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICA8L2NoaWxkPgogICAgICA8L29iamVj
dD4KICAgIDwvY2hpbGQ+CiAgPC9vYmplY3Q+CjwvaW50ZXJmYWNlPgo=
`,
	},

	"/definitions/SimpleNotification.xml": {
		local:   "definitions/SimpleNotification.xml",
		size:    359,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtNZXNzYWdlRGlhbG9nIiBpZD0iZGlhbG9nIj4K
ICAgIDxwcm9wZXJ0eSBuYW1lPSJ3aW5kb3ctcG9zaXRpb24iPkdUS19XSU5fUE9TX0NFTlRFUjwvcHJv
cGVydHk+CiAgICA8cHJvcGVydHkgbmFtZT0ibW9kYWwiPnRydWU8L3Byb3BlcnR5PgogICAgPHByb3Bl
cnR5IG5hbWU9ImJvcmRlcl93aWR0aCI+NzwvcHJvcGVydHk+CiAgICA8cHJvcGVydHkgbmFtZT0ibWVz
c2FnZS10eXBlIj5HVEtfTUVTU0FHRV9JTkZPPC9wcm9wZXJ0eT4KICAgIDxwcm9wZXJ0eSBuYW1lPSJi
dXR0b25zIj5HVEtfQlVUVE9OU19PSzwvcHJvcGVydHk+CiAgPC9vYmplY3Q+CjwvaW50ZXJmYWNlPgo=
`,
	},

	"/definitions/Test.xml": {
		local:   "definitions/Test.xml",
		size:    333,
		modtime: 1489449600,
		compressed: `
CjxpbnRlcmZhY2U+CiAgPG9iamVjdCBjbGFzcz0iR3RrV2luZG93IiBpZD0iY29udmVyc2F0aW9uIj4K
ICAgIDxwcm9wZXJ0eSBuYW1lPSJkZWZhdWx0LWhlaWdodCI+NjAwPC9wcm9wZXJ0eT4KICAgIDxwcm9w
ZXJ0eSBuYW1lPSJkZWZhdWx0LXdpZHRoIj41MDA8L3Byb3BlcnR5PgogICAgPGNoaWxkPgoJICA8b2Jq
ZWN0IGNsYXNzPSJHdGtCb3giIGlkPSJ2Ym94Ij4KCSAgICA8cHJvcGVydHkgbmFtZT0ib3JpZW50YXRp
b24iPkdUS19PUklFTlRBVElPTl9WRVJUSUNBTDwvcHJvcGVydHk+ICAKCSAgPC9vYmplY3Q+CiAgICA8
L2NoaWxkPgogIDwvb2JqZWN0Pgo8L2ludGVyZmFjZT4K
`,
	},

	"/definitions/TorHelper.xml": {
		local:   "definitions/TorHelper.xml",
		size:    631,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtNZXNzYWdlRGlhbG9nIiBpZD0iVG9ySGVscGVy
Ij4KICAgIDxwcm9wZXJ0eSBuYW1lPSJ3aW5kb3ctcG9zaXRpb24iPkdUS19XSU5fUE9TX0NFTlRFUl9B
TFdBWVM8L3Byb3BlcnR5PgogICAgPHByb3BlcnR5IG5hbWU9Im1vZGFsIj50cnVlPC9wcm9wZXJ0eT4K
ICAgIDxwcm9wZXJ0eSBuYW1lPSJib3JkZXJfd2lkdGgiPjc8L3Byb3BlcnR5PgogICAgPHByb3BlcnR5
IG5hbWU9InRleHQiIHRyYW5zbGF0YWJsZT0ieWVzIj5XYXJuaW5nITwvcHJvcGVydHk+CiAgICA8cHJv
cGVydHkgbmFtZT0ic2Vjb25kYXJ5X3RleHQiIHRyYW5zbGF0YWJsZT0ieWVzIj4KV2UgY291bGQgbm90
IGZpbmQgYSBydW5uaW5nIGluc3RhbGxhdGlvbiBvZiBUb3IuCgpJbiBvcmRlciB0byBoYXZlIHRoZSBz
YWZlc3QgZXhwZXJpZW5jZSwgd2UgcmVjb21tZW5kIHlvdSB0byBpbnN0YWxsIGl0LgoKVG9yIGlzIGEg
dG9vbCBmb3IgYW5vbnltaXR5LgoKV291bGQgeW91IGxpa2UgdG8gaW5zdGFsbCBpdD8KICAgIDwvcHJv
cGVydHk+CiAgICA8cHJvcGVydHkgbmFtZT0iYnV0dG9ucyI+R1RLX0JVVFRPTlNfWUVTX05PPC9wcm9w
ZXJ0eT4KICA8L29iamVjdD4KPC9pbnRlcmZhY2U+Cg==
`,
	},

	"/definitions/TorInstallHelper.xml": {
		local:   "definitions/TorInstallHelper.xml",
		size:    3772,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtEaWFsb2ciIGlkPSJkaWFsb2ciPgogICAgPHBy
b3BlcnR5IG5hbWU9IndpbmRvdy1wb3NpdGlvbiI+R1RLX1dJTl9QT1NfQ0VOVEVSPC9wcm9wZXJ0eT4K
ICAgIDxwcm9wZXJ0eSBuYW1lPSJib3JkZXItd2lkdGgiPjc8L3Byb3BlcnR5PgogICAgPHByb3BlcnR5
IG5hbWU9InRpdGxlIiB0cmFuc2xhdGFibGU9InllcyI+PC9wcm9wZXJ0eT4KICAgIDxjaGlsZCBpbnRl
cm5hbC1jaGlsZD0idmJveCI+CiAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0JveCIgaWQ9ImJveCI+CiAg
ICAgICAgPHByb3BlcnR5IG5hbWU9ImJvcmRlci13aWR0aCI+MTA8L3Byb3BlcnR5PgogICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJvcmllbnRhdGlvbiI+R1RLX09SSUVOVEFUSU9OX1ZFUlRJQ0FMPC9wcm9wZXJ0
eT4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCI+CiAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi10b3AiPjU8L3Byb3BlcnR5PgoJICAgIDxwcm9wZXJ0eSBu
YW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPgpJbnN0YWxsaW5nIFRvciBmb3IgQ295SU06CiAg
ICAgICAgICAgIDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxhdHRyaWJ1dGVzPgogICAgICAgICAgICAg
IDxhdHRyaWJ1dGUgbmFtZT0iZm9udC1kZXNjIiB2YWx1ZT0iJmx0O0VudGVyIFZhbHVlJmd0OyAxNCIv
PgoJICAgICAgPGF0dHJpYnV0ZSBuYW1lPSJ3ZWlnaHQiIHZhbHVlPSJzZW1pYm9sZCIvPgogICAgICAg
ICAgICA8L2F0dHJpYnV0ZXM+CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgIDxwYWNraW5nPgog
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZXhwYW5kIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9InBvc2l0aW9uIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgIDwvcGFja2luZz4KICAgICAg
ICA8L2NoaWxkPgogICAgICAgIDxjaGlsZD4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVs
Ij4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbl9mb2N1cyI+RmFsc2U8L3Byb3BlcnR5PgoJ
ICAgIDxwcm9wZXJ0eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPgoKICAgIDEuIEdvIHRv
IGh0dHBzOi8vd3d3LnRvcnByb2plY3Qub3JnLwoKICAgIDIuIERvd25sb2FkIGFuZCBpbnN0YWxsIFRv
ciAob3IgdGhlIFRvciBCcm93c2VyIEJ1bmRsZSkKCiAgICAzLiBTdGFydCBUb3IgKG9yIHRoZSBUb3Ig
QnJvd3NlcikKICAgICAgICAgICAgPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
InNlbGVjdGFibGUiPlRydWU8L3Byb3BlcnR5PgoJICA8L29iamVjdD4KICAgICAgICAgIDxwYWNraW5n
PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZXhwYW5kIj5GYWxzZTwvcHJvcGVydHk+CiAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9InBvc2l0aW9uIj4xPC9wcm9wZXJ0eT4KICAgICAgICAgIDwvcGFja2luZz4KICAg
ICAgICA8L2NoaWxkPgoJPGNoaWxkPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrRXZlbnRCb3gi
PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+VHJ1ZTwvcHJvcGVydHk+CiAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgPHNpZ25hbCBuYW1lPSJidXR0b24tcHJlc3MtZXZlbnQiIGhhbmRsZXI9Im9uX3ByZXNzX2xhYmVs
X3NpZ25hbCIgc3dhcHBlZD0ibm8iLz4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxv
YmplY3QgY2xhc3M9Ikd0a0xhYmVsIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5f
Zm9jdXMiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJn
aW4tdG9wIj4wPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4t
Ym90dG9tIj4xMDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2lu
LXJpZ2h0Ij42MDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwi
IHRyYW5zbGF0YWJsZT0ieWVzIj40LiBDbGljayBoZXJlIHRvIGNoZWNrIGlmIFRvciBpcyBub3cgcnVu
bmluZzwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8YXR0cmlidXRlcz4KICAgICAgICAgICAgICAg
ICAgPGF0dHJpYnV0ZSBuYW1lPSJ1bmRlcmxpbmUiIHZhbHVlPSJUcnVlIi8+CiAgICAgICAgICAgICAg
ICA8L2F0dHJpYnV0ZXM+CgkgICAgICA8L29iamVjdD4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAg
ICAgIDwvb2JqZWN0PgogICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJleHBhbmQiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImZpbGwi
PlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjI8L3By
b3BlcnR5PgogICAgICAgICAgPC9wYWNraW5nPgogICAgICAgIDwvY2hpbGQ+CiAgICAgICAgPGNoaWxk
PgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQm94IiBpZD0idG9yLXJ1bm5pbmctbm90aWZpY2F0
aW9uIj4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPlRydWU8L3Byb3BlcnR5Pgog
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuX2ZvY3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJvcmllbnRhdGlvbiI+dmVydGljYWw8L3Byb3BlcnR5PgogICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLWJvdHRvbSI+MTA8L3Byb3BlcnR5PgogICAgICAg
ICAgPC9vYmplY3Q+CiAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
ImV4cGFuZCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZmlsbCI+
VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MzwvcHJv
cGVydHk+CiAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgPC9jaGlsZD4KICAgICAgPC9vYmplY3Q+
CiAgICA8L2NoaWxkPgogICAgPGNoaWxkIGludGVybmFsLWNoaWxkPSJhY3Rpb25fYXJlYSI+CiAgICAg
IDxvYmplY3QgY2xhc3M9Ikd0a0J1dHRvbkJveCIgaWQ9ImJ1dHRvbl9ib3giPgogICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJvcmllbnRhdGlvbiI+R1RLX09SSUVOVEFUSU9OX0hPUklaT05UQUw8L3Byb3BlcnR5
PgogICAgICAgIDxjaGlsZD4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0J1dHRvbiIgaWQ9ImNs
b3NlIj4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+
Q2xvc2U8L3Byb3BlcnR5PgogICAgICAgICAgICA8c2lnbmFsIG5hbWU9ImNsaWNrZWQiIGhhbmRsZXI9
Im9uX2Nsb3NlX3NpZ25hbCIgLz4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgIDwvY2hpbGQ+CiAg
ICAgIDwvb2JqZWN0PgogICAgPC9jaGlsZD4KICA8L29iamVjdD4KPC9pbnRlcmZhY2U+Cg==
`,
	},

	"/definitions/TorNotRunningNotification.xml": {
		local:   "definitions/TorNotRunningNotification.xml",
		size:    2914,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtJbmZvQmFyIiBpZD0iaW5mb2JhciI+CiAgICA8
cHJvcGVydHkgbmFtZT0ibWVzc2FnZS10eXBlIj5HVEtfTUVTU0FHRV9FUlJPUjwvcHJvcGVydHk+CiAg
ICA8cHJvcGVydHkgbmFtZT0ic2hvdy1jbG9zZS1idXR0b24iPmZhbHNlPC9wcm9wZXJ0eT4KICAgIDxj
aGlsZCBpbnRlcm5hbC1jaGlsZD0iY29udGVudF9hcmVhIj4KICAgICAgPG9iamVjdCBjbGFzcz0iR3Rr
Qm94IiBpZD0iYm94Ij4KICAgICAgICA8cHJvcGVydHkgbmFtZT0ic3BhY2luZyI+MDwvcHJvcGVydHk+
CiAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhvbW9nZW5lb3VzIj5mYWxzZTwvcHJvcGVydHk+CiAgICAg
ICAgPHByb3BlcnR5IG5hbWU9Im9yaWVudGF0aW9uIj5HVEtfT1JJRU5UQVRJT05fSE9SSVpPTlRBTDwv
cHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhleHBhbmQiPlRydWU8L3Byb3BlcnR5Pgog
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tc3RhcnQiPjA8L3Byb3BlcnR5PgogICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJtYXJnaW4tZW5kIj4wPC9wcm9wZXJ0eT4KICAgICAgICA8Y2hpbGQ+CiAgICAg
ICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9Im1lc3NhZ2UiPgogICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0iZWxsaXBzaXplIj5QQU5HT19FTExJUFNJWkVfRU5EPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9ImhleHBhbmQiPlRydWU8L3Byb3BlcnR5PgoJICAgIDxwcm9wZXJ0
eSBuYW1lPSJ3cmFwIj50cnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1h
cmdpbi1yaWdodCI+MTA8L3Byb3BlcnR5PgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgPC9jaGls
ZD4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCdXR0b24iIGlkPSJi
dXR0b25fbW9yZV9pbmZvIj4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhleHBhbmQiPkZhbHNl
PC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+ZW5kPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InJlbGllZiI+bm9uZTwvcHJvcGVydHk+CiAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2YWxpZ24iPmNlbnRlcjwvcHJvcGVydHk+CiAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJib3JkZXJfd2lkdGgiPjA8L3Byb3BlcnR5PgogICAgICAgICAgICA8c3R5
bGU+CiAgICAgICAgICAgICAgPGNsYXNzIG5hbWU9InJhaXNlZCIvPgogICAgICAgICAgICAgIDxjbGFz
cyBuYW1lPSJjbG9zZSIvPgogICAgICAgICAgICA8L3N0eWxlPgogICAgICAgICAgICA8c2lnbmFsIG5h
bWU9ImNsaWNrZWQiIGhhbmRsZXI9Im9uX21vcmVfaW5mb19zaWduYWwiIC8+CiAgICAgICAgICAgIDxj
aGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtJbWFnZSIgaWQ9Im1vcmUtaW5mby1i
dG4iPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imljb24tbmFtZSI+ZGlhbG9nLXF1ZXN0
aW9uPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgPC9jaGlsZD4K
ICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJleHBhbmQiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9
InBvc2l0aW9uIj4xPC9wcm9wZXJ0eT4KICAgICAgICAgIDwvcGFja2luZz4KICAgICAgICA8L2NoaWxk
PgogICAgICAgIDxjaGlsZD4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0J1dHRvbiIgaWQ9ImJ1
dHRvbl9jbG9zZSI+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPjE8L3Byb3Bl
cnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iYm9yZGVyX3dpZHRoIj4yPC9wcm9wZXJ0eT4K
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InJlbGllZiI+bm9uZTwvcHJvcGVydHk+CiAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJ2YWxpZ24iPmNlbnRlcjwvcHJvcGVydHk+CiAgICAgICAgICAgIDxz
dHlsZT4KICAgICAgICAgICAgICA8Y2xhc3MgbmFtZT0icmFpc2VkIi8+CiAgICAgICAgICAgICAgPGNs
YXNzIG5hbWU9ImNsb3NlIi8+CiAgICAgICAgICAgIDwvc3R5bGU+CiAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJoZXhwYW5kIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJoYWxpZ24iPmVuZDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2YWxpZ24i
PmNlbnRlcjwvcHJvcGVydHk+CiAgICAgICAgICAgIDxzaWduYWwgbmFtZT0iY2xpY2tlZCIgaGFuZGxl
cj0ib25fY2xvc2Vfc2lnbmFsIiAvPgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9i
amVjdCBjbGFzcz0iR3RrSW1hZ2UiIGlkPSJjbG9zZV9pbWFnZSI+CiAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0idmlzaWJsZSI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0iaWNvbl9uYW1lIj53aW5kb3ctY2xvc2Utc3ltYm9saWM8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgPC9vYmplY3Q+CiAg
ICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+RmFsc2U8
L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjI8L3Byb3BlcnR5
PgogICAgICAgICAgPC9wYWNraW5nPgogICAgICAgIDwvY2hpbGQ+CiAgICAgIDwvb2JqZWN0PgogICAg
PC9jaGlsZD4KICA8L29iamVjdD4KPC9pbnRlcmZhY2U+Cg==
`,
	},

	"/definitions/TorRunningNotification.xml": {
		local:   "definitions/TorRunningNotification.xml",
		size:    1086,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtCb3giIGlkPSJpbmZvYmFyIj4KICAgIDxwcm9w
ZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5f
Zm9jdXMiPkZhbHNlPC9wcm9wZXJ0eT4KICAgIDxwcm9wZXJ0eSBuYW1lPSJzcGFjaW5nIj4xODwvcHJv
cGVydHk+CiAgICA8Y2hpbGQ+CiAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0ltYWdlIiBpZD0iaW1hZ2Ui
PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICA8
cHJvcGVydHkgbmFtZT0iY2FuX2ZvY3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5
IG5hbWU9Im1hcmdpbl9sZWZ0Ij4xNTwvcHJvcGVydHk+CiAgICAgIDwvb2JqZWN0PgogICAgICA8cGFj
a2luZz4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iZXhwYW5kIj5GYWxzZTwvcHJvcGVydHk+CiAgICAg
ICAgPHByb3BlcnR5IG5hbWU9ImZpbGwiPlRydWU8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJwb3NpdGlvbiI+MDwvcHJvcGVydHk+CiAgICAgIDwvcGFja2luZz4KICAgIDwvY2hpbGQ+CiAg
ICA8Y2hpbGQ+CiAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0ibWVzc2FnZSI+CiAgICAg
ICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJjYW5fZm9jdXMiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0i
bWFyZ2luX3RvcCI+NTwvcHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbl9ib3R0
b20iPjU8L3Byb3BlcnR5PgogICAgICA8L29iamVjdD4KICAgICAgPHBhY2tpbmc+CiAgICAgICAgPHBy
b3BlcnR5IG5hbWU9ImV4cGFuZCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJmaWxsIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjE8
L3Byb3BlcnR5PgogICAgICA8L3BhY2tpbmc+CiAgICA8L2NoaWxkPgogIDwvb2JqZWN0Pgo8L2ludGVy
ZmFjZT4K
`,
	},

	"/definitions/UnifiedLayout.xml": {
		local:   "definitions/UnifiedLayout.xml",
		size:    8749,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtMaXN0U3RvcmUiIGlkPSJsaXN0c3RvcmUiPgog
ICAgPGNvbHVtbnM+CiAgICAgIDwhLS0gY29sdW1uLW5hbWUgcGFnZV9pbmRleCAtLT4KICAgICAgPGNv
bHVtbiB0eXBlPSJnaW50Ii8+CiAgICAgIDwhLS0gY29sdW1uLW5hbWUgbmFtZSAtLT4KICAgICAgPGNv
bHVtbiB0eXBlPSJnY2hhcmFycmF5Ii8+CiAgICAgIDwhLS0gY29sdW1uLW5hbWUgamlkIC0tPgogICAg
ICA8Y29sdW1uIHR5cGU9ImdjaGFyYXJyYXkiLz4KICAgICAgPCEtLSBjb2x1bW4tbmFtZSBjb2xvciAt
LT4KICAgICAgPGNvbHVtbiB0eXBlPSJnY2hhcmFycmF5Ii8+CiAgICAgIDwhLS0gY29sdW1uLW5hbWUg
YmFja2dyb3VuZCAtLT4KICAgICAgPGNvbHVtbiB0eXBlPSJnY2hhcmFycmF5Ii8+CiAgICAgIDwhLS0g
Y29sdW1uLW5hbWUgd2VpZ2h0IC0tPgogICAgICA8Y29sdW1uIHR5cGU9ImdpbnQiLz4KICAgICAgPCEt
LSBjb2x1bW4tbmFtZSB0b29sdGlwIC0tPgogICAgICA8Y29sdW1uIHR5cGU9ImdjaGFyYXJyYXkiLz4K
ICAgICAgPCEtLSBjb2x1bW4tbmFtZSBzdGF0dXMgLS0+CiAgICAgIDxjb2x1bW4gdHlwZT0iR2RrUGl4
YnVmIi8+CiAgICAgIDwhLS0gY29sdW1uLW5hbWUgdW5kZXJsaW5lIC0tPgogICAgICA8Y29sdW1uIHR5
cGU9ImdpbnQiLz4KICAgIDwvY29sdW1ucz4KICA8L29iamVjdD4KICA8b2JqZWN0IGNsYXNzPSJHdGtI
ZWFkZXJCYXIiIGlkPSJoZWFkZXJiYXIiPgogICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPlRydWU8
L3Byb3BlcnR5PgogICAgPHByb3BlcnR5IG5hbWU9InRpdGxlIj5Db3lJTTwvcHJvcGVydHk+CiAgICA8
cHJvcGVydHkgbmFtZT0ic2hvd19jbG9zZV9idXR0b24iPlRydWU8L3Byb3BlcnR5PgogIDwvb2JqZWN0
PgogIDxvYmplY3QgY2xhc3M9Ikd0a0JveCIgaWQ9InJpZ2h0Ij4KICAgIDxwcm9wZXJ0eSBuYW1lPSJ2
aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPkZhbHNl
PC9wcm9wZXJ0eT4KICAgIDxwcm9wZXJ0eSBuYW1lPSJub19zaG93X2FsbCI+VHJ1ZTwvcHJvcGVydHk+
CiAgICA8Y2hpbGQ+CiAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a1NlcGFyYXRvciIgaWQ9InNlcGFyYXRv
cjEiPgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAg
ICA8cHJvcGVydHkgbmFtZT0iY2FuX2ZvY3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgPHByb3Bl
cnR5IG5hbWU9Im9yaWVudGF0aW9uIj52ZXJ0aWNhbDwvcHJvcGVydHk+CiAgICAgIDwvb2JqZWN0Pgog
ICAgICA8cGFja2luZz4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iZXhwYW5kIj5GYWxzZTwvcHJvcGVy
dHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9ImZpbGwiPlRydWU8L3Byb3BlcnR5PgogICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MDwvcHJvcGVydHk+CiAgICAgIDwvcGFja2luZz4KICAgIDwv
Y2hpbGQ+CiAgICA8Y2hpbGQ+CiAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0JveCIgaWQ9ImJveDEiPgog
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICA8cHJv
cGVydHkgbmFtZT0iY2FuX2ZvY3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5h
bWU9ImhleHBhbmQiPlRydWU8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJvcmllbnRh
dGlvbiI+dmVydGljYWw8L3Byb3BlcnR5PgogICAgICAgIDxjaGlsZD4KICAgICAgICAgIDxvYmplY3Qg
Y2xhc3M9Ikd0a0JveCIgaWQ9ImhlYWRlcl9ib3giPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
dmlzaWJsZSI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9j
dXMiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmpl
Y3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0iaGVhZGVyX2xhYmVsIj4KICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJjYW5fZm9jdXMiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxhdHRyaWJ1
dGVzPgogICAgICAgICAgICAgICAgICA8YXR0cmlidXRlIG5hbWU9IndlaWdodCIgdmFsdWU9ImJvbGQi
Lz4KICAgICAgICAgICAgICAgIDwvYXR0cmlidXRlcz4KICAgICAgICAgICAgICA8L29iamVjdD4KICAg
ICAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJleHBhbmQi
PlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImZpbGwiPlRydWU8
L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InBvc2l0aW9uIj4wPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAg
ICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCdXR0b24iIGlkPSJjbG9z
ZV9idXR0b24iPgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPlRydWU8L3By
b3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbl9mb2N1cyI+VHJ1ZTwvcHJv
cGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icmVjZWl2ZXNfZGVmYXVsdCI+VHJ1
ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icmVsaWVmIj5ub25lPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxzaWduYWwgbmFtZT0iY2xpY2tlZCIgaGFuZGxlcj0ib25f
Y2xpY2tlZCIgc3dhcHBlZD0ibm8iLz4KICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAg
ICAgICAgPG9iamVjdCBjbGFzcz0iR3RrSW1hZ2UiIGlkPSJpbWFnZTEiPgogICAgICAgICAgICAgICAg
ICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0iY2FuX2ZvY3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAg
ICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imljb25fbmFtZSI+d2luZG93LWNsb3NlLXN5bWJvbGljPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICAgICAgICA8L2NoaWxk
PgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9ImZpbGwiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHBy
b3BlcnR5IG5hbWU9InBhY2tfdHlwZSI+ZW5kPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgPC9wYWNraW5nPgog
ICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICA8cGFja2luZz4K
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0iZmlsbCI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MDwvcHJvcGVydHk+CiAgICAgICAgICA8L3BhY2tpbmc+CiAgICAg
ICAgPC9jaGlsZD4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtOb3Rl
Ym9vayIgaWQ9Im5vdGVib29rIj4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPlRy
dWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuX2ZvY3VzIj5UcnVlPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InNjcm9sbGFibGUiPlRydWU8L3Byb3Bl
cnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ic2hvd190YWJzIj5GYWxzZTwvcHJvcGVydHk+
CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJzaG93X2JvcmRlciI+RmFsc2U8L3Byb3BlcnR5Pgog
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luX3N0YXJ0Ij41PC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbl9lbmQiPjU8L3Byb3BlcnR5PgogICAgICAgICAgICA8
c2lnbmFsIG5hbWU9InN3aXRjaC1wYWdlIiBoYW5kbGVyPSJvbl9zd2l0Y2hfcGFnZSIgc3dhcHBlZD0i
bm8iLz4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJleHBhbmQiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0iZmlsbCI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3Np
dGlvbiI+MTwvcHJvcGVydHk+CiAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgPC9jaGlsZD4KICAg
ICAgPC9vYmplY3Q+CiAgICAgIDxwYWNraW5nPgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJleHBhbmQi
PkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iZmlsbCI+VHJ1ZTwvcHJvcGVy
dHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9InBvc2l0aW9uIj4xPC9wcm9wZXJ0eT4KICAgICAgPC9w
YWNraW5nPgogICAgPC9jaGlsZD4KICAgIDxjaGlsZD4KICAgICAgPG9iamVjdCBjbGFzcz0iR3RrU2Vw
YXJhdG9yIiBpZD0ic2VwYXJhdG9yMiI+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPlRy
dWU8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPkZhbHNlPC9wcm9w
ZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0ib3JpZW50YXRpb24iPnZlcnRpY2FsPC9wcm9wZXJ0
eT4KICAgICAgPC9vYmplY3Q+CiAgICAgIDxwYWNraW5nPgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJl
eHBhbmQiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iZmlsbCI+VHJ1ZTwv
cHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9InBvc2l0aW9uIj4yPC9wcm9wZXJ0eT4KICAg
ICAgPC9wYWNraW5nPgogICAgPC9jaGlsZD4KICAgIDxjaGlsZD4KICAgICAgPG9iamVjdCBjbGFzcz0i
R3RrQm94IiBpZD0iYm94MyI+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPlRydWU8L3By
b3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPkZhbHNlPC9wcm9wZXJ0eT4K
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ib3JpZW50YXRpb24iPnZlcnRpY2FsPC9wcm9wZXJ0eT4KICAg
ICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCIgaWQ9ImxhYmVsMSI+
CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbl9mb2N1cyI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5Db252ZXJzYXRpb25zPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgPGF0dHJpYnV0ZXM+CiAgICAgICAgICAgICAgPGF0dHJpYnV0ZSBu
YW1lPSJ3ZWlnaHQiIHZhbHVlPSJib2xkIi8+CiAgICAgICAgICAgIDwvYXR0cmlidXRlcz4KICAgICAg
ICAgIDwvb2JqZWN0PgogICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1l
PSJleHBhbmQiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImZpbGwi
PlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjA8L3By
b3BlcnR5PgogICAgICAgICAgPC9wYWNraW5nPgogICAgICAgIDwvY2hpbGQ+CiAgICAgICAgPGNoaWxk
PgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrVHJlZVZpZXciIGlkPSJ0cmVldmlldyI+CiAgICAg
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9ImNhbl9mb2N1cyI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJoc2Nyb2xsX3BvbGljeSI+bmF0dXJhbDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJtb2RlbCI+bGlzdHN0b3JlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3Bl
cnR5IG5hbWU9ImhlYWRlcnNfdmlzaWJsZSI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0iZW5hYmxlX3NlYXJjaCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0iYWN0aXZhdGVfb25fc2luZ2xlX2NsaWNrIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAg
ICAgICAgPHNpZ25hbCBuYW1lPSJyb3ctYWN0aXZhdGVkIiBoYW5kbGVyPSJvbl9hY3RpdmF0ZSIgc3dh
cHBlZD0ibm8iLz4KICAgICAgICAgICAgPGNoaWxkIGludGVybmFsLWNoaWxkPSJzZWxlY3Rpb24iPgog
ICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a1RyZWVTZWxlY3Rpb24iIGlkPSJ0cmVldmlldy1z
ZWxlY3Rpb24xIi8+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAg
ICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtUcmVlVmlld0NvbHVtbiIgaWQ9Imljb24iPgogICAgICAg
ICAgICAgICAgPHByb3BlcnR5IG5hbWU9InRpdGxlIiB0cmFuc2xhdGFibGU9InllcyI+aWNvbjwvcHJv
cGVydHk+CiAgICAgICAgICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xh
c3M9Ikd0a0NlbGxSZW5kZXJlclBpeGJ1ZiIgaWQ9Imljb24tY29sdW1uLXJlbmRlcmVkIi8+CiAgICAg
ICAgICAgICAgICAgIDxhdHRyaWJ1dGVzPgogICAgICAgICAgICAgICAgICAgIDxhdHRyaWJ1dGUgbmFt
ZT0iY2VsbC1iYWNrZ3JvdW5kIj40PC9hdHRyaWJ1dGU+CiAgICAgICAgICAgICAgICAgICAgPGF0dHJp
YnV0ZSBuYW1lPSJwaXhidWYiPjc8L2F0dHJpYnV0ZT4KICAgICAgICAgICAgICAgICAgPC9hdHRyaWJ1
dGVzPgogICAgICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAg
ICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xh
c3M9Ikd0a1RyZWVWaWV3Q29sdW1uIiBpZD0idHJlZXZpZXdjb2x1bW4xIj4KICAgICAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJzcGFjaW5nIj44PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJtaW5fd2lkdGgiPjEyMDwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0idGl0bGUiIHRyYW5zbGF0YWJsZT0ieWVzIj5jb2x1bW48L3Byb3BlcnR5PgogICAgICAg
ICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAg
ICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0NlbGxSZW5kZXJl
clRleHQiIGlkPSJ1aWQtY29sdW1uLXJlbmRlcmVyIi8+CiAgICAgICAgICAgICAgICAgIDxhdHRyaWJ1
dGVzPgogICAgICAgICAgICAgICAgICAgIDxhdHRyaWJ1dGUgbmFtZT0iYmFja2dyb3VuZCI+NDwvYXR0
cmlidXRlPgogICAgICAgICAgICAgICAgICAgIDxhdHRyaWJ1dGUgbmFtZT0iZm9yZWdyb3VuZCI+Mzwv
YXR0cmlidXRlPgogICAgICAgICAgICAgICAgICAgIDxhdHRyaWJ1dGUgbmFtZT0idGV4dCI+MTwvYXR0
cmlidXRlPgogICAgICAgICAgICAgICAgICAgIDxhdHRyaWJ1dGUgbmFtZT0id2VpZ2h0Ij41PC9hdHRy
aWJ1dGU+CiAgICAgICAgICAgICAgICAgICAgPGF0dHJpYnV0ZSBuYW1lPSJ1bmRlcmxpbmUiPjg8L2F0
dHJpYnV0ZT4KICAgICAgICAgICAgICAgICAgPC9hdHRyaWJ1dGVzPgogICAgICAgICAgICAgICAgPC9j
aGlsZD4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAg
IDwvb2JqZWN0PgogICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJl
eHBhbmQiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZmlsbCI+VHJ1
ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MTwvcHJvcGVy
dHk+CiAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgPC9jaGlsZD4KICAgICAgPC9vYmplY3Q+CiAg
ICAgIDxwYWNraW5nPgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJleHBhbmQiPkZhbHNlPC9wcm9wZXJ0
eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iZmlsbCI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgPHBy
b3BlcnR5IG5hbWU9InBvc2l0aW9uIj4zPC9wcm9wZXJ0eT4KICAgICAgPC9wYWNraW5nPgogICAgPC9j
aGlsZD4KICA8L29iamVjdD4KICA8b2JqZWN0IGNsYXNzPSJHdGtTaXplR3JvdXAiIGlkPSJzaXplZ3Jv
dXAiPgogICAgPHByb3BlcnR5IG5hbWU9Im1vZGUiPnZlcnRpY2FsPC9wcm9wZXJ0eT4KICAgIDx3aWRn
ZXRzPgogICAgICA8d2lkZ2V0IG5hbWU9ImhlYWRlcl9ib3giLz4KICAgICAgPHdpZGdldCBuYW1lPSJs
YWJlbDEiLz4KICAgIDwvd2lkZ2V0cz4KICA8L29iamVjdD4KPC9pbnRlcmZhY2U+Cg==
`,
	},

	"/definitions/UnverifiedWarning.xml": {
		local:   "definitions/UnverifiedWarning.xml",
		size:    4208,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtCb3giIGlkPSJ2ZXJpZnktaW5mb2JhciI+CiAg
ICA8cHJvcGVydHkgbmFtZT0iY2FuX2ZvY3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICA8cHJvcGVydHkg
bmFtZT0ib3JpZW50YXRpb24iPnZlcnRpY2FsPC9wcm9wZXJ0eT4KICAgIDxjaGlsZD4KICAgICAgPG9i
amVjdCBjbGFzcz0iR3RrQm94IiBpZD0idmVyaWZ5LWNsb3NlLWluZm9iYXIiPgogICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0i
Y2FuX2ZvY3VzIj5GYWxzZTwvcHJvcGVydHk+Cgk8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNz
PSJHdGtFdmVudEJveCI+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbl9mb2N1cyI+RmFsc2U8L3Byb3Bl
cnR5PgogICAgICAgICAgICA8c2lnbmFsIG5hbWU9ImJ1dHRvbi1wcmVzcy1ldmVudCIgaGFuZGxlcj0i
b25fcHJlc3NfaW1hZ2UiIHN3YXBwZWQ9Im5vIi8+CgkgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxv
YmplY3QgY2xhc3M9Ikd0a0ltYWdlIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNp
YmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9j
dXMiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwaXhlbF9z
aXplIj4xMjwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaWNvbl9uYW1l
Ij53aW5kb3ctY2xvc2U8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Imlj
b25fc2l6ZSI+MzwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2lu
X3JpZ2h0Ij4zPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW5f
dG9wIj4xPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW5fYm90
dG9tIj4xPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgPC9jaGls
ZD4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJleHBhbmQiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5h
bWU9ImZpbGwiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icGFja190
eXBlIj5lbmQ8L3Byb3BlcnR5PgoJICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MDwvcHJvcGVy
dHk+CiAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgPC9jaGlsZD4KICAgICAgPC9vYmplY3Q+CiAg
ICAgIDxwYWNraW5nPgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJleHBhbmQiPkZhbHNlPC9wcm9wZXJ0
eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iZmlsbCI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgPHBy
b3BlcnR5IG5hbWU9InBvc2l0aW9uIj4wPC9wcm9wZXJ0eT4KICAgICAgPC9wYWNraW5nPgogICAgPC9j
aGlsZD4KICAgIDxjaGlsZD4KICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQm94IiBpZD0idmVyaWZ5LW5v
dGlmaWNhdGlvbiI+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPlRydWU8L3Byb3BlcnR5
PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPkZhbHNlPC9wcm9wZXJ0eT4KICAgICAg
ICA8cHJvcGVydHkgbmFtZT0ic3BhY2luZyI+MzM8L3Byb3BlcnR5PgogICAgICAgIDxjaGlsZD4KICAg
ICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0ltYWdlIiBpZD0idmVyaWZ5LWltYWdlIj4KICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0iY2FuX2ZvY3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJtYXJnaW5fbGVmdCI+MTA8L3Byb3BlcnR5PgogICAgICAgICAgPC9vYmplY3Q+CiAgICAg
ICAgICA8cGFja2luZz4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+RmFsc2U8L3By
b3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZmlsbCI+VHJ1ZTwvcHJvcGVydHk+CiAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MDwvcHJvcGVydHk+CiAgICAgICAgICA8
L3BhY2tpbmc+CiAgICAgICAgPC9jaGlsZD4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0
IGNsYXNzPSJHdGtMYWJlbCIgaWQ9InZlcmlmeS1tZXNzYWdlIj4KICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9InZpc2libGUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
Y2FuX2ZvY3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJoYWxp
Z24iPnN0YXJ0PC9wcm9wZXJ0eT4KICAgICAgICAgICAgPGF0dHJpYnV0ZXM+CiAgICAgICAgICAgICAg
PGF0dHJpYnV0ZSBuYW1lPSJ3ZWlnaHQiIHZhbHVlPSJib2xkIi8+CiAgICAgICAgICAgIDwvYXR0cmli
dXRlcz4KCSAgPC9vYmplY3Q+CiAgICAgICAgICA8cGFja2luZz4KICAgICAgICAgICAgPHByb3BlcnR5
IG5hbWU9ImV4cGFuZCI+RmFsc2U8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0i
ZmlsbCI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+
MTwvcHJvcGVydHk+CiAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgPC9jaGlsZD4KICAgICAgICA8
Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCdXR0b24iIGlkPSJ2ZXJpZnktYnV0dG9u
Ij4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InZpc2libGUiPlRydWU8L3Byb3BlcnR5PgogICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuX2ZvY3VzIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9InJlY2VpdmVzX2RlZmF1bHQiPlRydWU8L3Byb3BlcnR5PgogICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luX3JpZ2h0Ij4xMDwvcHJvcGVydHk+CiAgICAgICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW5fdG9wIj4xNDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9w
ZXJ0eSBuYW1lPSJtYXJnaW5fYm90dG9tIj4xMjwvcHJvcGVydHk+CiAgICAgICAgICAgIDxjaGlsZD4K
ICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCI+CiAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0idmlzaWJsZSI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0iY2FuX2ZvY3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0ibWFyZ2luX2xlZnQiPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJtYXJnaW5fcmlnaHQiPjEwPC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJsYWJlbCIgdHJhbnNsYXRhYmxlPSJ5ZXMiPlZhbGlkYXRlIENoYW5uZWw8L3Byb3BlcnR5
PgogICAgICAgICAgICAgIDwvb2JqZWN0PgoJICAgIDwvY2hpbGQ+CiAgICAgICAgICA8L29iamVjdD4K
ICAgICAgICAgIDxwYWNraW5nPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iZXhwYW5kIj5GYWxz
ZTwvcHJvcGVydHk+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj5UcnVlPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InBhY2tfdHlwZSI+ZW5kPC9wcm9wZXJ0eT4KCSAg
ICA8cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjI8L3Byb3BlcnR5PgogICAgICAgICAgPC9wYWNraW5n
PgogICAgICAgIDwvY2hpbGQ+CiAgICAgIDwvb2JqZWN0PgogICAgICA8cGFja2luZz4KICAgICAgICA8
cHJvcGVydHkgbmFtZT0iZXhwYW5kIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5h
bWU9ImZpbGwiPlRydWU8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+
MTwvcHJvcGVydHk+CiAgICAgIDwvcGFja2luZz4KICAgIDwvY2hpbGQ+CiAgPC9vYmplY3Q+CjwvaW50
ZXJmYWNlPgo=
`,
	},

	"/definitions/VerificationFailed.xml": {
		local:   "definitions/VerificationFailed.xml",
		size:    2282,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtEaWFsb2ciIGlkPSJ2ZXJpZi1mYWlsdXJlLWRp
YWxvZyI+CiAgICA8cHJvcGVydHkgbmFtZT0id2luZG93LXBvc2l0aW9uIj5HVEtfV0lOX1BPU19DRU5U
RVI8L3Byb3BlcnR5PgogICAgPHByb3BlcnR5IG5hbWU9ImJvcmRlci13aWR0aCI+NzwvcHJvcGVydHk+
CiAgICA8cHJvcGVydHkgbmFtZT0idGl0bGUiIHRyYW5zbGF0YWJsZT0ieWVzIj48L3Byb3BlcnR5Pgog
ICAgPGNoaWxkIGludGVybmFsLWNoaWxkPSJ2Ym94Ij4KICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQm94
Ij4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iYm9yZGVyLXdpZHRoIj4xMDwvcHJvcGVydHk+CiAgICAg
ICAgPHByb3BlcnR5IG5hbWU9Im9yaWVudGF0aW9uIj5HVEtfT1JJRU5UQVRJT05fVkVSVElDQUw8L3By
b3BlcnR5PgogICAgICAgIDxjaGlsZD4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0dyaWQiPgog
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iaGFsaWduIj5HVEtfQUxJR05fQ0VOVEVSPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVs
Ij4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tdG9wIj41PC9wcm9wZXJ0eT4K
ICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJtYXJnaW4tYm90dG9tIj4xNTwvcHJvcGVydHk+
CgkJPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+VmVyaWZpY2F0aW9uIEZh
aWxlZDwvcHJvcGVydHk+CiAgICAgICAgICAgIDxhdHRyaWJ1dGVzPgogICAgICAgICAgICAgIDxhdHRy
aWJ1dGUgbmFtZT0iZm9udC1kZXNjIiB2YWx1ZT0iJmx0O0VudGVyIFZhbHVlJmd0OyAxNiIvPgoJICAg
ICAgPGF0dHJpYnV0ZSBuYW1lPSJ3ZWlnaHQiIHZhbHVlPSJzZW1pYm9sZCIvPgogICAgICAgICAgICA8
L2F0dHJpYnV0ZXM+CgkgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICA8cGFja2luZz4KICAgICAg
ICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LWF0dGFjaCI+MDwvcHJvcGVydHk+CiAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0idG9wLWF0dGFjaCI+MTwvcHJvcGVydHk+CiAgICAgICAgICAg
ICAgPC9wYWNraW5nPgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAgPC9vYmplY3Q+CiAgICAg
ICAgPC9jaGlsZD4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJl
bCIgaWQ9InZlcmlmLWZhaWx1cmUtbGFiZWwiPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFi
ZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj48L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFt
ZT0ibWFyZ2luLWJvdHRvbSI+MTA8L3Byb3BlcnR5PgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAg
PC9jaGlsZD4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtHcmlkIj4K
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNvbHVtbi1zcGFjaW5nIj42PC9wcm9wZXJ0eT4KICAg
ICAgICAgICAgPHByb3BlcnR5IG5hbWU9InJvdy1zcGFjaW5nIj4yPC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgPGNoaWxkIGludGVybmFsLWNoaWxkPSJhY3Rpb25fYXJlYSI+CiAgICAgICAgICAgICAgPG9iamVj
dCBjbGFzcz0iR3RrQnV0dG9uQm94IiBpZD0iYnV0dG9uX2JveCI+CiAgICAgICAgICAgICAgICA8cHJv
cGVydHkgbmFtZT0ib3JpZW50YXRpb24iPkdUS19PUklFTlRBVElPTl9IT1JJWk9OVEFMPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0i
R3RrQnV0dG9uIiBpZD0idmVyaWYtZmFpbHVyZS1idXR0b24iPgogICAgICAgICAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJjYW4tZGVmYXVsdCI+dHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbi10b3AiPjU8L3Byb3BlcnR5Pgo8cHJvcGVydHkgbmFtZT0i
bGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5UcnkgYWdhaW4gbGF0ZXI8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICAgICAg
PC9vYmplY3Q+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICA8
L2NoaWxkPgogICAgICA8L29iamVjdD4KICAgIDwvY2hpbGQ+CiAgPC9vYmplY3Q+CjwvaW50ZXJmYWNl
Pgo=
`,
	},

	"/definitions/VerificationSucceeded.xml": {
		local:   "definitions/VerificationSucceeded.xml",
		size:    1256,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtEaWFsb2ciIGlkPSJ2ZXJpZi1zdWNjZXNzLWRp
YWxvZyI+CiAgICA8cHJvcGVydHkgbmFtZT0id2luZG93LXBvc2l0aW9uIj5HVEtfV0lOX1BPU19DRU5U
RVI8L3Byb3BlcnR5PgogICAgPHByb3BlcnR5IG5hbWU9ImJvcmRlci13aWR0aCI+NzwvcHJvcGVydHk+
CiAgICA8cHJvcGVydHkgbmFtZT0idGl0bGUiIHRyYW5zbGF0YWJsZT0ieWVzIj48L3Byb3BlcnR5Pgog
ICAgPGNoaWxkIGludGVybmFsLWNoaWxkPSJ2Ym94Ij4KICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQm94
Ij4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iYm9yZGVyLXdpZHRoIj4xMDwvcHJvcGVydHk+CiAgICAg
ICAgPHByb3BlcnR5IG5hbWU9Im9yaWVudGF0aW9uIj5HVEtfT1JJRU5UQVRJT05fVkVSVElDQUw8L3By
b3BlcnR5PgogICAgICAgIDxjaGlsZD4KICAgICAgICAgIDxvYmplY3QgIGNsYXNzPSJHdGtJbWFnZSIg
aWQ9InZlcmlmLXN1Y2Nlc3MtaW1hZ2UiPgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgICA8cGFj
a2luZz4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InBhZGRpbmciPjIwPC9wcm9wZXJ0eT4KICAg
ICAgICAgIDwvcGFja2luZz4KICAgICAgICA8L2NoaWxkPgogICAgICAgIDxjaGlsZD4KICAgICAgICAg
IDxvYmplY3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0idmVyaWYtc3VjY2Vzcy1sYWJlbCIvPgogICAgICAg
IDwvY2hpbGQ+CiAgICAgICAgPGNoaWxkIGludGVybmFsLWNoaWxkPSJhY3Rpb25fYXJlYSI+CiAgICAg
ICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCdXR0b25Cb3giIGlkPSJidXR0b25fYm94Ij4KICAgICAgICAg
ICAgPHByb3BlcnR5IG5hbWU9Im9yaWVudGF0aW9uIj5HVEtfT1JJRU5UQVRJT05fSE9SSVpPTlRBTDwv
cHJvcGVydHk+CiAgICAgICAgICAgIDxjaGlsZD4KICAgICAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJH
dGtCdXR0b24iIGlkPSJ2ZXJpZi1zdWNjZXNzLWJ1dHRvbiI+CiAgICAgICAgICAgICAgICA8cHJvcGVy
dHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5PSzwvcHJvcGVydHk+CiAgICAgICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luLXRvcCI+MTA8L3Byb3BlcnR5PgoJICAgICAgPC9vYmpl
Y3Q+CiAgICAgICAgICAgIDwvY2hpbGQ+CiAgICAgICAgICA8L29iamVjdD4KICAgICAgICA8L2NoaWxk
PgogICAgICA8L29iamVjdD4KICAgIDwvY2hpbGQ+CiAgPC9vYmplY3Q+CjwvaW50ZXJmYWNlPgo=
`,
	},

	"/definitions/VerifyFingerprint.xml": {
		local:   "definitions/VerifyFingerprint.xml",
		size:    1371,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtEaWFsb2ciIGlkPSJkaWFsb2ciPgogICAgPHBy
b3BlcnR5IG5hbWU9IndpbmRvdy1wb3NpdGlvbiI+R1RLX1dJTl9QT1NfQ0VOVEVSPC9wcm9wZXJ0eT4K
ICAgIDxjaGlsZCBpbnRlcm5hbC1jaGlsZD0idmJveCI+CiAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0Jv
eCIgaWQ9ImJveCI+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9ImJvcmRlci13aWR0aCI+MTA8L3Byb3Bl
cnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJob21vZ2VuZW91cyI+ZmFsc2U8L3Byb3BlcnR5Pgog
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJvcmllbnRhdGlvbiI+R1RLX09SSUVOVEFUSU9OX1ZFUlRJQ0FM
PC9wcm9wZXJ0eT4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJl
bCIgaWQ9Im1lc3NhZ2UiLz4KICAgICAgICA8L2NoaWxkPgogICAgICAgIDxjaGlsZCBpbnRlcm5hbC1j
aGlsZD0iYWN0aW9uX2FyZWEiPgogICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQnV0dG9uQm94IiBp
ZD0iYnV0dG9uX2JveCI+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJvcmllbnRhdGlvbiI+R1RL
X09SSUVOVEFUSU9OX0hPUklaT05UQUw8L3Byb3BlcnR5PgogICAgICAgICAgICA8Y2hpbGQ+CiAgICAg
ICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQnV0dG9uIiBpZD0iYnV0dG9uX2NhbmNlbCI+CiAgICAg
ICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuLWRlZmF1bHQiPnRydWU8L3Byb3BlcnR5PgogICAg
ICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+Tm88L3By
b3BlcnR5PgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAg
ICAgICA8Y2hpbGQ+CiAgICAgICAgICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQnV0dG9uIiBpZD0iYnV0
dG9uX3ZlcmlmeSI+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0
YWJsZT0ieWVzIj5ZZXM8L3Byb3BlcnR5PgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAg
ICA8L2NoaWxkPgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgPC9jaGlsZD4KICAgICAgPC9vYmpl
Y3Q+CiAgICA8L2NoaWxkPgogICAgPGFjdGlvbi13aWRnZXRzPgogICAgICA8YWN0aW9uLXdpZGdldCBy
ZXNwb25zZT0ibm8iIGRlZmF1bHQ9InRydWUiPmJ1dHRvbl9jYW5jZWw8L2FjdGlvbi13aWRnZXQ+CiAg
ICAgIDxhY3Rpb24td2lkZ2V0IHJlc3BvbnNlPSJ5ZXMiPmJ1dHRvbl92ZXJpZnk8L2FjdGlvbi13aWRn
ZXQ+CiAgICA8L2FjdGlvbi13aWRnZXRzPgogIDwvb2JqZWN0Pgo8L2ludGVyZmFjZT4K
`,
	},

	"/definitions/VerifyFingerprintUnknown.xml": {
		local:   "definitions/VerifyFingerprintUnknown.xml",
		size:    1100,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtEaWFsb2ciIGlkPSJkaWFsb2ciPgogICAgPHBy
b3BlcnR5IG5hbWU9IndpbmRvdy1wb3NpdGlvbiI+R1RLX1dJTl9QT1NfQ0VOVEVSPC9wcm9wZXJ0eT4K
ICAgIDxjaGlsZCBpbnRlcm5hbC1jaGlsZD0idmJveCI+CiAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0Jv
eCIgaWQ9ImJveCI+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9ImJvcmRlci13aWR0aCI+MTA8L3Byb3Bl
cnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJob21vZ2VuZW91cyI+ZmFsc2U8L3Byb3BlcnR5Pgog
ICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJvcmllbnRhdGlvbiI+R1RLX09SSUVOVEFUSU9OX1ZFUlRJQ0FM
PC9wcm9wZXJ0eT4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8b2JqZWN0IGNsYXNzPSJHdGtMYWJl
bCIgaWQ9Im1lc3NhZ2UiIC8+CiAgICAgICAgPC9jaGlsZD4KICAgICAgICA8Y2hpbGQgaW50ZXJuYWwt
Y2hpbGQ9ImFjdGlvbl9hcmVhIj4KICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0J1dHRvbkJveCIg
aWQ9ImJ1dHRvbl9ib3giPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ib3JpZW50YXRpb24iPkdU
S19PUklFTlRBVElPTl9IT1JJWk9OVEFMPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPGNoaWxkPgogICAg
ICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0J1dHRvbiIgaWQ9ImJ1dHRvbl9vayI+CiAgICAgICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuLWRlZmF1bHQiPnRydWU8L3Byb3BlcnR5PgogICAgICAg
ICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImxhYmVsIiB0cmFuc2xhdGFibGU9InllcyI+T0s8L3Byb3Bl
cnR5PgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICA8L2NoaWxkPgogICAgICAgICAg
PC9vYmplY3Q+CiAgICAgICAgPC9jaGlsZD4KICAgICAgPC9vYmplY3Q+CiAgICA8L2NoaWxkPgogICAg
PGFjdGlvbi13aWRnZXRzPgogICAgICA8YWN0aW9uLXdpZGdldCByZXNwb25zZT0ib2siIGRlZmF1bHQ9
InRydWUiPmJ1dHRvbl9vazwvYWN0aW9uLXdpZGdldD4KICAgIDwvYWN0aW9uLXdpZGdldHM+CiAgPC9v
YmplY3Q+CjwvaW50ZXJmYWNlPgo=
`,
	},

	"/definitions/WaitingSMPComplete.xml": {
		local:   "definitions/WaitingSMPComplete.xml",
		size:    2133,
		modtime: 1489449600,
		compressed: `
PGludGVyZmFjZT4KICA8b2JqZWN0IGNsYXNzPSJHdGtCb3giIGlkPSJzbXAtd2FpdGluZy1pbmZvYmFy
Ij4KICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPkZhbHNlPC9wcm9wZXJ0eT4KICAgIDxwcm9w
ZXJ0eSBuYW1lPSJzcGFjaW5nIj4yNzwvcHJvcGVydHk+CiAgICA8Y2hpbGQ+CiAgICAgIDxvYmplY3Qg
Y2xhc3M9Ikd0a0ltYWdlIiBpZD0ic21wLXdhaXRpbmctaW1hZ2UiPgogICAgICAgIDxwcm9wZXJ0eSBu
YW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuX2Zv
Y3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbl9sZWZ0Ij4x
MDwvcHJvcGVydHk+CiAgICAgIDwvb2JqZWN0PgogICAgICA8cGFja2luZz4KICAgICAgICA8cHJvcGVy
dHkgbmFtZT0iZXhwYW5kIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9ImZp
bGwiPlRydWU8L3Byb3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJwb3NpdGlvbiI+MDwvcHJv
cGVydHk+CiAgICAgIDwvcGFja2luZz4KICAgIDwvY2hpbGQ+CiAgICA8Y2hpbGQ+CiAgICAgIDxvYmpl
Y3QgY2xhc3M9Ikd0a0xhYmVsIiBpZD0ic21wLXdhaXRpbmctbGFiZWwiPgogICAgICAgIDxwcm9wZXJ0
eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2Fu
X2ZvY3VzIj5GYWxzZTwvcHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9ImhhbGlnbiI+c3Rh
cnQ8L3Byb3BlcnR5PgogICAgICAgIDxhdHRyaWJ1dGVzPgogICAgICAgICAgPGF0dHJpYnV0ZSBuYW1l
PSJ3ZWlnaHQiIHZhbHVlPSJib2xkIi8+CiAgICAgICAgPC9hdHRyaWJ1dGVzPgogICAgICA8L29iamVj
dD4KICAgICAgPHBhY2tpbmc+CiAgICAgICAgPHByb3BlcnR5IG5hbWU9ImV4cGFuZCI+RmFsc2U8L3By
b3BlcnR5PgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJmaWxsIj5UcnVlPC9wcm9wZXJ0eT4KICAgICAg
ICA8cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjE8L3Byb3BlcnR5PgogICAgICA8L3BhY2tpbmc+CiAg
ICA8L2NoaWxkPgogICAgPGNoaWxkPgogICAgICA8b2JqZWN0IGNsYXNzPSJHdGtCdXR0b24iIGlkPSJz
bXAtd2FpdGluZy1idXR0b24iPgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxlIj5UcnVlPC9w
cm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iY2FuX2ZvY3VzIj5UcnVlPC9wcm9wZXJ0eT4K
ICAgICAgICA8cHJvcGVydHkgbmFtZT0icmVjZWl2ZXNfZGVmYXVsdCI+VHJ1ZTwvcHJvcGVydHk+CiAg
ICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbl9yaWdodCI+MTA8L3Byb3BlcnR5PgogICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJtYXJnaW5fdG9wIj4xNDwvcHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5h
bWU9Im1hcmdpbl9ib3R0b20iPjEyPC9wcm9wZXJ0eT4KICAgICAgICA8Y2hpbGQ+CiAgICAgICAgICA8
b2JqZWN0IGNsYXNzPSJHdGtMYWJlbCI+CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJ2aXNpYmxl
Ij5UcnVlPC9wcm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbl9mb2N1cyI+RmFs
c2U8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibWFyZ2luX2xlZnQiPjM1PC9w
cm9wZXJ0eT4KICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9Im1hcmdpbl9yaWdodCI+MzU8L3Byb3Bl
cnR5PgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5D
YW5jZWw8L3Byb3BlcnR5PgogICAgICAgICAgPC9vYmplY3Q+CiAgICAgICAgPC9jaGlsZD4KICAgICAg
PC9vYmplY3Q+CiAgICAgIDxwYWNraW5nPgogICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJleHBhbmQiPkZh
bHNlPC9wcm9wZXJ0eT4KICAgICAgICA8cHJvcGVydHkgbmFtZT0iZmlsbCI+VHJ1ZTwvcHJvcGVydHk+
CiAgICAgICAgPHByb3BlcnR5IG5hbWU9InBhY2tfdHlwZSI+ZW5kPC9wcm9wZXJ0eT4KICAgICAgICA8
cHJvcGVydHkgbmFtZT0icG9zaXRpb24iPjI8L3Byb3BlcnR5PgogICAgICA8L3BhY2tpbmc+CiAgICA8
L2NoaWxkPgogIDwvb2JqZWN0Pgo8L2ludGVyZmFjZT4K
`,
	},

	"/definitions/XMLConsole.xml": {
		local:   "definitions/XMLConsole.xml",
		size:    3087,
		modtime: 1489449600,
		compressed: `
PD94bWwgdmVyc2lvbj0iMS4wIiBlbmNvZGluZz0idXRmLTgiPz4KPGludGVyZmFjZT4KICA8b2JqZWN0
IGNsYXNzPSJHdGtUZXh0QnVmZmVyIiBpZD0iY29uc29sZUNvbnRlbnQiPgogICAgPHByb3BlcnR5IG5h
bWU9InRleHQiIHRyYW5zbGF0YWJsZT0ieWVzIj5IZWxsbyB3b3JsZDwvcHJvcGVydHk+CiAgPC9vYmpl
Y3Q+CiAgPG9iamVjdCBjbGFzcz0iR3RrRGlhbG9nIiBpZD0iWE1MQ29uc29sZSI+CiAgICA8cHJvcGVy
dHkgbmFtZT0id2luZG93LXBvc2l0aW9uIj5HVEtfV0lOX1BPU19DRU5URVI8L3Byb3BlcnR5PgogICAg
PHByb3BlcnR5IG5hbWU9ImJvcmRlcl93aWR0aCI+NjwvcHJvcGVydHk+CiAgICA8cHJvcGVydHkgbmFt
ZT0idGl0bGUiIHRyYW5zbGF0YWJsZT0ieWVzIj5YTVBQIENvbnNvbGU6IEFDQ09VTlRfTkFNRTwvcHJv
cGVydHk+CiAgICA8cHJvcGVydHkgbmFtZT0icmVzaXphYmxlIj5UcnVlPC9wcm9wZXJ0eT4KICAgIDxw
cm9wZXJ0eSBuYW1lPSJkZWZhdWx0LWhlaWdodCI+NDAwPC9wcm9wZXJ0eT4KICAgIDxwcm9wZXJ0eSBu
YW1lPSJkZWZhdWx0LXdpZHRoIj41MDA8L3Byb3BlcnR5PgogICAgPHByb3BlcnR5IG5hbWU9ImRlc3Ry
b3ktd2l0aC1wYXJlbnQiPnRydWU8L3Byb3BlcnR5PgogICAgPGNoaWxkIGludGVybmFsLWNoaWxkPSJ2
Ym94Ij4KICAgICAgPG9iamVjdCBjbGFzcz0iR3RrQm94IiBpZD0iVmJveCI+CiAgICAgICAgPHByb3Bl
cnR5IG5hbWU9ImhvbW9nZW5lb3VzIj5mYWxzZTwvcHJvcGVydHk+CiAgICAgICAgPHByb3BlcnR5IG5h
bWU9Im9yaWVudGF0aW9uIj5HVEtfT1JJRU5UQVRJT05fVkVSVElDQUw8L3Byb3BlcnR5PgogICAgICAg
IDxwcm9wZXJ0eSBuYW1lPSJzcGFjaW5nIj42PC9wcm9wZXJ0eT4KICAgICAgICA8Y2hpbGQ+CiAgICAg
ICAgICA8b2JqZWN0IGNsYXNzPSJHdGtTY3JvbGxlZFdpbmRvdyIgaWQ9Im1lc3NhZ2Vfc2Nyb2xsZWR3
aW5kb3ciPgogICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+VHJ1ZTwvcHJvcGVydHk+
CiAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJjYW5fZm9jdXMiPlRydWU8L3Byb3BlcnR5PgogICAg
ICAgICAgICA8cHJvcGVydHkgbmFtZT0ibm9fc2hvd19hbGwiPlRydWU8L3Byb3BlcnR5PgogICAgICAg
ICAgICA8cHJvcGVydHkgbmFtZT0iYm9yZGVyX3dpZHRoIj42PC9wcm9wZXJ0eT4KICAgICAgICAgICAg
PHByb3BlcnR5IG5hbWU9InNoYWRvd190eXBlIj5ldGNoZWQtaW48L3Byb3BlcnR5PgogICAgICAgICAg
ICA8cHJvcGVydHkgbmFtZT0ibWluX2NvbnRlbnRfaGVpZ2h0Ij41PC9wcm9wZXJ0eT4KICAgICAgICAg
ICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a1RleHRWaWV3IiBpZD0ibWVz
c2FnZV90ZXh0dmlldyI+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0idmlzaWJsZSI+dHJ1
ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0id3JhcC1tb2RlIj4yPC9w
cm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJlZGl0YWJsZSI+ZmFsc2U8L3By
b3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImN1cnNvci12aXNpYmxlIj5mYWxz
ZTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icGl4ZWxzLWJlbG93LWxp
bmVzIj41PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJsZWZ0LW1hcmdp
biI+NTwvcHJvcGVydHk+CiAgICAgICAgICAgICAgICA8cHJvcGVydHkgbmFtZT0icmlnaHQtbWFyZ2lu
Ij41PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJidWZmZXIiPmNvbnNv
bGVDb250ZW50PC9wcm9wZXJ0eT4KICAgICAgICAgICAgICA8L29iamVjdD4KICAgICAgICAgICAgPC9j
aGlsZD4KICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgPHBhY2tpbmc+CiAgICAgICAgICAgIDxw
cm9wZXJ0eSBuYW1lPSJleHBhbmQiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICA8cHJvcGVydHkg
bmFtZT0iZmlsbCI+VHJ1ZTwvcHJvcGVydHk+CiAgICAgICAgICA8L3BhY2tpbmc+CiAgICAgICAgPC9j
aGlsZD4KICAgICAgICA8Y2hpbGQgaW50ZXJuYWwtY2hpbGQ9ImFjdGlvbl9hcmVhIj4KICAgICAgICAg
IDxvYmplY3QgY2xhc3M9Ikd0a0J1dHRvbkJveCIgaWQ9ImJ1dHRvbl9ib3giPgogICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0ib3JpZW50YXRpb24iPkdUS19PUklFTlRBVElPTl9IT1JJWk9OVEFMPC9wcm9w
ZXJ0eT4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxvYmplY3QgY2xhc3M9Ikd0a0J1
dHRvbiIgaWQ9ImJ1dHRvbl9yZWZyZXNoIj4KICAgICAgICAgICAgICAgIDxwcm9wZXJ0eSBuYW1lPSJs
YWJlbCI+X1JlZnJlc2g8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InVz
ZS11bmRlcmxpbmUiPlRydWU8L3Byb3BlcnR5PgogICAgICAgICAgICAgICAgPHNpZ25hbCBuYW1lPSJj
bGlja2VkIiBoYW5kbGVyPSJvbl9yZWZyZXNoX3NpZ25hbCIgLz4KICAgICAgICAgICAgICA8L29iamVj
dD4KICAgICAgICAgICAgPC9jaGlsZD4KICAgICAgICAgICAgPGNoaWxkPgogICAgICAgICAgICAgIDxv
YmplY3QgY2xhc3M9Ikd0a0J1dHRvbiIgaWQ9ImJ1dHRvbl9jbG9zZSI+CiAgICAgICAgICAgICAgICA8
cHJvcGVydHkgbmFtZT0ibGFiZWwiIHRyYW5zbGF0YWJsZT0ieWVzIj5fQ2xvc2U8L3Byb3BlcnR5Pgog
ICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9InVzZS11bmRlcmxpbmUiPlRydWU8L3Byb3BlcnR5
PgogICAgICAgICAgICAgICAgPHByb3BlcnR5IG5hbWU9ImNhbi1kZWZhdWx0Ij50cnVlPC9wcm9wZXJ0
eT4KICAgICAgICAgICAgICAgIDxzaWduYWwgbmFtZT0iY2xpY2tlZCIgaGFuZGxlcj0ib25fY2xvc2Vf
c2lnbmFsIiAvPgogICAgICAgICAgICAgIDwvb2JqZWN0PgogICAgICAgICAgICA8L2NoaWxkPgogICAg
ICAgICAgPC9vYmplY3Q+CiAgICAgICAgPC9jaGlsZD4KICAgICAgPC9vYmplY3Q+CiAgICA8L2NoaWxk
PgogIDwvb2JqZWN0Pgo8L2ludGVyZmFjZT4K
`,
	},

	"/definitions/images/alert.svg": {
		local:   "definitions/images/alert.svg",
		size:    2810,
		modtime: 1489449600,
		compressed: `
PD94bWwgdmVyc2lvbj0iMS4wIiBlbmNvZGluZz0iVVRGLTgiIHN0YW5kYWxvbmU9Im5vIj8+Cjxzdmcg
d2lkdGg9IjE1cHgiIGhlaWdodD0iMTRweCIgdmlld0JveD0iMCAwIDE1IDE0IiB2ZXJzaW9uPSIxLjEi
IHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyIgeG1sbnM6eGxpbms9Imh0dHA6Ly93d3cu
dzMub3JnLzE5OTkveGxpbmsiPgogICAgPCEtLSBHZW5lcmF0b3I6IFNrZXRjaCA0MS4yICgzNTM5Nykg
LSBodHRwOi8vd3d3LmJvaGVtaWFuY29kaW5nLmNvbS9za2V0Y2ggLS0+CiAgICA8dGl0bGU+74GxIGNv
cHkgMjwvdGl0bGU+CiAgICA8ZGVzYz5DcmVhdGVkIHdpdGggU2tldGNoLjwvZGVzYz4KICAgIDxkZWZz
PjwvZGVmcz4KICAgIDxnIGlkPSJTTVAtZmxvd3MiIHN0cm9rZT0ibm9uZSIgc3Ryb2tlLXdpZHRoPSIx
IiBmaWxsPSJub25lIiBmaWxsLXJ1bGU9ImV2ZW5vZGQiPgogICAgICAgIDxnIGlkPSJCb2JfNGIiIHRy
YW5zZm9ybT0idHJhbnNsYXRlKC04MzkuMDAwMDAwLCAtMTY0LjAwMDAwMCkiIGZpbGw9IiNGODlCMUMi
PgogICAgICAgICAgICA8ZyBpZD0iR3JvdXAtMi1Db3B5IiB0cmFuc2Zvcm09InRyYW5zbGF0ZSg2MDAu
MDAwMDAwLCAxNDAuMDAwMDAwKSI+CiAgICAgICAgICAgICAgICA8cGF0aCBkPSJNMjQ3LjU3MDIxNywz
NS40OTY0NzY0IEwyNDcuNTcwMjE3LDMzLjkwNzg3MjQgQzI0Ny41NzAyMTcsMzMuODI5ODM1MyAyNDcu
NTQzNzQxLDMzLjc2NDM0MDkgMjQ3LjQ5MDc4NywzMy43MTEzODcxIEMyNDcuNDM3ODM0LDMzLjY1ODQz
MzQgMjQ3LjM3NTEyNiwzMy42MzE5NTY5IDI0Ny4zMDI2NjMsMzMuNjMxOTU2OSBMMjQ1LjY5NzMzNywz
My42MzE5NTY5IEMyNDUuNjI0ODc0LDMzLjYzMTk1NjkgMjQ1LjU2MjE2NiwzMy42NTg0MzM0IDI0NS41
MDkyMTMsMzMuNzExMzg3MSBDMjQ1LjQ1NjI1OSwzMy43NjQzNDA5IDI0NS40Mjk3ODMsMzMuODI5ODM1
MyAyNDUuNDI5NzgzLDMzLjkwNzg3MjQgTDI0NS40Mjk3ODMsMzUuNDk2NDc2NCBDMjQ1LjQyOTc4Mywz
NS41NzQ1MTM0IDI0NS40NTYyNTksMzUuNjQwMDA3OSAyNDUuNTA5MjEzLDM1LjY5Mjk2MTYgQzI0NS41
NjIxNjYsMzUuNzQ1OTE1MyAyNDUuNjI0ODc0LDM1Ljc3MjM5MTggMjQ1LjY5NzMzNywzNS43NzIzOTE4
IEwyNDcuMzAyNjYzLDM1Ljc3MjM5MTggQzI0Ny4zNzUxMjYsMzUuNzcyMzkxOCAyNDcuNDM3ODM0LDM1
Ljc0NTkxNTMgMjQ3LjQ5MDc4NywzNS42OTI5NjE2IEMyNDcuNTQzNzQxLDM1LjY0MDAwNzkgMjQ3LjU3
MDIxNywzNS41NzQ1MTM0IDI0Ny41NzAyMTcsMzUuNDk2NDc2NCBMMjQ3LjU3MDIxNywzNS40OTY0NzY0
IFogTTI0Ny41NTM0OTUsMzIuMzY5NDM0OCBMMjQ3LjcwMzk5NSwyOC41MzE3MDIgQzI0Ny43MDM5OTUs
MjguNDY0ODEzIDI0Ny42NzYxMjUsMjguNDExODYwMSAyNDcuNjIwMzg0LDI4LjM3Mjg0MTYgQzI0Ny41
NDc5MjEsMjguMzExNTI2NyAyNDcuNDgxMDMzLDI4LjI4MDg2OTcgMjQ3LjQxOTcxOCwyOC4yODA4Njk3
IEwyNDUuNTgwMjgyLDI4LjI4MDg2OTcgQzI0NS41MTg5NjcsMjguMjgwODY5NyAyNDUuNDUyMDc5LDI4
LjMxMTUyNjcgMjQ1LjM3OTYxNiwyOC4zNzI4NDE2IEMyNDUuMzIzODc1LDI4LjQxMTg2MDEgMjQ1LjI5
NjAwNSwyOC40NzAzODcgMjQ1LjI5NjAwNSwyOC41NDg0MjQxIEwyNDUuNDM4MTQ0LDMyLjM2OTQzNDgg
QzI0NS40MzgxNDQsMzIuNDI1MTc1NiAyNDUuNDY2MDE0LDMyLjQ3MTE2MSAyNDUuNTIxNzU0LDMyLjUw
NzM5MjUgQzI0NS41Nzc0OTUsMzIuNTQzNjI0IDI0NS42NDQzODMsMzIuNTYxNzM5NSAyNDUuNzIyNDIs
MzIuNTYxNzM5NSBMMjQ3LjI2OTIxOSwzMi41NjE3Mzk1IEMyNDcuMzQ3MjU2LDMyLjU2MTczOTUgMjQ3
LjQxMjc1LDMyLjU0MzYyNCAyNDcuNDY1NzA0LDMyLjUwNzM5MjUgQzI0Ny41MTg2NTgsMzIuNDcxMTYx
IDI0Ny41NDc5MjEsMzIuNDI1MTc1NiAyNDcuNTUzNDk1LDMyLjM2OTQzNDggTDI0Ny41NTM0OTUsMzIu
MzY5NDM0OCBaIE0yNDcuNDM2NDQsMjQuNTYwMTkxOSBMMjUzLjg1Nzc0NSwzNi4zMzI1ODM3IEMyNTQu
MDUyODM4LDM2LjY4Mzc1MDYgMjU0LjA0NzI2NCwzNy4wMzQ5MTIyIDI1My44NDEwMjMsMzcuMzg2MDc5
IEMyNTMuNzQ2MjYzLDM3LjU0NzcyNzMgMjUzLjYxNjY2OCwzNy42NzU5MjkxIDI1My40NTIyMzMsMzcu
NzcwNjg4NCBDMjUzLjI4Nzc5OCwzNy44NjU0NDc3IDI1My4xMTA4MjMsMzcuOTEyODI2NyAyNTIuOTIx
MzA1LDM3LjkxMjgyNjcgTDI0MC4wNzg2OTUsMzcuOTEyODI2NyBDMjM5Ljg4OTE3NywzNy45MTI4MjY3
IDIzOS43MTIyMDIsMzcuODY1NDQ3NyAyMzkuNTQ3NzY3LDM3Ljc3MDY4ODQgQzIzOS4zODMzMzIsMzcu
Njc1OTI5MSAyMzkuMjUzNzM3LDM3LjU0NzcyNzMgMjM5LjE1ODk3NywzNy4zODYwNzkgQzIzOC45NTI3
MzYsMzcuMDM0OTEyMiAyMzguOTQ3MTYyLDM2LjY4Mzc1MDYgMjM5LjE0MjI1NSwzNi4zMzI1ODM3IEwy
NDUuNTYzNTYsMjQuNTYwMTkxOSBDMjQ1LjY1ODMxOSwyNC4zODczOTU2IDI0NS43ODkzMDgsMjQuMjUw
ODMyNyAyNDUuOTU2NTMsMjQuMTUwNDk5MyBDMjQ2LjEyMzc1MywyNC4wNTAxNjU5IDI0Ni4zMDQ5MDcs
MjQgMjQ2LjUsMjQgQzI0Ni42OTUwOTMsMjQgMjQ2Ljg3NjI0NywyNC4wNTAxNjU5IDI0Ny4wNDM0Nywy
NC4xNTA0OTkzIEMyNDcuMjEwNjkyLDI0LjI1MDgzMjcgMjQ3LjM0MTY4MSwyNC4zODczOTU2IDI0Ny40
MzY0NCwyNC41NjAxOTE5IEwyNDcuNDM2NDQsMjQuNTYwMTkxOSBaIiBpZD0i74GxLWNvcHktMiI+PC9w
YXRoPgogICAgICAgICAgICA8L2c+CiAgICAgICAgPC9nPgogICAgPC9nPgo8L3N2Zz4=
`,
	},

	"/definitions/images/failure.svg": {
		local:   "definitions/images/failure.svg",
		size:    1153,
		modtime: 1489449600,
		compressed: `
PD94bWwgdmVyc2lvbj0iMS4wIiBlbmNvZGluZz0idXRmLTgiPz4KPCEtLSBHZW5lcmF0b3I6IEFkb2Jl
IElsbHVzdHJhdG9yIDIyLjAuMSwgU1ZHIEV4cG9ydCBQbHVnLUluIC4gU1ZHIFZlcnNpb246IDYuMDAg
QnVpbGQgMCkgIC0tPgo8c3ZnIHZlcnNpb249IjEuMSIgaWQ9IkxheWVyXzEiIHhtbG5zPSJodHRwOi8v
d3d3LnczLm9yZy8yMDAwL3N2ZyIgeG1sbnM6eGxpbms9Imh0dHA6Ly93d3cudzMub3JnLzE5OTkveGxp
bmsiIHg9IjBweCIgeT0iMHB4IgoJIHZpZXdCb3g9IjAgMCA1MiA1MyIgc3R5bGU9ImVuYWJsZS1iYWNr
Z3JvdW5kOm5ldyAwIDAgNTIgNTM7IiB4bWw6c3BhY2U9InByZXNlcnZlIj4KPHN0eWxlIHR5cGU9InRl
eHQvY3NzIj4KCS5zdDB7ZmlsbDpub25lO3N0cm9rZTojQ0MzNjM2O3N0cm9rZS1saW5lY2FwOnJvdW5k
O3N0cm9rZS1saW5lam9pbjpyb3VuZDtzdHJva2UtbWl0ZXJsaW1pdDoxMDt9Cgkuc3Qxe2ZpbGw6I0ZD
QUVBRTtzdHJva2U6I0NDMzYzNjtzdHJva2UtbGluZWNhcDpyb3VuZDtzdHJva2UtbGluZWpvaW46cm91
bmQ7c3Ryb2tlLW1pdGVybGltaXQ6MTA7fQo8L3N0eWxlPgo8Zz4KCTxjaXJjbGUgY2xhc3M9InN0MCIg
Y3g9IjI4LjYiIGN5PSIyNy41IiByPSIyMS4xIi8+Cgk8cGF0aCBjbGFzcz0ic3QwIiBkPSJNMTIuNSwy
My43YzAtMy43LDEuNy03LDQuNC05LjMiLz4KCTxwYXRoIGNsYXNzPSJzdDAiIGQ9Ik0xOC4zLDEzLjVj
MC4zLTAuMiwwLjYtMC40LDAuOS0wLjUiLz4KCTxwb2x5Z29uIGNsYXNzPSJzdDEiIHBvaW50cz0iMzcu
OCwyMi45IDMzLjksMTguOSAyOC42LDI0LjMgMjMuMiwxOC45IDE5LjMsMjIuOSAyNC42LDI4LjIgMTku
MywzMy42IDIzLjIsMzcuNSAyOC42LDMyLjIgCgkJMzMuOSwzNy41IDM3LjgsMzMuNiAzMi41LDI4LjIg
CSIvPgoJPHBhdGggY2xhc3M9InN0MCIgZD0iTTIxLjEsMTIuMWMxLjEtMC4zLDIuMi0wLjUsMy40LTAu
NSIvPgoJPHBhdGggY2xhc3M9InN0MCIgZD0iTTM1LjIsNDIuNmMtMS4xLDAuMy0yLjIsMC41LTMuNCww
LjUiLz4KCTxwYXRoIGNsYXNzPSJzdDAiIGQ9Ik00My45LDMxYzAsMy43LTEuNyw3LTQuNCw5LjMiLz4K
CTxwYXRoIGNsYXNzPSJzdDAiIGQ9Ik0zOC4xLDQxLjNjLTAuMywwLjItMC42LDAuNC0wLjksMC41Ii8+
CjwvZz4KPC9zdmc+Cg==
`,
	},

	"/definitions/images/filetransfer_receive.svg": {
		local:   "definitions/images/filetransfer_receive.svg",
		size:    3030,
		modtime: 1489449600,
		compressed: `
PD94bWwgdmVyc2lvbj0iMS4wIiBlbmNvZGluZz0idXRmLTgiPz4KPCEtLSBHZW5lcmF0b3I6IEFkb2Jl
IElsbHVzdHJhdG9yIDIyLjAuMSwgU1ZHIEV4cG9ydCBQbHVnLUluIC4gU1ZHIFZlcnNpb246IDYuMDAg
QnVpbGQgMCkgIC0tPgo8c3ZnIHZlcnNpb249IjEuMSIgaWQ9IkxheWVyXzEiIHhtbG5zPSJodHRwOi8v
d3d3LnczLm9yZy8yMDAwL3N2ZyIgeG1sbnM6eGxpbms9Imh0dHA6Ly93d3cudzMub3JnLzE5OTkveGxp
bmsiIHg9IjBweCIgeT0iMHB4IgoJIHZpZXdCb3g9IjAgMCA2MS41IDYzLjciIHN0eWxlPSJlbmFibGUt
YmFja2dyb3VuZDpuZXcgMCAwIDYxLjUgNjMuNzsiIHhtbDpzcGFjZT0icHJlc2VydmUiPgo8c3R5bGUg
dHlwZT0idGV4dC9jc3MiPgoJLnN0MHtmaWxsOiNGRkZCRjg7fQoJLnN0MXtmaWxsOiNGRkU1RDQ7fQoJ
LnN0MntmaWxsOiNGMEU5RkY7c3Ryb2tlOiM4OTgwQTA7c3Ryb2tlLXdpZHRoOjEuMjU7c3Ryb2tlLWxp
bmVjYXA6cm91bmQ7c3Ryb2tlLW1pdGVybGltaXQ6MTA7fQoJLnN0M3tmaWxsOiNGMEU5RkY7c3Ryb2tl
OiM4OTgwQTA7c3Ryb2tlLWxpbmVjYXA6cm91bmQ7c3Ryb2tlLWxpbmVqb2luOnJvdW5kO3N0cm9rZS1t
aXRlcmxpbWl0OjEwO30KCS5zdDR7ZmlsbDpub25lO3N0cm9rZTojODk4MEEwO3N0cm9rZS13aWR0aDox
LjI1O3N0cm9rZS1saW5lY2FwOnJvdW5kO3N0cm9rZS1taXRlcmxpbWl0OjEwO30KCS5zdDV7ZmlsbDpu
b25lO3N0cm9rZTojODk4MEEwO3N0cm9rZS13aWR0aDoxLjI1O3N0cm9rZS1saW5lY2FwOnJvdW5kO3N0
cm9rZS1saW5lam9pbjpyb3VuZDtzdHJva2UtbWl0ZXJsaW1pdDoxMDt9Cgkuc3Q2e2ZpbGw6bm9uZTtz
dHJva2U6Izg5ODBBMDtzdHJva2UtbGluZWNhcDpyb3VuZDtzdHJva2UtbGluZWpvaW46cm91bmQ7c3Ry
b2tlLW1pdGVybGltaXQ6MTA7fQoJLnN0N3tmaWxsOm5vbmU7c3Ryb2tlOiM4OTgwQTA7c3Ryb2tlLXdp
ZHRoOjAuNzU7c3Ryb2tlLWxpbmVjYXA6cm91bmQ7c3Ryb2tlLWxpbmVqb2luOnJvdW5kO3N0cm9rZS1t
aXRlcmxpbWl0OjEwO30KCS5zdDh7ZmlsbDojRjBFOUZGO3N0cm9rZTojODk4MEEwO3N0cm9rZS13aWR0
aDoxLjI1O3N0cm9rZS1saW5lY2FwOnJvdW5kO3N0cm9rZS1saW5lam9pbjpyb3VuZDtzdHJva2UtbWl0
ZXJsaW1pdDoxMDt9Cgkuc3Q5e2ZpbGw6I0YwRTlGRjt9Cjwvc3R5bGU+CjxjaXJjbGUgaWQ9IlhNTElE
XzhfIiBjbGFzcz0ic3QwIiBjeD0iMzIiIGN5PSIzMC45IiByPSIyNyIvPgo8cG9seWxpbmUgaWQ9IlJl
Y3RhbmdsZS00LUNvcHktMl8xXyIgY2xhc3M9InN0MSIgcG9pbnRzPSI0OCwxMi45IDQyLjcsMTIuOSA0
Mi43LDE4LjMgNDgsMTguMyAiLz4KPHBhdGggY2xhc3M9InN0MiIgZD0iTTI2LDE5aC00LjVjLTAuNyww
LTEuMywwLjUtMS4zLDEuMmwtMC4xLDEuM2MtMC4xLDAuNy0wLjYsMS4yLTEuMywxLjJoLTAuM2MtMC45
LDAtMS43LDAuNy0xLjgsMS42bC0wLjEsMS4xIgoJLz4KPHBhdGggY2xhc3M9InN0MiIgZD0iTTQ4LjMs
NDEuNWwyLjItMTcuM2MwLjEtMC45LTAuOS0xLjctMi4zLTEuN0gzMi40Yy0xLDAtMS44LTAuNi0xLjct
MS40bDAuMS0wLjdjMC4xLTAuNy0wLjYtMS40LTEuNy0xLjRoLTQuNSIKCS8+Cjxwb2x5bGluZSBjbGFz
cz0ic3QzIiBwb2ludHM9IjIxLjIsMjcuNCAyMS41LDI1LjcgNDYuOSwyNS43IDQ2LDMyICIvPgo8cGF0
aCBjbGFzcz0ic3Q0IiBkPSJNMTAuNiwyN2g3LjFjMC44LDAsMS41LDAuNSwxLjcsMS4zbDAuNCwxLjZj
MC4yLDAuNiwwLjcsMSwxLjMsMWgyMS4yYzAuOCwwLDEuNSwwLjUsMS43LDEuM2w0LjEsMTUuMkgxNS4z
CgljLTAuNiwwLTEuMS0wLjQtMS4zLTFMOS40LDI4LjciLz4KPGxpbmUgaWQ9IlhNTElEXzdfIiBjbGFz
cz0ic3Q1IiB4MT0iMy4xIiB5MT0iNDgiIHgyPSIxMC4yIiB5Mj0iNDgiLz4KPGxpbmUgaWQ9IlhNTElE
XzlfIiBjbGFzcz0ic3Q1IiB4MT0iNTIuMyIgeTE9IjQ4IiB4Mj0iNTQuMyIgeTI9IjQ4Ii8+CjxsaW5l
IGlkPSJYTUxJRF81XyIgY2xhc3M9InN0NiIgeDE9IjEzLjEiIHkxPSI0OCIgeDI9IjUwIiB5Mj0iNDgi
Lz4KPGxpbmUgaWQ9IlhNTElEXzRfIiBjbGFzcz0ic3Q3IiB4MT0iMTIuMiIgeTE9IjUxLjIiIHgyPSIx
OC4yIiB5Mj0iNTEuMiIvPgo8bGluZSBpZD0iWE1MSURfM18iIGNsYXNzPSJzdDciIHgxPSIyMS44IiB5
MT0iNTEuMiIgeDI9IjI3LjciIHkyPSI1MS4yIi8+CjxsaW5lIGlkPSJYTUxJRF8yXyIgY2xhc3M9InN0
NyIgeDE9IjMxLjEiIHkxPSI1MS4yIiB4Mj0iNDMuMiIgeTI9IjUxLjIiLz4KPGxpbmUgaWQ9IlhNTElE
XzFfIiBjbGFzcz0ic3Q3IiB4MT0iNDYuOSIgeTE9IjUxLjIiIHgyPSI1My44IiB5Mj0iNTEuMiIvPgo8
cmVjdCBpZD0iUmVjdGFuZ2xlLTQtQ29weS05XzFfIiB4PSI1Mi41IiB5PSIzNCIgY2xhc3M9InN0MSIg
d2lkdGg9IjIuNCIgaGVpZ2h0PSIzLjciLz4KPHBvbHlsaW5lIGlkPSJSZWN0YW5nbGUtNC1Db3B5LTE2
XzFfIiBjbGFzcz0ic3QxIiBwb2ludHM9IjE0LjMsMjIuMyAxNC4zLDE5LjggMTIuMSwxOS44IDEyLjEs
MjIuMyAiLz4KPHJlY3QgaWQ9IlJlY3RhbmdsZS00LUNvcHktMTlfMV8iIHg9IjE4LjUiIHk9IjE0LjQi
IGNsYXNzPSJzdDEiIHdpZHRoPSIyLjQiIGhlaWdodD0iMi40Ii8+CjxwYXRoIGNsYXNzPSJzdDgiIGQ9
Ik00Mi41LDIwLjFjLTAuOC0yLjUtMi4zLTQuNy00LjQtNi4zYy00LjktMy41LTExLjYtMi40LTE1Ljks
Mi40YzMuOC0xLjgsOC4yLTEuOCwxMS43LDAuNwoJYzEuNiwxLjEsMi44LDIuNiwzLjcsNC40bC0zLjks
MC45bDcsNS43bDQuOC04LjdMNDIuNSwyMC4xeiIvPgo8cGF0aCBjbGFzcz0ic3Q5IiBkPSJNMTIuNiwz
MC4xaDYuMmMwLjcsMCwxLjMsMC40LDEuNSwxbDAuMywxLjJjMC4yLDAuNCwwLjYsMC43LDEuMSwwLjdo
MTguNWMwLjcsMCwxLjMsMC40LDEuNSwxbDMuNiwxMS4zSDE2LjcKCWMtMC41LDAtMS0wLjMtMS4xLTAu
N2wtNC0xMy4yIi8+CjxsaW5lIGlkPSJYTUxJRF82XyIgY2xhc3M9InN0NSIgeDE9IjU2LjYiIHkxPSI0
OCIgeDI9IjU3LjgiIHkyPSI0OCIvPgo8L3N2Zz4K
`,
	},

	"/definitions/images/filetransfer_receive_send.svg": {
		local:   "definitions/images/filetransfer_receive_send.svg",
		size:    2786,
		modtime: 1489449600,
		compressed: `
PD94bWwgdmVyc2lvbj0iMS4wIiBlbmNvZGluZz0idXRmLTgiPz4KPCEtLSBHZW5lcmF0b3I6IEFkb2Jl
IElsbHVzdHJhdG9yIDIyLjAuMSwgU1ZHIEV4cG9ydCBQbHVnLUluIC4gU1ZHIFZlcnNpb246IDYuMDAg
QnVpbGQgMCkgIC0tPgo8c3ZnIHZlcnNpb249IjEuMSIgaWQ9IkxheWVyXzEiIHhtbG5zPSJodHRwOi8v
d3d3LnczLm9yZy8yMDAwL3N2ZyIgeG1sbnM6eGxpbms9Imh0dHA6Ly93d3cudzMub3JnLzE5OTkveGxp
bmsiIHg9IjBweCIgeT0iMHB4IgoJIHZpZXdCb3g9IjAgMCA2MCA2Mi4yIiBzdHlsZT0iZW5hYmxlLWJh
Y2tncm91bmQ6bmV3IDAgMCA2MCA2Mi4yOyIgeG1sOnNwYWNlPSJwcmVzZXJ2ZSI+CjxzdHlsZSB0eXBl
PSJ0ZXh0L2NzcyI+Cgkuc3Qwe2ZpbGw6I0ZGRkJGODt9Cgkuc3Qxe2ZpbGw6I0ZGRTVENDt9Cgkuc3Qy
e2ZpbGw6I0YwRTlGRjtzdHJva2U6Izg5ODBBMDtzdHJva2Utd2lkdGg6MS4yNTtzdHJva2UtbGluZWNh
cDpyb3VuZDtzdHJva2UtbWl0ZXJsaW1pdDoxMDt9Cgkuc3Qze2ZpbGw6I0YwRTlGRjtzdHJva2U6Izg5
ODBBMDtzdHJva2Utd2lkdGg6MS4yNTtzdHJva2UtbGluZWNhcDpyb3VuZDtzdHJva2UtbGluZWpvaW46
cm91bmQ7c3Ryb2tlLW1pdGVybGltaXQ6MTA7fQoJLnN0NHtmaWxsOm5vbmU7c3Ryb2tlOiM4OTgwQTA7
c3Ryb2tlLWxpbmVjYXA6cm91bmQ7c3Ryb2tlLW1pdGVybGltaXQ6MTA7fQoJLnN0NXtmaWxsOm5vbmU7
c3Ryb2tlOiM4OTgwQTA7c3Ryb2tlLXdpZHRoOjEuMjU7c3Ryb2tlLWxpbmVjYXA6cm91bmQ7c3Ryb2tl
LWxpbmVqb2luOnJvdW5kO3N0cm9rZS1taXRlcmxpbWl0OjEwO30KCS5zdDZ7ZmlsbDpub25lO3N0cm9r
ZTojODk4MEEwO3N0cm9rZS13aWR0aDowLjc1O3N0cm9rZS1saW5lY2FwOnJvdW5kO3N0cm9rZS1saW5l
am9pbjpyb3VuZDtzdHJva2UtbWl0ZXJsaW1pdDoxMDt9Cjwvc3R5bGU+CjxjaXJjbGUgaWQ9IlhNTElE
XzExNzZfIiBjbGFzcz0ic3QwIiBjeD0iMzAuMyIgY3k9IjMwLjciIHI9IjI3Ii8+Cjxwb2x5bGluZSBp
ZD0iUmVjdGFuZ2xlLTQtQ29weS0yIiBjbGFzcz0ic3QxIiBwb2ludHM9IjQ2LjQsMTIuNyA0MSwxMi43
IDQxLDE4LjEgNDYuNCwxOC4xICIvPgo8cGF0aCBjbGFzcz0ic3QyIiBkPSJNMjYuNCwyMC40aC03LjNj
LTAuNSwwLTEsMC41LTEsMS4yTDE4LDIyLjljLTAuMSwwLjctMC41LDEuMi0xLDEuMmgtMC4yYy0wLjcs
MC0xLjMsMC43LTEuNCwxLjZsLTAuMSwxIi8+CjxwYXRoIGNsYXNzPSJzdDIiIGQ9Ik00Ny4xLDQzbDEu
NS0xOC4xYzAuMS0xLTAuNi0xLjktMS44LTEuOUgyOC41Ii8+CjxwYXRoIGNsYXNzPSJzdDMiIGQ9Ik0x
OS45LDMwLjZoMjAuOWMwLjgsMCwxLjUsMC41LDEuNywxLjNsNCwxNC45SDE0LjJjLTAuNiwwLTEuMS0w
LjQtMS4zLTFMOC4yLDI4LjRjLTAuMi0wLjgsMC40LTEuNywxLjMtMS43aDcKCWMwLjgsMCwxLjUsMC41
LDEuNywxLjNsMC40LDEuNkMxOC44LDMwLjIsMTkuMywzMC42LDE5LjksMzAuNnoiLz4KPHBvbHlsaW5l
IGNsYXNzPSJzdDQiIHBvaW50cz0iMzguNiwyNi4xIDQ2LjEsMjYuMSA0NS4xLDM1LjcgIi8+Cjxwb2x5
bGluZSBjbGFzcz0ic3Q0IiBwb2ludHM9IjE5LjksMjguMiAyMC4xLDI1LjQgMjUuNCwyNS40ICIvPgo8
cGF0aCBjbGFzcz0ic3QzIiBkPSJNMjUuOSwyNS43YzEuNiwxLjUsMy42LDIuNiw1LjksM2M1LjIsMC44
LDEwLjEtMi4zLDExLjctNy4yYy0yLjMsMi42LTUuOCw0LjEtOS42LDMuNQoJYy0xLjctMC4yLTMuMi0w
LjktNC42LTEuOWwyLjctMmwtNy43LTEuN2wtMC41LDcuOUwyNS45LDI1Ljd6Ii8+CjxwYXRoIGNsYXNz
PSJzdDMiIGQ9Ik00Mi41LDEzLjNsLTAuMS0wLjFjLTEuNS0xLjUtMy41LTIuNi01LjktM2MtNS4yLTAu
OC0xMC4xLDIuMy0xMS43LDcuMWMyLjMtMi42LDUuOS00LDkuNi0zLjQKCWMxLjcsMC4yLDMuMiwxLDQu
NiwxLjlsMC4xLDAuMUwzNi4zLDE4bDcuNywxLjhsMC42LTcuOUw0Mi41LDEzLjN6Ii8+CjxsaW5lIGlk
PSJYTUxJRF8xMTg5XyIgY2xhc3M9InN0NSIgeDE9IjIuNCIgeTE9IjQ2LjgiIHgyPSI5LjYiIHkyPSI0
Ni44Ii8+CjxsaW5lIGlkPSJYTUxJRF8xMTg4XyIgY2xhc3M9InN0NSIgeDE9IjUyLjEiIHkxPSI0Ni44
IiB4Mj0iNTciIHkyPSI0Ni44Ii8+CjxsaW5lIGlkPSJYTUxJRF8xMTg3XyIgY2xhc3M9InN0NSIgeDE9
IjEyLjQiIHkxPSI0Ni44IiB4Mj0iNDkuMyIgeTI9IjQ2LjgiLz4KPGxpbmUgaWQ9IlhNTElEXzExODNf
IiBjbGFzcz0ic3Q2IiB4MT0iMTAuNiIgeTE9IjQ5LjkiIHgyPSIxNi41IiB5Mj0iNDkuOSIvPgo8bGlu
ZSBpZD0iWE1MSURfMTE4Ml8iIGNsYXNzPSJzdDYiIHgxPSIyMC4xIiB5MT0iNDkuOSIgeDI9IjI2IiB5
Mj0iNDkuOSIvPgo8bGluZSBpZD0iWE1MSURfMTE4MV8iIGNsYXNzPSJzdDYiIHgxPSIyOS40IiB5MT0i
NDkuOSIgeDI9IjQxLjYiIHkyPSI0OS45Ii8+CjxsaW5lIGlkPSJYTUxJRF8xMTgwXyIgY2xhc3M9InN0
NiIgeDE9IjQ1LjIiIHkxPSI0OS45IiB4Mj0iNTIuMSIgeTI9IjQ5LjkiLz4KPHJlY3QgaWQ9IlJlY3Rh
bmdsZS00LUNvcHktOSIgeD0iNTAuOSIgeT0iMzMuNyIgY2xhc3M9InN0MSIgd2lkdGg9IjIuNCIgaGVp
Z2h0PSIzLjciLz4KPHBvbHlsaW5lIGlkPSJSZWN0YW5nbGUtNC1Db3B5LTE2IiBjbGFzcz0ic3QxIiBw
b2ludHM9IjEyLjcsMjIuMSAxMi43LDE5LjcgMTAuNCwxOS43IDEwLjQsMjIuMSAiLz4KPHJlY3QgaWQ9
IlJlY3RhbmdsZS00LUNvcHktMTkiIHg9IjE2LjgiIHk9IjE0LjIiIGNsYXNzPSJzdDEiIHdpZHRoPSIy
LjQiIGhlaWdodD0iMi40Ii8+Cjwvc3ZnPgo=
`,
	},

	"/definitions/images/filetransfer_send.svg": {
		local:   "definitions/images/filetransfer_send.svg",
		size:    2609,
		modtime: 1489449600,
		compressed: `
PD94bWwgdmVyc2lvbj0iMS4wIiBlbmNvZGluZz0idXRmLTgiPz4KPCEtLSBHZW5lcmF0b3I6IEFkb2Jl
IElsbHVzdHJhdG9yIDIyLjAuMSwgU1ZHIEV4cG9ydCBQbHVnLUluIC4gU1ZHIFZlcnNpb246IDYuMDAg
QnVpbGQgMCkgIC0tPgo8c3ZnIHZlcnNpb249IjEuMSIgaWQ9IkxheWVyXzEiIHhtbG5zPSJodHRwOi8v
d3d3LnczLm9yZy8yMDAwL3N2ZyIgeG1sbnM6eGxpbms9Imh0dHA6Ly93d3cudzMub3JnLzE5OTkveGxp
bmsiIHg9IjBweCIgeT0iMHB4IgoJIHZpZXdCb3g9IjAgMCA2MS41IDYzLjciIHN0eWxlPSJlbmFibGUt
YmFja2dyb3VuZDpuZXcgMCAwIDYxLjUgNjMuNzsiIHhtbDpzcGFjZT0icHJlc2VydmUiPgo8c3R5bGUg
dHlwZT0idGV4dC9jc3MiPgoJLnN0MHtmaWxsOiNGRkU1RDQ7fQoJLnN0MXtmaWxsOiNGRkZCRjg7fQoJ
LnN0MntmaWxsOiNGMEU5RkY7c3Ryb2tlOiM4OTgwQTA7c3Ryb2tlLXdpZHRoOjEuMjU7c3Ryb2tlLWxp
bmVjYXA6cm91bmQ7c3Ryb2tlLW1pdGVybGltaXQ6MTA7fQoJLnN0M3tmaWxsOiNGMEU5RkY7c3Ryb2tl
OiM4OTgwQTA7c3Ryb2tlLWxpbmVjYXA6cm91bmQ7c3Ryb2tlLW1pdGVybGltaXQ6MTA7fQoJLnN0NHtm
aWxsOm5vbmU7c3Ryb2tlOiM4OTgwQTA7c3Ryb2tlLXdpZHRoOjEuMjU7c3Ryb2tlLWxpbmVjYXA6cm91
bmQ7c3Ryb2tlLWxpbmVqb2luOnJvdW5kO3N0cm9rZS1taXRlcmxpbWl0OjEwO30KCS5zdDV7ZmlsbDpu
b25lO3N0cm9rZTojODk4MEEwO3N0cm9rZS13aWR0aDowLjc1O3N0cm9rZS1saW5lY2FwOnJvdW5kO3N0
cm9rZS1saW5lam9pbjpyb3VuZDtzdHJva2UtbWl0ZXJsaW1pdDoxMDt9Cgkuc3Q2e2ZpbGw6I0YwRTlG
RjtzdHJva2U6Izg5ODBBMDtzdHJva2Utd2lkdGg6MS4yNTtzdHJva2UtbGluZWNhcDpyb3VuZDtzdHJv
a2UtbGluZWpvaW46cm91bmQ7c3Ryb2tlLW1pdGVybGltaXQ6MTA7fQo8L3N0eWxlPgo8cmVjdCBpZD0i
UmVjdGFuZ2xlLTQtQ29weS05IiB4PSI1Mi4xIiB5PSIzNS4zIiBjbGFzcz0ic3QwIiB3aWR0aD0iMi40
IiBoZWlnaHQ9IjMuNyIvPgo8Y2lyY2xlIGlkPSJYTUxJRF8xMTc2XyIgY2xhc3M9InN0MSIgY3g9IjMx
LjUiIGN5PSIzMi4zIiByPSIyNyIvPgo8cGF0aCBjbGFzcz0ic3QyIiBkPSJNMjQuOCwyMmgtNS40Yy0w
LjQsMC0wLjcsMC41LTAuNywxLjJsLTAuMSwxLjNjLTAuMSwwLjctMC4zLDEuMi0wLjcsMS4yaC0wLjFj
LTAuNSwwLTEsMC43LTEsMS42bC0wLjEsMSIvPgo8cGF0aCBjbGFzcz0ic3QyIiBkPSJNNDcuMiw0NC42
TDQ5LjksMjZjMC4xLTEtMS4xLTEuOS0zLjEtMS45SDMzLjciLz4KPHBvbHlsaW5lIGNsYXNzPSJzdDMi
IHBvaW50cz0iNDEsMjkuNSA0Ni4zLDI5LjUgNDUuNiwzNy4xICIvPgo8cG9seWxpbmUgY2xhc3M9InN0
MyIgcG9pbnRzPSIyMS4xLDI5LjggMjEuNCwyNyAyOS45LDI3ICIvPgo8bGluZSBpZD0iWE1MSURfMTE4
OV8iIGNsYXNzPSJzdDQiIHgxPSIzLjYiIHkxPSI0OC40IiB4Mj0iMTAuOCIgeTI9IjQ4LjQiLz4KPGxp
bmUgaWQ9IlhNTElEXzExODdfIiBjbGFzcz0ic3Q0IiB4MT0iMTMuNiIgeTE9IjQ4LjQiIHgyPSI1MC41
IiB5Mj0iNDguNCIvPgo8bGluZSBpZD0iWE1MSURfMTE4M18iIGNsYXNzPSJzdDUiIHgxPSIxMS44IiB5
MT0iNTEuNSIgeDI9IjE3LjciIHkyPSI1MS41Ii8+CjxsaW5lIGlkPSJYTUxJRF8xMTgyXyIgY2xhc3M9
InN0NSIgeDE9IjIxLjMiIHkxPSI1MS41IiB4Mj0iMjcuMiIgeTI9IjUxLjUiLz4KPGxpbmUgaWQ9IlhN
TElEXzExODFfIiBjbGFzcz0ic3Q1IiB4MT0iMzAuNiIgeTE9IjUxLjUiIHgyPSI0Mi44IiB5Mj0iNTEu
NSIvPgo8bGluZSBpZD0iWE1MSURfMTE4MF8iIGNsYXNzPSJzdDUiIHgxPSI0Ni40IiB5MT0iNTEuNSIg
eDI9IjUzLjMiIHkyPSI1MS41Ii8+CjxwYXRoIGNsYXNzPSJzdDYiIGQ9Ik0yMS4xLDMyLjJINDJjMC44
LDAsMS41LDAuNSwxLjcsMS4zbDQsMTQuOUgxNS40Yy0wLjYsMC0xLjEtMC40LTEuMy0xTDkuNCwzMGMt
MC4yLTAuOCwwLjQtMS43LDEuMy0xLjdoNwoJYzAuOCwwLDEuNSwwLjUsMS43LDEuM2wwLjQsMS42QzIw
LDMxLjgsMjAuNSwzMi4yLDIxLjEsMzIuMnoiLz4KPHBvbHlsaW5lIGlkPSJSZWN0YW5nbGUtNC1Db3B5
LTIiIGNsYXNzPSJzdDAiIHBvaW50cz0iNDcuNiwxNC4zIDQyLjIsMTQuMyA0Mi4yLDE5LjcgNDcuNiwx
OS43ICIvPgo8cG9seWxpbmUgaWQ9IlJlY3RhbmdsZS00LUNvcHktMTYiIGNsYXNzPSJzdDAiIHBvaW50
cz0iMTMuOSwyMy43IDEzLjksMjEuMyAxMS42LDIxLjMgMTEuNiwyMy43ICIvPgo8cmVjdCBpZD0iUmVj
dGFuZ2xlLTQtQ29weS0xOSIgeD0iMTgiIHk9IjE1LjgiIGNsYXNzPSJzdDAiIHdpZHRoPSIyLjQiIGhl
aWdodD0iMi40Ii8+CjxwYXRoIGNsYXNzPSJzdDYiIGQ9Ik0yMy45LDE3Yy0wLjEsMi42LDAuNSw1LjIs
MS45LDcuNGMzLjQsNC45LDEwLjEsNi4yLDE1LjcsMy4zYy00LjIsMC40LTguMy0xLjItMTAuNy00LjgK
CWMtMS4yLTEuNi0xLjctMy40LTEuOS01LjRsMy45LDAuNWwtNC41LTcuOGwtNy41LDYuNUwyMy45LDE3
eiIvPgo8bGluZSBpZD0iWE1MSURfMTE4OF8iIGNsYXNzPSJzdDQiIHgxPSI1NC41IiB5MT0iNDguNCIg
eDI9IjU4LjIiIHkyPSI0OC40Ii8+Cjwvc3ZnPgo=
`,
	},

	"/definitions/images/padlock.svg": {
		local:   "definitions/images/padlock.svg",
		size:    2835,
		modtime: 1489449600,
		compressed: `
PD94bWwgdmVyc2lvbj0iMS4wIiBlbmNvZGluZz0iVVRGLTgiIHN0YW5kYWxvbmU9Im5vIj8+Cjxzdmcg
d2lkdGg9IjExcHgiIGhlaWdodD0iMTNweCIgdmlld0JveD0iMCAwIDExIDEzIiB2ZXJzaW9uPSIxLjEi
IHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyIgeG1sbnM6eGxpbms9Imh0dHA6Ly93d3cu
dzMub3JnLzE5OTkveGxpbmsiPgogICAgPCEtLSBHZW5lcmF0b3I6IFNrZXRjaCA0MS4yICgzNTM5Nykg
LSBodHRwOi8vd3d3LmJvaGVtaWFuY29kaW5nLmNvbS9za2V0Y2ggLS0+CiAgICA8dGl0bGU+R3JvdXA8
L3RpdGxlPgogICAgPGRlc2M+Q3JlYXRlZCB3aXRoIFNrZXRjaC48L2Rlc2M+CiAgICA8ZGVmcz48L2Rl
ZnM+CiAgICA8ZyBpZD0iU01QLWZsb3dzIiBzdHJva2U9Im5vbmUiIHN0cm9rZS13aWR0aD0iMSIgZmls
bD0ibm9uZSIgZmlsbC1ydWxlPSJldmVub2RkIj4KICAgICAgICA8ZyBpZD0iQm9iXzQiIHRyYW5zZm9y
bT0idHJhbnNsYXRlKC04NDMuMDAwMDAwLCAtMTY0LjAwMDAwMCkiPgogICAgICAgICAgICA8ZyBpZD0i
R3JvdXAiIHRyYW5zZm9ybT0idHJhbnNsYXRlKDg0My4wMDAwMDAsIDE2NC4wMDAwMDApIj4KICAgICAg
ICAgICAgICAgIDxwYXRoIGQ9Ik0zLjA1NTU1NTU2LDUuOTA5MDkwOTEgTDcuOTQ0NDQ0NDQsNS45MDkw
OTA5MSBMNy45NDQ0NDQ0NCw0LjEzNjM2MzY0IEM3Ljk0NDQ0NDQ0LDMuNDgzODk4MjUgNy43MDU3MzE1
NSwyLjkyNjg0ODkgNy4yMjgyOTg2MSwyLjQ2NTE5ODg2IEM2Ljc1MDg2NTY3LDIuMDAzNTQ4ODMgNi4x
NzQ3NzE4OSwxLjc3MjcyNzI3IDUuNSwxLjc3MjcyNzI3IEM0LjgyNTIyODExLDEuNzcyNzI3MjcgNC4y
NDkxMzQzMywyLjAwMzU0ODgzIDMuNzcxNzAxMzksMi40NjUxOTg4NiBDMy4yOTQyNjg0NSwyLjkyNjg0
ODkgMy4wNTU1NTU1NiwzLjQ4Mzg5ODI1IDMuMDU1NTU1NTYsNC4xMzYzNjM2NCBMMy4wNTU1NTU1Niw1
LjkwOTA5MDkxIFogTTExLDYuNzk1NDU0NTUgTDExLDEyLjExMzYzNjQgQzExLDEyLjM1OTg0OTcgMTAu
OTEwODgwNSwxMi41NjkxMjc5IDEwLjczMjYzODksMTIuNzQxNDc3MyBDMTAuNTU0Mzk3MywxMi45MTM4
MjY2IDEwLjMzNzk2NDIsMTMgMTAuMDgzMzMzMywxMyBMMC45MTY2NjY2NjcsMTMgQzAuNjYyMDM1NzY0
LDEzIDAuNDQ1NjAyNzQzLDEyLjkxMzgyNjYgMC4yNjczNjExMTEsMTIuNzQxNDc3MyBDMC4wODkxMTk0
NzkyLDEyLjU2OTEyNzkgMCwxMi4zNTk4NDk3IDAsMTIuMTEzNjM2NCBMMCw2Ljc5NTQ1NDU1IEMwLDYu
NTQ5MjQxMTkgMC4wODkxMTk0NzkyLDYuMzM5OTYyOTggMC4yNjczNjExMTEsNi4xNjc2MTM2NCBDMC40
NDU2MDI3NDMsNS45OTUyNjQyOSAwLjY2MjAzNTc2NCw1LjkwOTA5MDkxIDAuOTE2NjY2NjY3LDUuOTA5
MDkwOTEgTDEuMjIyMjIyMjIsNS45MDkwOTA5MSBMMS4yMjIyMjIyMiw0LjEzNjM2MzY0IEMxLjIyMjIy
MjIyLDMuMDAzNzgyMjIgMS42NDIzNTY5MSwyLjAzMTI1NDA2IDIuNDgyNjM4ODksMS4yMTg3NSBDMy4z
MjI5MjA4NywwLjQwNjI0NTkzNyA0LjMyODY5Nzg1LDAgNS41LDAgQzYuNjcxMzAyMTUsMCA3LjY3NzA3
OTEzLDAuNDA2MjQ1OTM3IDguNTE3MzYxMTEsMS4yMTg3NSBDOS4zNTc2NDMwOSwyLjAzMTI1NDA2IDku
Nzc3Nzc3NzgsMy4wMDM3ODIyMiA5Ljc3Nzc3Nzc4LDQuMTM2MzYzNjQgTDkuNzc3Nzc3NzgsNS45MDkw
OTA5MSBMMTAuMDgzMzMzMyw1LjkwOTA5MDkxIEMxMC4zMzc5NjQyLDUuOTA5MDkwOTEgMTAuNTU0Mzk3
Myw1Ljk5NTI2NDI5IDEwLjczMjYzODksNi4xNjc2MTM2NCBDMTAuOTEwODgwNSw2LjMzOTk2Mjk4IDEx
LDYuNTQ5MjQxMTkgMTEsNi43OTU0NTQ1NSBMMTEsNi43OTU0NTQ1NSBaIiBpZD0i74CjLWNvcHktMiIg
ZmlsbD0iIzdFRDMyMSI+PC9wYXRoPgogICAgICAgICAgICAgICAgPHBhdGggZD0iTTguMzIzMjMxNzEs
OC4wMDQ5NDQ4NyBMNS4yMjA1NjI4MywxMS4wMDUwNDYyIEM1LjE1OTU3NjQxLDExLjA2NDAxNjUgNS4w
ODcxNTYxMSwxMS4wOTM1MDEzIDUuMDAzMjk5NzgsMTEuMDkzNTAxMyBDNC45MTk0NDM0NCwxMS4wOTM1
MDEzIDQuODQ3MDIzMTUsMTEuMDY0MDE2NSA0Ljc4NjAzNjcyLDExLjAwNTA0NjIgTDMuMTQ3MDM0NzQs
OS40MjAyMjYwOCBDMy4wODYwNDgzMSw5LjM2MTI1NTczIDMuMDU1NTU1NTYsOS4yOTEyMjk1IDMuMDU1
NTU1NTYsOS4yMTAxNDUyNyBDMy4wNTU1NTU1Niw5LjEyOTA2MTA1IDMuMDg2MDQ4MzEsOS4wNTkwMzQ4
MiAzLjE0NzAzNDc0LDkuMDAwMDY0NDcgTDMuNTY2MzE0MzEsOC41OTQ2NDUzNyBDMy42MjczMDA3NCw4
LjUzNTY3NTAzIDMuNjk5NzIxMDMsOC41MDYxOTAzIDMuNzgzNTc3MzcsOC41MDYxOTAzIEMzLjg2NzQz
MzcsOC41MDYxOTAzIDMuOTM5ODU0LDguNTM1Njc1MDMgNC4wMDA4NDA0Miw4LjU5NDY0NTM3IEw1LjAw
MzI5OTc4LDkuNTYzOTY1NTggTDcuNDY5NDI2MDIsNy4xNzkzNjQxNyBDNy41MzA0MTI0NSw3LjEyMDM5
MzgyIDcuNjAyODMyNzQsNy4wOTA5MDkwOSA3LjY4NjY4OTA4LDcuMDkwOTA5MDkgQzcuNzcwNTQ1NDEs
Ny4wOTA5MDkwOSA3Ljg0Mjk2NTcsNy4xMjAzOTM4MiA3LjkwMzk1MjEzLDcuMTc5MzY0MTcgTDguMzIz
MjMxNzEsNy41ODQ3ODMyNiBDOC4zODQyMTgxMyw3LjY0Mzc1MzYxIDguNDE0NzEwODksNy43MTM3Nzk4
NCA4LjQxNDcxMDg5LDcuNzk0ODY0MDcgQzguNDE0NzEwODksNy44NzU5NDgyOSA4LjM4NDIxODEzLDcu
OTQ1OTc0NTMgOC4zMjMyMzE3MSw4LjAwNDk0NDg3IEw4LjMyMzIzMTcxLDguMDA0OTQ0ODcgWiIgaWQ9
IlBhdGgiIGZpbGw9IiNGRkZGRkYiPjwvcGF0aD4KICAgICAgICAgICAgPC9nPgogICAgICAgIDwvZz4K
ICAgIDwvZz4KPC9zdmc+
`,
	},

	"/definitions/images/secure.svg": {
		local:   "definitions/images/secure.svg",
		size:    2255,
		modtime: 1489449600,
		compressed: `
PD94bWwgdmVyc2lvbj0iMS4wIiBlbmNvZGluZz0idXRmLTgiPz4KPCEtLSBHZW5lcmF0b3I6IEFkb2Jl
IElsbHVzdHJhdG9yIDIyLjAuMSwgU1ZHIEV4cG9ydCBQbHVnLUluIC4gU1ZHIFZlcnNpb246IDYuMDAg
QnVpbGQgMCkgIC0tPgo8c3ZnIHZlcnNpb249IjEuMSIgaWQ9IkxheWVyXzEiIHhtbG5zPSJodHRwOi8v
d3d3LnczLm9yZy8yMDAwL3N2ZyIgeG1sbnM6eGxpbms9Imh0dHA6Ly93d3cudzMub3JnLzE5OTkveGxp
bmsiIHg9IjBweCIgeT0iMHB4IgoJIHZpZXdCb3g9IjAgMCA2MS4xIDYxIiBzdHlsZT0iZW5hYmxlLWJh
Y2tncm91bmQ6bmV3IDAgMCA2MS4xIDYxOyIgeG1sOnNwYWNlPSJwcmVzZXJ2ZSI+CjxzdHlsZSB0eXBl
PSJ0ZXh0L2NzcyI+Cgkuc3Qwe2ZpbGw6I0ZGRjFGMTt9Cgkuc3Qxe2ZpbGw6I0Y4QzhCQzt9Cgkuc3Qy
e2ZpbGw6I0QzRDNEMztzdHJva2U6IzNDMzczNztzdHJva2UtbGluZWNhcDpyb3VuZDtzdHJva2UtbGlu
ZWpvaW46cm91bmQ7c3Ryb2tlLW1pdGVybGltaXQ6MTA7fQoJLnN0M3tmaWxsOm5vbmU7c3Ryb2tlOiMz
QzM3Mzc7c3Ryb2tlLXdpZHRoOjEuNTtzdHJva2UtbGluZWpvaW46cm91bmQ7c3Ryb2tlLW1pdGVybGlt
aXQ6MTA7fQoJLnN0NHtmaWxsOiNGRkYzRjM7c3Ryb2tlOiMzQzM3Mzc7c3Ryb2tlLWxpbmVjYXA6cm91
bmQ7c3Ryb2tlLWxpbmVqb2luOnJvdW5kO3N0cm9rZS1taXRlcmxpbWl0OjEwO30KCS5zdDV7ZmlsbDpu
b25lO3N0cm9rZTojM0MzNzM3O3N0cm9rZS13aWR0aDoxLjU7c3Ryb2tlLWxpbmVjYXA6cm91bmQ7c3Ry
b2tlLWxpbmVqb2luOnJvdW5kO3N0cm9rZS1taXRlcmxpbWl0OjEwO30KCS5zdDZ7ZmlsbDpub25lO3N0
cm9rZTojM0MzNzM3O3N0cm9rZS13aWR0aDowLjc1O3N0cm9rZS1saW5lY2FwOnJvdW5kO3N0cm9rZS1s
aW5lam9pbjpyb3VuZDtzdHJva2UtbWl0ZXJsaW1pdDoxMDt9Cjwvc3R5bGU+CjxnPgoJPGNpcmNsZSBp
ZD0iWE1MSURfMTE3Nl8iIGNsYXNzPSJzdDAiIGN4PSIyOS44IiBjeT0iMzAuNSIgcj0iMjciLz4KCTxw
b2x5bGluZSBpZD0iUmVjdGFuZ2xlLTQtQ29weS0yIiBjbGFzcz0ic3QxIiBwb2ludHM9IjQzLjgsMTMu
OCA0MC44LDEzLjggNDAuOCwxNi44IDQzLjgsMTYuOCAJIi8+Cgk8cmVjdCBpZD0iUmVjdGFuZ2xlLTQt
Q29weS05IiB4PSI0NS4xIiB5PSIyNy4xIiBjbGFzcz0ic3QxIiB3aWR0aD0iMi43IiBoZWlnaHQ9IjQu
MiIvPgoJPHBvbHlsaW5lIGlkPSJSZWN0YW5nbGUtNC1Db3B5LTE2IiBjbGFzcz0ic3QxIiBwb2ludHM9
IjE0LjgsMjMuOCAxNC44LDIxLjQgMTIuNCwyMS40IDEyLjQsMjMuOCAJIi8+Cgk8cmVjdCBpZD0iUmVj
dGFuZ2xlLTQtQ29weS0xOSIgeD0iMTYuOSIgeT0iMTQuNiIgY2xhc3M9InN0MSIgd2lkdGg9IjIuNyIg
aGVpZ2h0PSIyLjciLz4KCTxyZWN0IHg9IjE2LjgiIHk9IjI0LjIiIGNsYXNzPSJzdDIiIHdpZHRoPSIy
NiIgaGVpZ2h0PSIxOS43Ii8+Cgk8cGF0aCBjbGFzcz0ic3QzIiBkPSJNMzYuOSwyMi44di01LjRjMC0y
LTIuNS0zLjYtNS4zLTMuNkgyOGMtMi45LDAtNS40LDEuNi01LjQsMy42djUuNCIvPgoJPHBvbHlnb24g
Y2xhc3M9InN0NCIgcG9pbnRzPSIzNC42LDMwLjcgMzIuNSwyOC43IDI5LjgsMzEuNCAyNywyOC43IDI1
LDMwLjcgMjcuNywzMy40IDI1LDM2LjEgMjcsMzguMyAyOS44LDM1LjYgMzIuNSwzOC4zIAoJCTM0LjYs
MzYuMSAzMS45LDMzLjQgCSIvPgoJPGxpbmUgaWQ9IlhNTElEXzExODlfIiBjbGFzcz0ic3Q1IiB4MT0i
My42IiB5MT0iNDQuNCIgeDI9IjEwLjgiIHkyPSI0NC40Ii8+Cgk8bGluZSBpZD0iWE1MSURfMTE4OF8i
IGNsYXNzPSJzdDUiIHgxPSI1MiIgeTE9IjQ0LjQiIHgyPSI1OC4yIiB5Mj0iNDQuNCIvPgoJPGxpbmUg
aWQ9IlhNTElEXzExODdfIiBjbGFzcz0ic3Q1IiB4MT0iMTMuNiIgeTE9IjQ0LjQiIHgyPSI1MC41IiB5
Mj0iNDQuNCIvPgoJPGxpbmUgaWQ9IlhNTElEXzExODNfIiBjbGFzcz0ic3Q2IiB4MT0iMTEuOCIgeTE9
IjQ3LjQiIHgyPSIxNy43IiB5Mj0iNDcuNCIvPgoJPGxpbmUgaWQ9IlhNTElEXzExODJfIiBjbGFzcz0i
c3Q2IiB4MT0iMjEuMyIgeTE9IjQ3LjQiIHgyPSIyNy4yIiB5Mj0iNDcuNCIvPgoJPGxpbmUgaWQ9IlhN
TElEXzExODFfIiBjbGFzcz0ic3Q2IiB4MT0iMzAuNiIgeTE9IjQ3LjQiIHgyPSI0Mi44IiB5Mj0iNDcu
NCIvPgoJPGxpbmUgaWQ9IlhNTElEXzExODBfIiBjbGFzcz0ic3Q2IiB4MT0iNDYuNCIgeTE9IjQ3LjQi
IHgyPSI1My4zIiB5Mj0iNDcuNCIvPgo8L2c+Cjwvc3ZnPgo=
`,
	},

	"/definitions/images/smp.svg": {
		local:   "definitions/images/smp.svg",
		size:    5415,
		modtime: 1489449600,
		compressed: `
PD94bWwgdmVyc2lvbj0iMS4wIiBlbmNvZGluZz0iVVRGLTgiPz4KPHN2ZyB3aWR0aD0iODRweCIgaGVp
Z2h0PSI3NXB4IiB2aWV3Qm94PSIwIDAgODQgNzUiIHZlcnNpb249IjEuMSIgeG1sbnM9Imh0dHA6Ly93
d3cudzMub3JnLzIwMDAvc3ZnIiB4bWxuczp4bGluaz0iaHR0cDovL3d3dy53My5vcmcvMTk5OS94bGlu
ayI+CiAgICA8IS0tIEdlbmVyYXRvcjogU2tldGNoIDQzLjEgKDM5MDEyKSAtIGh0dHA6Ly93d3cuYm9o
ZW1pYW5jb2RpbmcuY29tL3NrZXRjaCAtLT4KICAgIDx0aXRsZT5Hcm91cCA1IENvcHk8L3RpdGxlPgog
ICAgPGRlc2M+Q3JlYXRlZCB3aXRoIFNrZXRjaC48L2Rlc2M+CiAgICA8ZGVmcz4KICAgICAgICA8bGlu
ZWFyR3JhZGllbnQgeDE9Ii0xLjExMDIyMzAyZS0xNCUiIHkxPSI1MCUiIHgyPSIxMDAlIiB5Mj0iNTAl
IiBpZD0ibGluZWFyR3JhZGllbnQtMSI+CiAgICAgICAgICAgIDxzdG9wIHN0b3AtY29sb3I9IiNGQkRB
NjEiIG9mZnNldD0iMCUiPjwvc3RvcD4KICAgICAgICAgICAgPHN0b3Agc3RvcC1jb2xvcj0iI0Y3NkIx
QyIgb2Zmc2V0PSIxMDAlIj48L3N0b3A+CiAgICAgICAgPC9saW5lYXJHcmFkaWVudD4KICAgICAgICA8
bGluZWFyR3JhZGllbnQgeDE9IjEwMCUiIHkxPSI1MCUiIHgyPSIyLjMwOTg2OTI2JSIgeTI9IjUwJSIg
aWQ9ImxpbmVhckdyYWRpZW50LTIiPgogICAgICAgICAgICA8c3RvcCBzdG9wLWNvbG9yPSIjNUJFREFD
IiBvZmZzZXQ9IjAlIj48L3N0b3A+CiAgICAgICAgICAgIDxzdG9wIHN0b3AtY29sb3I9IiMyREFCNDki
IG9mZnNldD0iMTAwJSI+PC9zdG9wPgogICAgICAgIDwvbGluZWFyR3JhZGllbnQ+CiAgICA8L2RlZnM+
CiAgICA8ZyBpZD0iU01QLWZsb3dzIiBzdHJva2U9Im5vbmUiIHN0cm9rZS13aWR0aD0iMSIgZmlsbD0i
bm9uZSIgZmlsbC1ydWxlPSJldmVub2RkIj4KICAgICAgICA8ZyBpZD0iR3JvdXAtNS1Db3B5IiB0cmFu
c2Zvcm09InRyYW5zbGF0ZSgyLjAwMDAwMCwgMy4wMDAwMDApIj4KICAgICAgICAgICAgPGcgaWQ9Ikdy
b3VwLTMiPgogICAgICAgICAgICAgICAgPGcgaWQ9Ikdyb3VwIiB0cmFuc2Zvcm09InRyYW5zbGF0ZSgw
LjM4NTk1MywgMC4wMDAwMDApIj4KICAgICAgICAgICAgICAgICAgICA8ZyBpZD0iR3JvdXAtMiIgdHJh
bnNmb3JtPSJ0cmFuc2xhdGUoMjUuMDAwMDAwLCAyNy4wMDAwMDApIiBmaWxsPSIjRjI1NjVCIj4KICAg
ICAgICAgICAgICAgICAgICAgICAgPHBhdGggZD0iTTIxLjgxODQ3ODEsMTEuODEzODUyNiBDMjEuNDY1
ODExNSwxMS44MTM4NTI2IDIxLjE3OTU4OTIsMTEuNDk0MzAyMyAyMS4xNzk1ODkyLDExLjEwMDU3MDYg
TDIxLjE3OTU4OTIsMC43MTMyODIwMzkgQzIxLjE3OTU4OTIsMC4zMTk1NTAzNTMgMjEuNDY1ODExNSww
IDIxLjgxODQ3ODEsMCBDMjIuMTcxMTQ0OCwwIDIyLjQ1NzM2NywwLjMxOTU1MDM1MyAyMi40NTczNjcs
MC43MTMyODIwMzkgTDIyLjQ1NzM2NywxMS4xMDA1NzA2IEMyMi40NTczNjcsMTEuNDk0MzAyMyAyMi4x
NzExNDQ4LDExLjgxMzg1MjYgMjEuODE4NDc4MSwxMS44MTM4NTI2IiBpZD0iRmlsbC0xMSI+PC9wYXRo
PgogICAgICAgICAgICAgICAgICAgICAgICA8cGF0aCBkPSJNMy4wNjczNzc3LDYuMDIxMjcwMjcgQzIu
NzI0NjM0NDEsNS42NTY5ODE1NyAyLjM0NDI0ODc5LDUuNDcyNzE5MjcgMS45MjYyMjA4NCw1LjQ3Mjcx
OTI3IEMxLjQ1MjcxOTk5LDUuNDcyNzE5MjcgMS4wNDY1NzkxLDUuNjU2OTgxNTcgMC43MDM4MzU4MDMs
Ni4wMjEyNzAyNyBDMC4zNjEwOTI1MSw2LjM4NzY3NjkyIDAuMTkyNjkyNjI3LDYuNzkyMjA2OCAwLjE5
MjY5MjYyNyw3LjI0MzMzMTc2IEMwLjE5MjY5MjYyNyw3Ljc0NzQwNTY1IDAuMzYxMDkyNTEsOC4xODM3
MDQ5IDAuNzAzODM1ODAzLDguNTQ3OTkzNTkgQzEuMDQ2NTc5MSw4LjkxMjI4MjI4IDEuNDUyNzE5OTks
OS4wOTQ0MjY2MyAxLjkyNjIyMDg0LDkuMDk0NDI2NjMgQzIuMzQ0MjQ4NzksOS4wOTQ0MjY2MyAyLjcy
NDYzNDQxLDguOTEyMjgyMjggMy4wNjczNzc3LDguNTQ3OTkzNTkgQzMuNDEwMTIxLDguMTgzNzA0OSAz
LjU3ODUyMDg4LDcuNzQ3NDA1NjUgMy41Nzg1MjA4OCw3LjI0MzMzMTc2IEMzLjU3ODUyMDg4LDYuNzky
MjA2OCAzLjQxMDEyMSw2LjM4NzY3NjkyIDMuMDY3Mzc3Nyw2LjAyMTI3MDI3IEwzLjA2NzM3NzcsNi4w
MjEyNzAyNyBaIiBpZD0iUGF0aCIgc3Ryb2tlPSIjRjI1NjVCIiBzdHJva2Utd2lkdGg9IjEuMyI+PC9w
YXRoPgogICAgICAgICAgICAgICAgICAgICAgICA8cGF0aCBkPSJNOS44MDMzNzMwNiw2LjAyMTI3MDI3
IEM5LjQ4ODM2NjIxLDUuNjU2OTgxNTcgOS4wOTQxMTIzNyw1LjQ3MjcxOTI3IDguNjIwNjExNTIsNS40
NzI3MTkyNyBDOC4xNDkwOTE4NCw1LjQ3MjcxOTI3IDcuNzU0ODM4LDUuNjU2OTgxNTcgNy40Mzk4MzEx
Niw2LjAyMTI3MDI3IEM3LjEyNDgyNDMxLDYuMzg3Njc2OTIgNi45NjYzMzAzMSw2Ljc5MjIwNjggNi45
NjYzMzAzMSw3LjI0MzMzMTc2IEM2Ljk2NjMzMDMxLDcuNzQ3NDA1NjUgNy4xMjQ4MjQzMSw4LjE4Mzcw
NDkgNy40Mzk4MzExNiw4LjU0Nzk5MzU5IEM3Ljc1NDgzOCw4LjkxMjI4MjI4IDguMTQ5MDkxODQsOS4w
OTQ0MjY2MyA4LjYyMDYxMTUyLDkuMDk0NDI2NjMgQzkuMDk0MTEyMzcsOS4wOTQ0MjY2MyA5LjQ4ODM2
NjIxLDguOTEyMjgyMjggOS44MDMzNzMwNiw4LjU0Nzk5MzU5IEMxMC4xMTgzNzk5LDguMTgzNzA0OSAx
MC4yNzY4NzM5LDcuNzQ3NDA1NjUgMTAuMjc2ODczOSw3LjI0MzMzMTc2IEMxMC4yNzY4NzM5LDYuNzky
MjA2OCAxMC4xMTgzNzk5LDYuMzg3Njc2OTIgOS44MDMzNzMwNiw2LjAyMTI3MDI3IEw5LjgwMzM3MzA2
LDYuMDIxMjcwMjcgWiIgaWQ9IlBhdGgiIHN0cm9rZT0iI0YyNTY1QiIgc3Ryb2tlLXdpZHRoPSIxLjMi
PjwvcGF0aD4KICAgICAgICAgICAgICAgICAgICAgICAgPHBhdGggZD0iTTE2LjUzOTM2ODQsNi4wMjEy
NzAyNyBDMTYuMTk2NjI1MSw1LjY1Njk4MTU3IDE1Ljc5MDQ4NDIsNS40NzI3MTkyNyAxNS4zMTg5NjQ1
LDUuNDcyNzE5MjcgQzE0Ljg5Njk3NDIsNS40NzI3MTkyNyAxNC41MTY1ODg2LDUuNjU2OTgxNTcgMTQu
MTc1ODI2NSw2LjAyMTI3MDI3IEMxMy44MzMwODMyLDYuMzg3Njc2OTIgMTMuNjYyNzAyMiw2Ljc5MjIw
NjggMTMuNjYyNzAyMiw3LjI0MzMzMTc2IEMxMy42NjI3MDIyLDcuNzQ3NDA1NjUgMTMuODMzMDgzMiw4
LjE4MzcwNDkgMTQuMTc1ODI2NSw4LjU0Nzk5MzU5IEMxNC41MTY1ODg2LDguOTEyMjgyMjggMTQuODk2
OTc0Miw5LjA5NDQyNjYzIDE1LjMxODk2NDUsOS4wOTQ0MjY2MyBDMTUuNzkwNDg0Miw5LjA5NDQyNjYz
IDE2LjE5NjYyNTEsOC45MTIyODIyOCAxNi41MzkzNjg0LDguNTQ3OTkzNTkgQzE2Ljg4MDEzMDUsOC4x
ODM3MDQ5IDE3LjA1MDUxMTYsNy43NDc0MDU2NSAxNy4wNTA1MTE2LDcuMjQzMzMxNzYgQzE3LjA1MDUx
MTYsNi43OTIyMDY4IDE2Ljg4MDEzMDUsNi4zODc2NzY5MiAxNi41MzkzNjg0LDYuMDIxMjcwMjcgTDE2
LjUzOTM2ODQsNi4wMjEyNzAyNyBaIiBpZD0iUGF0aCIgc3Ryb2tlPSIjRjI1NjVCIiBzdHJva2Utd2lk
dGg9IjEuMyI+PC9wYXRoPgogICAgICAgICAgICAgICAgICAgIDwvZz4KICAgICAgICAgICAgICAgICAg
ICA8cGF0aCBkPSJNMzUuMTQ5Mjg1LDY5LjMzMjM1NTEgQzE1LjczNzExNiw2OS4zMzIzNTUxIDAsNTMu
ODExNTM3MSAwLDM0LjY2NjE3NzUgQzAsMTUuNTIyNzQzOSAxNS43MzcxMTYsMCAzNS4xNDkyODUsMCBD
NTQuNTYxNDU0MSwwIDcwLjI5ODU3MDEsMTUuNTIyNzQzOSA3MC4yOTg1NzAxLDM0LjY2NjE3NzUgQzcw
LjI5ODU3MDEsNTMuODExNTM3MSA1NC41NjE0NTQxLDY5LjMzMjM1NTEgMzUuMTQ5Mjg1LDY5LjMzMjM1
NTEgWiIgaWQ9IlN0cm9rZS03LUNvcHktOCIgc3Ryb2tlPSJ1cmwoI2xpbmVhckdyYWRpZW50LTEpIiBz
dHJva2Utd2lkdGg9IjQuNTU5OTk5OTQiPjwvcGF0aD4KICAgICAgICAgICAgICAgIDwvZz4KICAgICAg
ICAgICAgICAgIDxwYXRoIGQ9Ik02Mi40NDQ3NDE3LDY1LjY1Nzc4ODQgQzYyLjA3NDk5NjIsNjYuMDI4
NzYyOCA2MS42MzU5Mjk5LDY2LjIxNDI0NzMgNjEuMTI3NTI5OCw2Ni4yMTQyNDczIEM2MC42MTkxMjk3
LDY2LjIxNDI0NzMgNjAuMTgwMDYzNSw2Ni4wMjg3NjI4IDU5LjgxMDMxNzksNjUuNjU3Nzg4NCBMNDku
ODczNDU2MSw1NS42ODc5IEM0OS41MDM3MTA2LDU1LjMxNjkyNTYgNDkuMzE4ODQwNiw1NC44NzY0IDQ5
LjMxODg0MDYsNTQuMzY2MzEwMiBDNDkuMzE4ODQwNiw1My44NTYyMjAzIDQ5LjUwMzcxMDYsNTMuNDE1
Njk0OCA0OS44NzM0NTYxLDUzLjA0NDcyMDMgTDUyLjQxNTQ0NCw1MC40OTQyODM4IEM1Mi43ODUxODk2
LDUwLjEyMzMwOTMgNTMuMjI0MjU1OCw0OS45Mzc4MjQ5IDUzLjczMjY1NTksNDkuOTM3ODI0OSBDNTQu
MjQxMDU2MSw0OS45Mzc4MjQ5IDU0LjY4MDEyMjMsNTAuMTIzMzA5MyA1NS4wNDk4Njc4LDUwLjQ5NDI4
MzggTDYxLjEyNzUyOTgsNTYuNTkyMTQ1NyBMNzYuMDc5MDQwNSw0MS41OTA5NDE2IEM3Ni40NDg3ODYs
NDEuMjE5OTY3MiA3Ni44ODc4NTIzLDQxLjAzNDQ4MjggNzcuMzk2MjUyNCw0MS4wMzQ0ODI4IEM3Ny45
MDQ2NTI1LDQxLjAzNDQ4MjggNzguMzQzNzE4OCw0MS4yMTk5NjcyIDc4LjcxMzQ2NDMsNDEuNTkwOTQx
NiBMODEuMjU1NDUyMiw0NC4xNDEzNzgyIEM4MS42MjUxOTc3LDQ0LjUxMjM1MjYgODEuODEwMDY3Nyw0
NC45NTI4NzgyIDgxLjgxMDA2NzcsNDUuNDYyOTY4IEM4MS44MTAwNjc3LDQ1Ljk3MzA1NzkgODEuNjI1
MTk3Nyw0Ni40MTM1ODM0IDgxLjI1NTQ1MjIsNDYuNzg0NTU3OSBMNjIuNDQ0NzQxNyw2NS42NTc3ODg0
IFoiIGlkPSJQYXRoLUNvcHktMiIgZmlsbD0idXJsKCNsaW5lYXJHcmFkaWVudC0yKSI+PC9wYXRoPgog
ICAgICAgICAgICA8L2c+CiAgICAgICAgICAgIDxnIGlkPSJHcm91cC00IiB0cmFuc2Zvcm09InRyYW5z
bGF0ZSgxNC4wMDAwMDAsIDE3LjAwMDAwMCkiIGZpbGw9IiNGMkE2NDEiPgogICAgICAgICAgICAgICAg
PGcgaWQ9Ikdyb3VwLTMiPgogICAgICAgICAgICAgICAgICAgIDxwYXRoIGQ9Ik00NCwzMy4wOTA5MDkx
IEwyNC43NSwzMy4wOTA5MDkxIEwyNC43NSwzNS44NDA5MDkxIEwzMywzNS44NDA5MDkxIEwzMywzOC41
OTA5MDkxIEwxMSwzOC41OTA5MDkxIEwxMSwzNS44NDA5MDkxIEwxOS4yNSwzNS44NDA5MDkxIEwxOS4y
NSwzMy4wOTA5MDkxIEwwLDMzLjA5MDkwOTEgTDAsMC4wOTA5MDkwOTA5IEw0NCwwLjA5MDkwOTA5MDkg
TDQ0LDMzLjA5MDkwOTEgWiBNMi43NSwzMC4zNDA5MDkxIEw0MS4yNSwzMC4zNDA5MDkxIEw0MS4yNSwy
Ljg0MDkwOTA5IEwyLjc1LDIuODQwOTA5MDkgTDIuNzUsMzAuMzQwOTA5MSBaIiBpZD0iWi1Db3B5LTIi
PjwvcGF0aD4KICAgICAgICAgICAgICAgIDwvZz4KICAgICAgICAgICAgPC9nPgogICAgICAgIDwvZz4K
ICAgIDwvZz4KPC9zdmc+
`,
	},

	"/definitions/images/smpsuccess.svg": {
		local:   "definitions/images/smpsuccess.svg",
		size:    15608,
		modtime: 1489449600,
		compressed: `
PD94bWwgdmVyc2lvbj0iMS4wIiBlbmNvZGluZz0iVVRGLTgiIHN0YW5kYWxvbmU9Im5vIj8+Cjxzdmcg
d2lkdGg9IjIzNHB4IiBoZWlnaHQ9IjU4cHgiIHZpZXdCb3g9IjAgMCAyMzQgNTgiIHZlcnNpb249IjEu
MSIgeG1sbnM9Imh0dHA6Ly93d3cudzMub3JnLzIwMDAvc3ZnIiB4bWxuczp4bGluaz0iaHR0cDovL3d3
dy53My5vcmcvMTk5OS94bGluayI+CiAgICA8IS0tIEdlbmVyYXRvcjogU2tldGNoIDQxLjIgKDM1Mzk3
KSAtIGh0dHA6Ly93d3cuYm9oZW1pYW5jb2RpbmcuY29tL3NrZXRjaCAtLT4KICAgIDx0aXRsZT5Hcm91
cCAxOTwvdGl0bGU+CiAgICA8ZGVzYz5DcmVhdGVkIHdpdGggU2tldGNoLjwvZGVzYz4KICAgIDxkZWZz
PgogICAgICAgIDxsaW5lYXJHcmFkaWVudCB4MT0iMTAwJSIgeTE9IjUwJSIgeDI9IjMuMDYxNjE3ZS0x
NSUiIHkyPSI1MCUiIGlkPSJsaW5lYXJHcmFkaWVudC0xIj4KICAgICAgICAgICAgPHN0b3Agc3RvcC1j
b2xvcj0iI0ZCREE2MSIgb2Zmc2V0PSIwJSI+PC9zdG9wPgogICAgICAgICAgICA8c3RvcCBzdG9wLWNv
bG9yPSIjRjc2QjFDIiBvZmZzZXQ9IjEwMCUiPjwvc3RvcD4KICAgICAgICA8L2xpbmVhckdyYWRpZW50
PgogICAgICAgIDxsaW5lYXJHcmFkaWVudCB4MT0iLTEuMTEwMjIzMDJlLTE0JSIgeTE9IjUwJSIgeDI9
IjEwMCUiIHkyPSI1MCUiIGlkPSJsaW5lYXJHcmFkaWVudC0yIj4KICAgICAgICAgICAgPHN0b3Agc3Rv
cC1jb2xvcj0iI0ZCREE2MSIgb2Zmc2V0PSIwJSI+PC9zdG9wPgogICAgICAgICAgICA8c3RvcCBzdG9w
LWNvbG9yPSIjRjc2QjFDIiBvZmZzZXQ9IjEwMCUiPjwvc3RvcD4KICAgICAgICA8L2xpbmVhckdyYWRp
ZW50PgogICAgPC9kZWZzPgogICAgPGcgaWQ9IlNNUC1mbG93cyIgc3Ryb2tlPSJub25lIiBzdHJva2Ut
d2lkdGg9IjEiIGZpbGw9Im5vbmUiIGZpbGwtcnVsZT0iZXZlbm9kZCI+CiAgICAgICAgPGcgaWQ9IkJv
Yl80IiB0cmFuc2Zvcm09InRyYW5zbGF0ZSgtNjE1LjAwMDAwMCwgLTI1Ni4wMDAwMDApIj4KICAgICAg
ICAgICAgPGcgaWQ9Ikdyb3VwLTI4LUNvcHkiIHRyYW5zZm9ybT0idHJhbnNsYXRlKDU2MC4wMDAwMDAs
IDIwMC4wMDAwMDApIj4KICAgICAgICAgICAgICAgIDxnIGlkPSJHcm91cC0yNiIgdHJhbnNmb3JtPSJ0
cmFuc2xhdGUoNDIuMDAwMDAwLCA1Ni4wMDAwMDApIj4KICAgICAgICAgICAgICAgICAgICA8ZyBpZD0i
R3JvdXAtMTkiIHRyYW5zZm9ybT0idHJhbnNsYXRlKDE0LjAwMDAwMCwgMC4wMDAwMDApIj4KICAgICAg
ICAgICAgICAgICAgICAgICAgPGcgaWQ9Ikdyb3VwIiB0cmFuc2Zvcm09InRyYW5zbGF0ZSgxMDcuOTA2
OTc3LCAwLjAwMDAwMCkiPgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHBhdGggZD0iTTUuNjIw
MTU1MDQsMTEuMTAzODk2MSBMMTQuNjEyNDAzMSwxMS4xMDM4OTYxIEwxNC42MTI0MDMxLDcuNzcyNzI3
MjcgQzE0LjYxMjQwMzEsNi41NDY2NjU5NSAxNC4xNzMzMzI5LDUuNDk5OTAyODggMTMuMjk1MTc5Myw0
LjYzMjQwNjY2IEMxMi40MTcwMjU2LDMuNzY0OTEwNDQgMTEuMzU3NDAyOCwzLjMzMTE2ODgzIDEwLjEx
NjI3OTEsMy4zMzExNjg4MyBDOC44NzUxNTUyOSwzLjMzMTE2ODgzIDcuODE1NTMyNDksMy43NjQ5MTA0
NCA2LjkzNzM3ODg4LDQuNjMyNDA2NjYgQzYuMDU5MjI1MjYsNS40OTk5MDI4OCA1LjYyMDE1NTA0LDYu
NTQ2NjY1OTUgNS42MjAxNTUwNCw3Ljc3MjcyNzI3IEw1LjYyMDE1NTA0LDExLjEwMzg5NjEgWiBNMjAu
MjMyNTU4MSwxMi43Njk0ODA1IEwyMC4yMzI1NTgxLDIyLjc2Mjk4NyBDMjAuMjMyNTU4MSwyMy4yMjU2
NTE3IDIwLjA2ODYzODYsMjMuNjE4OTEwNyAxOS43NDA3OTQ2LDIzLjk0Mjc3NiBDMTkuNDEyOTUwNiwy
NC4yNjY2NDEyIDE5LjAxNDg2MDIsMjQuNDI4NTcxNCAxOC41NDY1MTE2LDI0LjQyODU3MTQgTDEuNjg2
MDQ2NTEsMjQuNDI4NTcxNCBDMS4yMTc2OTc5MiwyNC40Mjg1NzE0IDAuODE5NjA3NTgyLDI0LjI2NjY0
MTIgMC40OTE3NjM1NjYsMjMuOTQyNzc2IEMwLjE2MzkxOTU0OSwyMy42MTg5MTA3IDAsMjMuMjI1NjUx
NyAwLDIyLjc2Mjk4NyBMMCwxMi43Njk0ODA1IEMwLDEyLjMwNjgxNTkgMC4xNjM5MTk1NDksMTEuOTEz
NTU2OCAwLjQ5MTc2MzU2NiwxMS41ODk2OTE2IEMwLjgxOTYwNzU4MiwxMS4yNjU4MjYzIDEuMjE3Njk3
OTIsMTEuMTAzODk2MSAxLjY4NjA0NjUxLDExLjEwMzg5NjEgTDIuMjQ4MDYyMDIsMTEuMTAzODk2MSBM
Mi4yNDgwNjIwMiw3Ljc3MjcyNzI3IEMyLjI0ODA2MjAyLDUuNjQ0NDY5ODggMy4wMjA4MjU2MSwzLjgx
Njk3MTkyIDQuNTY2Mzc1OTcsMi4yOTAxNzg1NyBDNi4xMTE5MjYzMywwLjc2MzM4NTIyMyA3Ljk2MTg3
NTUzLDAgMTAuMTE2Mjc5MSwwIEMxMi4yNzA2ODI2LDAgMTQuMTIwNjMxOCwwLjc2MzM4NTIyMyAxNS42
NjYxODIyLDIuMjkwMTc4NTcgQzE3LjIxMTczMjUsMy44MTY5NzE5MiAxNy45ODQ0OTYxLDUuNjQ0NDY5
ODggMTcuOTg0NDk2MSw3Ljc3MjcyNzI3IEwxNy45ODQ0OTYxLDExLjEwMzg5NjEgTDE4LjU0NjUxMTYs
MTEuMTAzODk2MSBDMTkuMDE0ODYwMiwxMS4xMDM4OTYxIDE5LjQxMjk1MDYsMTEuMjY1ODI2MyAxOS43
NDA3OTQ2LDExLjU4OTY5MTYgQzIwLjA2ODYzODYsMTEuOTEzNTU2OCAyMC4yMzI1NTgxLDEyLjMwNjgx
NTkgMjAuMjMyNTU4MSwxMi43Njk0ODA1IEwyMC4yMzI1NTgxLDEyLjc2OTQ4MDUgWiIgaWQ9Iu+Aoy1j
b3B5LTIiIGZpbGw9IiM3RUQzMjEiPjwvcGF0aD4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxw
YXRoIGQ9Ik0xNS4zMDkxMTU0LDE1LjA0MjI1OSBMOS42MDIzMDM3MywyMC42Nzk4MTIxIEM5LjQ5MDEy
OTk2LDIwLjc5MDYyNDUgOS4zNTY5MjU2MiwyMC44NDYwMjk4IDkuMjAyNjg2NjksMjAuODQ2MDI5OCBD
OS4wNDg0NDc3NywyMC44NDYwMjk4IDguOTE1MjQzNDIsMjAuNzkwNjI0NSA4LjgwMzA2OTY2LDIwLjY3
OTgxMjEgTDUuNzg4NDE0ODQsMTcuNzAxNzQzNSBDNS42NzYyNDEwOCwxNy41OTA5MzExIDUuNjIwMTU1
MDQsMTcuNDU5MzQzMyA1LjYyMDE1NTA0LDE3LjMwNjk3NjMgQzUuNjIwMTU1MDQsMTcuMTU0NjA5MiA1
LjY3NjI0MTA4LDE3LjAyMzAyMTUgNS43ODg0MTQ4NCwxNi45MTIyMDkxIEw2LjU1OTYwNTYxLDE2LjE1
MDM3NzYgQzYuNjcxNzc5MzcsMTYuMDM5NTY1MiA2LjgwNDk4MzcyLDE1Ljk4NDE1OTggNi45NTkyMjI2
NCwxNS45ODQxNTk4IEM3LjExMzQ2MTU3LDE1Ljk4NDE1OTggNy4yNDY2NjU5MSwxNi4wMzk1NjUyIDcu
MzU4ODM5NjgsMTYuMTUwMzc3NiBMOS4yMDI2ODY2OSwxNy45NzE4NDc0IEwxMy43Mzg2OTA2LDEzLjQ5
MDg5MzEgQzEzLjg1MDg2NDMsMTMuMzgwMDgwNyAxMy45ODQwNjg3LDEzLjMyNDY3NTMgMTQuMTM4MzA3
NiwxMy4zMjQ2NzUzIEMxNC4yOTI1NDY1LDEzLjMyNDY3NTMgMTQuNDI1NzUwOSwxMy4zODAwODA3IDE0
LjUzNzkyNDYsMTMuNDkwODkzMSBMMTUuMzA5MTE1NCwxNC4yNTI3MjQ2IEMxNS40MjEyODkyLDE0LjM2
MzUzNyAxNS40NzczNzUyLDE0LjQ5NTEyNDggMTUuNDc3Mzc1MiwxNC42NDc0OTE4IEMxNS40NzczNzUy
LDE0Ljc5OTg1ODkgMTUuNDIxMjg5MiwxNC45MzE0NDY2IDE1LjMwOTExNTQsMTUuMDQyMjU5IEwxNS4z
MDkxMTU0LDE1LjA0MjI1OSBaIiBpZD0iUGF0aCIgZmlsbD0iI0ZGRkZGRiI+PC9wYXRoPgogICAgICAg
ICAgICAgICAgICAgICAgICA8L2c+CiAgICAgICAgICAgICAgICAgICAgICAgIDxwYXRoIGQ9Ik0xNzUu
NDcyNzMsMjkuNDU5NTk2IEw1Ni44NDg2Mzk5LDI5LjQ1OTU5NiIgaWQ9IlN0cm9rZS0xMy1Db3B5IiBz
dHJva2U9InVybCgjbGluZWFyR3JhZGllbnQtMSkiIHN0cm9rZS13aWR0aD0iMS41NzQ2NCIgc3Ryb2tl
LWxpbmVjYXA9InJvdW5kIiB0cmFuc2Zvcm09InRyYW5zbGF0ZSgxMTYuMTU2NTE3LCAyOS4yMTExMjkp
IHNjYWxlKC0xLCAxKSB0cmFuc2xhdGUoLTExNi4xNTY1MTcsIC0yOS4yMTExMjkpICI+PC9wYXRoPgog
ICAgICAgICAgICAgICAgICAgICAgICA8cmVjdCBpZD0iUmVjdGFuZ2xlLTMtQ29weS05IiBmaWxsPSIj
RTlFOUVBIiB4PSIxNi4xODYwNDY1IiB5PSIxNi4yODU3MTQzIiB3aWR0aD0iMS4zNDg4MzcyMSIgaGVp
Z2h0PSIyMS43MTQyODU3IiByeD0iMC42NzQ0MTg2MDUiPjwvcmVjdD4KICAgICAgICAgICAgICAgICAg
ICAgICAgPGcgaWQ9Ikdyb3VwLTE2IiB0cmFuc2Zvcm09InRyYW5zbGF0ZSgxNzUuMzQ4ODM3LCAxLjM1
NzE0MykiPgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHBhdGggZD0iTTI4LjUzOTM3NzksNTUu
NjQyODU3MSBDMTMuMDEzODYzOCw1NS42NDI4NTcxIDAuNDI3NTkzMDI5LDQzLjE4NjU4NTMgMC40Mjc1
OTMwMjksMjcuODIxNDI4NiBDMC40Mjc1OTMwMjksMTIuNDU3ODE3NSAxMy4wMTM4NjM4LDAgMjguNTM5
Mzc3OSwwIEM0NC4wNjQ4OTIsMCA1Ni42NTExNjI4LDEyLjQ1NzgxNzUgNTYuNjUxMTYyOCwyNy44MjE0
Mjg2IEM1Ni42NTExNjI4LDQzLjE4NjU4NTMgNDQuMDY0ODkyLDU1LjY0Mjg1NzEgMjguNTM5Mzc3OSw1
NS42NDI4NTcxIFoiIGlkPSJTdHJva2UtNyIgc3Ryb2tlPSJ1cmwoI2xpbmVhckdyYWRpZW50LTIpIiBz
dHJva2Utd2lkdGg9IjEuMTMzNzQwOCIgZmlsbD0iI0U4RThFNyI+PC9wYXRoPgogICAgICAgICAgICAg
ICAgICAgICAgICAgICAgPHBhdGggZD0iTTI3LjQxNDkwNjUsNTMuOTU2NzEgQzEyLjgyMDkyMzMsNTMu
OTU2NzEgMC45ODk4Mjg3MjcsNDIuMjU1MzYzNyAwLjk4OTgyODcyNywyNy44MjE0Mjg2IEMwLjk4OTgy
ODcyNywxMy4zODg5NDU0IDEyLjgyMDkyMzMsMS42ODYxNDcxOSAyNy40MTQ5MDY1LDEuNjg2MTQ3MTkg
QzQyLjAwODg4OTgsMS42ODYxNDcxOSA1My44Mzk5ODQzLDEzLjM4ODk0NTQgNTMuODM5OTg0MywyNy44
MjE0Mjg2IEM1My44Mzk5ODQzLDQyLjI1NTM2MzcgNDIuMDA4ODg5OCw1My45NTY3MSAyNy40MTQ5MDY1
LDUzLjk1NjcxIFoiIGlkPSJTdHJva2UtNy1Db3B5IiBmaWxsPSIjRkZGRkZGIj48L3BhdGg+CiAgICAg
ICAgICAgICAgICAgICAgICAgICAgICA8cmVjdCBpZD0iUmVjdGFuZ2xlLTMtQ29weS03IiBmaWxsPSIj
RTlFOUVBIiB0cmFuc2Zvcm09InRyYW5zbGF0ZSg0MC4zNDYzMjgsIDM3LjkzODMxMikgcm90YXRlKC05
MC4wMDAwMDApIHRyYW5zbGF0ZSgtNDAuMzQ2MzI4LCAtMzcuOTM4MzEyKSAiIHg9IjM5LjUwMjk3NCIg
eT0iMzEuMTkzNzIyOSIgd2lkdGg9IjEuNjg2NzA3MDkiIGhlaWdodD0iMTMuNDg5MTc3NSIgcng9IjAu
ODQzMzUzNTQ2Ij48L3JlY3Q+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICA8ZyBpZD0iR3JvdXAt
MTgiIHRyYW5zZm9ybT0idHJhbnNsYXRlKDkuNjAxODkyLCAxNi4zMDkxMTMpIj4KICAgICAgICAgICAg
ICAgICAgICAgICAgICAgICAgICA8cmVjdCBpZD0iUmVjdGFuZ2xlLTMiIGZpbGw9IiNFOUU5RUEiIHg9
IjMzLjU1NTYxNCIgeT0iMC41NTIzNTg1NjEiIHdpZHRoPSIxLjY4NjcwNzA5IiBoZWlnaHQ9IjE5LjEw
OTY2ODEiIHJ4PSIwLjg0MzM1MzU0NiI+PC9yZWN0PgogICAgICAgICAgICAgICAgICAgICAgICAgICAg
ICAgIDxyZWN0IGlkPSJSZWN0YW5nbGUtMy1Db3B5IiBmaWxsPSIjRTlFOUVBIiB0cmFuc2Zvcm09InRy
YW5zbGF0ZSgyOS42MTk5NjQsIDIwLjUwNTEwMCkgcm90YXRlKC05MC4wMDAwMDApIHRyYW5zbGF0ZSgt
MjkuNjE5OTY0LCAtMjAuNTA1MTAwKSAiIHg9IjI4Ljc3NjYxMDYiIHk9IjEwLjk1MDI2NjIiIHdpZHRo
PSIxLjY4NjcwNzA5IiBoZWlnaHQ9IjE5LjEwOTY2ODEiIHJ4PSIwLjg0MzM1MzU0NiI+PC9yZWN0Pgog
ICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxyZWN0IGlkPSJSZWN0YW5nbGUtMy1Db3B5LTYi
IGZpbGw9IiNFOUU5RUEiIHRyYW5zZm9ybT0idHJhbnNsYXRlKDI1LjEyMjA3OSwgMTguODE4OTUzKSBy
b3RhdGUoLTkwLjAwMDAwMCkgdHJhbnNsYXRlKC0yNS4xMjIwNzksIC0xOC44MTg5NTMpICIgeD0iMjQu
Mjc4NzI1IiB5PSI5LjI2NDExOTAyIiB3aWR0aD0iMS42ODY3MDcwOSIgaGVpZ2h0PSIxOS4xMDk2Njgx
IiByeD0iMC44NDMzNTM1NDYiPjwvcmVjdD4KICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgICA8
cmVjdCBpZD0iUmVjdGFuZ2xlLTMtQ29weS04IiBmaWxsPSIjRTlFOUVBIiB0cmFuc2Zvcm09InRyYW5z
bGF0ZSgzMC4xODIyMDAsIDEuOTU3NDgxKSByb3RhdGUoLTkwLjAwMDAwMCkgdHJhbnNsYXRlKC0zMC4x
ODIyMDAsIC0xLjk1NzQ4MSkgIiB4PSIyOS4zMzg4NDYzIiB5PSItMi41Mzg5MTEyOCIgd2lkdGg9IjEu
Njg2NzA3MDkiIGhlaWdodD0iOC45OTI3ODQ5OSIgcng9IjAuODQzMzUzNTQ2Ij48L3JlY3Q+CiAgICAg
ICAgICAgICAgICAgICAgICAgICAgICAgICAgPHBhdGggZD0iTTE5Ljg1NzA1OTEsMS41NDA5NTM2NyBD
MTkuNzA0NjE5OCwxLjU0MDk1MzY3IDE5LjYyOTk1NTYsMS42NDA3NDg3NSAxOS42Mjk5NTU2LDEuODQz
NDU3NTMgQzE5LjYyOTk1NTYsMS45OTYyNjg3NiAxOS43MDQ2MTk4LDIuMDcyNjc0MzggMTkuODU3MDU5
MSwyLjA3MjY3NDM4IEMyMC4wNTc3MTkxLDIuMDcyNjc0MzggMjAuMTYwMzgyNCwxLjk5NjI2ODc2IDIw
LjE2MDM4MjQsMS44NDM0NTc1MyBDMjAuMTYwMzgyNCwxLjY0MDc0ODc1IDIwLjA1NzcxOTEsMS41NDA5
NTM2NyAxOS44NTcwNTkxLDEuNTQwOTUzNjcgTDE5Ljg1NzA1OTEsMS41NDA5NTM2NyBaIE03LjA1MjE1
MjQsMTcuNTY1ODYyIEwzMi44ODkwNjk0LDE3LjU2NTg2MiBMMzIuODg5MDY5NCwzLjEzNDU1NjUgTDMw
LjU0MDI1ODgsMy4xMzQ1NTY1IEwyOS43ODI3Mjg1LDMuMTM0NTU2NSBMMjguNzIxODc0OSwzLjEzNDU1
NjUgTDI3Ljk2NDM0NDYsMy4xMzQ1NTY1IEwyNi42NzYzODc1LDMuMTM0NTU2NSBMMjUuOTE3MzAxNywz
LjEzNDU1NjUgTDcuMDUyMTUyNCwzLjEzNDU1NjUgTDcuMDUyMTUyNCwxNy41NjU4NjIgWiBNNi40NDU1
MDU5NCwxOC4xNzM5ODgzIEwzMy40OTQxNjA0LDE4LjE3Mzk4ODMgTDMzLjQ5NDE2MDQsMi41Mjc5ODk0
NyBMNi40NDU1MDU5NCwyLjUyNzk4OTQ3IEw2LjQ0NTUwNTk0LDE4LjE3Mzk4ODMgWiBNNS4xNTc1NDg4
NSwxOS4zODg2ODE3IEwzNC44NTk4OTI3LDE5LjM4ODY4MTcgTDM0Ljg1OTg5MjcsMS45OTYyNjg3NiBD
MzQuODU5ODkyNywxLjQzOTU5OTI4IDM0LjU4MTQ1NzUsMS4xNjA0ODQ4OSAzNC4wMjQ1ODcyLDEuMTYw
NDg0ODkgTDUuOTkxMjk4ODUsMS4xNjA0ODQ4OSBDNS40MzQ0Mjg1MiwxLjE2MDQ4NDg5IDUuMTU3NTQ4
ODUsMS40Mzk1OTkyOCA1LjE1NzU0ODg1LDEuOTk2MjY4NzYgTDUuMTU3NTQ4ODUsMTkuMzg4NjgxNyBa
IE0yLjk2MTE3NzU4LDIyLjQyNzc1NCBMMzYuNjAyMDU2OCwyMi40Mjc3NTQgTDM3LjQzNTgwNjgsMjEu
ODE5NjI3NyBMMi4wNTEyMDc4OSwyMS44MTk2Mjc3IEwyLjk2MTE3NzU4LDIyLjQyNzc1NCBaIE0zOC42
NDc1NDQzLDIwLjIyNDQ2NTYgQzM4LjY0NzU0NDMsMjAuMDIzMzE2MSAzOC41NDQ4ODEsMTkuOTIwNDAy
NCAzOC4zNDQyMjEsMTkuOTIwNDAyNCBMMS4yMTc0NTc4OSwxOS45MjA0MDI0IEMxLjAxNTI0MjQxLDE5
LjkyMDQwMjQgMC45MTQxMzQ2NjQsMjAuMDIzMzE2MSAwLjkxNDEzNDY2NCwyMC4yMjQ0NjU2IEwwLjkx
NDEzNDY2NCwyMC45ODU0MDMxIEMwLjkxNDEzNDY2NCwyMS4wMzUzMDA3IDAuOTkwMzU0MzQ3LDIxLjEz
NTA5NTcgMS4xNDI3OTM3MSwyMS4yODc5MDcgTDEuMTQyNzkzNzEsMjEuMjExNTAxNCBMMS4yMTc0NTc4
OSwyMS4yODc5MDcgTDM4LjM0NDIyMSwyMS4yODc5MDcgQzM4LjU0NDg4MSwyMS4yODc5MDcgMzguNjQ3
NTQ0MywyMS4xODgxMTE5IDM4LjY0NzU0NDMsMjAuOTg1NDAzMSBMMzguNjQ3NTQ0MywyMC4yMjQ0NjU2
IFogTTM5LjE3Nzk3MSwyMC4yMjQ0NjU2IEwzOS4xNzc5NzEsMjAuOTg1NDAzMSBDMzkuMTc3OTcxLDIx
LjQ0MDcxODIgMzguOTUwODY3NSwyMS43MTgyNzMzIDM4LjQ5NTEwNDksMjEuODE5NjI3NyBMMzYuODI5
MTYwNCwyMi45NTc5MTU0IEwzNi44MjkxNjA0LDIzLjAzNDMyMSBMMi42NTc4NTQzNSwyMy4wMzQzMjEg
TDIuNjU3ODU0MzUsMjIuOTU3OTE1NCBMMC45OTAzNTQzNDcsMjEuODE5NjI3NyBDMC41ODQzNjc4NzIs
MjEuNjY4Mzc1OCAwLjM4MzcwNzg5LDIxLjM4OTI2MTQgMC4zODM3MDc4OSwyMC45ODU0MDMxIEwwLjM4
MzcwNzg5LDIwLjIyNDQ2NTYgQzAuMzgzNzA3ODksMTkuNjY3Nzk2MSAwLjY2MDU4NzU1NSwxOS4zODg2
ODE3IDEuMjE3NDU3ODksMTkuMzg4NjgxNyBMNC41NTI0NTc5LDE5LjM4ODY4MTcgTDQuNTUyNDU3OSwx
Ljk5NjI2ODc2IEM0LjU1MjQ1NzksMS42NDA3NDg3NSA0LjY5MDg5NzczLDEuMzEzMjk2MTIgNC45Njc3
Nzc0LDEuMDA3NjczNjYgQzUuMjQ0NjU3MDYsMC43MDUxNjk3OTIgNS41ODUzMTIzOCwwLjU1MjM1ODU2
MSA1Ljk5MTI5ODg1LDAuNTUyMzU4NTYxIEwzNC4wMjQ1ODcyLDAuNTUyMzU4NTYxIEMzNC4zNzc2ODY1
LDAuNTUyMzU4NTYxIDM0LjY5NTAwOTMsMC42OTI2OTU0MDUgMzQuOTcxODg4OSwwLjk3MDI1MDQ5OCBD
MzUuMjUwMzI0MSwxLjI0OTM2NDg5IDM1LjM4ODc2MzksMS41OTA4NTEyMSAzNS4zODg3NjM5LDEuOTk2
MjY4NzYgTDM1LjM4ODc2MzksMTkuMzg4NjgxNyBMMzguMzQ0MjIxLDE5LjM4ODY4MTcgQzM4Ljg5OTUz
NTksMTkuMzg4NjgxNyAzOS4xNzc5NzEsMTkuNjY3Nzk2MSAzOS4xNzc5NzEsMjAuMjI0NDY1NiBMMzku
MTc3OTcxLDIwLjIyNDQ2NTYgWiIgaWQ9IkZpbGwtMyIgZmlsbD0iIzAwMDAwMCI+PC9wYXRoPgogICAg
ICAgICAgICAgICAgICAgICAgICAgICAgPC9nPgogICAgICAgICAgICAgICAgICAgICAgICA8L2c+CiAg
ICAgICAgICAgICAgICAgICAgICAgIDxnIGlkPSJHcm91cC0xNyIgdHJhbnNmb3JtPSJ0cmFuc2xhdGUo
MC4wMDAwMDAsIDEuMzU3MTQzKSI+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICA8cGF0aCBkPSJN
MjguMTExNzg0OSw1NS42NDI4NTcxIEMxMi41ODYyNzA4LDU1LjY0Mjg1NzEgMCw0My4xODY1ODUzIDAs
MjcuODIxNDI4NiBDMCwxMi40NTc4MTc1IDEyLjU4NjI3MDgsMCAyOC4xMTE3ODQ5LDAgQzQzLjYzNzI5
OSwwIDU2LjIyMzU2OTgsMTIuNDU3ODE3NSA1Ni4yMjM1Njk4LDI3LjgyMTQyODYgQzU2LjIyMzU2OTgs
NDMuMTg2NTg1MyA0My42MzcyOTksNTUuNjQyODU3MSAyOC4xMTE3ODQ5LDU1LjY0Mjg1NzEgWiIgaWQ9
IlN0cm9rZS03LUNvcHktMiIgc3Ryb2tlPSJ1cmwoI2xpbmVhckdyYWRpZW50LTIpIiBzdHJva2Utd2lk
dGg9IjEuMTMzNzQwOCIgZmlsbD0iI0U4RThFNyIgdHJhbnNmb3JtPSJ0cmFuc2xhdGUoMjguMTExNzg1
LCAyNy44MjE0MjkpIHNjYWxlKC0xLCAxKSB0cmFuc2xhdGUoLTI4LjExMTc4NSwgLTI3LjgyMTQyOSkg
Ij48L3BhdGg+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICA8cGF0aCBkPSJNMjkuMjM2MjU2Myw1
My45NTY3MSBDMTQuNjQyMjczLDUzLjk1NjcxIDIuODExMTc4NDksNDIuMjU1MzYzNyAyLjgxMTE3ODQ5
LDI3LjgyMTQyODYgQzIuODExMTc4NDksMTMuMzg4OTQ1NCAxNC42NDIyNzMsMS42ODYxNDcxOSAyOS4y
MzYyNTYzLDEuNjg2MTQ3MTkgQzQzLjgzMDIzOTUsMS42ODYxNDcxOSA1NS42NjEzMzQxLDEzLjM4ODk0
NTQgNTUuNjYxMzM0MSwyNy44MjE0Mjg2IEM1NS42NjEzMzQxLDQyLjI1NTM2MzcgNDMuODMwMjM5NSw1
My45NTY3MSAyOS4yMzYyNTYzLDUzLjk1NjcxIFoiIGlkPSJTdHJva2UtNy1Db3B5LTMiIGZpbGw9IiNG
RkZGRkYiIHRyYW5zZm9ybT0idHJhbnNsYXRlKDI5LjIzNjI1NiwgMjcuODIxNDI5KSBzY2FsZSgtMSwg
MSkgdHJhbnNsYXRlKC0yOS4yMzYyNTYsIC0yNy44MjE0MjkpICI+PC9wYXRoPgogICAgICAgICAgICAg
ICAgICAgICAgICAgICAgPHJlY3QgaWQ9IlJlY3RhbmdsZS0zLUNvcHktMTAiIGZpbGw9IiNFOUU5RUEi
IHRyYW5zZm9ybT0idHJhbnNsYXRlKDE5LjY3ODI0OSwgMzUuNjkwMTE1KSByb3RhdGUoLTkwLjAwMDAw
MCkgdHJhbnNsYXRlKC0xOS42NzgyNDksIC0zNS42OTAxMTUpICIgeD0iMTguODM0ODk1OSIgeT0iMjYu
MTM1MjgxNCIgd2lkdGg9IjEuNjg2NzA3MDkiIGhlaWdodD0iMTkuMTA5NjY4MSIgcng9IjAuODQzMzUz
NTQ2Ij48L3JlY3Q+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICA8cmVjdCBpZD0iUmVjdGFuZ2xl
LTMtQ29weS0xMSIgZmlsbD0iI0U5RTlFQSIgdHJhbnNmb3JtPSJ0cmFuc2xhdGUoMTcuOTkxNTQyLCAz
Ni44MTQyMTQpIHJvdGF0ZSgtOTAuMDAwMDAwKSB0cmFuc2xhdGUoLTE3Ljk5MTU0MiwgLTM2LjgxNDIx
NCkgIiB4PSIxNy4xNDgxODg4IiB5PSIzMC4wNjk2MjQ4IiB3aWR0aD0iMS42ODY3MDcwOSIgaGVpZ2h0
PSIxMy40ODkxNzc1IiByeD0iMC44NDMzNTM1NDYiPjwvcmVjdD4KICAgICAgICAgICAgICAgICAgICAg
ICAgICAgIDxyZWN0IGlkPSJSZWN0YW5nbGUtMy1Db3B5LTEyIiBmaWxsPSIjRTlFOUVBIiB0cmFuc2Zv
cm09InRyYW5zbGF0ZSgyMy42MTM4OTksIDM0LjAwMzk2OCkgcm90YXRlKC05MC4wMDAwMDApIHRyYW5z
bGF0ZSgtMjMuNjEzODk5LCAtMzQuMDAzOTY4KSAiIHg9IjIyLjc3MDU0NTgiIHk9IjI0LjQ0OTEzNDIi
IHdpZHRoPSIxLjY4NjcwNzA5IiBoZWlnaHQ9IjE5LjEwOTY2ODEiIHJ4PSIwLjg0MzM1MzU0NiI+PC9y
ZWN0PgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHJlY3QgaWQ9IlJlY3RhbmdsZS0zLUNvcHkt
MTMiIGZpbGw9IiNFOUU5RUEiIHRyYW5zZm9ybT0idHJhbnNsYXRlKDE5LjExNjAxNCwgMTcuMTQyNDk2
KSByb3RhdGUoLTkwLjAwMDAwMCkgdHJhbnNsYXRlKC0xOS4xMTYwMTQsIC0xNy4xNDI0OTYpICIgeD0i
MTguMjcyNjYwMiIgeT0iMTIuNjQ2MTAzOSIgd2lkdGg9IjEuNjg2NzA3MDkiIGhlaWdodD0iOC45OTI3
ODQ5OSIgcng9IjAuODQzMzUzNTQ2Ij48L3JlY3Q+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICA8
cGF0aCBkPSJNMjkuNTkzNTkzOCwxNi43MjU5Njg4IEMyOS40NDExNTQ0LDE2LjcyNTk2ODggMjkuMzY2
NDkwMywxNi44MjU3NjM5IDI5LjM2NjQ5MDMsMTcuMDI4NDcyNyBDMjkuMzY2NDkwMywxNy4xODEyODM5
IDI5LjQ0MTE1NDQsMTcuMjU3Njg5NiAyOS41OTM1OTM4LDE3LjI1NzY4OTYgQzI5Ljc5NDI1MzgsMTcu
MjU3Njg5NiAyOS44OTY5MTcsMTcuMTgxMjgzOSAyOS44OTY5MTcsMTcuMDI4NDcyNyBDMjkuODk2OTE3
LDE2LjgyNTc2MzkgMjkuNzk0MjUzOCwxNi43MjU5Njg4IDI5LjU5MzU5MzgsMTYuNzI1OTY4OCBMMjku
NTkzNTkzOCwxNi43MjU5Njg4IFogTTE2Ljc4ODY4NzEsMzIuNzUwODc3MiBMNDIuNjI1NjA0MSwzMi43
NTA4NzcyIEw0Mi42MjU2MDQxLDE4LjMxOTU3MTcgTDQwLjI3Njc5MzUsMTguMzE5NTcxNyBMMzkuNTE5
MjYzMSwxOC4zMTk1NzE3IEwzOC40NTg0MDk2LDE4LjMxOTU3MTcgTDM3LjcwMDg3OTMsMTguMzE5NTcx
NyBMMzYuNDEyOTIyMiwxOC4zMTk1NzE3IEwzNS42NTM4MzY0LDE4LjMxOTU3MTcgTDE2Ljc4ODY4NzEs
MTguMzE5NTcxNyBMMTYuNzg4Njg3MSwzMi43NTA4NzcyIFogTTE2LjE4MjA0MDYsMzMuMzU5MDAzNSBM
NDMuMjMwNjk1LDMzLjM1OTAwMzUgTDQzLjIzMDY5NSwxNy43MTMwMDQ2IEwxNi4xODIwNDA2LDE3Ljcx
MzAwNDYgTDE2LjE4MjA0MDYsMzMuMzU5MDAzNSBaIE0xNC44OTQwODM1LDM0LjU3MzY5NjkgTDQ0LjU5
NjQyNzMsMzQuNTczNjk2OSBMNDQuNTk2NDI3MywxNy4xODEyODM5IEM0NC41OTY0MjczLDE2LjYyNDYx
NDUgNDQuMzE3OTkyMiwxNi4zNDU1MDAxIDQzLjc2MTEyMTgsMTYuMzQ1NTAwMSBMMTUuNzI3ODMzNSwx
Ni4zNDU1MDAxIEMxNS4xNzA5NjMyLDE2LjM0NTUwMDEgMTQuODk0MDgzNSwxNi42MjQ2MTQ1IDE0Ljg5
NDA4MzUsMTcuMTgxMjgzOSBMMTQuODk0MDgzNSwzNC41NzM2OTY5IFogTTEyLjY5NzcxMjIsMzcuNjEy
NzY5MiBMNDYuMzM4NTkxNSwzNy42MTI3NjkyIEw0Ny4xNzIzNDE1LDM3LjAwNDY0MjkgTDExLjc4Nzc0
MjYsMzcuMDA0NjQyOSBMMTIuNjk3NzEyMiwzNy42MTI3NjkyIFogTTQ4LjM4NDA3ODksMzUuNDA5NDgw
NyBDNDguMzg0MDc4OSwzNS4yMDgzMzEzIDQ4LjI4MTQxNTcsMzUuMTA1NDE3NiA0OC4wODA3NTU3LDM1
LjEwNTQxNzYgTDEwLjk1Mzk5MjYsMzUuMTA1NDE3NiBDMTAuNzUxNzc3MSwzNS4xMDU0MTc2IDEwLjY1
MDY2OTMsMzUuMjA4MzMxMyAxMC42NTA2NjkzLDM1LjQwOTQ4MDcgTDEwLjY1MDY2OTMsMzYuMTcwNDE4
MyBDMTAuNjUwNjY5MywzNi4yMjAzMTU4IDEwLjcyNjg4OSwzNi4zMjAxMTA5IDEwLjg3OTMyODQsMzYu
NDcyOTIyMiBMMTAuODc5MzI4NCwzNi4zOTY1MTY1IEwxMC45NTM5OTI2LDM2LjQ3MjkyMjIgTDQ4LjA4
MDc1NTcsMzYuNDcyOTIyMiBDNDguMjgxNDE1NywzNi40NzI5MjIyIDQ4LjM4NDA3ODksMzYuMzczMTI3
MSA0OC4zODQwNzg5LDM2LjE3MDQxODMgTDQ4LjM4NDA3ODksMzUuNDA5NDgwNyBaIE00OC45MTQ1MDU3
LDM1LjQwOTQ4MDcgTDQ4LjkxNDUwNTcsMzYuMTcwNDE4MyBDNDguOTE0NTA1NywzNi42MjU3MzM0IDQ4
LjY4NzQwMjEsMzYuOTAzMjg4NSA0OC4yMzE2Mzk2LDM3LjAwNDY0MjkgTDQ2LjU2NTY5NTEsMzguMTQy
OTMwNiBMNDYuNTY1Njk1MSwzOC4yMTkzMzYyIEwxMi4zOTQzODksMzguMjE5MzM2MiBMMTIuMzk0Mzg5
LDM4LjE0MjkzMDYgTDEwLjcyNjg4OSwzNy4wMDQ2NDI5IEMxMC4zMjA5MDI1LDM2Ljg1MzM5MDkgMTAu
MTIwMjQyNiwzNi41NzQyNzY1IDEwLjEyMDI0MjYsMzYuMTcwNDE4MyBMMTAuMTIwMjQyNiwzNS40MDk0
ODA3IEMxMC4xMjAyNDI2LDM0Ljg1MjgxMTMgMTAuMzk3MTIyMiwzNC41NzM2OTY5IDEwLjk1Mzk5MjYs
MzQuNTczNjk2OSBMMTQuMjg4OTkyNiwzNC41NzM2OTY5IEwxNC4yODg5OTI2LDE3LjE4MTI4MzkgQzE0
LjI4ODk5MjYsMTYuODI1NzYzOSAxNC40Mjc0MzI0LDE2LjQ5ODMxMTMgMTQuNzA0MzEyMSwxNi4xOTI2
ODg4IEMxNC45ODExOTE3LDE1Ljg5MDE4NSAxNS4zMjE4NDcsMTUuNzM3MzczNyAxNS43Mjc4MzM1LDE1
LjczNzM3MzcgTDQzLjc2MTEyMTgsMTUuNzM3MzczNyBDNDQuMTE0MjIxMiwxNS43MzczNzM3IDQ0LjQz
MTU0MzksMTUuODc3NzEwNiA0NC43MDg0MjM2LDE2LjE1NTI2NTcgQzQ0Ljk4Njg1ODgsMTYuNDM0Mzgw
MSA0NS4xMjUyOTg2LDE2Ljc3NTg2NjQgNDUuMTI1Mjk4NiwxNy4xODEyODM5IEw0NS4xMjUyOTg2LDM0
LjU3MzY5NjkgTDQ4LjA4MDc1NTcsMzQuNTczNjk2OSBDNDguNjM2MDcwNSwzNC41NzM2OTY5IDQ4Ljkx
NDUwNTcsMzQuODUyODExMyA0OC45MTQ1MDU3LDM1LjQwOTQ4MDcgTDQ4LjkxNDUwNTcsMzUuNDA5NDgw
NyBaIiBpZD0iRmlsbC0zLUNvcHkiIGZpbGw9IiMwMDAwMDAiPjwvcGF0aD4KICAgICAgICAgICAgICAg
ICAgICAgICAgPC9nPgogICAgICAgICAgICAgICAgICAgICAgICA8cmVjdCBpZD0iUmVjdGFuZ2xlLTQi
IGZpbGw9IiNGRkVEQ0YiIHg9IjYwLjMxOTc5MDEiIHk9IjE0LjU1MTU4NzMiIHdpZHRoPSI4LjY4Nzg4
MDE4IiBoZWlnaHQ9IjguNzQ2MDMxNzUiPjwvcmVjdD4KICAgICAgICAgICAgICAgICAgICAgICAgPHJl
Y3QgaWQ9IlJlY3RhbmdsZS00LUNvcHktMyIgZmlsbD0iI0ZGRDI4QSIgeD0iNjMuNTYzMjk3NCIgeT0i
MzUuMzIzNDEyNyIgd2lkdGg9IjYuMzE4NDU4MzIiIGhlaWdodD0iNi4zNjA3NTAzNiI+PC9yZWN0Pgog
ICAgICAgICAgICAgICAgICAgICAgICA8cmVjdCBpZD0iUmVjdGFuZ2xlLTQtQ29weSIgZmlsbD0iI0Y5
Q0M4NSIgeD0iNzYuNTM3MzI2OSIgeT0iMjEuMTExMTExMSIgd2lkdGg9IjQuNzM4ODQzNzQiIGhlaWdo
dD0iNC43NzA1NjI3NyI+PC9yZWN0PgogICAgICAgICAgICAgICAgICAgICAgICA8cmVjdCBpZD0iUmVj
dGFuZ2xlLTQtQ29weS00IiBmaWxsPSIjRjhDOEJDIiB4PSI5OS4yNDE4Nzg2IiB5PSIzMC45NTAzOTY4
IiB3aWR0aD0iMy4xNTkyMjkxNiIgaGVpZ2h0PSIzLjE4MDM3NTE4Ij48L3JlY3Q+CiAgICAgICAgICAg
ICAgICAgICAgICAgIDxyZWN0IGlkPSJSZWN0YW5nbGUtNC1Db3B5LTYiIGZpbGw9IiNGOEM4QkMiIHg9
IjkyLjc1NDg2MzgiIHk9IjI0LjM5MDg3MyIgd2lkdGg9IjIuMzY5NDIxODciIGhlaWdodD0iMi4zODUy
ODEzOSI+PC9yZWN0PgogICAgICAgICAgICAgICAgICAgICAgICA8cmVjdCBpZD0iUmVjdGFuZ2xlLTQt
Q29weS01IiBmaWxsPSIjRjhDOEJDIiB4PSI4My4wMjQzNDE3IiB5PSIzNC4yMzAxNTg3IiB3aWR0aD0i
My45NDkwMzY0NSIgaGVpZ2h0PSIzLjk3NTQ2ODk4Ij48L3JlY3Q+CiAgICAgICAgICAgICAgICAgICAg
ICAgIDxyZWN0IGlkPSJSZWN0YW5nbGUtNC1Db3B5LTE0IiBmaWxsPSIjRjhDOEJDIiB4PSI4Ni4yNjc4
NDkxIiB5PSIxNy44MzEzNDkyIiB3aWR0aD0iMi4zNjk0MjE4NyIgaGVpZ2h0PSIyLjM4NTI4MTM5Ij48
L3JlY3Q+CiAgICAgICAgICAgICAgICAgICAgICAgIDxnIGlkPSJHcm91cC0xNSIgdHJhbnNmb3JtPSJ0
cmFuc2xhdGUoMTUwLjU5NzQxMiwgMjkuMzEwNTE2KSBzY2FsZSgtMSwgMSkgdHJhbnNsYXRlKC0xNTAu
NTk3NDEyLCAtMjkuMzEwNTE2KSB0cmFuc2xhdGUoMTI5LjUxNDYxNCwgMTUuNjQ0ODQxKSI+CiAgICAg
ICAgICAgICAgICAgICAgICAgICAgICA8cmVjdCBpZD0iUmVjdGFuZ2xlLTQtQ29weS0yIiBmaWxsPSIj
RkZFRENGIiB4PSIwIiB5PSIwIiB3aWR0aD0iOC42ODc4ODAxOCIgaGVpZ2h0PSI4Ljc0NjAzMTc1Ij48
L3JlY3Q+CiAgICAgICAgICAgICAgICAgICAgICAgICAgICA8cmVjdCBpZD0iUmVjdGFuZ2xlLTQtQ29w
eS03IiBmaWxsPSIjRkZEMjhBIiB4PSI0LjMyNDY3NjUiIHk9IjIwLjc3MTgyNTQiIHdpZHRoPSI2LjMx
ODQ1ODMyIiBoZWlnaHQ9IjYuMzYwNzUwMzYiPjwvcmVjdD4KICAgICAgICAgICAgICAgICAgICAgICAg
ICAgIDxyZWN0IGlkPSJSZWN0YW5nbGUtNC1Db3B5LTkiIGZpbGw9IiNGOUNDODUiIHg9IjE2LjIxNzUz
NjkiIHk9IjUuNDY2MjY5ODQiIHdpZHRoPSI0LjczODg0Mzc0IiBoZWlnaHQ9IjQuNzcwNTYyNzciPjwv
cmVjdD4KICAgICAgICAgICAgICAgICAgICAgICAgICAgIDxyZWN0IGlkPSJSZWN0YW5nbGUtNC1Db3B5
LTEzIiBmaWxsPSIjRjhDOEJDIiB4PSIzOC45MjIwODg1IiB5PSIxNi4zOTg4MDk1IiB3aWR0aD0iMy4x
NTkyMjkxNiIgaGVpZ2h0PSIzLjE4MDM3NTE4Ij48L3JlY3Q+CiAgICAgICAgICAgICAgICAgICAgICAg
ICAgICA8cmVjdCBpZD0iUmVjdGFuZ2xlLTQtQ29weS0xNiIgZmlsbD0iI0Y4QzhCQyIgeD0iMzIuNDM1
MDczNyIgeT0iOC43NDYwMzE3NSIgd2lkdGg9IjIuMzY5NDIxODciIGhlaWdodD0iMi4zODUyODEzOSI+
PC9yZWN0PgogICAgICAgICAgICAgICAgICAgICAgICAgICAgPHJlY3QgaWQ9IlJlY3RhbmdsZS00LUNv
cHktMTgiIGZpbGw9IiNGOEM4QkMiIHg9IjIyLjcwNDU1MTYiIHk9IjE5LjY3ODU3MTQiIHdpZHRoPSIz
Ljk0OTAzNjQ1IiBoZWlnaHQ9IjMuOTc1NDY4OTgiPjwvcmVjdD4KICAgICAgICAgICAgICAgICAgICAg
ICAgICAgIDxyZWN0IGlkPSJSZWN0YW5nbGUtNC1Db3B5LTE5IiBmaWxsPSIjRjhDOEJDIiB4PSIyNS45
NDgwNTkiIHk9IjMuMjc5NzYxOSIgd2lkdGg9IjIuMzY5NDIxODciIGhlaWdodD0iMi4zODUyODEzOSI+
PC9yZWN0PgogICAgICAgICAgICAgICAgICAgICAgICA8L2c+CiAgICAgICAgICAgICAgICAgICAgPC9n
PgogICAgICAgICAgICAgICAgPC9nPgogICAgICAgICAgICA8L2c+CiAgICAgICAgPC9nPgogICAgPC9n
Pgo8L3N2Zz4=
`,
	},

	"/definitions/images/success.svg": {
		local:   "definitions/images/success.svg",
		size:    1103,
		modtime: 1489449600,
		compressed: `
PD94bWwgdmVyc2lvbj0iMS4wIiBlbmNvZGluZz0idXRmLTgiPz4KPCEtLSBHZW5lcmF0b3I6IEFkb2Jl
IElsbHVzdHJhdG9yIDIyLjAuMSwgU1ZHIEV4cG9ydCBQbHVnLUluIC4gU1ZHIFZlcnNpb246IDYuMDAg
QnVpbGQgMCkgIC0tPgo8c3ZnIHZlcnNpb249IjEuMSIgaWQ9IkxheWVyXzEiIHhtbG5zPSJodHRwOi8v
d3d3LnczLm9yZy8yMDAwL3N2ZyIgeG1sbnM6eGxpbms9Imh0dHA6Ly93d3cudzMub3JnLzE5OTkveGxp
bmsiIHg9IjBweCIgeT0iMHB4IgoJIHZpZXdCb3g9IjAgMCA1MiA1MyIgc3R5bGU9ImVuYWJsZS1iYWNr
Z3JvdW5kOm5ldyAwIDAgNTIgNTM7IiB4bWw6c3BhY2U9InByZXNlcnZlIj4KPHN0eWxlIHR5cGU9InRl
eHQvY3NzIj4KCS5zdDB7ZmlsbDpub25lO3N0cm9rZTojODlBRjhGO3N0cm9rZS1taXRlcmxpbWl0OjEw
O30KCS5zdDF7ZmlsbDojRTFGOUUzO3N0cm9rZTojODlBRjhGO3N0cm9rZS1taXRlcmxpbWl0OjEwO30K
CS5zdDJ7ZmlsbDpub25lO3N0cm9rZTojODlBRjhGO3N0cm9rZS1saW5lY2FwOnJvdW5kO3N0cm9rZS1s
aW5lam9pbjpyb3VuZDtzdHJva2UtbWl0ZXJsaW1pdDoxMDt9Cjwvc3R5bGU+CjxnPgoJPGNpcmNsZSBj
bGFzcz0ic3QwIiBjeD0iMjguOSIgY3k9IjI3LjMiIHI9IjIxLjEiLz4KCTxwb2x5Z29uIGNsYXNzPSJz
dDEiIHBvaW50cz0iMTcuMSwyOC44IDI1LjIsMzYuOSA0MC4xLDIxLjkgMzYuOCwxOC42IDI1LjIsMzAu
MyAyMC4zLDI1LjQgCSIvPgoJPHBhdGggY2xhc3M9InN0MiIgZD0iTTIxLjQsMTEuOWMxLjEtMC4zLDIu
Mi0wLjUsMy40LTAuNSIvPgoJPHBhdGggY2xhc3M9InN0MiIgZD0iTTEyLjgsMjMuNWMwLTMuNywxLjct
Nyw0LjQtOS4zIi8+Cgk8cGF0aCBjbGFzcz0ic3QyIiBkPSJNMTguNiwxMy4zYzAuMy0wLjIsMC42LTAu
NCwwLjktMC41Ii8+Cgk8cGF0aCBjbGFzcz0ic3QyIiBkPSJNMzUuNSw0Mi40Yy0xLjEsMC4zLTIuMiww
LjUtMy40LDAuNSIvPgoJPHBhdGggY2xhc3M9InN0MiIgZD0iTTQ0LjIsMzAuOGMwLDMuNy0xLjcsNy00
LjQsOS4zIi8+Cgk8cGF0aCBjbGFzcz0ic3QyIiBkPSJNMzguNCw0MS4xYy0wLjMsMC4yLTAuNiwwLjQt
MC45LDAuNSIvPgo8L2c+Cjwvc3ZnPgo=
`,
	},

	"/definitions/images/waiting.svg": {
		local:   "definitions/images/waiting.svg",
		size:    2496,
		modtime: 1489449600,
		compressed: `
PD94bWwgdmVyc2lvbj0iMS4wIiBlbmNvZGluZz0idXRmLTgiPz4KPCEtLSBHZW5lcmF0b3I6IEFkb2Jl
IElsbHVzdHJhdG9yIDIyLjAuMSwgU1ZHIEV4cG9ydCBQbHVnLUluIC4gU1ZHIFZlcnNpb246IDYuMDAg
QnVpbGQgMCkgIC0tPgo8c3ZnIHZlcnNpb249IjEuMSIgaWQ9IkxheWVyXzEiIHhtbG5zPSJodHRwOi8v
d3d3LnczLm9yZy8yMDAwL3N2ZyIgeG1sbnM6eGxpbms9Imh0dHA6Ly93d3cudzMub3JnLzE5OTkveGxp
bmsiIHg9IjBweCIgeT0iMHB4IgoJIHZpZXdCb3g9IjAgMCA1Ny43IDU3IiBzdHlsZT0iZW5hYmxlLWJh
Y2tncm91bmQ6bmV3IDAgMCA1Ny43IDU3OyIgeG1sOnNwYWNlPSJwcmVzZXJ2ZSI+CjxzdHlsZSB0eXBl
PSJ0ZXh0L2NzcyI+Cgkuc3Qwe2ZpbGw6I0ZGRjFGMTt9Cgkuc3Qxe2ZpbGw6I0Y4QzhCQzt9Cgkuc3Qy
e2ZpbGw6bm9uZTtzdHJva2U6IzY1NzI3RjtzdHJva2Utd2lkdGg6MS41O3N0cm9rZS1saW5lY2FwOnJv
dW5kO3N0cm9rZS1saW5lam9pbjpyb3VuZDtzdHJva2UtbWl0ZXJsaW1pdDoxMDt9Cgkuc3Qze2ZpbGw6
bm9uZTtzdHJva2U6IzY1NzI3RjtzdHJva2Utd2lkdGg6MC43NTtzdHJva2UtbGluZWNhcDpyb3VuZDtz
dHJva2UtbGluZWpvaW46cm91bmQ7c3Ryb2tlLW1pdGVybGltaXQ6MTA7fQoJLnN0NHtmaWxsOiM2NTcy
N0Y7fQo8L3N0eWxlPgo8Y2lyY2xlIGlkPSJYTUxJRF8xMTc2XyIgY2xhc3M9InN0MCIgY3g9IjI3LjYi
IGN5PSIyOC41IiByPSIyNyIvPgo8cG9seWxpbmUgaWQ9IlJlY3RhbmdsZS00LUNvcHktMiIgY2xhc3M9
InN0MSIgcG9pbnRzPSI0MS42LDExLjggMzguNiwxMS44IDM4LjYsMTQuOCA0MS42LDE0LjggIi8+Cjxy
ZWN0IGlkPSJSZWN0YW5nbGUtNC1Db3B5LTkiIHg9IjQyLjkiIHk9IjI1LjEiIGNsYXNzPSJzdDEiIHdp
ZHRoPSIyLjciIGhlaWdodD0iNC4yIi8+Cjxwb2x5bGluZSBpZD0iUmVjdGFuZ2xlLTQtQ29weS0xNiIg
Y2xhc3M9InN0MSIgcG9pbnRzPSIxMi42LDIxLjggMTIuNiwxOS40IDEwLjIsMTkuNCAxMC4yLDIxLjgg
Ii8+CjxyZWN0IGlkPSJSZWN0YW5nbGUtNC1Db3B5LTE5IiB4PSIxNC43IiB5PSIxMi43IiBjbGFzcz0i
c3QxIiB3aWR0aD0iMi43IiBoZWlnaHQ9IjIuNyIvPgo8bGluZSBpZD0iWE1MSURfMTE4OV8iIGNsYXNz
PSJzdDIiIHgxPSIxLjQiIHkxPSI0Mi40IiB4Mj0iOC42IiB5Mj0iNDIuNCIvPgo8bGluZSBpZD0iWE1M
SURfMTE4OF8iIGNsYXNzPSJzdDIiIHgxPSI1MS4xIiB5MT0iNDIuNCIgeDI9IjU2IiB5Mj0iNDIuNCIv
Pgo8bGluZSBpZD0iWE1MSURfMTE4N18iIGNsYXNzPSJzdDIiIHgxPSIxMS40IiB5MT0iNDIuNCIgeDI9
IjQ4LjMiIHkyPSI0Mi40Ii8+CjxsaW5lIGlkPSJYTUxJRF8xMTgzXyIgY2xhc3M9InN0MyIgeDE9Ijku
NiIgeTE9IjQ1LjQiIHgyPSIxNS41IiB5Mj0iNDUuNCIvPgo8bGluZSBpZD0iWE1MSURfMTE4Ml8iIGNs
YXNzPSJzdDMiIHgxPSIxOS4xIiB5MT0iNDUuNCIgeDI9IjI1IiB5Mj0iNDUuNCIvPgo8bGluZSBpZD0i
WE1MSURfMTE4MV8iIGNsYXNzPSJzdDMiIHgxPSIyOC40IiB5MT0iNDUuNCIgeDI9IjQwLjYiIHkyPSI0
NS40Ii8+CjxsaW5lIGlkPSJYTUxJRF8xMTgwXyIgY2xhc3M9InN0MyIgeDE9IjQ0LjIiIHkxPSI0NS40
IiB4Mj0iNTEuMSIgeTI9IjQ1LjQiLz4KPHBhdGggY2xhc3M9InN0NCIgZD0iTTMzLjYsMjYuOGMwLjUt
MC4yLDEuNi0xLjIsMS42LTMuMWwwLjEtMy45aDAuNWMwLjIsMCwwLjUtMC4xLDAuNS0wLjR2LTIuMWMw
LTAuMi0wLjEtMC41LTAuNC0wLjVsLTE3LjEtMC4xCgljLTAuMiwwLTAuNSwwLjEtMC41LDAuNHYyLjFj
MCwwLjIsMC4xLDAuNSwwLjQsMC41aDAuNWwtMC4xLDRjMCwxLjksMC41LDIuNCwxLjUsMi45bDQuNCwy
Ljd2MS4xYy0xLjIsMC42LTQsMi4yLTQuNCwyLjYKCWMtMC40LDAuMi0xLjYsMS4yLTEuNiwzdjMuM2gt
MC41Yy0wLjIsMC0wLjUsMC4xLTAuNSwwLjR2MmMwLDAuMiwwLjEsMC41LDAuNCwwLjVsMTcuMSwwLjJj
MC4yLDAsMC41LTAuMSwwLjUtMC40di0yLjEKCWMwLTAuMi0wLjEtMC41LTAuNC0wLjVoLTAuNXYtMy4z
YzAtMS45LTAuNS0yLjQtMS41LTIuOWwtNC40LTIuN3YtMS4xQzMwLjUsMjguNiwzMy4zLDI3LDMzLjYs
MjYuOHogTTI4LjEsMzEuNmw0LjcsMi45CgljMC40LDAuMiwwLjYsMC40LDAuNywwLjVjMC4xLDAuMiww
LjIsMC42LDAuMSwxLjF2My4zbC0xMi44LTAuMVYzNmMwLTEuMSwwLjYtMS40LDAuOS0xLjZjMC40LTAu
MiwyLjktMS42LDQuNy0yLjcKCWMwLjItMC4xLDAuNC0wLjQsMC41LTAuNmwwLjEtMmMwLTAuMi0wLjIt
MC42LTAuNC0wLjZsLTQuOS0zLjFsMCwwYy0wLjQtMC4yLTAuNi0wLjQtMC43LTAuNWMtMC4xLTAuMi0w
LjItMC42LTAuMS0xLjFsMC4xLTQKCWwxMi44LDAuMWwtMC4xLDMuOWMwLDEuMS0wLjYsMS40LTAuOSwx
LjZjLTAuNCwwLjItMi45LDEuNi00LjcsMi43Yy0wLjIsMC4xLTAuNCwwLjQtMC41LDAuNnYyQzI3LjYs
MzEuMiwyNy45LDMxLjMsMjguMSwzMS42eiIvPgo8L3N2Zz4K
`,
	},

	"/definitions/images/warning.svg": {
		local:   "definitions/images/warning.svg",
		size:    1203,
		modtime: 1489449600,
		compressed: `
PD94bWwgdmVyc2lvbj0iMS4wIiBlbmNvZGluZz0idXRmLTgiPz4KPCEtLSBHZW5lcmF0b3I6IEFkb2Jl
IElsbHVzdHJhdG9yIDIyLjAuMSwgU1ZHIEV4cG9ydCBQbHVnLUluIC4gU1ZHIFZlcnNpb246IDYuMDAg
QnVpbGQgMCkgIC0tPgo8c3ZnIHZlcnNpb249IjEuMSIgaWQ9IkxheWVyXzEiIHhtbG5zPSJodHRwOi8v
d3d3LnczLm9yZy8yMDAwL3N2ZyIgeG1sbnM6eGxpbms9Imh0dHA6Ly93d3cudzMub3JnLzE5OTkveGxp
bmsiIHg9IjBweCIgeT0iMHB4IgoJIHZpZXdCb3g9IjAgMCA1NiA1NyIgc3R5bGU9ImVuYWJsZS1iYWNr
Z3JvdW5kOm5ldyAwIDAgNTYgNTc7IiB4bWw6c3BhY2U9InByZXNlcnZlIj4KPHN0eWxlIHR5cGU9InRl
eHQvY3NzIj4KCS5zdDB7ZmlsbDojQUE4MzgxO3N0cm9rZTojN0I1MTREO3N0cm9rZS13aWR0aDoxLjU7
c3Ryb2tlLWxpbmVjYXA6cm91bmQ7c3Ryb2tlLWxpbmVqb2luOnJvdW5kO3N0cm9rZS1taXRlcmxpbWl0
OjEwO30KCS5zdDF7ZmlsbDpub25lO3N0cm9rZTojN0I1MTREO3N0cm9rZS1saW5lY2FwOnJvdW5kO3N0
cm9rZS1saW5lam9pbjpyb3VuZDtzdHJva2UtbWl0ZXJsaW1pdDoxMDt9Cjwvc3R5bGU+CjxnPgoJPGNp
cmNsZSBjbGFzcz0ic3QwIiBjeD0iMjcuNCIgY3k9IjM4LjIiIHI9IjEuNCIvPgoJPHBhdGggY2xhc3M9
InN0MCIgZD0iTTI4LjYsMTkuNlYzMmMwLDAuNy0wLjYsMS40LTEuMywxLjRjLTAuNywwLTEuMy0wLjUt
MS4zLTEuM1YxOS42di0wLjFjMC0wLjcsMC42LTEuMywxLjMtMS4zCgkJUzI4LjYsMTguOSwyOC42LDE5
LjZ6Ii8+Cgk8Y2lyY2xlIGNsYXNzPSJzdDEiIGN4PSIyNy4zIiBjeT0iMjkuMyIgcj0iMjEuMSIvPgoJ
PHBhdGggY2xhc3M9InN0MSIgZD0iTTExLjIsMjUuNWMwLTMuNywxLjctNyw0LjQtOS4zIi8+Cgk8cGF0
aCBjbGFzcz0ic3QxIiBkPSJNMTcsMTUuM2MwLjMtMC4yLDAuNi0wLjQsMC45LTAuNSIvPgoJPHBhdGgg
Y2xhc3M9InN0MSIgZD0iTTE5LjgsMTMuOWMxLjEtMC4zLDIuMi0wLjUsMy40LTAuNSIvPgoJPHBhdGgg
Y2xhc3M9InN0MSIgZD0iTTMzLjksNDQuNGMtMS4xLDAuMy0yLjIsMC41LTMuNCwwLjUiLz4KCTxwYXRo
IGNsYXNzPSJzdDEiIGQ9Ik00Mi42LDMyLjhjMCwzLjctMS43LDctNC40LDkuMyIvPgoJPHBhdGggY2xh
c3M9InN0MSIgZD0iTTM2LjgsNDMuMWMtMC4zLDAuMi0wLjYsMC40LTAuOSwwLjUiLz4KPC9nPgo8L3N2
Zz4K
`,
	},

	"/": {
		isDir: true,
		local: "",
	},

	"/definitions": {
		isDir: true,
		local: "definitions",
	},

	"/definitions/images": {
		isDir: true,
		local: "definitions/images",
	},
}
